///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef XMLPARAM_H
#define XMLPARAM_H

#include "libxmlparam_global.h"
#include <QDomDocument>
#include <QDomNodeList>
#include <string>
#include <vector>


// Convenient type definitions for lists of quantities
typedef std::vector<bool> BoolArray;
typedef std::vector<int> IntArray;
typedef std::vector<float> FloatArray;
typedef std::vector<std::string> StringArray;


/**
 * A class for loading / saving global parameters in XML format.
 *
 *
 *	How to use this library:
 *
 *  In your main() function, pass the command-line arguments and
 *  a variable storing the parameters (xmlParam par) into the method
 *
 *      bool initParamXML(int argc, char *argv[], xmlParam &par);
 *
 *  This method returns true if the reading/writing of parameters is completed;
 *  it will return false if there is some error.
 *
 *  Also, declare and implement a function to link global variables to the
 *  parameters structure:
 *
 *	void setXMLParamList(xmlParam &p) {
 *		// populate the parameter list with your desired global variables
 *		p.addParam("endFrame", "int", (void*)&endFrame);
 *		p.addParam("exposure", "float", (void*)&exposure);
 *		p.addParam("basePath", "string", (void*)&basePath);
 *	}
 *
 *  Allowed types are: bool, int, float, string, boolArray, intArray, floatArray, stringArray.
 *
 *
 */


/**
 * A class for loading / saving global parameters in XML format.
 */
class LIBXMLPARAM_EXPORT xmlParam {
public:
	// Default Constructor
	xmlParam();

	// Constructor with name
	xmlParam(char *name);

	// Helper Methods
	void addParam(std::string name, std::string type, void* ptr, const char *comment = NULL);
	void resetParams();
	void printParams();
	void printParamsINI();
	QDomDocument makeParamXML();
	bool writeParamXML(const char* filename, QDomDocument &doc);
	bool readParamXML(const char* filename, QDomDocument &doc);
	void loadParamXML(QDomDocument &doc);
	std::string getTopName();
	bool initParamXML(int argc, char *argv[]);

private:
	void printUsage(char *argv0);
	std::vector<std::string> paramList;
	std::vector<std::string> paramTypes;
	std::vector<void*> paramPtr;
	std::vector<std::string> paramComment;
	std::string topName;
};


// Declaration of setXMLParamList (this must be defined and implemented in your main.cpp file!)
extern void setXMLParamList(xmlParam &p);


#endif // XMLPARAM_H
