///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __ABSTRACT_DEFORM_H__
#define __ABSTRACT_DEFORM_H__

#include <QThread>
#include "subject.h"

// Enumerate deformation types.
enum DeformTypes {
	DEFORM_NONE = 0,
	DEFORM_EMBEDDED,
	DEFORM_WEIGHTED,
	DEFORM_EM,
	DEFORM_ALL
};

namespace Deform {

	class DrawOptions;
	class SolveOptions;

	//////////////////////////////////////////////////////////////////////////
	// Abstract class representing deformation algorithms.
	//////////////////////////////////////////////////////////////////////////
	class DeformAlgorithm : public Subject {
	public:
		DeformAlgorithm ( const DeformTypes type ) : deformType(type), iterationFinished(true) { };
		virtual ~DeformAlgorithm() { };

		virtual void solveDeformation( const SolveOptions &options ) = 0;			/**< Solve for the deformation */
		virtual void draw( const DrawOptions &options ) const = 0;					/**< Draw the deformed mesh and related visualizations */

		DeformTypes getType() const {
			return deformType;
		};

		bool finished() const {
			return iterationFinished;
		};

	protected:
		bool iterationFinished;						/**< This variable is set true when the solve iteration has finished */

	private:
		const DeformTypes deformType;				/**< Name of the algorithm (used for printing purposes) */
	};


	/**
	* Passing solver options to the various DeformAlgorithms.
	*/
	class SolveOptions
	{
	public:
		SolveOptions() {
			init();
		};

		int maxItr;				/**< Maximum number of iterations in the solver */

	private:
		inline void init() {
			maxItr = 0;
		};
	};

	class SolveThread : public QThread
	{
	public:
		SolveThread() : deform(NULL), options() { }
		SolveThread( DeformAlgorithm * const deform , const SolveOptions &options ) : deform(deform), options(options) { }

		// Run the solver thread
		virtual void run() {
			// Set termination to be enabled
			this->setTerminationEnabled(true);

			if (deform) {
				deform->solveDeformation(options);
			}

			// Exit the solver thread
			this->exit();
		}

	private:
		DeformAlgorithm * const deform;		/**< Algorithm to execute (managed externally) */
		SolveOptions options;				/**< Solver options */
	};
}

#endif
