///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "Colormap.h"
#include "util.h"

using namespace std;

Colormap::Colormap()
{
	makeDefault();
}

void Colormap::makeDefault() {
	// Default colormap is "autumn" from Matlab
	CtrlPts.resize(2);
	CtrlColor.resize(2);
	CtrlPts[0] = 0; CtrlColor[0] = RealColor(1,0,0);
	CtrlPts[1] = 1; CtrlColor[1] = RealColor(1,1,0);
}

bool Colormap::save( const char *str )
{
	// Save the information to a file
	QFile file(str);
	if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
		cout << "Warning in Colormap::save(): Cannot open save file " << str << "!" << endl;
		return false;
	}

	cout << "Saving colormap...";

	QTextStream out(&file);
	for (unsigned int i = 0; i < CtrlPts.size(); i++) {
		out << CtrlPts[i] << ", " << CtrlColor[i] << "\n";
	}
	file.close();

	cout << "done!" << endl;
	return true;
}

bool Colormap::load( const char *str )
{
	QFile file(str);
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		cout << "Warning in Colormap::load(): Cannot open file " << str << "!" << endl;
		return false;
	}

	CtrlPts.clear();
	CtrlColor.clear();
	cout << "Loading colormap...";

	real_type pt, r, g, b;
	QTextStream in(&file);
	while (!in.atEnd()) {
		QString line = in.readLine();
		sscanf_s(qPrintable(line), "%f, [%f,%f,%f]", &pt, &r, &g, &b);
		
		if ((CtrlPts.size() > 0) && (pt <= CtrlPts[CtrlPts.size()-1])) {
			cerr << endl << "Error in Colormap::load(): Control points must be in strictly increasing order!  Creating default colormap." << endl;
			makeDefault();
			return false;
		}
		CtrlPts.push_back(pt);
		CtrlColor.push_back(RealColor(r,g,b));
	}

	file.close();

	cout << "done!" << endl;
	return true;
}

RealColor Colormap::getColor( real_type x ) const
{
	// Clamp the value between 0 and 1
	x = std::max((real_type)0.0, std::min((real_type)1.0, x));

	// Search for the interval
	index_type i = 0;
	while (CtrlPts[i] < x) {
		i++;
	}

	if (i == 0) {
		return CtrlColor[0];
	}
	else if (i > CtrlPts.size()-1) {
		return CtrlColor[CtrlColor.size()-1];
	}
	else {
		// Linearly interpolate
		real_type a, b, len;
		a = CtrlPts[i-1];
		b = CtrlPts[i];
		len = b - a;
		
		return (CtrlColor[i]*((x-a)/(b-a)) + CtrlColor[i-1]*((b-x)/(b-a)));
	}
}
