///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "CorrespViewer.h"
#include "scene.h"
#include "viewer.h"
#include <algorithm>
#include "FileSequence.h"

using namespace std;

CorrespViewerElement::CorrespViewerElement(CorrespViewer *v, Scene *s, int displayMode) :
parent(v), scene(s), displayMode(displayMode), pointSelect(3), selectionMode_(NONE),
contextMenu(NULL), saveCorrespondencesAct(NULL), loadCorrespondencesAct(NULL),
clearCorrespondencesAct(NULL), showMotionSegmentationAct(NULL), showModelsAct(NULL),
switchCorrespViewTypeAct(NULL), dummyTitleAct(NULL), loadCorrespResultAct(NULL),
saveCorrespResultAct(NULL), showVerifiedRegionsAct(NULL), saveScreenshotAct(NULL),
showExtraEdgesAct(NULL), displaySubsetNum(10), displaySubsetIndex(0), displaySubsetTotal(0),
displaySubsetMode(TRANSFORM_SET_PAIRS), showMotionSegmentationAsPointsAct(NULL),
switchSelectModeAct(NULL), cList(NULL), displayDeformEMResultAct(NULL)
{
}

CorrespViewerElement::~CorrespViewerElement()
{
	if (contextMenu) delete contextMenu;
	if (dummyTitleAct) delete dummyTitleAct;
	if (saveCorrespondencesAct) delete saveCorrespondencesAct;
	if (loadCorrespondencesAct) delete loadCorrespondencesAct;
	if (clearCorrespondencesAct) delete clearCorrespondencesAct;
	if (showMotionSegmentationAct) delete showMotionSegmentationAct;
	if (showModelsAct) delete showModelsAct;
	if (switchCorrespViewTypeAct) delete switchCorrespViewTypeAct;
	if (saveCorrespResultAct) delete saveCorrespResultAct;
	if (loadCorrespResultAct) delete loadCorrespResultAct;
	if (showVerifiedRegionsAct) delete showVerifiedRegionsAct;
	if (saveScreenshotAct) delete saveScreenshotAct;
	if (showExtraEdgesAct) delete showExtraEdgesAct;
	if (showMotionSegmentationAsPointsAct) delete showMotionSegmentationAsPointsAct;
	if (switchSelectModeAct) delete switchSelectModeAct;
	if (displayDeformEMResultAct) delete displayDeformEMResultAct;
}

void CorrespViewerElement::init()
{
	cout << " --> Initializing ";
	switch(displayMode)
	{
	case DISPLAY_SOURCE:
		cout << "Source";
		break;
	case DISPLAY_TARGET:
		cout << "Target";
	    break;
	default:
	    break;
	}
	cout << " Viewer...";

	createContextMenu();

	//////////////////////////////////////////////////////////////////////////
	cout << " view";
	// Disable the exit key
	setShortcut(EXIT_VIEWER, 0);

	if (displayMode == DISPLAY_SOURCE) {
		restoreStateFromFile();
	}

	// Center the view on the source and target
	real3_type box_min, box_max;
	for (int d = 0; d < 3; d++) {
		box_min[d] = min(scene->source->box_min[d], scene->target->box_min[d]);
		box_max[d] = max(scene->source->box_max[d], scene->target->box_max[d]);
	}
	setSceneBoundingBox(qglviewer::Vec(box_min[0], box_min[1], box_min[2]),
						qglviewer::Vec(box_max[0], box_max[1], box_max[2]));
	//camera()->setZClippingCoefficient(100);
	//camera()->setZNearCoefficient(1e-5);
	showEntireScene();

	//////////////////////////////////////////////////////////////////////////
	cout << " gl";
	// Set up the material
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);

	// Used to display semi-transparent selection rectangle
	glBlendFunc(GL_ONE, GL_ONE);

	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));

	// Set the max select buffer size
	setSelectBufferSize(100000);

	// Start the animation
	startAnimation();

	//////////////////////////////////////////////////////////////////////////
	cout << "...done!" << endl;
}

void CorrespViewerElement::draw()
{
	if (!scene->getAnimation()) {
		return;
	}

	if (displayDeformEMResultAct->isChecked()) {
		Deform::DrawOptions drawOptions;

		drawOptions.showSourceModel = false;
		drawOptions.showTargetModel = (displayMode == DISPLAY_SOURCE);
		drawOptions.showDeformedModel = (displayMode == DISPLAY_SOURCE);
		drawOptions.showEffRad = false;
		drawOptions.renderTransparent = false;
		drawOptions.showComponent = -1;
		drawOptions.showGraph = (displayMode == DISPLAY_TARGET);
		drawOptions.showGraphDeformed = true;
		drawOptions.showGraphFaces = true;
		drawOptions.showExtendedGrid = -1;
		drawOptions.showWireframe = false;
		drawOptions.openGLPointSize = 5.0;
		drawOptions.sourceModelColor = RealColor(1.0, 0.7, 0.7);
		drawOptions.deformedModelColor = RealColor(0.7, 1.0, 0.7);
		drawOptions.targetModelColor = RealColor(0.7, 0.7, 1.0);
		drawOptions.ignoreDisplayInterface = (displayMode == DISPLAY_TARGET);

		scene->drawDeformation(drawOptions);

		if (displayMode == DISPLAY_SOURCE) {
			QFont useFont;
			useFont.setBold(true);
			useFont.setPixelSize(20);

			// Draw solver status on bottom of this window
			glColor3f(0.0,0.0,0.0);
			drawText(20, height() - 20, tr(scene->getSolverStatus().c_str()), useFont);
			//cout << scene->getSolverStatus().c_str();
		}
	}

	// Draws rectangular selection area. Could be done in postDraw() instead.
	if (selectionMode_ == TOGGLE)
		drawSelectionRectangle();
	else if (selectionMode_ == ADD) {
		setOpenGLMaterialColor(scene->selectedColor);
		
		if (displayMode == DISPLAY_SOURCE && parent->setNewSourcePt) {
			drawSphere(parent->newSourcePt, scene->pointSize());
		}
		else if (displayMode == DISPLAY_TARGET && parent->setNewTargetPt) {
			drawSphere(parent->newTargetPt, scene->pointSize());
		}
	}
}

void CorrespViewerElement::updateDisplay()
{
	updateGL();
}

void CorrespViewerElement::drawWithNames()
{
	if (scene->getAnimation()) {

		if (parent->selectMode == SELECT_VERTICES) {
			// Draw mesh vertices
			glPointSize(5.0);
			if (displayMode == DISPLAY_SOURCE) {
				PointList_Utility::drawWithNames(scene->source->vertex, POINT_SQUARES);
			}
			else if (displayMode == DISPLAY_TARGET) {
				PointList_Utility::drawWithNames(scene->target->vertex, POINT_SQUARES);
			}
		}
		else if (parent->selectMode == SELECT_CORRESP) {
			// Draw correspondences
			if (cList) {
				if (displayMode == DISPLAY_SOURCE) {
					PointList_Utility::drawWithNamesWithWeight(cList->sourcePoints, POINT_SPHERES, cList->weight, scene->pointSize());
				}
				else if (displayMode == DISPLAY_TARGET) {
					PointList_Utility::drawWithNamesWithWeight(cList->targetPoints, POINT_SPHERES, cList->weight, scene->pointSize());
				}
			}
		}
	}
}

void CorrespViewerElement::drawSelectionRectangle() const
{
	startScreenCoordinatesSystem();
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);

	glColor4f(0.0, 0.0, 0.3, 0.3);
	glBegin(GL_QUADS);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glLineWidth(2.0);
	glColor4f(0.4, 0.4, 0.5, 0.5);
	glBegin(GL_LINE_LOOP);
	glVertex2i(rectangle_.left(),  rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.top());
	glVertex2i(rectangle_.right(), rectangle_.bottom());
	glVertex2i(rectangle_.left(),  rectangle_.bottom());
	glEnd();

	glDisable(GL_BLEND);
	glEnable(GL_LIGHTING);
	stopScreenCoordinatesSystem();
}

void CorrespViewerElement::mousePressEvent( QMouseEvent *e )
{
	// Start selection. Mode is ADD with Shift key and TOGGLE with Alt key.
	rectangle_ = QRect(e->pos(), e->pos());

	// Reset the add selection mode if an add doesn't immediately follow with another add.
	if (selectionMode_ == ADD) {
		if (((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ControlModifier))
			|| ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ShiftModifier))) {
			parent->setNewSourcePt = parent->setNewTargetPt = false;
			selectionMode_ = NONE;
		}
	}

	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ControlModifier)) {
		selectionMode_ = TOGGLE;
	}
	else if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::ShiftModifier)) {
		// Define selection window dimensions
		setSelectRegionWidth(1);
		setSelectRegionHeight(1);

		// Compute rectangle center and perform selection
		parent->selected.clear();
		select(rectangle_.center());

		if (parent->selected.size() == 1) {
			selectionMode_ = MOVE;
		}
		else {
			selectionMode_ = NONE;
		}
	}
	else if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
		selectionMode_ = ADD;
	}
	else {
		QGLViewer::mousePressEvent(e);
	}
}

void CorrespViewerElement::mouseMoveEvent(QMouseEvent* e)
{
	if (selectionMode_ == TOGGLE) {
		// Updates rectangle_ coordinates and redraws rectangle
		rectangle_.setBottomRight(e->pos());
		parent->updateDisplay();
	}
	else if (selectionMode_ == MOVE) {
		updateGL();

		// Move the selected vertex along the shape
		bool found;
		Geometry::Vec3f pickedPoint = VecToVec3f(camera()->pointUnderPixel(e->pos(), found));

		if (cList && found) {
			cout << "Setting position " << parent->selected[0] << " to " << pickedPoint << endl;

			// Modify the selected point's position
			if (displayMode == DISPLAY_SOURCE) {
				cList->setSourcePos(parent->selected[0], pickedPoint);
			}
			else if (displayMode == DISPLAY_TARGET) {
				cList->setTargetPos(parent->selected[0], pickedPoint);
			}
		}

		parent->updateDisplay();
	}
	else {
		QGLViewer::mouseMoveEvent(e);
	}
}

void CorrespViewerElement::mouseReleaseEvent(QMouseEvent* e)
{
	if (selectionMode_ == TOGGLE) {
		// Actual selection on the rectangular area.
		// Possibly swap left/right and top/bottom to make rectangle_ valid.
		rectangle_ = rectangle_.normalized();

		int rw, rh;
		rw = rectangle_.width();
		rh = rectangle_.height();

		if (rw <= 0) rw = 1;
		if (rh <= 0) rh = 1;

		// Define selection window dimensions
		setSelectRegionWidth(rw);
		setSelectRegionHeight(rh);

		// Compute rectangle center and perform selection
		select(rectangle_.center());

		// reset selection mode
		selectionMode_ = NONE;

		// Update display to show new selected objects
		parent->updateDisplay();
	}
	else if (selectionMode_ == MOVE) {
		// reset selection mode
		selectionMode_ = NONE;

		parent->updateDisplay();
	}
	else if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
		updateGL();

		// Add the picked point
		bool found;
		Geometry::Vec3f pickedPoint = VecToVec3f(camera()->pointUnderPixel(e->pos(), found));

		if (found) {
			//cout << "Adding position " << pickedPoint << endl;

			// Modify the selected point's position
			if (displayMode == DISPLAY_SOURCE) {
				parent->newSourcePt = pickedPoint;
				parent->setNewSourcePt = true;
			}
			else if (displayMode == DISPLAY_TARGET) {
				parent->newTargetPt = pickedPoint;
				parent->setNewTargetPt = true;
			}
		}

		// If both the new source and target point is set, then add the new correspondence
		if (parent->setNewSourcePt && parent->setNewTargetPt) {
			cList->add(parent->newSourcePt, parent->newTargetPt);
			parent->setNewSourcePt = parent->setNewTargetPt = false;
			selectionMode_ = NONE;
			cout << "Adding point pair..." << endl;
		}

		parent->updateDisplay();
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}

void CorrespViewerElement::endSelection( const QPoint& )
{
	// Flush GL buffers
	glFlush();

	// Get the number of objects that were seen through the pick matrix frustum. Reset GL_RENDER mode.
	GLint nbHits = glRenderMode(GL_RENDER);

	//cout << "Number of objects selected: " << nbHits << endl;

	// Interpret results : each object created 4 values in the selectBuffer().
	// (selectBuffer())[4*i+3] is the id pushed on the stack.
	if (nbHits > 0) {
		// If the rectangle is small, then select a single object of minimum depth.
		if ((rectangle_.width() < pointSelect) || (rectangle_.height() < pointSelect)) {
			GLuint zMin = (selectBuffer())[1];
			int minId = (selectBuffer())[3];

			for (int i=1; i<nbHits; ++i) {
				if ((selectBuffer())[4*i+1] < zMin) {
					zMin = (selectBuffer())[4*i+1];
					minId = (selectBuffer())[4*i+3];
				}
			}

			if (displayMode == DISPLAY_SOURCE) {
				parent->toggleIdSelection(minId);
			}
			else if (displayMode == DISPLAY_TARGET) {
				parent->toggleTargetIdSelection(minId);
			}
		}
		else {
			for (int i=0; i<nbHits; ++i) {
				if (displayMode == DISPLAY_SOURCE) {
					parent->toggleIdSelection((selectBuffer())[4*i+3]);
				}
				else if (displayMode == DISPLAY_TARGET) {
					parent->toggleTargetIdSelection((selectBuffer())[4*i+3]);
				}
			}
		}
	}
	else {
		// Clear the selection
		if (displayMode == DISPLAY_SOURCE) {
			parent->selected.clear();
		}
		else if (displayMode == DISPLAY_TARGET) {
			parent->targetSelected.clear();
		}
	}
}

void CorrespViewerElement::keyPressEvent( QKeyEvent *e )
{
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	if ((e->key()==Qt::Key_Delete) && (modifiers==Qt::NoButton)) {
		// Remove the selected points from the correspondence
		if (cList) {
			cList->remove(parent->selected);
			parent->selected.clear();
			parent->updateDisplay();
		}
	}
	else if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = parent->geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);

		// Open the menu at this location
		contextMenu->exec(center);
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Backspace) && (modifiers==Qt::NoButton)) {
		switchCorrespViewType();
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_P) && (modifiers==Qt::NoButton)) {
		parent->updateDisplay();
		saveScreenshot(false);
	}
	else if ((e->key()==Qt::Key_0) && (modifiers==Qt::KeypadModifier)) {
		// Reset the display subset index
		displaySubsetIndex = 0;
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Plus) && (modifiers==Qt::KeypadModifier)) {
		// Increment the current display subset
		displaySubsetIndex += displaySubsetNum;
		if (displaySubsetIndex >= displaySubsetTotal) {
			displaySubsetIndex = 0;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Minus) && (modifiers==Qt::KeypadModifier)) {
		// Decrement the current display subset
		displaySubsetIndex -= displaySubsetNum;
		if (displaySubsetIndex < 0) {
			displaySubsetIndex = displaySubsetTotal - displaySubsetNum;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Asterisk) && (modifiers==Qt::KeypadModifier)) {
		// Increment the current display number
		displaySubsetNum++;
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_Slash) && (modifiers==Qt::KeypadModifier)) {
		// Decrement the current display number
		displaySubsetNum--;
		if (displaySubsetNum < 1) {
			displaySubsetNum = 1;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::NoButton)) {
		displaySubsetMode++;
		if (displaySubsetMode > 3) {
			displaySubsetMode = 3;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::NoButton)) {
		displaySubsetMode--;
		if (displaySubsetMode < 1) {
			displaySubsetMode = 1;
		}
		parent->updateDisplay();
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::CTRL)) {
		scene->regenerateColors();
		parent->updateDisplay();
	}
	else {
		QGLViewer::keyPressEvent(e);
	}
}

void CorrespViewerElement::createContextMenu()
{
	// Create menu bar.
	QString menuTitle;

	if (displayMode == DISPLAY_SOURCE) {
		menuTitle = tr("&Source Model View");
	}
	else if (displayMode == DISPLAY_TARGET) {
		menuTitle = tr("&Target Model View");
	}

	contextMenu = new QMenu(tr("&Menu"), this);

	dummyTitleAct = new QAction(menuTitle, this);
	contextMenu->addAction(dummyTitleAct);
	dummyTitleAct->setDisabled(true);

	contextMenu->addSeparator();

	saveCorrespondencesAct = new QAction(tr("&Save Correspondences..."), this);
	contextMenu->addAction(saveCorrespondencesAct);
	connect(saveCorrespondencesAct, SIGNAL(triggered()), this, SLOT(saveCorrespondences()));

	loadCorrespondencesAct = new QAction(tr("&Load Correspondences..."), this);
	contextMenu->addAction(loadCorrespondencesAct);
	connect(loadCorrespondencesAct, SIGNAL(triggered()), this, SLOT(loadCorrespondences()));

	clearCorrespondencesAct = new QAction(tr("&Clear Correspondences..."), this);
	contextMenu->addAction(clearCorrespondencesAct);
	connect(clearCorrespondencesAct, SIGNAL(triggered()), this, SLOT(clearCorrespondences()));

	contextMenu->addSeparator();

	showModelsAct = new QAction(tr("&Show Models..."), this);
	showModelsAct->setCheckable(true);
	contextMenu->addAction(showModelsAct);
	showModelsAct->setChecked(true);

	showClusterMembersAct = new QAction(tr("&Show Cluster Members..."), this);
	showClusterMembersAct->setCheckable(true);
	contextMenu->addAction(showClusterMembersAct);

	showVerifiedRegionsAct = new QAction(tr("&Show Verified Regions..."), this);
	showVerifiedRegionsAct->setCheckable(true);
	contextMenu->addAction(showVerifiedRegionsAct);

	showMotionSegmentationAct = new QAction(tr("&Show Motion Clustering Result..."), this);
	showMotionSegmentationAct->setCheckable(true);
	contextMenu->addAction(showMotionSegmentationAct);

	switchCorrespViewTypeAct = new QAction(tr("&Switch Correspondence Direction..."), this);
	contextMenu->addAction(switchCorrespViewTypeAct);
	connect(switchCorrespViewTypeAct, SIGNAL(triggered()), this, SLOT(switchCorrespViewType()));

	showExtraEdgesAct = new QAction(tr("&Show Extra Smoothness Constraints..."), this);
	showExtraEdgesAct->setCheckable(true);
	contextMenu->addAction(showExtraEdgesAct);
	showExtraEdgesAct->setChecked(true);

	showMotionSegmentationAsPointsAct = new QAction(tr("&Show Motion Clustering Result As Points..."), this);
	showMotionSegmentationAsPointsAct->setCheckable(true);
	contextMenu->addAction(showMotionSegmentationAsPointsAct);

	switchSelectModeAct = new QAction(tr("&Switch Mouse Selection Mode..."), this);
	contextMenu->addAction(switchSelectModeAct);
	connect(switchSelectModeAct, SIGNAL(triggered()), parent, SLOT(SwitchSelectMode()));

	displayDeformEMResultAct = new QAction(tr("&Display DeformEM Result..."), this);
	displayDeformEMResultAct->setCheckable(true);
	contextMenu->addAction(displayDeformEMResultAct);
	displayDeformEMResultAct->setChecked(true);

	contextMenu->addSeparator();

	saveScreenshotAct = new QAction(tr("&Save Screenshot (Press P)"), this);
	contextMenu->addAction(saveScreenshotAct);
	connect(saveScreenshotAct, SIGNAL(triggered()), this, SLOT(saveScreenshot()));

}


void CorrespViewerElement::saveCorrespondences()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the correspondeces...", scene->lastDir, QString("*.txt"));
	if (savefileName.length() == 0) {
		cout << "Save canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	if (cList) {
		cList->save(std::string(qPrintable(savefileName)));
	}
}

void CorrespViewerElement::loadCorrespondences()
{
	QString loadfileName;
	loadfileName = QFileDialog::getOpenFileName(this, "Choose a file name to load the correspondeces...", scene->lastDir, tr("*.txt"));
	if (loadfileName.length() == 0) {
		cout << "Loading canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(loadfileName);
	}

	if (cList) {
		cList->load(std::string(qPrintable(loadfileName)));
	}
}

void CorrespViewerElement::clearCorrespondences()
{
	if (cList) {
		cList->clear();
		parent->selected.clear();
	}
}

void CorrespViewerElement::switchCorrespViewType()
{
	if (parent->showCorrespType == SHOW_SOURCE_TO_TARGET) {
		parent->showCorrespType = SHOW_TARGET_TO_SOURCE;
	}
	else if (parent->showCorrespType == SHOW_TARGET_TO_SOURCE) {
		parent->showCorrespType = SHOW_SOURCE_TO_TARGET;
	}
}

void CorrespViewerElement::DisplayCorrespondenceResult( bool show )
{
	showMotionSegmentationAct->setChecked(show);
}

void CorrespViewerElement::saveScreenshot( bool promptFile /*= true */ )
{
	QString saveFile;

	if (promptFile) {
		saveFile = QFileDialog::getSaveFileName(this, tr("Save Screenshot"), scene->lastDir, tr("png (*.png)")); 
		if (saveFile.length()) {
			scene->lastDir = grabPath(saveFile);
		}
	}
	else {
		// Use the autosave filename (replace extension w/ png)
		saveFile = tr(scene->autoSaveCorrespondenceResult.c_str());
		saveFile.replace(tr(".data"), tr("_shot.png"));
	}

	if (!QFile::exists(saveFile)) {
		// Grab both views and save it in a file
		QPixmap::grabWindow(parent->winId()).toImage().save(saveFile);

		// Show status message
		displayMessage(tr("Took screenshot!"));
	}
}

void CorrespViewerElement::drawTextMessage( int i, QColor color, QString text )
{
	// Draw a text message on the screen
	// Draw it on column i
	glDisable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glDepthMask (GL_TRUE);
	glDisable(GL_BLEND);
	glDisable(GL_CULL_FACE);

	glColor4f(color.redF(), color.greenF(), color.blueF(), 1.0);
	renderText(10, height() - 12*i, text);
}

void CorrespViewerElement::animate()
{
	// Do nothing for now
}

CorrespViewer::CorrespViewer( Scene *s, Viewer *v ) :
sourceView(NULL), targetView(NULL), scene(s), sceneChanged(false), masterViewer(v),
setNewSourcePt(false), setNewTargetPt(false), selectMode(SELECT_VERTICES),
showCorrespType(SHOW_SOURCE_TO_TARGET)
{
	init();
}

CorrespViewer::~CorrespViewer()
{
	// Restore camera
	targetView->setCamera(tempCamera);
	sourceView->close();
	targetView->close();
	delete sourceView;
	delete targetView;

	// Detach from subject
	scene->Detach(this);
}

void CorrespViewer::init()
{
	// Destroy the object when closing
	setAttribute(Qt::WA_DeleteOnClose);

	// Set the window title.
	setWindowTitle("Correspondence Viewer");

	cout << "Constructing correspondence viewer..." << endl;

	sourceView = new CorrespViewerElement(this, scene, DISPLAY_SOURCE);
	targetView = new CorrespViewerElement(this, scene, DISPLAY_TARGET);

	connect(this, SIGNAL(updateViews()), sourceView, SLOT(updateDisplay()));
	connect(this, SIGNAL(updateViews()), targetView, SLOT(updateDisplay()));

	// Set the same camera
	tempCamera = targetView->camera();
	targetView->setCamera(sourceView->camera());

	// Attach to subject and master viewer
	scene->Attach(this);

	// Initialize the elements of the splitter
	this->addWidget(sourceView);
	this->addWidget(targetView);

	// Set the window geometry
	setGeometry(10, 120, 1260, 560);
}

void CorrespViewer::Update( Subject* theChangedSubject )
{
	// Implement updating of OpenGL here by setting a variable and looping in an animation loop
	sceneMutex.lock();
	sceneChanged = true;
	sceneMutex.unlock();

	//cerr << "Updated by scene at: " << theChangedSubject << endl;
	emit updateViews();
}

void CorrespViewer::updateDisplay() {
	emit updateViews();
}

void CorrespViewer::toggleIdSelection( index_type id )
{
	idxList::iterator found = std::find(selected.begin(), selected.end(), id);
	if (found == selected.end()) {
		selected.push_back(id);
	}
	else {
		selected.erase(found);
	}
}

void CorrespViewer::toggleTargetIdSelection( index_type id )
{
	idxList::iterator found = std::find(targetSelected.begin(), targetSelected.end(), id);
	if (found == targetSelected.end()) {
		targetSelected.push_back(id);
	}
	else {
		targetSelected.erase(found);
	}
}

void CorrespViewer::DisplayCorrespondenceResult( bool show )
{
	sourceView->DisplayCorrespondenceResult(show);
	targetView->DisplayCorrespondenceResult(show);
}

void CorrespViewer::SwitchSelectMode()
{
	if (selectMode == SELECT_VERTICES) {
		selectMode = SELECT_CORRESP;
	}
	else if (selectMode == SELECT_CORRESP) {
		selectMode = SELECT_VERTICES;
	}
}
