///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __CORRESP_VIEWER_H__
#define __CORRESP_VIEWER_H__

#include <QGLViewer/qglviewer.h>
#include <QMutex>
#include <QSplitter>
#include "observer.h"
#include "Correspondence.h"
#include <QMenu>

class Viewer;
class Scene;
class Subject;
class CorrespViewer;

enum selectionMode {
	SELECT_VERTICES = 1,
	SELECT_CORRESP
};

enum showCorrespModes {
	SHOW_SOURCE_TO_TARGET = 1,
	SHOW_TARGET_TO_SOURCE
};

enum TransformSets {
	TRANSFORM_SET_PAIRS = 1,
	TRANSFORM_SET_CLUSTERS,
	TRANSFORM_SET_REGIONS
};

class CorrespViewerElement : public QGLViewer
{
	Q_OBJECT

public:
	CorrespViewerElement(CorrespViewer *v, Scene* s, int displayMode);		/**< Constructor: initialize members here. */
	virtual ~CorrespViewerElement();										/**< Destructor: destroy members here. */

public slots:
	void updateDisplay();

	// Correspondences actions
	void saveCorrespondences();
	void loadCorrespondences();
	void clearCorrespondences();

	// Switch correspondence viewing type
	void switchCorrespViewType();

	// Show correspondence result
	void DisplayCorrespondenceResult( bool show );

	// Grab screenshot
	void saveScreenshot( bool promptFile = true );

protected :
	virtual void draw();
	virtual void init();
	virtual void animate();
	virtual void drawWithNames();
	void drawSelectionRectangle() const;
	virtual void mousePressEvent(QMouseEvent *e);
	virtual void mouseMoveEvent(QMouseEvent *e);
	virtual void mouseReleaseEvent(QMouseEvent *e);
	virtual void keyPressEvent(QKeyEvent *e);
	void endSelection(const QPoint&);

private:
	// Pointer to the entire scene
	Scene *scene;
	CorrespViewer *parent;

	// Correspondence Data
	CorrespondenceList *cList;

	// Mode
	const int displayMode;

	// Subset display
	int displaySubsetMode;
	int displaySubsetNum, displaySubsetIndex, displaySubsetTotal;

	//////////////////////////////////////////////////////////////////////////
	// Selection variables
	// Current rectangular selection
	QRect rectangle_;

	// Different selection modes
	enum SelectionMode { NONE, TOGGLE, MOVE, ADD };
	SelectionMode selectionMode_;

	// Point select threshold
	const int pointSelect;

	void drawTransformedSubset( int transformSet );
	void drawTextMessage( int i, QColor color, QString text );

	// Menu variables and functions
	void createContextMenu();

	QMenu* contextMenu;
	QAction *dummyTitleAct;
	QAction *saveCorrespondencesAct;
	QAction *loadCorrespondencesAct;
	QAction *clearCorrespondencesAct;
	QAction *showMotionSegmentationAct;
	QAction *showModelsAct;
	QAction *switchCorrespViewTypeAct;
	QAction *loadCorrespResultAct;
	QAction *saveCorrespResultAct;
	QAction *showClusterMembersAct;
	QAction *showVerifiedRegionsAct;
	QAction *showExtraEdgesAct;
	QAction *saveScreenshotAct;
	QAction *showMotionSegmentationAsPointsAct;
	QAction *switchSelectModeAct;
	QAction *displayDeformEMResultAct;
};


class CorrespViewer : public QSplitter, public Observer {

	Q_OBJECT

public slots:
	void updateDisplay();
	void SwitchSelectMode();					/**< Switch selection mode */

public:
	CorrespViewer( Scene *s, Viewer *v );
	virtual ~CorrespViewer();

	virtual void Update(Subject* theChangedSubject);		/**< Update the display according to the subject */

	void toggleIdSelection(index_type id);
	void toggleTargetIdSelection(index_type id);

	// Selected points
	int selectMode;
	idxList selected, targetSelected;

	// Adding new correspondences
	bool setNewSourcePt, setNewTargetPt;
	Geometry::Vec3f newSourcePt, newTargetPt;

	// Showing forward or backward correspondences
	int showCorrespType;

	// Show correspondence result
	void DisplayCorrespondenceResult( bool show );

protected:
	virtual void init();

private:
	// Pointer to the entire scene
	Scene *scene;
	Viewer *masterViewer;
	qglviewer::Camera *tempCamera;

	// Source and Target views
	CorrespViewerElement *sourceView, *targetView;

	// Scene change notify variable
	QMutex sceneMutex;
	bool sceneChanged;

signals:
	void updateViews();
};

#endif
