///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __CORRESPONDENCE_H__
#define __CORRESPONDENCE_H__

#include "PointList.h"

class SE3;
class ShapeModel;

/**
 * Be careful modifying points in the correspondence list!!!!
 * This is because when you modify it, it doesn't change the ANN search structure,
 * So make sure you clear the ANN search structure when you modify the correspondences!!!
 */
class CorrespondenceList {
public:

	//////////////////////////////////////////////////////////////////////////
	// Access
	size_type size() const;										/**< The total size of points. */
	real3_type sourcePos( index_type i ) const;					/**< Retrieve the i^th source point. */
	real3_type targetPos( index_type i ) const;					/**< Retrieve the i^th target point. */
	bool valid( index_type i ) const;							/**< Determine if weight is positive. */
	real_type getWeight( index_type i ) const;					/**< Retrieve weight of the i^th correspondence. */

	//////////////////////////////////////////////////////////////////////////
	// Change
	void clear();												/**< Clear all correspondence pairs. */
	void resize( size_type size );								/**< Resize the list of correspondences. */
	void reserve( size_type size );								/**< Reserve space for the list of correspondences. */
	void add( real3_type sourcePt, real3_type targetPt,
		      index_type sourceIndex = 0, index_type targetIndex  = 0);	/**< Add a new correspondence pair, with indices. */
	void remove( idxList list );								/**< Remove a list of pairs. */

	void setSourcePos( index_type i, real3_type pt );
	void setTargetPos( index_type i, real3_type pt );

	//////////////////////////////////////////////////////////////////////////
	// Loading and saving
	bool load(std::string filename);
	bool save(std::string filename) const;

	void loadBinary( std::ifstream &in );
	void saveBinary( std::ofstream &out ) const;

	//////////////////////////////////////////////////////////////////////////
	// Data
	PointList sourcePoints, targetPoints;						/**< Source and target point locations */
	idxList sourceIndex, targetIndex;							/**< Indices for the source and target points */
	std::vector<real_type> weight;								/**< Weight for each correspondence */
};

#endif
