///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __DEFORM_EM_H__
#define __DEFORM_EM_H__

#include "deform.h"
#include "MarkedUniformGrid.h"
#include <hash_map>
#include "types.h"
#include <deque>

using namespace Deform;

class AffineTransform3 {
public:
	AffineTransform3() : lr(0.0,0.0,0.0), tr(0.0,0.0,0.0) { }
	AffineTransform3(Geometry::Vec3d c, Geometry::Vec3d t) : lr(c), tr(t) { }

	// A linearized rotation (cross product) and a translation
	Geometry::Vec3d lr, tr;
};
typedef std::vector<AffineTransform3> AffineDeform;


// Class definition
class ShapeModel;
class Scene;
class GCoptimizationGeneralGraph;
class QAbstractItemModel;

// When stepping through to solve the deformation, specify what type of step to take
enum SolveStepType {
	DEFORM_EM_INITIALIZE = 1,
	DEFORM_EM_UPDATE_CORRESP,
	DEFORM_EM_E_STEP,
	DEFORM_EM_M_STEP,
	DEFORM_EM_AUTO_STEP,
	DEFORM_EM_INTERACTIVE
};

// Viewing the solution
enum ViewSolutionType {
	VIEW_SOURCE = 1,
	VIEW_TARGET,
	VIEW_DEFORMED,
	VIEW_INVERSE_DEFORMED,
	VIEW_ADJ_CORRESP,
	VIEW_CORRESP_SOURCE,
	VIEW_CORRESP_TARGET,
	VIEW_NUM_COLUMNS
};

enum SolverControlType {
	SOLVER_RESUME = 1,
	SOLVER_PAUSE,
	SOLVER_STEP,
	SOLVER_STOP
};

class DeformEMOptions {
public:
	DeformEMOptions();

	void copyChangeableOptions( const DeformEMOptions *options );

	// Initialization variables
	int numBones;
	int numInflNeighbors;
	int gridDivisions;
	bool markCellsWithPoints;

	// Solution variables
	bool solveInitialRigid;
	bool solveFinalNNLS;
	FloatArray EStepWeights, DiscreteMStepWeights, MStepWeights;			/**< Weights for each error term in each step */

	// Iteration related variables
	int maxEstepItr, maxProjItr, maxIRLSItr;

	// Correspondence related variables
	int numCorrespondences;							/**< Number of correspondences to use */
	int updateCorrespType;							/**< Whether to update closest point correspondence during the solving process */
	real_type maxCorrespondenceDistance;			/**< Maximum distance between correspondences, in terms of grid resolution */
	real_type boundaryCorrespondenceDistance;		/**< Maximum distance between correspondences, in terms of grid resolution, for points on boundary */

	// Accumulation related variables
	bool useAccumulatedGeometry;
	int inverse_deformation_maxSearchLevel;
	int inverse_deformation_gridExpandStep;

	// Label swapping related
	real_type maxErrorPerCorrespondenceThreshold;
};



// Keep track of the accumulated skin geometry
class AccumulatedGeometry {
public:
	AccumulatedGeometry();
	AccumulatedGeometry( const ShapeModel &shape );

	void clearData();
	void resetInitialShape( const ShapeModel &shape );										/**< Add initial geometry */
	void accumulateShape( const ShapeModel &shape, real_type maxDistanceThreshold, real_type minDistanceThreshold );			/**< Accumulate the geometry */
	void draw( const DrawOptions &options ) const;			/**< Draw the accumulated geometry */

	void transformVertex( const AccumulatedGeometry &geometry, const SE3 &tf, index_type i );

	PointList vertex, normal;								/**< Vertex & Normal Data */
	SpinImageStack spinImageStack;							/**< Per-vertex spin image data */
	real3_type box_min, box_max;							/**< Bounding box */
	idx3List triangle_vertex_index;							/**< Triangle Data */
};


class Candidate {
public:
	Candidate( real3_type srcPos, real3_type tgtPos, index_type tgtIdx, real_type score, index_type clIndex ) :
	  sourcePos(srcPos), targetPos(tgtPos), targetIndex(tgtIdx), score(score), clIndex(clIndex) { };

	  // Data members  
	  real3_type sourcePos;
	  real3_type targetPos;
	  index_type targetIndex;
	  real_type score;
	  index_type clIndex;
};


class DeformEM_GUI_Communicator : public QObject
{
	Q_OBJECT
public:
	static void updateSolutionView( QAbstractItemModel* model, int numRows );

public slots:
	void updateGUI( int numRows );

signals:
	void solutionIncreased( int numRows );
};


class DeformEM : public DeformAlgorithm
{
public:
	class FrameSolution;														/**< Pre-declare FrameSolution class */

	DeformEM();
	virtual ~DeformEM();

	void initialize( const ShapeModel *m, Scene *s, ShapeModel *target, PointList *sample, const DeformEMOptions &options );
	void initialize( ShapeModel *source, ShapeModel *alignedTarget, int gridDivisions, int numBonesPerGridPoint, int numExpandInfluence, RigidDeform &tfList, std::vector<int> vertexLabels, RigidDeform &invTfList, std::vector<int> invVertexLabels, CorrespondenceList *con, FloatArray Eweights, FloatArray Mweights );
	virtual void draw( const DrawOptions &options ) const;
	void drawAccumulatedGeometry( const DrawOptions &options ) const;

	void solveStep();
	void setStepType( int type );
	virtual void solveDeformation( const SolveOptions &options );
	//void applyDeformation( const RigidDeform &deform );
	void applyDeformation_AccumulatedGeometry( const RigidDeform &deform, AccumulatedGeometry &geometry );
	void applyDeformation_FrameGeometry( const RigidDeform &deform, AccumulatedGeometry &geometry, const ShapeModel *source );
	void concatenateDeformation( const RigidDeform &add, RigidDeform &cat );
	void setInitialDeformation( const RigidDeform &def );						/**< Initialize the deformation */
	void solveWeights_finalNNLSUpdate();
	void generateGridColors();

	// Accumulation model
	void accumulateFrame();														/**< Apply inverse and accumulate new scan */
	ShapeModel* requestNextFrame();												/**< Update the target model with the next shape to align. */
	MarkedUniformGrid3* constructExpandedVertexWeightGrid( real3_type newMinPos, real3_type newMaxPos, const std::vector<PointList*> &accumulationPoints, const std::vector< std::vector<bool> > &accPointsValid );

	// Simultaneous alignment model
	void updateGrid( ShapeModel *newFrame );
	void updateCorrespondences_fromAdjacent( FrameSolution &updateSoln );

	void printCurrentDeformation() const;										/**< Print the current transformations on screen */
	void setOptions( const DeformEMOptions &options );							/**< Setting options */
	void updateCorrespondences( int type, FrameSolution &updateSoln );			/**< Update correspondences */
	void setCorrespondenceThresholds( int iterationNumber );					/**< Update thresholds for outlier correspondence rejection */
	virtual ShapeModel* inverseDeformation( const ShapeModel &shape );			/**< Apply inverse of the solved deformation to a shape */

	void saveResult( const char *filename ) const;								/**< Save the registration result. */
	void loadResult( const char *filename );									/**< Load the registration result. */
	void writeStats( std::ofstream &out ) const;

	// Displaying the solution
	std::pair<QAbstractItemModel*,DeformEM_GUI_Communicator*> solutionViewModel();
	void updateSolutionViewModel();

	// Controlling solver behavior
	void controlSolver( int controlType );

	// Debugging continuous weights: revert to previously saved discrete weights
	void saveDiscreteWeights();
	void revertDiscreteWeights();

	// Interactive solver
	void requestInteractiveCorrespUpdate();

	// Solver status string
	std::string solverStatus;

	// Return pointer to aligned result
	bool writeDeformedResult( const char *file ) const;

private:

	static const int dim = DIMENSION;

	// Constructing reduced M-step system.
	std::vector< std::vector<int> > bonesPerCell, cellsPerBone, bonesPerPoint, pointsPerBone;
	std::vector< stdext::hash_map<int,int> > boneCellLinearIndex, bonePointLinearIndex;

	// Solution variables
	Scene *scene;														/**< Used for updating the next frame */
	MarkedUniformGrid3 *vWeightGrid;									/**< Vertex weight function. */
	mutable QMutex changingFrames;

	// Geometry accumulation variables
	AccumulatedGeometry accumulatedGeometry;							/**< Store reference pose accumulated geometry from multiple frames. */

	// Iteration related variables
	int stepType;
	int currentSolveFrame;												/**< The current frame that we're solving */
	int globalIterationNumber;
	real_type stopEpsilon_cubeRt, stopEpsilon_sqRt, stopEpsilon;

	// Correspondence related variables
	//(deprecated) std::vector<bool> useCorrespondenceSubset;							/**< Use only a subset of the correspondences */
	real_type boundaryCorrespondenceDistance;							/**< Maximum Euclidean distance threshold for boundary points to reject outliers */
	real_type maxCorrespondenceDistance;								/**< Maximum Euclidean distance threshold to reject outliers */
	real_type minNormalCosAngle;										/**< Maximum threshold angle between normals to reject outliers */

	// All options
	DeformEMOptions solveOptions;
	QAbstractItemModel *solutionView;
	DeformEM_GUI_Communicator *solutionViewCommunicator;


	// Graph and labeling related variables
	std::vector<int> excludedLabels;									/**< Keep track of labels that are not used */
	stdext::hash_map<int,std::vector<int> > cellAdjacency;				/**< Storing adjacency information (updated each time a new grid is constructed; can throw away old ones)*/
	
	/**< Debugging: looking at where we insert new labels */
	std::vector< std::pair<real3_type,real3_type> > newBoneBoxes;
	std::vector<PointList> showOutliers;
	std::vector< std::pair<real3_type,real_type> > newBoneSpheres;

	// Controlling solver behavior
	bool controlCheck();
	QMutex controlMutex;
	int controlSignal;


	//////////////////////////////////////////////////////////////////////////
	// Pairwise solution variables
	CorrespondenceList *cl;				/**< List of correspondences for the deformation (managed externally) */
	const ShapeModel *input;			/**< Input mesh (managed externally) */
	ShapeModel *deformed;				/**< Deformed mesh */		


	//////////////////////////////////////////////////////////////////////////
	// Per-frame solution variables
	typedef std::map<int3_type,idxList,int3Comp> int3mapList;
	typedef std::vector<real3_type> clErrorType;

	class FrameSolution
	{
	public:
		// Prevent us from calling the default constructor
		FrameSolution() : cList(NULL), targetModel(NULL), aList(NULL), inverseMappedShape(NULL), sourceModel(NULL) { };
		virtual ~FrameSolution() {
			if (cList) delete cList;
			if (aList) delete aList;
			if (inverseMappedShape) delete inverseMappedShape;
		};

		void setData( const RigidDeform &def, const ShapeModel* sourceModel, const ShapeModel *targetModel, const ShapeModel* initialDeformedGeometry ) {
			this->solvedDeformation = def;
			this->sourceModel = sourceModel;
			this->targetModel = targetModel;
			this->deformedAccumulatedGeometry.resetInitialShape(*initialDeformedGeometry);
		}

		void generateInitialCorrespondences( int numCorrespondences );
		void generateAdjacentCorrespondences( const ShapeModel *shape, int numCorrespondences );

		// Data members
		RigidDeform solvedDeformation;							        /**< Storing the solved deformations (store for each frame) */
		CorrespondenceList* cList;										/**< Storing the correspondences (store for each frame) */
		const ShapeModel* targetModel;									/**< Storing the target models (store for each frame) */
		int3mapList correspondencesPerCell;								/**< Correspondences per cell in each frame (store for each frame) */
		clErrorType correspondenceError;								/**< Keep track of the correspondence error in each frame (3 errors: point-to-point, point-to-plane, normal-to-normal) */

		// Deformed geometry cache (used for precomputing deformed positions and normals for each frame's registration)
		AccumulatedGeometry deformedAccumulatedGeometry;				/**< Store intermediary result of deforming accumulated geometry. */

		CorrespondenceList* aList;										/**< Storing correspondences between adjacent frames */
		ShapeModel* inverseMappedShape;									/**< Keep track of the shape mapped inversely to the reference pose */
		std::vector<bool> cList_ok;										/**< Keep track of which correspondences in cList are actually ok to use */
		const ShapeModel* sourceModel;									/**< Keep track of the source model */
	};
	std::deque<FrameSolution> solutionPerFrame;							/**< Store all solutions in all frames */

	// Methods to access the current solution
	FrameSolution& currentSolution() {
		return solutionPerFrame[currentSolveFrame];
	}
	RigidDeform& currentDeform() {
		return currentSolution().solvedDeformation;
	}
	CorrespondenceList* currentCL() {
		return currentSolution().cList;
	}
	int3mapList& currentCorrespondencesPerCell() {
		return currentSolution().correspondencesPerCell;
	}
	clErrorType& currentCorrespondenceError() {
		return currentSolution().correspondenceError;
	}
	AccumulatedGeometry& currentDeformedAccumulatedGeometry() {
		return currentSolution().deformedAccumulatedGeometry;
	}
	const FrameSolution& currentSolution() const {
		return solutionPerFrame[currentSolveFrame];
	}
	const RigidDeform& currentDeform() const {
		return currentSolution().solvedDeformation;
	}
	const CorrespondenceList* currentCL() const {
		return currentSolution().cList;
	}
	const ShapeModel* currentTarget() const {
		return currentSolution().targetModel;
	}
	const int3mapList& currentCorrespondencesPerCell() const {
		return currentSolution().correspondencesPerCell;
	}
	const clErrorType& currentCorrespondenceError() const {
		return currentSolution().correspondenceError;
	}
	const AccumulatedGeometry& currentDeformedAccumulatedGeometry() const {
		return currentSolution().deformedAccumulatedGeometry;
	}
	const ShapeModel* currentSource() const {
		return currentSolution().sourceModel;
	}


	// Deprecated
	//RigidDeform rigidDeform;											/**< Store the current solved deformation */
	//ShapeModel *targetModel;											/**< Used for updating correspondences (memory managed externally) */
	//std::vector< std::vector<data_type> > solvedWeights;				/**< Storing the solved weights */


	//////////////////////////////////////////////////////////////////////////
	// Debugging: Displaying the initial bones and vertex weights
	PointList initialBoneLocation;
	std::vector<real_type> initialBoneRadius;

	// Debugging: save the weights so that we can revert
	std::vector<data_type> savedDiscreteWeights;

	//////////////////////////////////////////////////////////////////////////
	real_type EvaluateGaussian(real3_type x, real3_type center, real_type d_max);
	bool checkConstraints( CorrespondenceList *checkCL ) const;
	bool checkWeights() const;
	void printTransforms(int itr, const RigidDeform &def) const;
	void printWeights(int itr) const;

	void expandBoneInfluence();										/**< Expand all bones's regions of influence by 1 grid cell */
	void setBoneInfluence_EntireGrid();								/**< Initialize bone's influence sets to the entire grid */
	void setBoneInfluence_EntireGrid_RemoveExcludedBones();


	//////////////////////////////////////////////////////////////////////////
	// Discrete labeling for vertex weights
	static real_type GraphCutDataFn( int p, int l, void *data );
	static real_type GraphCutSmoothFn( int p1, int p2, int l1, int l2, void *data );
	real_type solveDiscreteLabelling();								/**< Perform graph cut optimization */
	void updateDiscreteWeights( GCoptimizationGeneralGraph *gc );

	// Get and set the label at a given grid point
	int getPointLabel( const MarkedUniformGrid3 *grid, int3_type gridIndex ) const;
	bool setPointLabel( MarkedUniformGrid3 *grid, int3_type gridIndex, int label ) const;


	// Labeling heuristics
	void analyzeLabellingResult();									/**< Analyze how labels were used */
	bool reuseExtraLabels();										/**< Reuse labels that were not used */
	//////////////////////////////////////////////////////////////////////////

	class LinearSystemE {
	public:
		LinearSystemE( const size_type numBones );
		LinearSystemE( const size_type numBones, const std::vector<int> &excludedBoneList );

		// Solution organization:
		// Linear system solves for an affine transformation (cross product and a translation):
		//
		// [ bone 1 rot , bone 2 rot, ..., bone n rot   ,    bone 1 tr, bone 2 tr, ... , bone n tr ]
		//
		// --> Total of (2*dim)*numBones variables in the system

		// The constraint objective
		void makeConstraintTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def);
		void makeConstraintTerm_PointToPlane(real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def, const ShapeModel &targetModel);
		real_type findConstraintEnergy( RigidDeform &rd, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, const RigidDeform &currentDef );
		real_type findConstraintEnergy( AffineDeform &ad, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, const RigidDeform &currentDef );

		// The smoothness objective
		void makeSmoothnessTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid, RigidDeform &def);

		// Solution methods
		taucsType* solve(taucsType &residual, taucsType &gradResidual);
		AffineDeform solnToAffine( taucsType *soln, size_type numBones );
		RigidDeform projectToRigid( taucsType *soln, int maxItr, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *vc, const RigidDeform &currentDef );
	
		// Utility methods
		void clear();

		// Data members
		TermList terms;
		const size_type numBones;

		// Vertex weights for each constraint
		const CorrespondenceList *constraints;
		std::vector<idxList> nodeList;
		std::vector<distList> vertexWeight;

		// Bone exclusion code
		const size_type numValidBones;
		void computeBoneIndexTranslation();
		void translateIntegrals( std::vector< std::vector<double> > &cellIntegrals ) const;
		void translateVertexWeights( std::vector<idxList> &nodeList, std::vector<distList> &weightList ) const;
		AffineDeform translateDeformation( const AffineDeform &def ) const;
		RigidDeform translateDeformation( const RigidDeform &def ) const;
		void translateDeformationBack( RigidDeform &def ) const;
		std::vector<int> excludedBones;					/**< Option to exclude certain bones from being optimized (list of bones) */
		std::vector<bool> excludeBone;					/**< Option to exclude certain bones from being optimized (boolean array marking excluded bones) */
		std::vector<int> back_to_original_index, reduced_index;		/**< When certain bones are excluded, compute indices to translate back and forth */
	};

	class LinearSystemM {
	public:
		LinearSystemM(const size_type numBones, const size_type numGridPoints);

		// Solution organization:
		//
		// [ -----  weights for bone 1 ----- , -----  weights for bone 2 ----- , ... , -----  weights for bone n ----- ]
		//
		// --> total gridPoints*numBones variables

		// The constraint objective
		void makeConstraintTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def, const std::vector< std::vector<int> > &affectingBones );
		void makeSmoothTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid);
		void makeUnityTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid);

		// Solving and applying the system
		taucsType* solve( taucsType &residual, taucsType &gradResidual );
		void updateWeights( taucsType *soln, MarkedUniformGrid3 *vWeightGrid );

		// Utility methods
		void clear();
	
		// Data members
		TermList terms;
		const size_type numBones, numGridPoints;
	};

	class LinearSystemM_Reduced {
	public:
		LinearSystemM_Reduced( const size_type numBones, const size_type numGridPoints,
								const std::vector< std::vector<int> > &pointsPerBone,
								const std::vector< std::vector<int> > &cellsPerBone,
								const std::vector< std::vector<int> > &bonesPerPoint,
								const std::vector< std::vector<int> > &bonesPerCell,
								const std::vector< stdext::hash_map<int,int> > &boneCellLinearIndex,
								const std::vector< stdext::hash_map<int,int> > &bonePointLinearIndex );

		// Solution organization:
		//
		// [ ----- affected weights for bone 1 ----- , ----- affected weights for bone 2 ----- , ... , ----- affected weights for bone n ----- ]
		//
		// --> total variables: sum (over i) of (# affected weights in bone i)

		// The constraint objective
		void makeConstraintTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def, const std::vector< std::vector<int> > &affectingBones );
		void makeSmoothTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid);
		void makeUnityTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid);
		void makeZeroTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid);
		void makeDiscreteTerm(real_type weight, MarkedUniformGrid3 *vWeightGrid);

		// Solving and applying the system
		taucsType* solve( int maxIRLSItr, taucsType &resNorm, taucsType &gradResNorm );
		void updateWeights( taucsType *soln, MarkedUniformGrid3 *vWeightGrid );
		taucsType* solve_OLD( taucsType &residual, taucsType &gradResidual );

		// Utility methods
		void clear();
		int findVariableCellIndex( int bone, int gridCellIndex );
		int findVariableIndex( int bone, int gridPointIndex );
		int getBoneOffset( int bone );
		
		// Data members
		TermList terms;
		const size_type numBones, numGridPoints;
		const std::vector< std::vector<int> > &pointsPerBone, &cellsPerBone, &bonesPerPoint, &bonesPerCell;
		const std::vector< stdext::hash_map<int,int> > &boneCellLinearIndex, &bonePointLinearIndex;
		std::vector<int> boneOffsets;
	};
};

#endif
