///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "FindCorrespondence.h"
#include <SylvainTools/simple_mesh.h>
#include <iostream>
#include <fstream>
#include "PointListN.h"
#include "util.h"
#include <clapack.h>
#include <deque>
#include "geom_util.h"
#include "MarkedUniformGrid.h"
#include <GraphCutLib/GCoptimization.h>

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;

extern "C" {
	void dpotrs_(char *uplo,
				int *n, int *nrhs,
				double *a, int *lda,
				double *b, int *ldb,
				int *info);

	void dpotrf_(char *uplo,
				int *n, double *a, int *lda,
				int *info);

	void dpocon_(char *uplo,
				int *n,
				double *a, int *lda,
				double *anorm, double *rcond,
				double *work, int *iwork,
				int *info);
	
	void dgesv_(int *n, int *nrhs,
				double *a, int *lda, int *ipiv,
				double *b, int *ldb,
				int *info);
}

template<int N, typename Real>
double* ToFortranStyle( Geometry::Square_matrix<N,Real> M )
{
	// Write the matrix in column-major format.

	double *out = new double[N*N];

	for (int i = 0; i < N; i++) {
		for (int j = 0; j < N; j++) {
			out[i+j*N] = M(i,j);
		}
	}

	return out;
}

template<int N, typename Real>
double MatrixNorm1( Geometry::Square_matrix<N,Real> M )
{
	double result = -DBL_MAX;

	// Compute column sums
	for (int j = 0; j < N; j++) {
		double colSum = 0;
		for (int i = 0; i < N; i++) {
			colSum += M(i,j);
		}

		if (colSum > result) {
			result = colSum;
		}
	}

	return result;
}


template<int N,typename Real>
void Cholesky_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v )
{
	char uplo;
	int n, nrhs, lda, ldb, info;
	double *a, *x, *work;
	double rcond, anorm;
	int *iwork;

	uplo = 'U'; /* U/L indicated that the upper/lower triangle of the
				symmetric matrix is stored. */

	nrhs = 1; /* The Fortran routine that I am calling will solve a
			  system of equations with multiple right hand sides.
			  nrhs gives the number of right hand sides, and then b
			  must be a matrix of dimension n by nrhs. */
	n = N;
	lda = N;	// Leading dimension of a and b
	ldb = N;
	a = ToFortranStyle(M); // Convert A to a Fortran style matrix

	/* The Fortran routine replaces the input information in b with the
	results.  Since we are interested in the results, we put the
	initial conditions in the output array and use that in the
	call. */
	x = new double[N];
	for (int i = 0; i < N; i++) {
		x[i] = v[i];
	}

	//// Compute the 1-norm of the input matrix
	//anorm = MatrixNorm1(M);

	//// Estimate the condition number for the matrix; print out warning if it is high
	//work = new double[3*N];
	//iwork = new int[N];

	//// Now the function calls
	dpotrf_(&uplo, &n, a, &lda, &info);
	//dpocon_(&uplo, &n, a, &lda, &anorm, &rcond, work, iwork, &info);

	////cout << "Matrix 1-norm is: " << anorm << endl;
	////cout << "Condition number is: " << rcond << endl;

	//if (rcond < DBL_EPSILON) {
	//	cerr << "Warning in Cholesky_Solve(): Matrix is very close to singular! rcond = " << rcond << "." << endl;
	//}

	// Solve the system
	dpotrs_(&uplo, &n, &nrhs, a, &lda, x, &ldb, &info);

	// Replace the input in v with the output x
	for (int i = 0; i < N; i++) {
		v[i] = x[i];
	}

	// Clean up the memory before returning to the calling program
	delete[] a;
	delete[] x;
	//delete[] work;
	//delete[] iwork;
}


template<int N,typename Real>
void General_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v )
{
	int n, nrhs, lda, ldb, info;
	double *a, *x;
	int *ipiv;
	
	nrhs = 1; /* The Fortran routine that I am calling will solve a
			  system of equations with multiple right hand sides.
			  nrhs gives the number of right hand sides, and then b
			  must be a matrix of dimension n by nrhs. */

	n = N;
	lda = N;	// Leading dimension of a and b
	ldb = N;
	a = ToFortranStyle(M); // Convert A to a Fortran style matrix

	/* The Fortran routine replaces the input information in b with the
	results.  Since we are interested in the results, we put the
	initial conditions in the output array and use that in the
	call. */
	ipiv = new int[N];
	x = new double[N];
	for (int i = 0; i < N; i++) {
		x[i] = v[i];
	}

	// Solve the system
	dgesv_(&n, &nrhs, a, &lda, ipiv, x, &ldb, &info);

	// Replace the input in v with the output x
	for (int i = 0; i < N; i++) {
		v[i] = x[i];
	}

	// Clean up the memory before returning to the calling program
	delete[] a;
	delete[] x;
	delete[] ipiv;
}


real_type FitSphereWeight( const real3_type &query, const real3_type &center, real_type &radSq ) {
	// Compute the squared distance between the center of the evaluation kernel and the query point
	real3_type diff = query - center;
	real_type x2 = diff*diff;

	if (x2 < radSq) {
		// Normalize the domain between 0 and 1
		x2 = 1.0 - (x2 * (1.0 / radSq));
		x2 = x2*x2;
		
		// Evaluate the function (1-x^2)^4
		return x2*x2;
	}

	return 0;
}

bool FitSphere( const PointList &pts, const PointList &normals, real3_type &center, real_type &radSq )
{
	// Fit a 3D sphere to the list of points and normals
	// Use constant weight for the points for now
	const int dim = DIMENSION;
	const real_type beta = 1e6;
	real_type sumWeights = 0;
	real_type nsq, elt;

	// Initialize matrices.  These will be initially zero.
	Geometry::Square_matrix<dim+2, float> DTD;
	Geometry::Vec<dim+2, float> DTb;

	for (index_type i = 0; i < pts.size(); i++) {
		// Construct the design matrix and rhs
		real3_type pt = pts[i];
		real3_type normal = normals[i];
		Geometry::Vec<dim+2,float> ri;
		real_type weight;
		
		weight = FitSphereWeight(pt, center, radSq);
		sumWeights += weight;

		//cout << "Point  #" << i << ": " << pt << endl;
		//cout << "Normal #" << i << ": " << normal << endl;
		//cout << "Weight #" << i << ": " << weight << endl;

		// Add the constraint for the point
		nsq = pt*pt;
		DTD(0,0) += weight;
		DTD(0,dim+1) += weight*nsq;
		DTD(dim+1,dim+1) += weight*nsq*nsq;
		for (int d = 0; d < dim; d++) {
			DTD(0,d+1) += weight*pt[d];
			for (int e = 0; e <= d; e++) {
				DTD(e+1,d+1) += weight*pt[e]*pt[d];
			}
			DTD(d+1,dim+1) += weight*nsq*pt[d];
		}

		weight = weight*beta;
		// Constraint for the normal
		for (int d = 0; d < dim; d++) {
			elt = 2*pt[d];
			DTD(d+1,d+1) += weight;
			DTD(d+1,dim+1) += weight*elt;
			DTD(dim+1,dim+1) += weight*elt*elt;
			DTb[d+1] += weight*normal[d];
			DTb[dim+1] += weight*elt*normal[d];
		}
	}

	if (sumWeights < REAL_TYPE_EPSILON) {
		cerr << "Warning in FitSphere(): Sum of weights " << sumWeights << " is very close to zero!" << endl;
	}

	//cout << "DTD: " << endl << DTD << endl;
	//cout << "DTb: " << DTb << endl << endl;

	// Solve the resulting system
	Cholesky_Solve(DTD, DTb);

	//cout << "Solve result: " << DTb << endl;

	// Store the center
	for (int d = 0; d < dim; d++) {
		center[d] = DTb[d+1];
	}

	// Detect whether if it's a planar feature
	if (abs(DTb[dim+1]) < REAL_TYPE_EPSILON) {
		radSq = DTb[0];

		return false;
	}
	else {
		center = -0.5 / DTb[dim+1] * center;
		radSq = center*center - (DTb[0] / DTb[dim+1]);
		//cout << "Center: " << center << ", radius^2: " << radSq << endl;

		return true;
	}
}




void LocalCoordinateFrame( real3_type n, real3_type &u, real3_type &v )
{
	// Construct a local tangent plane (u,v) perpendicular to n.
	// Helper function for ComputePrincipalCurvatures().

	// 1. Find the minimum absolute component value
	int minComp = 0;
	for (int i = 1; i < 3; i++) {
		if (abs(n[i]) < abs(n[minComp])) {
			minComp = i;
		}
	}

	// 2. Use cross products to find a local frame.
	u[0] = u[1] = u[2] = 0.0;
	u[minComp] = 1.0;
	v = n ^ u;
	v.normalize();
	u = v ^ n;
	u.normalize();
}

bool FitCurvatureTensorLS( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result )
{
	const size_type npts = pts.size();
	const int dim = DIMENSION;
	real3_type dp, dn;
	real_type dpu, dpv, dnu, dnv;

	if (npts < 3) {
		cout << "Error in FitCurvatureTensorLS(): Not enough constraints for curvature tensor!" << endl;
		return false;
	}

	// Initialize matrices.  These will be initially zero.
	Geometry::Square_matrix<dim, float> ATA;
	Geometry::Vec<dim, float> ATb;

	// Dense sampling method: for all pairs of points
	for (index_type i = 0; i < (npts-1); i++) {
		for (index_type j = i+1; j < npts; j++) {
			//////////////////////////////////////////////////////////////////////////
			// Finite difference of vertex and normal
			dp = pts[j] - pts[i];
			dn = normals[j] - normals[i];
			// Dot product with local coordinate frame
			dpu = dp*u; dpv = dp*v;
			dnu = dn*u; dnv = dn*v;

			//////////////////////////////////////////////////////////////////////////
			// Add to the design matrix: upper-left diagonal
			ATA(0,0) += dpu*dpu;
			ATA(0,1) += dpu*dpv;
			ATA(2,2) += dpv*dpv;

			// Add to the RHS
			ATb[0] += dpu*dnu;
			ATb[1] += dpv*dnu + dpu*dnv;
			ATb[2] += dpv*dnv;
		}
	}

	// Copy over the values to lower-right block of ATA
	ATA(1,1) = ATA(0,0) + ATA(2,2);
	ATA(1,2) = ATA(0,1);

	// Solve the resulting system
	Cholesky_Solve(ATA, ATb);

	//////////////////////////////////////////////////////////////////////////
	// 3. Find eigenvalues / eigenvectors of curvature tensor
	real_type a,b,c,d;
	a = ATb[0]; b = ATb[1]; c = ATb[2];
	d = sqrt((a-c)*(a-c) + 4.0*b*b);
	result.kmax = 0.5*(a + c + d);
	result.kmin = 0.5*(a + c - d);
	real2_type eigmax(-2.0*b, a - c - d);
	real2_type eigmin(-2.0*b, a - c + d);
	eigmax.normalize();
	eigmin.normalize();
	result.vmax = eigmax[0]*u + eigmax[1]*v; result.vmax.normalize();
	result.vmin = eigmin[0]*u + eigmin[1]*v; result.vmin.normalize();

	//// Check result
	//cout << "% Resulting curvature tensor: " << endl;
	//cout << "A = [" << a << ", " << b << "; " << b << ", " << c << "];" << endl;
	//cout << "% Maximum eigenvalue / eigenvector: " << endl;
	//cout << "kmax = " << result.kmax << ";" << endl;
	//cout << "vmax = " << eigmax << ";" << endl;
	//cout << "% Minimum eigenvalue / eigenvector: " << endl;
	//cout << "kmin = " << result.kmin << ";" << endl;
	//cout << "vmin = " << eigmin << ";" << endl;

	return true;
}

bool FitCurvatureTensorRobust( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result )
{
	const int dim = DIMENSION;
	const size_type numPairs = pts.size() * (pts.size() - 1) / 2;
	const size_type medianIndex = numPairs;
	real3_type dp, dn;
	real_type w1, w2;
	int numItr = 0;
	const int maxItr = 20;
	bool converged = false;
	real_type invSig = 1.0;
	real_type diff;
	const real_type convergeThreshold = 1e-3;
	real_type sigThreshold = REAL_TYPE_MAX;

	// Initialize matrices.  These will be initially zero.
	Geometry::Square_matrix<dim+1, real_type> ATA;
	Geometry::Vec<dim+1, real_type> ATb, lastATb;
	vector< Geometry::Vec<dim+1, real_type> > input;
	vector<real_type> residuals, partialResiduals;
	residuals.resize(2*numPairs);
	partialResiduals.resize(medianIndex+1);
	input.resize(numPairs);

	// Dense sampling method: use all pairs of points for curvature estimate
	index_type pos = 0;
	for (index_type i = 0; i < (pts.size()-1); i++) {
		for (index_type j = i+1; j < pts.size(); j++) {
			//////////////////////////////////////////////////////////////////////////
			// Finite difference of vertex and normal
			dp = pts[j] - pts[i];
			dn = normals[j] - normals[i];

			// Dot product with local coordinate frame & store
			input[pos][0] = dp*u;
			input[pos][1] = dp*v;
			input[pos][2] = dn*u;
			input[pos][3] = dn*v;

			// Initialize residuals
			residuals[2*i] = 1.0;
			residuals[2*i+1] = 1.0;
			
			pos++;
		}
	}

	while (!converged) {
		// Set matrices to zero
		for (index_type i = 0; i < dim+1; i++) {
			for (index_type j = 0; j < dim+1; j++) {
				ATA(i,j) = 0.0;
			}
			ATb[i] = 0.0;
		}

		// Dense sampling method: use all pairs of points for curvature estimate
		for (index_type i = 0; i < input.size(); i++) {
			if (numItr == 0) {
				w1 = w2 = 1.0;
			}
			else {
				if (residuals[2*i] > sigThreshold) {
					w1 = w2 = 0.0;
				}
				else {
					w1 = residuals[2*i]*invSig;   w1 = 1.0 + w1*w1; w1 = 2.0 / (w1*w1); w1 = w1*w1;
					w2 = residuals[2*i+1]*invSig; w2 = 1.0 + w2*w2; w2 = 2.0 / (w2*w2); w2 = w2*w2;
				}
			}

			// Add to the design matrix: upper-left diagonal
			ATA(0,0) += w1*input[i][0]*input[i][0];
			ATA(0,1) += w1*input[i][0]*input[i][1];
			ATA(1,1) += w1*input[i][1]*input[i][1];
			ATA(2,2) += w2*input[i][0]*input[i][0];
			ATA(2,3) += w2*input[i][0]*input[i][1];
			ATA(3,3) += w2*input[i][1]*input[i][1];

			// Add to the RHS
			ATb[0] += w1*input[i][0]*input[i][2];
			ATb[1] += w1*input[i][1]*input[i][2];
			ATb[2] += w2*input[i][0]*input[i][3];
			ATb[3] += w2*input[i][1]*input[i][3];
		}

		//cout << "ATA: " << endl << ATA << endl;
		//cout << "ATb: " << ATb << endl << endl;

		// Solve the resulting system
		Cholesky_Solve(ATA, ATb);

		//cout << "Solve result: " << ATb << endl;

		// Compute residuals, find median
		for (index_type i = 0; i < input.size(); i++) {
			residuals[2*i] = abs(input[i][2] - (input[i][0]*ATb[0] + input[i][1]*ATb[1]));
			residuals[2*i+1] = abs(input[i][3] - (input[i][0]*ATb[2] + input[i][1]*ATb[3]));
			w1 = residuals[2*i]*invSig;   w1 = w1*w1;
			w2 = residuals[2*i+1]*invSig; w2 = w2*w2;
		}
		partial_sort_copy(residuals.begin(), residuals.end(), partialResiduals.begin(), partialResiduals.end());
		w1 = 1.4826 * partialResiduals[medianIndex];
		if ((2.0*w1) < sigThreshold) {
			sigThreshold = 2.0*w1;
			invSig = 1.0/w1;
		}

		// Determine convergence
		if (numItr > 0) {
			diff = (ATb - lastATb).norm();
			if ((numItr >= maxItr) || (diff < convergeThreshold)) {
				converged = true;
			}
		}
		lastATb = ATb;
		numItr++;
	}

	//////////////////////////////////////////////////////////////////////////
	// 3. Find eigenvalues / eigenvectors of curvature tensor
	real_type a,b,c,d,dsq;
	real3_type vmax, vmin;
	a = ATb[0]; b = ATb[1]; c = ATb[2]; d = ATb[3];
	dsq = sqrt((a-d)*(a-d) + 4.0*b*c);
	result.vmax = 0.5*(a - d + dsq)*u + b*v; result.vmax.normalize();
	result.vmin = 0.5*(a - d - dsq)*u + c*v; result.vmin.normalize();
	result.kmax = 0.5*(a + d + dsq);
	result.kmin = 0.5*(a + d - dsq);
	
	return true;
}


QString ComputePrincipalCurvatures( PointList &vertices, const PointList &normals, const distList &scales, CurvInfoList &cl, bool robust )
{
	const size_type	minimumNeighbors = 7;
	const size_type initialNN = 6;
	const real_type nnFactor = 5.0;
	const size_type numScales = scales.size();
	real3_type x, n, u, v;				/**< Vertex & local coordinate frame */
	idxList inRange;
	distList distRange;
	real_type radSq;
	CurvInfo ct;						/**< Curvature tensor */

	cout << "Computing " << (robust?"robust ":"") << "principal curvatures..." << endl;
	cout << "Scales:";
	for (index_type i = 0; i < scales.size(); i++) {
		cout << " " << scales[i];
	}
	cout << endl;

	cl.resize(vertices.size());

	TextProgress(vertices.size(), true);
	for (index_type vi = 0; vi < vertices.size(); vi++) {
		//////////////////////////////////////////////////////////////////////////
		// 1. Compute local frame at the vertex
		x = vertices[vi];
		ct.n = normals[vi];
		LocalCoordinateFrame(ct.n, u, v);
		cl[vi].resize(numScales);

		for (index_type si = 0; si < numScales; si++) {
			//////////////////////////////////////////////////////////////////////////
			if (scales[si] == 0) {
				// Use adaptive neighborhood size: multiple of the distance to the kth nearest neighbor
				vertices.queryKNN(vertices[vi], initialNN, inRange, distRange, true);
				radSq = nnFactor*distRange[initialNN-1];
				radSq = radSq*radSq;
			}
			else {
				// Use pre-specified fixed neighborhood size
				radSq = scales[si];
			}

			//////////////////////////////////////////////////////////////////////////
			// 2. Gather the points within the search radius
			PointList rangePts, rangeNormals;
			vertices.queryRange(x, radSq, inRange, distRange, true);

			if (inRange.size() < minimumNeighbors) {
				vertices.queryKNN(x, minimumNeighbors, inRange, distRange, true);
			}

			rangePts.reserve(inRange.size());
			rangeNormals.reserve(inRange.size());
			for (index_type j = 0; j < inRange.size(); j++) {
				// If the normal differs by an angle more than pi/2
				if (normals[vi] * normals[inRange[j]] <= 0) {
					// Reject
					continue;
				}
				rangePts.add(vertices[inRange[j]]);
				rangeNormals.add(normals[inRange[j]]);
			}

			//////////////////////////////////////////////////////////////////////////
			// 3. Construct and solve least-squares problem
			if (robust) {
				FitCurvatureTensorRobust(rangePts, rangeNormals, u, v, ct);
			}
			else {
				FitCurvatureTensorLS(rangePts, rangeNormals, u, v, ct);
			}

			// Store result
			cl[vi][si] = ct;

			// Check if there's an overflow
			if (!_finite(ct.kmin) || !_finite(ct.kmax)) {
				cerr << "Warning in ComputePrincipalCurvatures(): found NaN value in principal curvature for vertex: " << vi << ", scale: " << scales[si] << "." << endl;

				//// Recompute for debugging
				//cout << "% Local frame (n, u, v):" << endl;
				//cout << "n = " << normals[vi] << "';" << endl;
				//cout << "u = " << u << "';" << endl;
				//cout << "v = " << v << "';" << endl;
				//cout << "% Total " << rangePts.size() << " points:" << endl;
				//for (index_type j = 0; j < rangePts.size(); j++) {
				//	cout << "pts(:," << j+1 << ") = " << rangePts[j] << "';" << endl;
				//	cout << "normals(:," << j+1 << ") = " << rangeNormals[j] << "';" << endl;
				//}
				FitCurvatureTensorLS(rangePts, rangeNormals, u, v, ct);
			}
		}

		TextProgress(vi);
	}

	cout << "...done!" << endl;

	//////////////////////////////////////////////////////////////////////////
	// Return a description of the method used to compute the principal curvatures
	QString result;
	if (robust) {
		result += QString("Robust method");
	}
	else {
		result += QString("Least-squares method");
	}
	result += QString(" using");
	for (index_type i = 0; i < numScales; i++) {
		if (scales[i] == 0) {
			result += QString(" Scale %1: adaptive operating region of %2 * distance to %3 nearest neighbor per vertex")
				.arg(i).arg(nnFactor).arg(initialNN);
		}
		else {
			result += QString(" Scale %1: fixed operating region of radius %2").arg(i).arg(scales[i]);
		}
	}
	return result;
}


data_type ComputeMeanCurvature( const real3_type &query, PointList& allPoints, PointList& allNormals, const distList &scales )
{
	real3_type sample;
	real_type radSq, radSq99;
	data_type features(scales.size());

	for (index_type s = 0; s < scales.size(); s++) {
		// Perform a range query in all points on the original mesh
		idxList inRange;
		distList distRange;
		PointList rangePts, rangeNormals;
		sample = query;
		radSq = scales[s];

		// Compute an "effective radius" where 99% of the weight is within this radius
		// This is to save computation time in the range query
		radSq99 = scales[s] * (FitSphereWeight99Factor * FitSphereWeight99Factor);

		perfTimers["Range query"].AddStart();

		allPoints.queryRange(sample, radSq99, inRange, distRange, true);

		perfTimers["Range query"].Stop();

		// Gather the point locations and the normals
		rangePts.resize(inRange.size());
		rangeNormals.resize(inRange.size());
		for (index_type j = 0; j < inRange.size(); j++) {
			// Throw out the point if the normal is zero
			if (allNormals[inRange[j]].norm() >= REAL_TYPE_EPSILON) {
				rangePts[j] = allPoints[inRange[j]];
				rangeNormals[j] = allNormals[inRange[j]];
			}
		}

		perfTimers["Fit circle"].AddStart();

		// Fit a circle at this position
		if (FitSphere(rangePts, rangeNormals, sample, radSq)) {
			// Store the mean curvature (inverse of the radius) as the feature
			// Also scale the curvatures logarithmically (add one to avoid -Inf singularity for zero curvature)
			features.val[s] = log(1.0 / sqrt(radSq) + 1.0);
		}
		else {
			// If the fit is degenerate, then the curvature is 0.
			features.val[s] = 0;
		}

		perfTimers["Fit circle"].Stop();
	}

	return features;
}




idxList randomSample( int n, int max )
{
	idxList indices;

	if ((n <= 0) || (n >= max)) {
		// Just use all the vertices in the mesh
		indices.resize(max);
		for (index_type i = 0; i < max; i++) {
			indices[i] = i;
		}
	}
	else {
		indices.resize(n);

		// Otherwise, randomly select vertices from the mesh
		for (index_type i = 0; i < n; i++) {
			indices[i] = floor(((double)max) * genrand_real1());
		}
	}

	return indices;
}



inline void sortTwo( int &p, int &q )
{
	if (p > q) {
		int temp = q;
		q = p;
		p = temp;
	}
}

/**
 * Save the local frames at each vertex of a mesh to a binary file.  For reading in MATLAB.
 */
void saveMeshFrames( const char *filename, const ShapeModel* mesh, const CurvInfoList &cl )
{
	float temp;
	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cout << "Error in saveFrames(): could not open file " << filename << " for writing!" << endl;
		return;
	}

	if (!mesh || (cl.size() == 0)) {
		cout << "Error in saveFrames(): No mesh / curvature information found!" << endl;
		out.close();
		return;
	}

	if (mesh->vertex.size() != cl.size()) {
		cout << "Number of vertices in mesh does not match curvature info!" << endl;
		out.close();
		return;
	}

	cout << "Saving per-vertex frames...";

	//////////////////////////////////////////////////////////////////////////
	// Write the data
	unsigned int numVertices = cl.size();
	unsigned int numScales = cl[0].size();
	out.write((char*) &numVertices, sizeof(unsigned int));
	out.write((char*) &numScales, sizeof(unsigned int));

	for (index_type i = 0; i < numVertices; i++) {
		// Write the current vertex
		for (int c = 0; c < 3; c++) {
			temp = mesh->vertex[i][c];
			out.write((char*) &temp, sizeof(float));
		}

		for (index_type j = 0; j < numScales; j++) {
			CurvInfo ci = cl[i][j];

			for (int c = 0; c < 3; c++) {
				temp = ci.n[c];
				out.write((char*) &temp, sizeof(float));
			}
			for (int c = 0; c < 3; c++) {
				temp = ci.vmax[c];
				out.write((char*) &temp, sizeof(float));
			}
			for (int c = 0; c < 3; c++) {
				temp = ci.vmin[c];
				out.write((char*) &temp, sizeof(float));
			}
			temp = ci.kmax;
			out.write((char*) &temp, sizeof(float));
			temp = ci.kmin;
			out.write((char*) &temp, sizeof(float));
		}
	}

	out.close();

	cout << "done!" << endl;
}

SE3::SE3() : rot(Geometry::Matrix33f::identity()), tr(0,0,0)
{

}

SE3::SE3( Geometry::Matrix33f R, real3_type t ) : rot(R), tr(t)
{

}

void SE3::setIdentity()
{
	rot = Geometry::Matrix33f::identity();
	tr[0] = tr[1] = tr[2] = 0.0;
}

void SE3::setZero()
{
	for (int i = 0; i < dim; i++) {
		tr[i] = 0.0;
		for (int j = 0; j < dim; j++) {
			rot(i,j) = 0.0;
		}
	}
}

data_type SE3::log() const
{
	// Compute axis-angle representation
	data_type result(twistDim);
	real3_type axis;
	real_type angle;

	// Convert rotation to axis-angle representation
	toAxisAngle(rot, axis, angle);

	if (abs(angle) < REAL_TYPE_EPSILON) {
		// Assign linear velocity = translation & zero angular velocity
		for (int d = 0; d < dim; d++) {
			result.val[d] = tr[d];
			result.val[d+dim] = 0.0;
		}
	}
	else {
		// Debugging: check to see if the axis is unit length
		//real_type axisNorm = axis.norm();
		//if ((axisNorm < (1.0 - REAL_TYPE_EPSILON)) || (axisNorm > (1.0 + REAL_TYPE_EPSILON))) {
		//	cerr << "Warning in SE3::log(): Axis is not unit length! Length = " << axisNorm << "." << endl;
		//}

		// Given the translation, solve for the linear velocity.
		real3_type v = tr;
		Geometry::Matrix33f axisop, ssmAxis, temp;
		temp = -rot;
		for (int d = 0; d < dim; d++) {
			temp(d,d) += 1.0;
		}
		ssmAxis(0,0) = 0.0;      ssmAxis(0,1) = -axis[2]; ssmAxis(0,2) = axis[1];
		ssmAxis(1,0) = axis[2];  ssmAxis(1,1) = 0.0;      ssmAxis(1,2) = -axis[0];
		ssmAxis(2,0) = -axis[1]; ssmAxis(2,1) = axis[0];  ssmAxis(2,2) = 0.0;
		axisop(0,0) = axis[0]*axis[0];
		axisop(1,1) = axis[1]*axis[1];
		axisop(2,2) = axis[2]*axis[2];
		axisop(1,0) = axisop(0,1) = axis[1]*axis[0];
		axisop(2,0) = axisop(0,2) = axis[2]*axis[0];
		axisop(1,2) = axisop(2,1) = axis[1]*axis[2];
		temp = (temp * ssmAxis) + angle*axisop;

		// Solve the system
		General_Solve(temp, v);

		// Assign linear velocity & angular velocity
		for (int d = 0; d < dim; d++) {
			result.val[d] = angle*v[d];
			result.val[d+dim] = angle*axis[d];
		}
	}

	return result;
}

void SE3::applyGL() const
{
	static float m[16];
	OpenGLMatrix(m);
	glMultMatrixf(m);
}

void SE3::OpenGLMatrix( float *o ) const
{
	o[0] = rot(0,0); o[4] = rot(0,1);  o[8] = rot(0,2); o[12] = tr[0];
	o[1] = rot(1,0); o[5] = rot(1,1);  o[9] = rot(1,2); o[13] = tr[1];
	o[2] = rot(2,0); o[6] = rot(2,1); o[10] = rot(2,2); o[14] = tr[2];
	o[3] = 0.0;      o[7] = 0.0;      o[11] = 0.0;      o[15] = 1.0;
}

void SE3::loadBinary( std::ifstream &in )
{
	readBinary(in, rot);
	readBinary(in, tr);
}

void SE3::saveBinary( std::ofstream &out ) const
{
	writeBinary(out, rot);
	writeBinary(out, tr);
}

/**
 * Axis-angle conversion from matrix.  Assumes that the rotation is determinant +1.
 * Grabbed from http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToAngle/index.htm.
 */
void toAxisAngle( const Geometry::Matrix33f& R, real3_type &axis, real_type &angle )
{
	const real_type epsilon = 0.1;
	const real_type eepsilon = 0.01;
	const real_type ax = R(2,1)-R(1,2);
	const real_type ay = R(0,2)-R(2,0);
	const real_type az = R(1,0)-R(0,1);
	const real_type trace = R(0,0) + R(1,1) + R(2,2);

	// If the matrix is symmetric
	if ((abs(ax) < eepsilon) && (abs(ay) < eepsilon) && (abs(az) < eepsilon))
	{

		// If the matrix is identity
		if ((abs(R(2,1)+R(1,2)) < epsilon) &&
			(abs(R(0,2)+R(2,0)) < epsilon) &&
			(abs(R(1,0)+R(0,1)) < epsilon) &&
			(abs(trace - 3.0) < epsilon)) {
			axis[0] = 1.0;
			axis[1] = 0.0;
			axis[2] = 0.0;
			angle = 0.0;
		}
		else {
			// Otherwise, this matrix is a 180 degree rotation
			angle = M_PI;

			if (R(0,0) >= R(1,1)) {
				if (R(0,0) >= R(2,2)) {
					//00 is max diag 
					axis[0] = sqrt(0.5 * (R(0,0)+1)); 
					axis[1] = R(0,1) * 0.5 / axis[0];
					axis[2] = R(0,2) * 0.5 / axis[0];
				} else {
					//22 is max
					axis[2] = sqrt(0.5 * (R(2,2)+1));
					axis[0] = R(0,2) * 0.5 / axis[2];
					axis[1] = R(1,2) * 0.5 / axis[2];
				}
			} else {
				if (R(1,1) >= R(2,2)) {
					//11 is max diag 
					axis[1] = sqrt(0.5 * (R(1,1)+1));
					axis[0] = R(0,1) * 0.5 / axis[1];
					axis[2] = R(1,2) * 0.5 / axis[1];
				} else {
					//22 is max
					axis[2] = sqrt(0.5 * (R(2,2)+1));
					axis[0] = R(0,2) * 0.5 / axis[2];
					axis[1] = R(1,2) * 0.5 / axis[2];
				}
			}

			//for (int d = 0; d < 3; d++) {
			//	axis[d] = (R(d,d) + 1.0) / 2.0;
			//	if (axis[d] > 0)
			//		axis[d] = sqrt(axis[d]);
			//	else
			//		axis[d] = 0.0;
			//}
			//bool xZero = (abs(axis[0]) < eepsilon);
			//bool yZero = (abs(axis[1]) < eepsilon);
			//bool zZero = (abs(axis[2]) < eepsilon);
			//bool xyPositive = (R(0,1) > 0);
			//bool xzPositive = (R(0,2) > 0);
			//bool yzPositive = (R(1,2) > 0);
			//if (xZero && !yZero && !zZero) {
			//	if (!yzPositive)
			//		axis[1] = -axis[1];
			//} else if (yZero && !zZero) {
			//	if (!xzPositive)
			//		axis[2] = -axis[2];
			//} else if (zZero) {
			//	if (!xyPositive)
			//		axis[0] = -axis[0];
			//}
		}
	}
	else {
		real_type s = sqrt(ax*ax + ay*ay + az*az);

		if (abs(s) < eepsilon)
			s = 1.0;

		s = 1.0 / s;
		axis[0] = ax * s;
		axis[1] = ay * s;
		axis[2] = az * s;
		angle = acos((trace - 1.0) / 2.0);
	}
}

real_type det( Geometry::Matrix33f M )
{
	real_type plus = M(0,0)*M(1,1)*M(2,2) + M(0,1)*M(1,2)*M(2,0) + M(0,2)*M(1,0)*M(2,1);
	real_type minus = M(0,0)*M(1,2)*M(2,1) + M(0,1)*M(1,0)*M(2,2) + M(0,2)*M(1,1)*M(2,0);
	return plus - minus;
}

SE3 exp( data_type twist )
{
	// Extract linear and angular velocity.
	real_type angle;
	real3_type v, w, t;
	for (int d = 0; d < 3; d++) {
		v[d] = twist.val[d];
		w[d] = twist.val[d+3];
	}
	angle = w.norm();
	if (angle < REAL_TYPE_EPSILON) {
		// If the angle is zero, return identity matrix
		return SE3();
	}
	w = w / angle;
	v = v / angle;

	// Compute result
	Geometry::Matrix33f R, ssm, axisop;
	ssm(0,0) = 0.f;   ssm(0,1) = -w[2]; ssm(0,2) = w[1];
	ssm(1,0) = w[2];  ssm(1,1) = 0.f;   ssm(1,2) = -w[0];
	ssm(2,0) = -w[1]; ssm(2,1) = w[0];  ssm(2,2) = 0.f;
	axisop(0,0) = w[0]*w[0];
	axisop(1,1) = w[1]*w[1];
	axisop(2,2) = w[2]*w[2];
	axisop(1,0) = axisop(0,1) = w[1]*w[0];
	axisop(2,0) = axisop(0,2) = w[2]*w[0];
	axisop(1,2) = axisop(2,1) = w[1]*w[2];

	R = ssm*sin(angle) + ssm*ssm*(1.f - cos(angle));
	t = (-R*ssm + angle*axisop) * v;

	// Assign result
	for (int d = 0; d < 3; d++) {
		R(d,d) = R(d,d) + 1;
	}
	return SE3(R, t);
}

std::ostream& operator<<( std::ostream& out, const SE3 &tf )
{
	// Print rotation and translation
	out << std::endl << tf.rot << std::endl << tf.tr << std::endl;
	return out;
}


SE3 registerQuat( const PointList &left, const PointList &right )
{
	real3_type leftCenter, rightCenter;
	Geometry::Matrix33f M;
	size_type n = left.size();

	if (!left.size() || (left.size() != right.size())) {
		cerr << "Error in registerQuat(): No points, or number of points b/t sets do not match!" << endl;
		return SE3();
	}

	//////////////////////////////////////////////////////////////////////////
	// 1. Compute statistics
	for (index_type i = 0; i < n; i++) {
		leftCenter += left[i];
		rightCenter += right[i];
		for (index_type j = 0; j < 3; j++) {
			for (index_type k = 0; k < 3; k++) {
				M(j,k) += left[i][j]*right[i][k];
			}
		}
	}
	leftCenter = leftCenter / ((real_type)n);
	rightCenter = rightCenter / ((real_type)n);
	for (index_type j = 0; j < 3; j++) {
		for (index_type k = 0; k < 3; k++) {
			M(j,k) -= ((real_type)n) * leftCenter[j]*rightCenter[k];
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// 2. Compute eigenvalues
	Geometry::Matrix44f N;
	N(0,0) =   M(0,0) + M(1,1) + M(2,2);
	N(1,1) =   M(0,0) - M(1,1) - M(2,2);
	N(2,2) = - M(0,0) + M(1,1) - M(2,2);
	N(3,3) = - M(0,0) - M(1,1) + M(2,2);
	N(0,1) = N(1,0) = M(1,2) - M(2,1);
	N(0,2) = N(2,0) = M(2,0) - M(0,2);
	N(0,3) = N(3,0) = M(0,1) - M(1,0);
	N(1,2) = N(2,1) = M(0,1) + M(1,0);
	N(1,3) = N(3,1) = M(2,0) + M(0,2);
	N(2,3) = N(3,2) = M(1,2) + M(2,1);

	// Compute the eigenvector
	double *evals = new double[4];
	double **evecs = new double*[4];
	for (int i = 0; i < 4; i++) {
		evecs[i] = new double[4];
	}
	N.eigSymmetric(evals, evecs);

	// Find the index of the maximum eigenvalue
	double maxEval = -DBL_MAX;
	int maxIndex = -1;
	for (int i=0; i<4; i++) {
		if (evals[i] > maxEval) {
			maxIndex = i;
			maxEval = evals[i];
		}
		//cout << "Eigenvalue  " << i << ": " << evals[i] << "\n";
		//cout << "Eigenvector " << i << ": ("
		//	<< evecs[0][i] << ", "
		//	<< evecs[1][i] << ", "
		//	<< evecs[2][i] << ", "
		//	<< evecs[3][i] << ")\n";
	}

	//////////////////////////////////////////////////////////////////////////
	// 3. Construct rotation matrix
	SE3 tf;

	// 1. Compute the rotation according to the maximum eigenvalue's eigenvector
	real_type q0, qx, qy, qz;
	real4_type q;
	for (int i = 0; i < 4; i++) q[i] = evecs[i][maxIndex];
	q.normalize();
	q0 = q[0]; qx = q[1]; qy = q[2]; qz = q[3];

	tf.rot(0,0) = q0*q0 + qx*qx - qy*qy - qz*qz;
	tf.rot(1,1) = q0*q0 - qx*qx + qy*qy - qz*qz;
	tf.rot(2,2) = q0*q0 - qx*qx - qy*qy + qz*qz;
	tf.rot(0,1) = 2.0*(qx*qy - q0*qz);
	tf.rot(1,0) = 2.0*(qy*qx + q0*qz);
	tf.rot(0,2) = 2.0*(qx*qz + q0*qy);
	tf.rot(2,0) = 2.0*(qz*qx - q0*qy);
	tf.rot(1,2) = 2.0*(qy*qz - q0*qx);
	tf.rot(2,1) = 2.0*(qz*qy + q0*qx);

	// 2. Compute the translation as the right centroid minus the rotated left centroid
	tf.tr = rightCenter - (tf.rot * leftCenter);

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: Print matrices
	//cout << "M: " << endl << M << endl;
	//cout << "N: " << endl << N << endl;
	//cout << "(R,t): " << endl << tf << endl;

	//////////////////////////////////////////////////////////////////////////
	// Delete temporary variables
	delete[] evals;
	for (int i = 0; i < 4; i++) {
		delete[] evecs[i];
	}
	delete[] evecs;

	return tf;
}

SE3 inv( const SE3 &tf )
{
	// Take the inverse of a rigid transformation
	SE3 inverse;
	inverse.rot = tf.rot.transpose();
	inverse.tr = - inverse.rot * tf.tr;

	return inverse;
}
