///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __FIND_CORRESPONDENCE_H__
#define __FIND_CORRESPONDENCE_H__

#include "PointListN.h"
#include "Correspondence.h"
#include "ShapeModel.h"

class GCoptimizationGeneralGraph;

// Define correspondence types.
enum CorrespondenceTypes {
	CORRESPONDENCE_NONE = 0,
	CORRESPONDENCE_LOAD_FROM_FILE,
	CORRESPONDENCE_GROUND_TRUTH,
	CORRESPONDENCE_CLOSEST_POINTS,
	CORRESPONDENCE_CLOSEST_POINTS_SPIN_IMAGE_WEIGHTS,
	CORRESPONDENCE_GRAPH_CUT,
	CORRESPONDENCE_SPIN_IMAGES
};

template<int N,typename Real>
void Cholesky_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v );
template<int N,typename Real>
void General_Solve( const Geometry::Square_matrix<N,Real> &M, Geometry::Vec<N, Real> &v );

real_type det(Geometry::Matrix33f M);

/**
 * Fit a sphere given points and normals.  If the fit is degenerate, then radSq is set to the 
 * distance to the origin and center contains the normal of the plane fitting the points.
 * @param pts      A PointList containing the points to fit.
 * @param normals  A PointList containing the corresponding normals to fit.
 * @param center   This is both input and output; input: center of the weighting kernel, output: center of the fitted sphere
 * @param radSq    This is both input and output; input: radius^2 of the weighting kernel, output: radius^2 of the fitted sphere
 * @return	       This returns false if the fit was degenerate .
 */
bool FitSphere( const PointList &pts, const PointList &normals, real3_type &center, real_type &radSq );


/**
 * Compute per-vertex principal curvatures using the method of [KSN*07].
 */
bool FitCurvatureTensorLS( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result );
bool FitCurvatureTensorRobust( const PointList &pts, const PointList &normals, real3_type u, real3_type v, CurvInfo &result );
QString ComputePrincipalCurvatures( PointList &vertices, const PointList &normals, const distList &scales, CurvInfoList &cl, bool robust );

/**
 * Save the local frames at each vertex of a mesh to a binary file.  For reading in MATLAB.
 */
void saveMeshFrames( const char *filename, const ShapeModel* mesh, const CurvInfoList &cl );

/**
 * Compute the mean curvature at the query vertex given the points and the normals.
 * @param query       The query point at which to compute the curvature.
 * @param allPoints   List of all points.
 * @param allNormals  List of all normals.
 * @param scales      The list of scales (squared radius for the weight function) at which to compute the curvature.
 */
data_type ComputeMeanCurvature( const real3_type &query, PointList& allPoints, PointList& allNormals, const distList &scales );

/**
 * Returns argument x such that the sum of the weight between [0,x] constitutes 90% of the total weight.
 * This is used to scale visualizations of the weight function as a sphere.
 */
const real_type FitSphereWeight90Factor = 0.5;
const real_type FitSphereWeight99Factor = 0.708;

idxList randomSample( int n, int max );


// Implement Lie groups and algebras

/**
 * Element of the Lie group SE(3).
 */
class SE3 {
public:
	SE3();
	SE3(Geometry::Matrix33f R, real3_type t);

	void setIdentity();
	void setZero();
	data_type log() const;
	inline real3_type operator*( const real3_type &x ) const {
		return rot*x + tr;
	}

	/**
	* Construct a 4x4 OpenGL matrix corresponding to this transformation.
	*/
	void OpenGLMatrix( float *o ) const;
	void applyGL() const;

	Geometry::Matrix33f rot;
	real3_type tr;

	// Reading and writing binary data to file
	void loadBinary( std::ifstream &in );
	void saveBinary( std::ofstream &out ) const;

private:
	static const int dim = 3;
	static const int twistDim = 6;
};

typedef std::vector<SE3> RigidDeform;

SE3 inv( const SE3 &tf );
SE3 exp( data_type twist );
void toAxisAngle( const Geometry::Matrix33f& R, real3_type &axis, real_type &angle );
std::ostream& operator<<( std::ostream& out, const SE3 &tf );
SE3 registerQuat( const PointList &left, const PointList &right );

inline SE3 operator*( const SE3 &lhs, const SE3 &rhs )
{
	SE3 result;
	result.rot = lhs.rot * rhs.rot;
	result.tr = lhs.rot*rhs.tr + lhs.tr;
	return result;
}

/**
 * Create an arbitrary orthonormal local coordinate frame with normal n.
 */
void LocalCoordinateFrame( real3_type n, real3_type &u, real3_type &v );


/**
 * Constants for checking if the graph cut vertex is from the source or target.
 */
enum CLASSIFY_GRAPH_CUT_VERTEX {
	GRAPH_CUT_FROM_SOURCE = 1,
	GRAPH_CUT_FROM_TARGET
};

/**
 * Different graph cut solution types: swap or expansion?
 */
enum GRAPH_CUT_SOLUTION_TYPE {
	GRAPH_CUT_SWAP = 1,
	GRAPH_CUT_EXPANSION
};

/**
 * Type of terms in the graph cut error function
 */
enum GRAPH_CUT_ERROR_TYPE {
	GRAPH_CUT_DATA_POINT_TO_POINT = 1,
	GRAPH_CUT_DATA_POINT_TO_PLANE,
	GRAPH_CUT_DATA_HYBRID,
	GRAPH_CUT_DATA_D6
};

/**
 * Type of graph cut instances
 */
enum GRAPH_CUT_INSTANCE_TYPE {
	GRAPH_CUT_INSTANCE_SYMMETRIC = 1,
	GRAPH_CUT_INSTANCE_FORWARD,
	GRAPH_CUT_INSTANCE_BACKWARD
};

#endif

