///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "MarkedUniformGrid.h"
#include "ShapeModel.h"
#include "util.h"

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;
using namespace Geometry;

// Triangle-box intersection code from Tomas Moller
int triBoxOverlap(float boxcenter[3],float boxhalfsize[3],float triverts[3][3]);

MarkedUniformGrid3::MarkedUniformGrid3( int dataDim /*= 1*/ ) :
UniformGrid3(dataDim), totalMarkedCells(0), totalMarkedPoints(0)
{

}

void MarkedUniformGrid3::resetMark()
{
	// We will mark each cell and each grid point
	if (UniformGrid3::totalGridPoints() > 0) {
		markedPointLinearIndex.clear();
		markedPoints.clear();
		markedPoints.resize(UniformGrid3::totalGridPoints());
		for (index_type i = 0; i < markedPoints.size(); i++) {
			markedPoints[i] = false;
		}
		totalMarkedPoints = 0;
	}

	if (UniformGrid3::totalGridCells() > 0) {
		markedCellLinearIndex.clear();
		markedCells.clear();
		markedCells.resize(UniformGrid3::totalGridCells());
		for (index_type i = 0; i < markedCells.size(); i++) {
			markedCells[i] = false;
		}
		totalMarkedCells = 0;
	}
}

void MarkedUniformGrid3::drawSortedSubset( int comp, const std::vector<int> &sortedLinearIndex )
{
	glPointSize(10.f);
	UniformGrid3::draw(comp);

	if (!sortedLinearIndex.size()) {
		return;
	}

	GLboolean lighting;
	glGetBooleanv(GL_LIGHTING, &lighting);
	glDisable(GL_LIGHTING);

	// Draw a box for all the marked cells that are selected
	glLineWidth(1.f);
	glColor3f(0.f, 0.f, 0.f);

	int3_type cellIndex;
	real3_type cellMin, cellMax;
	for (index_type i = 0; i < sortedLinearIndex.size(); i++) {
		cellIndex = translateCellLinearIndex(sortedLinearIndex[i]);
		getCellBox(cellIndex, cellMin, cellMax);
		drawBoxOpenGL(cellMin, cellMax);
	}

	//int currentPos = 0;
	//int3_type cellIndex = scanCellStart();
	//real3_type cellMin, cellMax;
	//do {
	//	int cellLinearIndex = getCellLinearIndex(cellIndex);

	//	// Check if the current cell is selected
	//	if (cellLinearIndex == sortedLinearIndex[currentPos]) {
	//		getCellBox(cellIndex, cellMin, cellMax);
	//		drawBoxOpenGL(cellMin, cellMax);
	//		currentPos++;

	//		if (currentPos >= sortedLinearIndex.size()) {
	//			break;
	//		}
	//	}
	//	else if (cellLinearIndex > sortedLinearIndex[currentPos]) {
	//		cerr << "Error in MarkedUniformGrid3::drawSortedSubset(): (this case should never happen) somehow skipped drawing a selected cell!" << endl;
	//	}
	//} while(scanNextCell(cellIndex));

	// Restore previous state
	if (lighting) glEnable(GL_LIGHTING);
}

void MarkedUniformGrid3::markCells( const Mesh3_real &mesh )
{
	//real3_type query;
	//int3_type index;
	//real3_type remainder;

	//for (index_type i = 0; i < mesh.vertexSize(); i++) {
	//	query = mesh.vertex[i];
	//	queryCell(query, index, remainder);

	//	// Mark the cell containing the point
	//	markedCells[getCellLinearIndex(index)] = true;

	//	// Mark the points of this cell
	//	for (int i = 0; i < (1 << dim); i++) {
	//		int3_type offset = offset2sub(i);

	//		// Look up the grid point value and add to the total
	//		markedPoints[getLinearIndex(index + offset)] = true;
	//	}
	//}

	// Triangle-box intersection code
	int3_type cellIndex, cellIndexMin, cellIndexMax;
	real3_type cellMin, cellMax;
	real3_type triMin, triMax;
	real3_type temp;
	float boxcenter[3];
	float boxhalfsize[3];
	float triverts[3][3];

	for (int d = 0; d < 3; d++) {
		triMin[d] = REAL_TYPE_MAX;
		triMax[d] = -REAL_TYPE_MAX;
	}

	for (index_type ti = 0; ti < mesh.triangle_vertex_index.size(); ti++) {
		const index3_type &tvi = mesh.triangle_vertex_index[ti];

		for(int td = 0; td < 3; td++) {
			for (int d = 0; d < 3; d++) {
				triverts[td][d] = mesh.vertex[tvi[td]][d];

				// Compute the bounding box of the triangle
				if (triMin[d] > triverts[td][d]) {
					triMin[d] = triverts[td][d];
				}
				if (triMax[d] < triverts[td][d]) {
					triMax[d] = triverts[td][d];
				}
			}
		}

		// Find the cells containing the bounding box of the triangle
		queryCell(triMin, cellIndexMin, temp);
		queryCell(triMax, cellIndexMax, temp);

		// Iterate over cells in the bounding box of this triangle
		for (index_type x = cellIndexMin[0]; x <= cellIndexMax[0]; x++) {
			for (index_type y = cellIndexMin[1]; y <= cellIndexMax[1]; y++) {
				for (index_type z = cellIndexMin[2]; z <= cellIndexMax[2]; z++) {
					cellIndex[0] = x; cellIndex[1] = y; cellIndex[2] = z;

					if (!isMarked(cellIndex)) {
						getCellBox(cellIndex, cellMin, cellMax);
						for (int d = 0; d < 3; d++) {
							boxcenter[d] = (cellMax[d] + cellMin[d]) / 2.f;
							boxhalfsize[d] = (cellMax[d] - cellMin[d]) / 2.f;
						}

						if (triBoxOverlap(boxcenter, boxhalfsize, triverts)) {
							markCell(cellIndex);
						}
					}
				}
			}
		}

		//// For each grid cell, test if the triangle intersects that box
		//cellIndex = scanCellStart();
		//do {
		//	getCellBox(cellIndex, cellMin, cellMax);
		//	for (int d = 0; d < 3; d++) {
		//		boxcenter[d] = (cellMax[d] + cellMin[d]) / 2.f;
		//		boxhalfsize[d] = (cellMax[d] - cellMin[d]) / 2.f;
		//	}

		//	if (triBoxOverlap(boxcenter, boxhalfsize, triverts)) {
		//		markedCells[getCellLinearIndex(cellIndex)] = true;

		//		// Mark the points of this cell
		//		for (int i = 0; i < (1 << dim); i++) {
		//			int3_type offset = offset2sub(i);

		//			// Look up the grid point value and add to the total
		//			markedPoints[getLinearIndex(cellIndex + offset)] = true;
		//		}
		//	}
		//} while(scanNextCell(cellIndex));
	}

	countMarked();
}

void MarkedUniformGrid3::markCells( const PointList &points )
{
	// Mark cells containing each point
	cout << "*** Marking grid cells using points... ";
	for (int i = 0; i < points.size(); i++) {
		markCell(points[i]);
	}
	cout << " done!" << endl;

	countMarked();
}

void MarkedUniformGrid3::markCells( const std::vector<PointList*> &points, const std::vector< std::vector<bool> > &valid )
{
	if (points.size() != valid.size()) {
		cerr << "Error in MarkedUniformGrid3::markCells(): argument sizes don't match!" << endl;
		return;
	}


	cout << "*** Marking grid cells using array of points and valid flags... ";
	for (int i = 0; i < points.size(); ++i) {
		if (points[i]->size() != valid[i].size()) {
			cerr << "Error in MarkedUniformGrid3::markCells(): argument sizes in " << i << " don't match!" << endl;
			return;
		}

		// Mark cells containing each point
		for (int j = 0; j < points[i]->size(); ++j) {
			if (valid[i][j]) {
				markCell((*points[i])[j]);
			}
		}
	}
	cout << " done!" << endl;

	countMarked();
}

int3_type MarkedUniformGrid3::scanStart() const
{
	int3_type gridIndex = UniformGrid3::scanStart();

	// Scan the points in the usual order until we encounter a marked point
	while (!markedPoints[UniformGrid3::getLinearIndex(gridIndex)]) {
		if (!UniformGrid3::scanNext(gridIndex)) {
			cerr << "Warning in MarkedUniformGrid3::scanStart(): could not find a marked grid point!";
			return int3_type(0,0,0);
		}
	}

	return gridIndex;
}

bool MarkedUniformGrid3::scanNext( int3_type &gridIndex ) const
{
	// Scan the next grid point until we encounter a marked point
	do {
		if (!UniformGrid3::scanNext(gridIndex)) {
			return false;
		}
	} while (!markedPoints[UniformGrid3::getLinearIndex(gridIndex)]);

	return true;
}

int3_type MarkedUniformGrid3::scanCellStart() const
{
	int3_type cellIndex = UniformGrid3::scanCellStart();

	// Scan the cells in the usual order until we encounter a marked point
	while (!isMarked(cellIndex)) {
		if (!UniformGrid3::scanNextCell(cellIndex)) {
			cerr << "Warning in MarkedUniformGrid3::scanCellStart(): could not find a marked cell!";
			return int3_type(0,0,0);
		}
	}

	return cellIndex;
}

bool MarkedUniformGrid3::scanNextCell( int3_type &cellIndex ) const
{
	// Scan the next cell until we encounter a marked point
	do {
		if (!UniformGrid3::scanNextCell(cellIndex)) {
			return false;
		}
	} while (!isMarked(cellIndex));

	return true;
}

void MarkedUniformGrid3::countMarked()
{
	totalMarkedCells = totalMarkedPoints = 0;
	markedCellLinearIndex.clear();
	markedPointLinearIndex.clear();
	markedCellLinearIndex.resize(UniformGrid3::totalGridCells());
	markedPointLinearIndex.resize(UniformGrid3::totalGridPoints());

	index_type lind;
	int3_type gridIndex = UniformGrid3::scanStart();
	do {
		lind = UniformGrid3::getLinearIndex(gridIndex);

		if (markedPoints[lind]) {
			markedPointLinearIndex[lind] = totalMarkedPoints++;
		}
		else {
			markedPointLinearIndex[lind] = -1;
		}

	} while(UniformGrid3::scanNext(gridIndex));

	int3_type cellIndex = UniformGrid3::scanCellStart();
	do {
		lind = UniformGrid3::getCellLinearIndex(cellIndex);

		if (markedCells[lind]) {
			markedCellLinearIndex[lind] = totalMarkedCells++;
		}
		else {
			markedCellLinearIndex[lind] = -1;
		}

	} while(UniformGrid3::scanNextCell(cellIndex));

	cout << "Total marked points: " << totalMarkedPoints << " out of " << UniformGrid3::totalGridPoints() << endl;
	cout << "Total marked cells: " << totalMarkedCells << " out of " << UniformGrid3::totalGridCells() << endl;
}

index_type MarkedUniformGrid3::totalGridPoints() const
{
	return totalMarkedPoints;
}

index_type MarkedUniformGrid3::totalGridCells() const
{
	return totalMarkedCells;
}

index_type MarkedUniformGrid3::getLinearIndex( const int3_type &gridIndex ) const
{
	// Check bounds
	if (!validPoint(gridIndex)) {
		cerr << "Error in MarkedUniformGrid3::getLinearIndex(): invalid point " << gridIndex << " requested!" << endl;
	}
	return markedPointLinearIndex[UniformGrid3::getLinearIndex(gridIndex)];
}

index_type MarkedUniformGrid3::getCellLinearIndex( const int3_type &cellIndex ) const
{
	// Check bounds
	if (!validCell(cellIndex)) {
		cerr << "Error in MarkedUniformGrid3::getCellLinearIndex(): invalid cell " << cellIndex << " requested!" << endl;
	}
	return markedCellLinearIndex[UniformGrid3::getCellLinearIndex(cellIndex)];
}


bool MarkedUniformGrid3::isMarked( real3_type query ) const
{
	int3_type cellIndex;
	real3_type remainder;
	queryCell(query, cellIndex, remainder);

	return isMarked(cellIndex);
}

bool MarkedUniformGrid3::isMarked( int3_type cellIndex ) const
{
	return markedCells[UniformGrid3::getCellLinearIndex(cellIndex)];
}

void MarkedUniformGrid3::markCell( real3_type query )
{
	int3_type cellIndex;
	real3_type remainder;
	queryCell(query, cellIndex, remainder);

	markCell(cellIndex);
}

void MarkedUniformGrid3::markCell( int3_type cellIndex )
{
	// First, check if the cell is indeed within bounds
	if (!inCellBounds(cellIndex)) {
		cerr << "Error in MarkedUniformGrid3::markCell(): trying to mark a cell out of bounds!" << endl;
		return;
	}

	// Mark the cell
	markedCells[UniformGrid3::getCellLinearIndex(cellIndex)] = true;

	// Mark the points of this cell
	for (int i = 0; i < (1 << dim); i++) {
		int3_type offset = offset2sub(i);

		// Look up the grid point value and add to the total
		markedPoints[UniformGrid3::getLinearIndex(cellIndex + offset)] = true;
	}
}

void MarkedUniformGrid3::findEdges()
{
	linearIndexToIndex.clear();
	allEdges.clear();

	// Iterate over all grid points
	index_type linearIndex, linearIndexN;
	int3_type gridIndex = scanStart();
	do {
		linearIndex = getLinearIndex(gridIndex);
		linearIndexToIndex[linearIndex] = gridIndex;

		if (linearIndex == 4294967295) {
			cerr << "Error: negative index detected!" << endl;
		}

		//////////////////////////////////////////////////////////////////////////
		////// Commented out because: This detects edges that are not part of a marked cell
		////int3List neighbors = findNeighbors(gridIndex);
		////for (unsigned int i = 0; i < neighbors.size(); i++) {
		////	linearIndexN = getLinearIndex(neighbors[i]);
		////	if (linearIndexN == 4294967295) {
		////		cerr << "Error: negative neighbor index detected !" << endl;
		////	}

		////	// Sort the pair
		////	int i1, i2;
		////	if (linearIndex <= linearIndexN) {
		////		i1 = linearIndex;
		////		i2 = linearIndexN;
		////	}
		////	else {
		////		i1 = linearIndexN;
		////		i2 = linearIndex;
		////	}

		////	std::vector<int> temp;
		////	temp.push_back(i2);
		////	sortedListUnionMerge(temp, allEdges[i1]);
		////}

	} while(scanNext(gridIndex));


	//////////////////////////////////////////////////////////////////////////
	// Iterate over all grid cells

	// Add all edges of each grid cell (12 edges in 3D)
	static const int gridEdges[12][2] = {{0,1},{0,2},{0,4},{1,3},{1,5},{2,3},{2,6},{3,7},{4,5},{4,6},{5,7},{6,7}};

	cellLinearIndexToIndex.clear();
	int3_type gridCellIndex = scanCellStart();
	do {
		cellLinearIndexToIndex[getCellLinearIndex(gridCellIndex)] = gridCellIndex;

		idxList cellPts;
		getCellGridPoints(gridCellIndex, cellPts);
		for (unsigned int i = 0; i < 12; i++) {
			linearIndex = cellPts[gridEdges[i][0]];
			linearIndexN = cellPts[gridEdges[i][1]];

			// Sort the pair
			int i1, i2;
			if (linearIndex <= linearIndexN) {
				i1 = linearIndex;
				i2 = linearIndexN;
			}
			else {
				i1 = linearIndexN;
				i2 = linearIndex;
			}

			std::vector<int> temp;
			temp.push_back(i2);
			sortedListUnionMerge(temp, allEdges[i1]);
		}
	} while (scanNextCell(gridCellIndex));
}

void MarkedUniformGrid3::drawEdges( bool deformed /*= true */ )
{
	// Draw all the edges
	stdext::hash_map<int,std::vector<int> >::const_iterator it;
	int3_type idx1, idx2;
	real3_type pos1, pos2;

	glBegin(GL_LINES);
	for (it = allEdges.begin(); it != allEdges.end(); it++) {
		int i1 = it->first;
		const std::vector<int> &neighbors = it->second;

		for (int j = 0; j < neighbors.size(); j++) {
			int i2 = neighbors[j];

			// Find the full indices
			idx1 = translateLinearIndex(i1);
			idx2 = translateLinearIndex(i2);

			// Plot the positions
			if (deformed) {
				pos1 = deformedGridLocation(idx1);
				pos2 = deformedGridLocation(idx2);
			}
			else {
				pos1 = gridLocation(idx1);
				pos2 = gridLocation(idx2);
			}

			glVertex3f(pos1[0],pos1[1],pos1[2]);
			glVertex3f(pos2[0],pos2[1],pos2[2]);
		}
	}
	glEnd();
}


void MarkedUniformGrid3::drawFaces( int comp /*= -1*/, bool deformed /*= true */, bool showWireframe /*= false */ )
{
	// Find the projected position under each face
	static const int hexFaces[6][4] = {{0,1,2,3},{4,6,5,7},{0,4,1,5},{2,3,6,7},{0,2,4,6},{1,5,3,7}};
	static const int drawQuadOrder[4] = {0, 1, 3, 2};
	real3_type temp[4];

	if (showWireframe) {
		glDisable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else {
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset (1., 1.);
	}

	glBegin(GL_QUADS);

	// Iterate through each cell and scan cell faces to see if they're on the boundary
	int3_type gridCellIndex = scanCellStart();
	do {
		// Neighbor in each dimension (4-connected neighbors)
		int faceIndex = 0;
		for (int d = 0; d < dim; d++) {
			// Top neighbor & Bottom neighbor
			for (int ni = -1; ni <= 1; ni +=2) {
				int3_type neighborIndex = gridCellIndex;
				neighborIndex[d] += ni;

				if (!validCell(neighborIndex)) {
					// This is a boundary face, draw it
					for (int j = 0; j < 4; j++) {
						int3_type pointIndex = gridCellIndex + offset2sub(hexFaces[faceIndex][drawQuadOrder[j]]);
						drawGridPoint(pointIndex, comp, deformed);
					}
				}
				faceIndex++;
			}
		}
	} while(scanNextCell(gridCellIndex));

	glEnd();

	if (showWireframe) {
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}
	else {
		glDisable(GL_POLYGON_OFFSET_FILL);
	}
}

int3_type MarkedUniformGrid3::translateLinearIndex( const index_type &pointIndex ) const
{
	// Look up the index in the hash table
	stdext::hash_map<int,int3_type>::const_iterator it;
	it = linearIndexToIndex.find(pointIndex);

	if (it == linearIndexToIndex.end()) {
		cerr << "Error in MarkedUniformGrid3::translateLinearIndex(): no mapping to grid point index!" << endl;
		return int3_type();
	}
	else {
		return it->second;
	}
}

int3_type MarkedUniformGrid3::translateCellLinearIndex( const index_type &cellIndex ) const
{
	// Look up the index in the hash table
	stdext::hash_map<int,int3_type>::const_iterator it;
	it = cellLinearIndexToIndex.find(cellIndex);

	if (it == cellLinearIndexToIndex.end()) {
		cerr << "Error in MarkedUniformGrid3::translateCellLinearIndex(): no mapping to grid cell index!" << endl;
		return int3_type();
	}
	else {
		return it->second;
	}
}

void MarkedUniformGrid3::computeIntegrals( const int3_type cellIndex, std::vector< std::vector<double> > &result ) const
{
	static real3_type gridMin, gridMax;
	static double xmin, ymin, zmin, xmax, ymax, zmax, vol;
	static double t1, t2, t15, t37, t53, t54, t55, t58, t71, t93, t109, t112, t125, t147, t163, t166, t169, t174, t179, t244, t249, t278, t300, t320, t331, t344, t362;
	static double t20, t25, t26, t31, t34, t50, t78, t81, t82, t85, t88, t106, t128, t131, t134, t137, t140, t160, t184, t189, t192, t197, t200, t202, t204, t206, t207, t209, t211, t213, t215, t239, t258, t260, t262, t264, t266, t272, t275, t281, t286, t288, t292, t315, t334, t341, t355, t383, t397, t411, t432;
	static double V000, V001, V010, V011, V100, V101, V110, V111;

	// Find position of the grid point
	getCellBox(cellIndex, gridMin, gridMax);
	xmin = gridMin[0]; ymin = gridMin[1]; zmin = gridMin[2];
	xmax = gridMax[0]; ymax = gridMax[1]; zmax = gridMax[2];
	vol = (xmax-xmin)*(ymax-ymin)*(zmax-zmin);

	// Initialize result
	result.clear();
	result.resize(dataDim);

	// For each bone
	for (int bi = 0; bi < dataDim; bi++) {
		// Query the values of the grid
		V000 = getValue(cellIndex + int3_type(0,0,0), bi);
		V001 = getValue(cellIndex + int3_type(0,0,1), bi);
		V010 = getValue(cellIndex + int3_type(0,1,0), bi);
		V011 = getValue(cellIndex + int3_type(0,1,1), bi);
		V100 = getValue(cellIndex + int3_type(1,0,0), bi);
		V101 = getValue(cellIndex + int3_type(1,0,1), bi);
		V110 = getValue(cellIndex + int3_type(1,1,0), bi);
		V111 = getValue(cellIndex + int3_type(1,1,1), bi);

		// Compute the integral
		t1 = xmin-xmax;
		t2 = xmin*xmin;
		t15 = xmin*xmax;
		t37 = xmax*xmax;
		t53 = zmin-zmax;
		t54 = ymin-ymax;
		t55 = t53*t54;
		t58 = ymin*ymin;
		t71 = ymin*ymax;
		t93 = ymax*ymax;
		t109 = t53*t1;
		t112 = zmin*zmin;
		t125 = zmin*zmax;
		t147 = zmax*zmax;
		t163 = t54*t1;
		t166 = xmin*ymax;
		t169 = ymin*xmax;
		t174 = xmin*ymin;
		t179 = ymax*xmax;
		t244 = ymin*zmax;
		t249 = zmin*ymax;
		t278 = ymax*zmax;
		t300 = ymin*zmin;
		t320 = zmin*xmax;
		t331 = xmin*zmin;
		t344 = zmax*xmax;
		t362 = xmin*zmax;

		t20 = xmin*V010;
		t25 = 3.0*t2*V011+t2*V111+t2*V101+t2*V110+3.0*t2*V001+t2*V100+3.0*t2*V000+3.0*t2*V010+2.0*t15*V110+2.0*t15*V101+2.0*t20*xmax+2.0*t15*V100;
		t26 = xmin*V001;
		t31 = xmin*V000;
		t34 = xmin*V011;
		t50 = 2.0*t26*xmax+2.0*t15*V111+2.0*t31*xmax+2.0*t34*xmax+V010*t37+V011*t37+3.0*t37*V111+3.0*t37*V101+V000*t37+V001*t37+3.0*t37*V100+3.0*t37*V110;
		t78 = ymin*V000;
		t81 = t58*V110+3.0*t58*V000+3.0*t58*V100+3.0*t58*V101+t58*V111+t58*V011+3.0*t58*V001+t58*V010+2.0*t71*V110+2.0*t71*V010+2.0*t71*V111+2.0*t78*ymax;
		t82 = ymin*V001;
		t85 = ymin*V100;
		t88 = ymin*V101;
		t106 = 2.0*t82*ymax+2.0*t85*ymax+2.0*t88*ymax+2.0*t71*V011+3.0*t93*V010+t93*V000+3.0*t93*V110+3.0*t93*V011+t93*V100+t93*V101+t93*V001+3.0*t93*V111;
		t128 = zmin*V000;
		t131 = zmin*V010;
		t134 = zmin*V100;
		t137 = t112*V111+t112*V011+3.0*t112*V100+t112*V101+3.0*t112*V110+3.0*t112*V000+3.0*t112*V010+t112*V001+2.0*t125*V111+2.0*t128*zmax+2.0*t131*zmax+2.0*t134*zmax;
		t140 = zmin*V110;
		t160 = 2.0*t125*V001+2.0*t140*zmax+2.0*t125*V011+2.0*t125*V101+t147*V100+t147*V010+t147*V110+3.0*t147*V011+3.0*t147*V001+3.0*t147*V101+3.0*t147*V111+t147*V000;
		t184 = ymax*V011;
		t189 = ymax*V010;
		t192 = V100*xmax;
		t197 = V101*xmax;
		t200 = V111*xmin;
		t202 = V011*xmax;
		t204 = V010*xmax;
		t206 = 4.0*t166*V010+4.0*t169*V101+2.0*t82*xmax+2.0*t174*V101+4.0*t169*V100+4.0*t179*V110+2.0*t169*V111+2.0*t184*xmax+2.0*t78*xmax+2.0*t189*xmax+2.0*t192*ymax+4.0*t179*V111+2.0*t197*ymax+t200*ymin+t202*ymin+t204*ymin;
		t207 = V100*xmin;
		t209 = V110*xmin;
		t211 = V000*xmax;
		t213 = V001*xmax;
		t215 = V101*xmin;
		t239 = t207*ymax+t209*ymin+t211*ymax+t213*ymax+t215*ymax+2.0*t26*ymax+2.0*t169*V110+2.0*t20*ymin+2.0*t34*ymin+2.0*t31*ymax+2.0*t166*V110+4.0*t166*V011+2.0*t174*V100+4.0*t174*V001+4.0*t174*V000+2.0*t166*V111;
		t258 = V000*ymax;
		t260 = V110*ymin;
		t262 = V001*ymax;
		t264 = V010*ymin;
		t266 = V100*ymax;
		t272 = zmax*V101;
		t275 = zmax*V001;
		t281 = 2.0*t244*V011+4.0*t244*V101+2.0*t249*V011+4.0*t249*V110+2.0*t128*ymax+4.0*t249*V010+t258*zmax+t260*zmax+t262*zmin+t264*zmax+t266*zmax+2.0*t78*zmax+4.0*t244*V001+2.0*t272*ymax+2.0*t275*ymax+2.0*t278*V010;
		t286 = V011*ymin;
		t288 = V101*ymax;
		t292 = V111*ymin;
		t315 = 2.0*t278*V110+2.0*t85*zmax+t286*zmin+t288*zmin+4.0*t278*V011+t292*zmin+2.0*t249*V111+4.0*t278*V111+2.0*t88*zmin+2.0*t300*V010+2.0*t134*ymax+4.0*t300*V100+4.0*t300*V000+2.0*t300*V110+2.0*t82*zmin+2.0*t244*V111;
		t334 = V110*xmax;
		t341 = zmax*V011;
		t355 = 2.0*t320*V111+4.0*t320*V110+2.0*t26*zmin+2.0*t128*xmax+t200*zmin+t204*zmax+4.0*t331*V000+2.0*t334*zmax+2.0*t275*xmax+2.0*t331*V110+2.0*t341*xmax+4.0*t344*V111+4.0*t344*V101+2.0*t192*zmax+2.0*t34*zmin+2.0*t320*V101;
		t383 = 2.0*t20*zmax+2.0*t331*V100+4.0*t331*V010+4.0*t362*V011+4.0*t362*V001+t202*zmin+t207*zmax+t215*zmin+t211*zmax+2.0*t31*zmax+t213*zmin+4.0*t320*V100+2.0*t131*xmax+2.0*t362*V111+2.0*t362*V101+t209*zmax;
		t397 = 2.0*t26+t215+t207+2.0*t34+t200+t209+2.0*t20+2.0*t31+t202+t211+t204+2.0*t192+2.0*t197+2.0*V111*xmax+t213+2.0*t334;
		t411 = t264+2.0*t82+2.0*t78+t292+2.0*t88+2.0*t85+t286+t260+2.0*t184+t266+t288+2.0*t189+2.0*ymax*V110+t258+t262+2.0*ymax*V111;
		t432 = 2.0*t128+2.0*t140+V001*zmin+2.0*t131+V111*zmin+V011*zmin+2.0*t134+V101*zmin+V010*zmax+2.0*t272+2.0*zmax*V111+V000*zmax+V110*zmax+V100*zmax+2.0*t341+2.0*t275;

		result[bi].resize(10);
		result[bi][0] = -1.0*t1*(t25+t50)*t55/48.0;
		result[bi][1] = -1.0*t54*(t81+t106)*t109/48.0;
		result[bi][2] = -1.0*t53*(t137+t160)*t163/48.0;
		result[bi][3] = -1.0*t163*(t206+t239)*t53/72.0;
		result[bi][4] = -1.0*t55*(t281+t315)*t1/72.0;
		result[bi][5] = -1.0*t109*(t355+t383)*t54/72.0;
		result[bi][6] = -1.0*t1*t397*t55/24.0;
		result[bi][7] = -1.0*t54*t411*t109/24.0;
		result[bi][8] = -1.0*t53*t432*t163/24.0;
		result[bi][9] = -1.0*(V100+V001+V110+V011+V111+V000+V101+V010)*t53*t163/8.0;
	}
}


int MarkedUniformGrid3::computeOverlapIntegrals( const int3_type cellIndex, const RigidDeform &def, std::vector< std::pair<index2_type, std::vector<double> > > &result, bool justCount /* = false*/ ) const
{
	static real3_type gridMin, gridMax;
	static double xmin, ymin, zmin, xmax, ymax, zmax, vol;
	static double V000_i, V001_i, V010_i, V011_i, V100_i, V101_i, V110_i, V111_i;
	static double V000_j, V001_j, V010_j, V011_j, V100_j, V101_j, V110_j, V111_j;
	static double Ri00, Ri01, Ri02, Ri10, Ri11, Ri12, Ri20, Ri21, Ri22, tix, tiy, tiz;
	static double Rj00, Rj01, Rj02, Rj10, Rj11, Rj12, Rj20, Rj21, Rj22, tjx, tjy, tjz;

	// Initialize result
	result.clear();

	if (!validCell(cellIndex)) {
		cerr << "Warning in MarkedUniformGrid3::computeOverlapIntegrals(): cellIndex " << cellIndex << " is not valid!" << endl;
		return 0;
	}

	// Find position of the grid point
	getCellBox(cellIndex, gridMin, gridMax);
	xmin = gridMin[0]; ymin = gridMin[1]; zmin = gridMin[2];
	xmax = gridMax[0]; ymax = gridMax[1]; zmax = gridMax[2];
	vol = (xmax-xmin)*(ymax-ymin)*(zmax-zmin);

	// Find all bones that affect this cell
	std::vector<bool> affectingBones;
	affectingBones.resize(dataDim, false);
	idxList cellGridPoints;
	getCellGridPoints(cellIndex, cellGridPoints);
	for (int i = 0; i < cellGridPoints.size(); i++) {
		data_type val = data[cellGridPoints[i]];
		for (int j = 0; j < dataDim; j++) {
			if (val.val[j] > 0.0) {
				affectingBones[j] = true;
			}
		}
		//cout << "** value: " << val << endl;
	}
	std::vector<int> affectingBoneList;
	for (int i = 0; i < dataDim; i++) {
		if (affectingBones[i]) {
			affectingBoneList.push_back(i);
		}
	}
	if (affectingBoneList.size() <= 1) {
		// Only one bone affects this cell!
		return 0;
	}

	// Construct a list of bone pairs
	std::vector<index2_type> bonePairs;
	for (int i = 0; i < affectingBoneList.size()-1; i++) {
		for (int j = i + 1; j < affectingBoneList.size(); j++) {
			bonePairs.push_back(index2_type(affectingBoneList[i],affectingBoneList[j]));
		}
	}
	//cout << "List of bones affecting this node: ";
	//for (int i = 0; i < affectingBoneList.size(); i++) {
	//	cout << affectingBoneList[i] << ",";
	//}
	//cout << endl;
	//cout << "List of bone pairs affecting this node: ";
	//for (int i = 0; i < bonePairs.size(); i++) {
	//	cout << bonePairs[i] << ",";
	//}
	//cout << endl;

	if (justCount) {
		// Return the bone pairs, but don't calculate the integrals
		std::vector<double> integrals;
		for (int i = 0; i < bonePairs.size(); i++) {
			result.push_back(make_pair(bonePairs[i], integrals));
		}

		return bonePairs.size();
	}

	// For each pair
	for (int i = 0; i < bonePairs.size(); i++) {
		// Which pair?
		int bi = bonePairs[i][0];
		int bj = bonePairs[i][1];

		// Find the rigid deformation
		Ri00 = def[bi].rot(0,0);   Rj00 = def[bj].rot(0,0);
		Ri01 = def[bi].rot(0,1);   Rj01 = def[bj].rot(0,1);
		Ri02 = def[bi].rot(0,2);   Rj02 = def[bj].rot(0,2);
		Ri10 = def[bi].rot(1,0);   Rj10 = def[bj].rot(1,0);
		Ri11 = def[bi].rot(1,1);   Rj11 = def[bj].rot(1,1);
		Ri12 = def[bi].rot(1,2);   Rj12 = def[bj].rot(1,2);
		Ri20 = def[bi].rot(2,0);   Rj20 = def[bj].rot(2,0);
		Ri21 = def[bi].rot(2,1);   Rj21 = def[bj].rot(2,1);
		Ri22 = def[bi].rot(2,2);   Rj22 = def[bj].rot(2,2);
		tix = def[bi].tr[0];	   tjx = def[bj].tr[0];
		tiy = def[bi].tr[1];	   tjy = def[bj].tr[1];
		tiz = def[bi].tr[2];	   tjz = def[bj].tr[2];
		
		// Query the values of the grid
		V000_i = getValue(cellIndex + int3_type(0,0,0), bi);  V000_j = getValue(cellIndex + int3_type(0,0,0), bj);
		V001_i = getValue(cellIndex + int3_type(0,0,1), bi);  V001_j = getValue(cellIndex + int3_type(0,0,1), bj);
		V010_i = getValue(cellIndex + int3_type(0,1,0), bi);  V010_j = getValue(cellIndex + int3_type(0,1,0), bj);
		V011_i = getValue(cellIndex + int3_type(0,1,1), bi);  V011_j = getValue(cellIndex + int3_type(0,1,1), bj);
		V100_i = getValue(cellIndex + int3_type(1,0,0), bi);  V100_j = getValue(cellIndex + int3_type(1,0,0), bj);
		V101_i = getValue(cellIndex + int3_type(1,0,1), bi);  V101_j = getValue(cellIndex + int3_type(1,0,1), bj);
		V110_i = getValue(cellIndex + int3_type(1,1,0), bi);  V110_j = getValue(cellIndex + int3_type(1,1,0), bj);
		V111_i = getValue(cellIndex + int3_type(1,1,1), bi);  V111_j = getValue(cellIndex + int3_type(1,1,1), bj);
		
		std::vector<double> integrals;
		integrals.resize(27, 0.0);

		// Compute normalization factor
		real_type tau_ij;
		{
			double t1,t2,t21,t22,t23,t24,t25,t26,t27,t28,t3,t4;
			t28 = 2.0*V011_j+4.0*V100_j;
			t27 = 4.0*V001_j+2.0*V110_j;
			t26 = 4.0*V110_j+2.0*V001_j;
			t25 = 2.0*V101_j+4.0*V010_j;
			t24 = 4.0*V101_j+2.0*V010_j;
			t23 = 2.0*V000_j+4.0*V111_j;
			t22 = 4.0*V000_j+2.0*V111_j;
			t21 = 2.0*V100_j+4.0*V011_j;
			t4 = t21+t22;
			t3 = t23+t27;
			t2 = t25+t28;
			t1 = t24+t26;
			tau_ij = (ymin-ymax)*(zmax-zmin)*(xmin-xmax)*((V110_j+8.0*V001_j+t4+t24)*V001_i+(V001_j+8.0*V110_j+t2+t23)*V110_i+(V011_j+8.0*V100_j+t1+t22)*V100_i+(V111_j+8.0*V000_j+t2+t27)*V000_i+(V010_j+8.0*V101_j+t3+t28)*V101_i+(V101_j+8.0*V010_j+t4+t26)*V010_i+(8.0*V111_j+V000_j+t1+t21)*V111_i+(8.0*V011_j+V100_j+t3+t25)*V011_i)/216.0;
			tau_ij = 1 / tau_ij;
		}

		// Compute single term integrals
		{
			double t10,t100,t101,t102,t104,t105,t106,t107,t108,t109,t11,t110,t111,t114,t115,t116,t117,t118,t119,t12,t120,t121,t122,t123,t124,t125,t126,t127,t13,t14,t15,t16,t17,t18,t19,t20,t21,t22,t23,t24,t25,t26,t27,t28,t29,t30,t31,t32,t33,t34,t35,t36,t37,t38,t39,t4,t40,t41,t42,t43,t44,t45,t46,t47,t48,t49,t5,t50,t51,t52,t53,t54,t55,t56,t57,t58,t59,t6,t60,t61,t62,t63,t64,t65,t66,t67,t68,t69,t7,t70,t71,t73,t74,t75,t76,t77,t78,t8,t80,t81,t83,t84,t85,t86,t87,t88,t89,t9,t90,t91,t92,t93,t94,t95,t96,t97,t99;
			t80 = 4.0*V010_j;
			t127 = 8.0*V101_j+t80;
			t126 = 3.0*V101_j+t80;
			t95 = 4.0*V101_j;
			t125 = t95+3.0*V010_j;
			t107 = 4.0*V000_j;
			t124 = 3.0*V111_j+t107;
			t123 = t107+8.0*V111_j;
			t122 = V101_j+t80;
			t121 = V101_j+12.0*V010_j;
			t120 = V001_j+12.0*V110_j;
			t102 = 4.0*V110_j;
			t119 = V001_j+t102;
			t106 = 4.0*V011_j;
			t118 = V100_j+t106;
			t117 = 3.0*V001_j;
			t116 = 6.0*V101_j;
			t115 = 6.0*V000_j;
			t114 = 3.0*V110_j;
			t111 = 2.0*V101_j;
			t110 = 3.0*V100_j;
			t109 = 6.0*V111_j;
			t108 = 3.0*V011_j;
			t105 = 8.0*V100_j;
			t104 = 8.0*V001_j;
			t101 = 2.0*V111_j;
			t100 = 2.0*V010_j;
			t99 = 6.0*V011_j;
			t97 = 2.0*V001_j;
			t96 = 8.0*V010_j;
			t94 = 2.0*V110_j;
			t93 = 3.0*V000_j;
			t92 = 2.0*V011_j;
			t91 = 6.0*V100_j;
			t90 = 4.0*V111_j;
			t89 = 6.0*V010_j;
			t88 = 2.0*V000_j;
			t87 = 4.0*V100_j;
			t86 = 8.0*V011_j;
			t85 = 2.0*V100_j;
			t84 = 4.0*V001_j;
			t83 = 8.0*V000_j;
			t81 = 8.0*V110_j;
			t78 = 12.0*V000_j;
			t77 = 12.0*V011_j;
			t76 = 12.0*V100_j;
			t75 = 12.0*V001_j;
			t74 = 12.0*V111_j;
			t73 = 12.0*V101_j;
			t71 = t109+t88;
			t70 = t115+t101;
			t69 = V100_j+t88;
			t68 = t92+t91;
			t67 = t116+t100;
			t66 = V011_j+t100;
			t65 = t111+t89;
			t64 = t88+t87;
			t63 = t94+6.0*V001_j;
			t62 = t102+t101;
			t61 = t92+t90;
			t60 = t111+t84;
			t59 = 6.0*V110_j+t97;
			t58 = V111_j+t111;
			t57 = V000_j+t100;
			t56 = t99+t85;
			t55 = t92+t84;
			t54 = t101+t88;
			t53 = t106+t100;
			t52 = t97+t94;
			t51 = t92+t85;
			t50 = t95+t85;
			t49 = V010_j+t88;
			t48 = V110_j+t85;
			t47 = (xmin-xmax)*(zmax-zmin)*(ymin-ymax)/432.0;
			t46 = t101+t95+t49;
			t45 = t93+t90+t63;
			t44 = t110+t106+t59;
			t43 = t117+t115+t62;
			t42 = t84+t81+t127;
			t41 = t52+t122;
			t40 = t106+t105+t96+t95;
			t39 = t59+t124;
			t38 = t48+t55;
			t37 = t104+t102+t123;
			t36 = t90+t87+t86+t83;
			t35 = t51+t119;
			t34 = t108+t87+t63;
			t33 = t63+t126;
			t32 = V011_j+t87+t52;
			t31 = t111+t90+t57;
			t30 = t59+t125;
			t29 = t107+t100+t58;
			t28 = V110_j+t101+t53;
			t27 = t78+t89+t58;
			t26 = t116+t76+t66;
			t25 = V000_j+t74+t56;
			t24 = t65+t120;
			t23 = t99+t75+t48;
			t22 = t77+t109+t69;
			t21 = V010_j+t73+t68;
			t20 = t56+t121;
			t19 = t38+t46;
			t18 = t29+t38;
			t17 = t111+t87+t54+t66+t119;
			t16 = V111_j+t107+t41+t51;
			t15 = t29+t32;
			t14 = V010_j+V000_j+t50+t52+t61;
			t13 = t31+t35;
			t12 = t35+t54+t122;
			t11 = t41+t54+t118;
			t10 = t32+t46;
			t9 = t31+t52+t118;
			t8 = t28+t60+t69;
			t7 = (t83+16.0*V100_j+t42+t61)*V100_i+(t105+16.0*V101_j+t37+t53)*V101_i+(16.0*V110_j+t97+t40+t123)*V110_i+(16.0*V000_j+t104+t40+t62)*V000_i+(16.0*V001_j+t94+t36+t127)*V001_i+(16.0*V111_j+t86+t42+t64)*V111_i+(t96+16.0*V011_j+t37+t50)*V011_i+(t81+16.0*V010_j+t36+t60)*V010_i;
			t6 = (t17*V100_i+t14*V101_i+(t109+t108+t24+t64)*V110_i+t16*V000_i+t8*V001_i+(t25+t30)*V111_i+(t89+t114+t22+t60)*V011_i+(t20+t39)*V010_i)*ymax+((t26+t43)*V100_i+(t21+t45)*V101_i+t17*V110_i+(V111_j+t78+t33+t68)*V000_i+(t116+t110+t115+t75+t28)*V001_i+t14*V111_i+t8*V011_i+t16*V010_i)*ymin;
			t5 = ((V011_j+t76+t30+t70)*V100_i+t10*V101_i+(t93+t91+t24+t61)*V110_i+(t114+t91+t27+t55)*V000_i+t18*V001_i+t13*V111_i+t11*V011_i+(t44+t70+t121)*V010_i)*zmin+(t10*V100_i+(t109+t73+t34+t49)*V101_i+t13*V110_i+t18*V000_i+(t23+t67+t124)*V001_i+(t117+t102+t25+t67)*V111_i+(t22+t33)*V011_i+t11*V010_i)*zmax;
			t4 = ((t26+t39)*V100_i+(t84+t114+t21+t71)*V101_i+(t68+t71+t120+t126)*V110_i+t15*V000_i+t19*V001_i+(t116+t74+t44+t57)*V111_i+t9*V011_i+t12*V010_i)*xmax+(t15*V100_i+t19*V101_i+t12*V110_i+(t27+t34)*V000_i+(t23+t70+t125)*V001_i+t9*V111_i+(t77+V100_j+t45+t65)*V011_i+(t20+t43)*V010_i)*xmin;
			integrals[0] = (t7*tix+t5*Ri02+t4*Ri00+t6*Ri01)*t47;
			integrals[1] = (t7*tiy+t4*Ri10+t6*Ri11+t5*Ri12)*t47;
			integrals[2] = (t7*tiz+t6*Ri21+t5*Ri22+t4*Ri20)*t47;
			integrals[3] = (t7*tjx+t6*Rj01+t5*Rj02+t4*Rj00)*t47;
			integrals[4] = (t7*tjy+t5*Rj12+t4*Rj10+t6*Rj11)*t47;
			integrals[5] = (t7*tjz+t5*Rj22+t4*Rj20+t6*Rj21)*t47;
		}

		// Compute product term integrals
		{
			double t100,t101,t102,t103,t104,t105,t106,t107,t108,t109,t11,t110,t111,t112,t113,t114,t115,t116,t117,t118,t119,t12,t120,t121,t122,t123,t124,t125,t126,t127,t128,t129,t130,t131,t132,t133,t134,t135,t136,t137,t138,t139,t140,t141,t142,t143,t144,t145,t146,t147,t148,t149,t150,t151,t152,t153,t154,t155,t156,t157,t158,t159,t160,t161,t162,t163,t164,t165,t166,t167,t168,t169,t170,t171,t172,t173,t174,t175,t176,t177,t178,t179,t180,t181,t182,t183,t184,t185,t186,t187,t188,t189,t19,t190,t191,t192,t193,t194,t195,t196;
			double t197,t198,t199,t20,t200,t201,t202,t203,t204,t205,t206,t207,t208,t209,t21,t210,t211,t212,t213,t214,t215,t216,t217,t218,t219,t22,t220,t221,t222,t223,t224,t225,t226,t227,t228,t229,t230,t231,t232,t233,t234,t235,t236,t237,t238,t239,t240,t241,t242,t243,t244,t245,t246,t247,t248,t249,t250,t251,t252,t253,t254,t255,t256,t257,t258,t259,t260,t261,t262,t263,t264,t265,t266,t267,t268,t269,t270,t271,t272,t273,t274,t275,t276,t277,t278,t279,t280,t281,t282,t283,t284,t285,t286,t287,t288,t289,t290,t291,t292,t293;
			double t294,t295,t296,t297,t298,t299,t300,t301,t302,t303,t304,t305,t306,t307,t308,t309,t310,t311,t312,t313,t314,t315,t316,t317,t318,t319,t320,t321,t322,t323,t324,t325,t326,t327,t328,t329,t330,t331,t332,t333,t334,t335,t336,t337,t338,t339,t34,t340,t341,t342,t343,t344,t345,t346,t347,t348,t349,t35,t350,t351,t352,t353,t354,t355,t356,t357,t358,t359,t36,t360,t361,t362,t363,t364,t365,t366,t367,t368,t369,t37,t370,t371,t372,t373,t374,t375,t376,t377,t378,t379,t38,t380,t381,t382,t383,t384,t385,t386,t387,t388;
			double t389,t39,t390,t391,t392,t393,t394,t395,t396,t397,t398,t399,t40,t400,t401,t402,t403,t404,t405,t406,t407,t408,t409,t41,t410,t411,t412,t413,t414,t415,t416,t417,t418,t419,t42,t420,t421,t422,t423,t424,t425,t426,t427,t428,t429,t43,t430,t431,t432,t433,t434,t435,t436,t437,t438,t439,t44,t440,t441,t442,t443,t444,t445,t446,t447,t448,t449,t45,t450,t451,t452,t453,t454,t455,t456,t457,t458,t459,t46,t460,t461,t462,t463,t464,t465,t466,t467,t468,t469,t47,t470,t471,t472,t473,t474,t475,t476,t477,t478;
			double t479,t48,t480,t481,t483,t487,t49,t490,t491,t492,t493,t494,t495,t496,t497,t498,t50,t502,t51,t52,t522,t523,t524,t525,t526,t527,t528,t529,t53,t530,t531,t532,t533,t534,t535,t536,t537,t538,t539,t54,t540,t541,t542,t543,t544,t545,t546,t547,t548,t549,t55,t550,t551,t552,t553,t554,t555,t556,t557,t558,t559,t56,t560,t561,t562,t563,t564,t565,t566,t567,t568,t569,t57,t570,t571,t572,t573,t574,t575,t576,t577,t578,t579,t58,t580,t581,t582,t583,t584,t585,t586,t587,t588,t589,t59,t590,t591,t592,t593,t594,t595;
			double t596,t597,t60,t606,t607,t608,t609,t61,t610,t611,t612,t613,t614,t615,t616,t617,t618,t619,t62,t620,t621,t622,t623,t624,t625,t626,t627,t628,t629,t63,t630,t631,t632,t633,t634,t635,t636,t637,t638,t639,t64,t640,t641,t642,t643,t644,t645,t646,t647,t648,t649,t65,t650,t651,t652,t653,t654,t655,t656,t657,t658,t659,t66,t660,t661,t662,t663,t664,t665,t666,t667,t668,t669,t67,t670,t671,t672,t673,t674,t675,t676,t677,t678,t679,t68,t680,t681,t682,t683,t684,t685,t686,t687,t688,t69,t70,t71,t72,t73,t74,t75,t76,t77,t78,t79,t80,t81,t82,t83,t84,t85,t86,t87,t88,t89,t90,t91,t92,t93,t94,t95,t96,t97,t98,t99;
			t566 = 10.0*V010_i;
			t688 = t566+45.0*V101_i;
			t622 = 6.0*V110_i;
			t687 = 96.0*V001_i+t622;
			t579 = 40.0*V101_i;
			t562 = 30.0*V010_i;
			t686 = t579+t562;
			t525 = 16.0*V110_i;
			t494 = 32.0*V010_i;
			t685 = t525+t494;
			t563 = 10.0*V011_i;
			t684 = t563+120.0*V100_i;
			t568 = 10.0*V101_i;
			t683 = t568+120.0*V010_i;
			t682 = 90.0*V110_i+5.0*V001_i;
			t524 = 16.0*V100_i;
			t493 = 32.0*V110_i;
			t681 = t524+t493;
			t560 = 30.0*V000_i;
			t597 = 40.0*V111_i;
			t680 = t560+t597;
			t679 = 5.0*V110_i+90.0*V001_i;
			t590 = 10.0*V001_i;
			t678 = t590+45.0*V110_i;
			t575 = 10.0*V000_i;
			t677 = t575+120.0*V111_i;
			t676 = t575+45.0*V111_i;
			t621 = 6.0*V011_i;
			t675 = t621+t524;
			t674 = 90.0*V000_i+5.0*V111_i;
			t595 = 40.0*V010_i;
			t591 = 80.0*V110_i;
			t673 = t595+t591;
			t672 = 5.0*V100_i+90.0*V011_i;
			t576 = 40.0*V110_i;
			t564 = 80.0*V100_i;
			t671 = t576+t564;
			t584 = 40.0*V100_i;
			t565 = 80.0*V101_i;
			t670 = t584+t565;
			t669 = 5.0*V000_i+90.0*V111_i;
			t557 = 30.0*V101_i;
			t668 = t595+t557;
			t585 = 10.0*V111_i;
			t667 = t585+120.0*V000_i;
			t625 = 6.0*V111_i;
			t666 = 96.0*V000_i+t625;
			t665 = t621+96.0*V100_i;
			t637 = 6.0*V000_i;
			t664 = t637+96.0*V111_i;
			t663 = t585+45.0*V000_i;
			t662 = 5.0*V010_i+90.0*V101_i;
			t578 = 80.0*V111_i;
			t661 = t579+t578;
			t660 = 5.0*V011_i+90.0*V100_i;
			t659 = 5.0*V101_i+90.0*V010_i;
			t586 = 10.0*V110_i;
			t658 = t586+45.0*V001_i;
			t498 = 96.0*V101_i;
			t624 = 6.0*V010_i;
			t657 = t498+t624;
			t553 = 30.0*V011_i;
			t656 = t584+t553;
			t655 = 45.0*V100_i+t563;
			t596 = 10.0*V100_i;
			t654 = 120.0*V011_i+t596;
			t581 = 40.0*V011_i;
			t589 = 80.0*V001_i;
			t653 = t581+t589;
			t652 = 45.0*V011_i+t596;
			t651 = t568+45.0*V010_i;
			t497 = 32.0*V100_i;
			t650 = t525+t497;
			t649 = 120.0*V001_i+t586;
			t632 = 6.0*V101_i;
			t648 = t632+96.0*V010_i;
			t559 = 30.0*V001_i;
			t647 = t576+t559;
			t642 = 6.0*V001_i;
			t646 = 96.0*V110_i+t642;
			t645 = 3.0*V000_i;
			t644 = 2.0*V010_i;
			t643 = 4.0*V011_i;
			t641 = 4.0*V110_i;
			t640 = 6.0*V100_i;
			t639 = 8.0*V111_i;
			t638 = 2.0*V110_i;
			t636 = 4.0*V111_i;
			t635 = 3.0*V100_i;
			t634 = 4.0*V010_i;
			t633 = 8.0*V011_i;
			t631 = 3.0*V010_i;
			t630 = 3.0*V011_i;
			t629 = 8.0*V110_i;
			t628 = 3.0*V001_i;
			t627 = 2.0*V011_i;
			t626 = 3.0*V110_i;
			t623 = 3.0*V111_i;
			t620 = 4.0*V101_i;
			t619 = 8.0*V001_i;
			t618 = 4.0*V000_i;
			t617 = 2.0*V100_i;
			t616 = 4.0*V001_i;
			t615 = 4.0*V100_i;
			t614 = 8.0*V000_i;
			t613 = 2.0*V101_i;
			t612 = 3.0*V101_i;
			t611 = 2.0*V111_i;
			t610 = 8.0*V010_i;
			t609 = 8.0*V101_i;
			t608 = 2.0*V000_i;
			t607 = 8.0*V100_i;
			t606 = 2.0*V001_i;
			t594 = 80.0*V011_i;
			t593 = 20.0*V110_i;
			t592 = 20.0*V101_i;
			t588 = 40.0*V000_i;
			t587 = 20.0*V111_i;
			t583 = 80.0*V000_i;
			t582 = 20.0*V011_i;
			t580 = 20.0*V001_i;
			t577 = 20.0*V010_i;
			t574 = 20.0*V100_i;
			t573 = 40.0*V001_i;
			t572 = 20.0*V000_i;
			t571 = 60.0*V100_i;
			t570 = 30.0*V110_i;
			t569 = 60.0*V010_i;
			t567 = 80.0*V010_i;
			t561 = 60.0*V110_i;
			t558 = 60.0*V111_i;
			t556 = 30.0*V100_i;
			t555 = 12.0*V101_i;
			t554 = 12.0*V001_i;
			t552 = 30.0*V111_i;
			t551 = 12.0*V000_i;
			t550 = 48.0*V101_i;
			t549 = 12.0*V111_i;
			t548 = 12.0*V110_i;
			t547 = 48.0*V100_i;
			t546 = 24.0*V000_i;
			t545 = 12.0*V100_i;
			t544 = 12.0*V011_i;
			t543 = 48.0*V001_i;
			t542 = 24.0*V101_i;
			t541 = 12.0*V010_i;
			t540 = 24.0*V001_i;
			t539 = 24.0*V100_i;
			t538 = 48.0*V000_i;
			t537 = 24.0*V011_i;
			t536 = 48.0*V111_i;
			t535 = 24.0*V110_i;
			t534 = 24.0*V111_i;
			t533 = 24.0*V010_i;
			t532 = 48.0*V011_i;
			t531 = 48.0*V010_i;
			t530 = 48.0*V110_i;
			t529 = 16.0*V010_i;
			t528 = 16.0*V101_i;
			t527 = 16.0*V001_i;
			t526 = 16.0*V111_i;
			t523 = 16.0*V011_i;
			t522 = 16.0*V000_i;
			t502 = 96.0*V011_i;
			t496 = 32.0*V101_i;
			t495 = 32.0*V011_i;
			t492 = 32.0*V001_i;
			t491 = 32.0*V000_i;
			t490 = 32.0*V111_i;
			t487 = 120.0*V110_i;
			t483 = 120.0*V101_i;
			t481 = (zmax-zmin)*(xmin-xmax)*(ymin-ymax);
			t480 = t641+t607;
			t479 = t634+t633;
			t478 = t626+t606;
			t477 = t639+t620;
			t476 = t643+t606;
			t475 = V101_i+t624;
			t474 = t628+t621;
			t473 = t630+t624;
			t472 = t631+t621;
			t471 = t645+t624;
			t470 = t634+t608;
			t469 = t634+t629;
			t468 = t620+t611;
			t467 = t630+t617;
			t466 = t635+t627;
			t465 = t618+t606;
			t464 = t632+t628;
			t463 = V101_i+t611;
			t462 = V100_i+V110_i;
			t461 = t632+t623;
			t460 = V111_i+t637;
			t459 = V010_i+V110_i;
			t458 = t644+t618;
			t457 = t640+t626;
			t456 = V000_i+t644;
			t455 = t638+t615;
			t454 = t645+t642;
			t453 = t618+t610;
			t452 = t631+t622;
			t451 = t637+t631;
			t450 = V000_i+t625;
			t449 = t616+t614;
			t448 = t633+t616;
			t447 = t644+t612;
			t446 = t636+t609;
			t445 = t642+t639;
			t444 = V001_i+t622;
			t443 = t615+t609;
			t442 = t626+t625;
			t441 = V011_i+V001_i;
			t440 = t634+t614;
			t439 = V011_i+t608;
			t438 = t623+t621;
			t437 = t620+t607;
			t436 = V001_i+t644;
			t435 = V010_i+t606;
			t434 = t642+t630;
			t433 = t645+t640;
			t432 = t629+t615;
			t431 = t642+t612;
			t430 = t638+t628;
			t429 = t635+t632;
			t428 = V111_i+V110_i;
			t427 = t627+t611;
			t426 = V100_i+t638;
			t425 = V010_i+V000_i;
			t424 = t641+t617;
			t423 = t643+t610;
			t422 = t631+t613;
			t421 = t608+t606;
			t420 = t627+t606;
			t419 = t613+t611;
			t418 = t637+t635;
			t417 = V001_i+t627;
			t416 = t636+t613;
			t415 = V010_i+t638;
			t414 = t610+t522;
			t413 = t639+t523;
			t412 = t633+t527;
			t411 = t548+t625;
			t410 = t619+t523;
			t409 = t637+t541;
			t408 = t639+t528;
			t407 = t545+t622;
			t406 = t632+t533;
			t405 = t607+t522;
			t404 = t619+t528;
			t403 = t555+t640;
			t402 = t551+t624;
			t401 = t632+t545;
			t400 = t632+t549;
			t399 = t549+t622;
			t398 = t637+t545;
			t397 = t544+t624;
			t396 = t640+t548;
			t395 = t541+t622;
			t394 = t614+t527;
			t393 = t643+t551;
			t392 = t554+t621;
			t391 = t555+t610;
			t390 = t609+t526;
			t389 = t633+t529;
			t388 = t614+t529;
			t387 = t541+t609;
			t386 = t642+t551;
			t385 = t559+15.0*V110_i;
			t384 = t562+15.0*V101_i;
			t383 = t534+t530;
			t382 = t555+t534;
			t381 = t542+t533;
			t380 = t573+t572;
			t379 = t588+t577;
			t378 = t555+t539;
			t377 = t550+t534;
			t376 = t555+t540;
			t375 = t576+t574;
			t374 = t550+t541;
			t373 = t551+t533;
			t372 = t555+t531;
			t371 = t572+t558;
			t370 = t570+15.0*V001_i;
			t369 = t553+15.0*V100_i;
			t368 = t548+t539;
			t367 = t577+60.0*V101_i;
			t366 = t537+t531;
			t365 = t582+t571;
			t364 = t541+t537;
			t363 = t546+t543;
			t362 = t549+t535;
			t361 = t545+t535;
			t360 = t560+15.0*V111_i;
			t359 = t533+t528;
			t358 = t545+t532;
			t357 = t548+t543;
			t356 = t547+t535;
			t355 = t539+t537;
			t354 = t597+t593;
			t353 = t536+t535;
			t352 = t580+t561;
			t351 = t554+t542;
			t350 = t556+15.0*V011_i;
			t349 = t546+t541;
			t348 = t549+t542;
			t347 = t537+t524;
			t346 = t540+t535;
			t345 = t557+15.0*V010_i;
			t344 = t534+t532;
			t343 = t544+t533;
			t342 = t546+t534;
			t341 = t587+t579;
			t340 = t546+t531;
			t339 = t592+t569;
			t338 = t547+t544;
			t337 = t551+t549;
			t336 = t542+t529;
			t335 = t544+t540;
			t334 = t593+60.0*V001_i;
			t333 = t554+t530;
			t332 = t552+15.0*V000_i;
			t331 = t592+t584;
			t330 = t597+t582;
			t329 = t587+60.0*V000_i;
			t328 = t551+t525;
			t327 = t581+t580;
			t326 = t554+t546;
			t325 = t551+t536;
			t324 = t588+t580;
			t323 = t539+t523;
			t322 = t574+60.0*V011_i;
			t321 = t549+t538;
			t320 = t540+t538;
			t319 = t547+t542;
			t318 = t481/2160.0;
			t317 = t481/4320.0;
			t316 = t645+t641+t445;
			t315 = t642+t637+t428;
			t314 = t440+t461;
			t313 = t444+t450;
			t312 = t416+t456;
			t311 = t446+t451;
			t310 = t430+t460;
			t309 = t463+t470;
			t308 = t422+t467;
			t307 = t644+t627+t428;
			t306 = t426+t463;
			t305 = t419+t462;
			t304 = V100_i+t608+t438;
			t303 = t640+t623+t439;
			t302 = t619+t618+t442;
			t301 = t419+t441;
			t300 = t466+t475;
			t299 = t447+t466;
			t298 = t617+t608+t459;
			t297 = V111_i+V011_i+t613+t606;
			t296 = t644+t608+t419;
			t295 = t632+t626+t435;
			t294 = t638+t617+t420;
			t293 = t415+t464;
			t292 = V111_i+t617+t439;
			t291 = t426+t476;
			t290 = t467+t475;
			t289 = t417+t424;
			t288 = V100_i+t621+t450;
			t287 = t411+t473;
			t286 = t644+t643+t403;
			t285 = t625+t544+t431;
			t284 = t554+t632+t418;
			t283 = t620+t617+t397;
			t282 = t627+t616+t407;
			t281 = t616+t608+t399;
			t280 = t400+t474;
			t279 = t396+t471;
			t278 = t395+t438;
			t277 = t625+t612+t396;
			t276 = t641+t611+t386;
			t275 = t409+t416;
			t274 = t392+t455;
			t273 = t555+t625+t434;
			t272 = t398+t464;
			t271 = t396+t410;
			t270 = t335+t480;
			t269 = t351+t469;
			t268 = t363+t442;
			t267 = t387+t675;
			t266 = t643+t619+t368;
			t265 = t353+t454;
			t264 = t636+t629+t326;
			t263 = t389+t401;
			t262 = t554+t537+t432;
			t261 = t343+t437;
			t260 = t407+t412;
			t259 = t637+t526+t387;
			t258 = t625+t522+t391;
			t257 = t362+t449;
			t256 = t402+t408;
			t255 = t544+t618+t534+t607;
			t254 = t642+t525+t391;
			t253 = t345+t660;
			t252 = t364+t403;
			t251 = t545+t542+t397;
			t250 = t338+t646;
			t249 = t384+t682;
			t248 = t640+t551+t376;
			t247 = t345+t669;
			t246 = t543+t537+t402;
			t245 = t629+t526+t363;
			t244 = t353+t394;
			t243 = t368+t392;
			t242 = t640+t502+t357;
			t241 = t540+t622+t337;
			t240 = t360+t679;
			t239 = t352+t686;
			t238 = t582+t566+t380;
			t237 = t343+t383;
			t236 = t321+t374;
			t235 = t326+t366;
			t234 = t370+t688;
			t233 = t340+t361;
			t232 = t332+t651;
			t231 = t319+t326;
			t230 = t539+t530+t373;
			t229 = t321+t358;
			t228 = t350+t658;
			t227 = t344+t376;
			t226 = t334+t668;
			t225 = t338+t357;
			t224 = t549+t491+t347;
			t223 = t367+t647;
			t222 = t595+t594+t354;
			t221 = t350+t385;
			t220 = t320+t378;
			t219 = t581+t556+t367;
			t218 = t333+t358;
			t217 = t590+t572+t331;
			t216 = t369+t663;
			t215 = t541+t496+t347;
			t214 = t537+t536+t351;
			t213 = t360+t678;
			t212 = t588+t552+t367;
			t211 = t341+t379;
			t210 = t370+t652;
			t209 = t320+t343;
			t208 = t332+t655;
			t207 = t334+t339;
			t206 = t540+t490+t328;
			t205 = t321+t333;
			t204 = t340+t382;
			t203 = t573+t565+t673;
			t202 = t548+t533+t344;
			t201 = t333+t374;
			t200 = t325+t381;
			t199 = t355+t372;
			t198 = t335+t377;
			t197 = t375+t661;
			t196 = t358+t374;
			t195 = t384+t385;
			t194 = t319+t362;
			t193 = t573+t570+t339;
			t192 = t335+t356;
			t191 = t653+t671;
			t190 = t545+t495+t336;
			t189 = t581+t567+t324;
			t188 = t378+t383;
			t187 = t583+t573+t331;
			t186 = t349+t356;
			t185 = t548+t492+t323;
			t184 = t322+t677;
			t183 = t322+t683;
			t182 = t365+t667;
			t181 = t329+t649;
			t180 = t566+t483+t334;
			t179 = t590+t487+t365;
			t178 = t371+t654;
			t177 = t329+t684;
			t176 = t610+t641+t248+t413;
			t175 = t264+t267;
			t174 = t619+t522+t261+t411;
			t173 = t271+t382+t453;
			t172 = t259+t262;
			t171 = t260+t348+t440;
			t170 = t639+t539+t254+t393;
			t169 = t260+t349+t446;
			t168 = t256+t270;
			t167 = t257+t263;
			t166 = t364+t394+t399+t443;
			t165 = t261+t328+t445;
			t164 = t271+t373+t477;
			t163 = t263+t264;
			t162 = t258+t270;
			t161 = t254+t255;
			t160 = t258+t266;
			t159 = t259+t361+t448;
			t158 = t624+t548+t255+t404;
			t157 = t262+t388+t400;
			t156 = t257+t267;
			t155 = t614+t549+t269+t675;
			t154 = t526+t633+t269+t398;
			t153 = t256+t266;
			t152 = 5.0*t305+5.0*t454+5.0*t472;
			t151 = 5.0*t299+5.0*t313;
			t150 = 5.0*t295+5.0*t303;
			t149 = 5.0*t301+5.0*t433+5.0*t452;
			t148 = 5.0*V101_i+5.0*t638+5.0*t292+5.0*t436;
			t147 = 5.0*t611+5.0*V011_i+5.0*t293+5.0*t433;
			t146 = 5.0*t305+5.0*t420+5.0*t425;
			t145 = 5.0*t295+5.0*t450+5.0*t466;
			t144 = 5.0*t608+5.0*t623+5.0*t290+5.0*t444;
			t143 = 5.0*t427+5.0*t429+5.0*t454+5.0*t459;
			t142 = 5.0*t645+5.0*t611+5.0*t300+5.0*t444;
			t141 = 5.0*t622+5.0*t612+5.0*t303+5.0*t436;
			t140 = 5.0*t304+5.0*t475+5.0*t478;
			t139 = 5.0*t300+5.0*t460+5.0*t478;
			t138 = 5.0*t613+5.0*V110_i+5.0*t292+5.0*t435;
			t137 = 5.0*t299+5.0*t315;
			t136 = 5.0*t293+5.0*t450+5.0*t467;
			t135 = 5.0*t290+5.0*t310;
			t134 = 5.0*t297+5.0*t451+5.0*t457;
			t133 = 5.0*t288+5.0*t422+5.0*t478;
			t132 = 5.0*t308+5.0*t313;
			t131 = 5.0*t306+5.0*t417+5.0*t456;
			t130 = 5.0*V011_i+5.0*t640+5.0*t310+5.0*t447;
			t129 = 5.0*t308+5.0*t315;
			t128 = 5.0*V000_i+5.0*t617+5.0*t301+5.0*t415;
			t127 = 5.0*t298+5.0*t434+5.0*t461;
			t126 = 5.0*t288+5.0*t430+5.0*t447;
			t125 = 5.0*V100_i+5.0*V101_i+5.0*t307+5.0*t421;
			t124 = 5.0*t306+5.0*t471+5.0*t474;
			t123 = 5.0*V110_i+5.0*t644+5.0*t304+5.0*t431;
			t122 = 10.0*V101_i+10.0*t634+10.0*t618+10.0*V111_i+10.0*t294;
			t121 = 10.0*t291+10.0*t309;
			t120 = 10.0*t297+10.0*t455+10.0*t458;
			t119 = 10.0*t289+10.0*t309;
			t118 = 10.0*t615+10.0*V011_i+10.0*t415+10.0*t421+10.0*t468;
			t117 = 10.0*t643+10.0*t616+10.0*t296+10.0*t462;
			t116 = 10.0*t636+10.0*t620+10.0*t294+10.0*t425;
			t115 = 10.0*t291+10.0*t312;
			t114 = 10.0*t613+10.0*t617+10.0*t616+10.0*t618+10.0*t307;
			t113 = 10.0*t641+10.0*t615+10.0*t296+10.0*t441;
			t112 = 10.0*t616+10.0*t620+10.0*t298+10.0*t427;
			t111 = 10.0*t289+10.0*t312;
			t110 = (t592+t596+t222+t380)*V011_j+(t575+t577+t197+t327)*V111_j+(t595+t582+t593+t585+t187)*V000_j+(t586+t574+t211+t653)*V001_j+(t217+t330+t673)*V110_j+(t563+t580+t211+t671)*V100_j+(t238+t354+t670)*V101_j+(t587+t568+t189+t375)*V010_j;
			t109 = xmax*((t532+t607+t206+t359)*V011_j+(t190+t244)*V111_j+(t209+t408+t650)*V000_j+(t215+t245)*V001_j+(t522+t494+t188+t410)*V110_j+(t491+t527+t194+t389)*V100_j+(t185+t377+t414)*V101_j+(t235+t390+t681)*V010_j);
			t108 = zmin*((t227+t405+t685)*V011_j+(t214+t388+t681)*V111_j+(t538+t639+t185+t359)*V000_j+(t543+t629+t224+t336)*V001_j+(t523+t490+t230+t404)*V110_j+(t496+t526+t186+t412)*V100_j+(t198+t414+t650)*V101_j+(t527+t495+t233+t390)*V010_j);
			t107 = ymax*((t528+t492+t202+t405)*V011_j+(t215+t244)*V111_j+(t220+t413+t685)*V000_j+(t190+t245)*V001_j+(t522+t497+t237+t404)*V110_j+(t526+t493+t231+t389)*V100_j+(t550+t610+t206+t323)*V101_j+(t609+t527+t535+t531+t224)*V010_j);
			t106 = (160.0*V011_i+t567+t588+t589+t197)*V011_j+(160.0*V111_i+t594+t572+t584+t203)*V111_j+(160.0*V000_i+t567+t191+t341)*V000_j+(t583+160.0*V001_i+t222+t670)*V001_j+(160.0*V110_i+t564+t189+t661)*V110_j+(t583+160.0*V100_i+t203+t330)*V100_j+(t578+160.0*V101_i+t191+t379)*V101_j+(160.0*V010_i+t591+t594+t597+t187)*V010_j;
			t105 = V011_j*t152+V111_j*t146+(t228+t384+t674)*V000_j+(t216+t345+t679)*V001_j+V110_j*t148+V100_j*t130+V101_j*t143+V010_j*t135;
			t104 = V011_j*t133+(t247+t369+t678)*V111_j+V000_j*t148+V001_j*t146+(t249+t350+t676)*V110_j+V100_j*t141+V101_j*t145+V010_j*t144;
			t103 = (t332+t384+t658+t672)*V011_j+V111_j*t126+V000_j*t129+(t240+t345+t652)*V001_j+V110_j*t131+V100_j*t138+V101_j*t127+V010_j*t124;
			t102 = V011_j*t124+V111_j*t131+(t221+t651+t674)*V000_j+V001_j*t129+V110_j*t142+V100_j*t134+V101_j*t138+(t216+t370+t659)*V010_j;
			t101 = V011_j*t140+V111_j*t132+V000_j*t139+V001_j*t125+(t232+t350+t682)*V110_j+V100_j*t149+V101_j*t128+(t213+t369+t659)*V010_j;
			t100 = (t195+t672+t676)*V011_j+(t210+t247)*V111_j+V000_j*t125+V001_j*t123+V110_j*t132+V100_j*t128+V101_j*t136+V010_j*t140;
			t99 = V011_j*t146+V111_j*t145+V000_j*t130+V001_j*t143+V110_j*t141+(t234+t360+t660)*V100_j+(t208+t385+t662)*V101_j+V010_j*t148;
			t98 = V011_j*t125+V111_j*t128+(t195+t655+t674)*V000_j+V001_j*t137+V110_j*t149+(t253+t370+t663)*V100_j+V101_j*t147+V010_j*t139;
			t97 = V011_j*t123+V111_j*t136+V000_j*t137+(t240+t369+t688)*V001_j+V110_j*t128+V100_j*t147+(t228+t332+t662)*V101_j+V010_j*t125;
			t96 = (t232+t385+t672)*V011_j+V111_j*t133+V000_j*t135+V001_j*t152+V110_j*t144+V100_j*t148+V101_j*t146+(t210+t360+t659)*V010_j;
			t95 = V011_j*t131+V111_j*t151+V000_j*t134+V001_j*t138+(t208+t249)*V110_j+(t213+t253)*V100_j+V101_j*t150+V010_j*t142;
			t94 = V011_j*t126+(t234+t369+t669)*V111_j+V000_j*t138+V001_j*t127+V110_j*t151+V100_j*t150+(t221+t662+t676)*V101_j+V010_j*t131;
			t93 = (t178+t193)*V011_j+(t184+t239)*V111_j+V000_j*t122+V001_j*t117+(t558+t569+t487+t553+t217)*V110_j+V100_j*t113+V101_j*t116+(t552+t561+t183+t324)*V010_j;
			t92 = V011_j*t117+V111_j*t116+(t182+t226)*V000_j+(t181+t219)*V001_j+V110_j*t113+(t177+t223)*V100_j+(t560+t571+t180+t330)*V101_j+V010_j*t122;
			t91 = (t178+t226)*V011_j+(t184+t223)*V111_j+V000_j*t114+(t212+t322+t649)*V001_j+V110_j*t111+V100_j*t118+(t180+t371+t656)*V101_j+V010_j*t121;
			t90 = V011_j*t121+V111_j*t111+(t182+t193)*V000_j+V001_j*t114+(t179+t339+t680)*V110_j+(t177+t239)*V100_j+V101_j*t118+(t556+t561+t327+t329+t683)*V010_j;
			t89 = V011_j*t115+(t219+t352+t677)*V111_j+V000_j*t120+V001_j*t112+(t179+t371+t668)*V110_j+(t212+t352+t684)*V100_j+(t483+t571+t570+t558+t238)*V101_j+V010_j*t119;
			t88 = (t207+t654+t680)*V011_j+V111_j*t115+(t207+t656+t667)*V000_j+(t181+t322+t686)*V001_j+V110_j*t119+V100_j*t120+V101_j*t112+(t183+t329+t647)*V010_j;
			t20 = ymax*ymax;
			t11 = ymin*ymin;
			t87 = (t176*V011_j+t154*V111_j+(t225+t381+t666)*V000_j+(t236+t355+t687)*V001_j+t170*V110_j+(t236+t346+t665)*V100_j+(t225+t342+t657)*V101_j+t163*V010_j)*t11+((t502+t531+t248+t353)*V011_j+(t218+t381+t664)*V111_j+t174*V000_j+t166*V001_j+(t199+t325+t646)*V110_j+t156*V100_j+t158*V101_j+(t218+t342+t648)*V010_j)*t20;
			t21 = zmax*zmax;
			t22 = zmin*zmin;
			t86 = (t164*V011_j+t159*V111_j+(t192+t372+t666)*V000_j+t169*V001_j+(t204+t250)*V110_j+(t205+t381+t665)*V100_j+t153*V101_j+(t205+t355+t648)*V010_j)*t22+((t200+t242)*V011_j+(t196+t346+t664)*V111_j+t162*V000_j+(t196+t342+t687)*V001_j+t161*V110_j+t155*V100_j+(t536+t498+t246+t368)*V101_j+t157*V010_j)*t21;
			t19 = xmax*xmax;
			t12 = xmin*xmin;
			t85 = ((t204+t242)*V011_j+t172*V111_j+(t199+t357+t666)*V000_j+(t229+t381+t687)*V001_j+t165*V110_j+t175*V100_j+t168*V101_j+(t229+t346+t648)*V010_j)*t12+(t173*V011_j+(t201+t355+t664)*V111_j+t160*V000_j+t171*V001_j+(t200+t250)*V110_j+(t201+t342+t665)*V100_j+(t192+t325+t657)*V101_j+t167*V010_j)*t19;
			t84 = t104*xmax+t96*xmin;
			t83 = t101*zmin+t100*zmax;
			t82 = t102*zmin+t103*zmax;
			t81 = t95*xmax+t102*xmin;
			t80 = t98*zmin+t97*zmax;
			t79 = t96*ymax+t105*ymin;
			t78 = t94*xmax+t103*xmin;
			t77 = t104*ymax+t99*ymin;
			t76 = t101*ymax+t98*ymin;
			t75 = t100*ymax+t97*ymin;
			t74 = t95*zmin+t94*zmax;
			t73 = t99*xmax+t105*xmin;
			t72 = t93*ymax+t92*ymin;
			t71 = t90*zmin+t91*zmax;
			t70 = t89*xmax+t88*xmin;
			t69 = t72*Ri11;
			t68 = t72*Ri21;
			t67 = t72*Rj01;
			t66 = t72*Rj21;
			t65 = t71*Rj12;
			t64 = ((t635+t622+t275+t448)*V011_j+(t279+t476+t477)*V111_j+(t538+t623+t243+t406)*V000_j+(t619+t624+t393+t457+t468)*V001_j+(t230+t280)*V110_j+(t186+t273)*V100_j+(t282+t311)*V101_j+(t233+t285)*V010_j)*t22+((t227+t395+t418)*V011_j+(t214+t279)*V111_j+(t274+t314)*V000_j+(t246+t348+t457)*V001_j+(t280+t432+t470)*V110_j+(t273+t458+t480)*V100_j+(t198+t398+t452)*V101_j+(t285+t424+t453)*V010_j)*t21+(t157*V011_j+t161*V111_j+t169*V000_j+t162*V001_j+t159*V110_j+t153*V100_j+t155*V101_j+t164*V010_j)*zmin*zmax;
			t63 = ((t636+t638+t284+t479)*V011_j+(t286+t316)*V111_j+(t220+t278)*V000_j+(t251+t268)*V001_j+(t636+t627+t272+t469)*V110_j+(t231+t287)*V100_j+(t550+t539+t241+t472)*V101_j+(t640+t612+t276+t423)*V010_j)*t11+((t202+t284)*V011_j+(t252+t265)*V111_j+(t613+t615+t278+t449)*V000_j+(t283+t302)*V001_j+(t237+t272)*V110_j+(t287+t437+t465)*V100_j+(t281+t443+t472)*V101_j+(t535+t640+t337+t366+t431)*V010_j)*t20+(t166*V011_j+t158*V111_j+t163*V000_j+t176*V001_j+t156*V110_j+t170*V100_j+t154*V101_j+t174*V010_j)*ymax*ymin;
			t62 = ((t635+t532+t241+t406)*V011_j+(t283+t316)*V111_j+(t622+t623+t209+t401)*V000_j+(t252+t268)*V001_j+(t275+t432+t474)*V110_j+(t276+t437+t473)*V100_j+(t274+t311)*V101_j+(t235+t277)*V010_j)*t12+((t281+t429+t479)*V011_j+(t251+t265)*V111_j+(t282+t314)*V000_j+(t286+t302)*V001_j+(t188+t409+t474)*V110_j+(t194+t386+t473)*V100_j+(t243+t377+t451)*V101_j+(t277+t423+t465)*V010_j)*t19+(t172*V011_j+t173*V111_j+t175*V000_j+t168*V001_j+t167*V110_j+t160*V100_j+t171*V101_j+t165*V010_j)*xmax*xmin;
			t61 = t84*ymax+t73*ymin;
			t60 = t83*ymax+t80*ymin;
			t59 = t88*tiy+t79*Ri11+t82*Ri12;
			t58 = t91*tiz+t75*Ri21+t78*Ri20;
			t57 = t91*tjx+t75*Rj01+t78*Rj00;
			t56 = t88*tjy+t82*Rj12+t79*Rj11;
			t55 = t91*tjz+t75*Rj21+t78*Rj20;
			t54 = (t89*tiy+t77*Ri11+t74*Ri12)*xmax;
			t53 = (t89*tjy+t74*Rj12+t77*Rj11)*xmax;
			t52 = (t90*tiz+t76*Ri21+t81*Ri20)*zmin;
			t51 = (t90*tjx+t76*Rj01+t81*Rj00)*zmin;
			t50 = (t90*tjz+t76*Rj21+t81*Rj20)*zmin;
			t49 = t106*tiz+t68+t70*Ri20+t71*Ri22;
			t48 = t106*tjx+t67+t70*Rj00+t71*Rj02;
			t47 = t106*tjy+t65+t72*Rj11+t70*Rj10;
			t46 = t106*tjz+t66+t70*Rj20+t71*Rj22;
			t45 = t86*Ri22+t52+(Ri22*t108+t58)*zmax;
			t44 = t85*Ri20+(t89*tiz+t77*Ri21+t74*Ri22)*xmax+(t88*tiz+Ri20*t109+t79*Ri21+t82*Ri22)*xmin;
			t43 = t86*Rj02+t51+(Rj02*t108+t57)*zmax;
			t42 = t85*Rj00+(t89*tjx+t77*Rj01+t74*Rj02)*xmax+(t88*tjx+Rj00*t109+t79*Rj01+t82*Rj02)*xmin;
			t41 = t86*Rj12+(t90*tjy+t76*Rj11+t81*Rj10)*zmin+(t91*tjy+Rj12*t108+t75*Rj11+t78*Rj10)*zmax;
			t40 = t85*Rj10+t53+(Rj10*t109+t56)*xmin;
			t39 = t86*Rj22+t50+(Rj22*t108+t55)*zmax;
			t38 = t85*Rj20+(t89*tjz+t77*Rj21+t74*Rj22)*xmax+(t88*tjz+Rj20*t109+t79*Rj21+t82*Rj22)*xmin;
			t37 = t87*Ri21+(t93*tiz+t84*Ri20+t83*Ri22)*ymax+(t92*tiz+Ri21*t107+t73*Ri20+t80*Ri22)*ymin;
			t36 = t87*Rj01+(t93*tjx+t84*Rj00+t83*Rj02)*ymax+(t92*tjx+Rj01*t107+t73*Rj00+t80*Rj02)*ymin;
			t35 = t87*Rj11+(t93*tjy+t83*Rj12+t84*Rj10)*ymax+(t92*tjy+Rj11*t107+t80*Rj12+t73*Rj10)*ymin;
			t34 = t87*Rj21+(t93*tjz+t84*Rj20+t83*Rj22)*ymax+(t92*tjz+Rj21*t107+t73*Rj20+t80*Rj22)*ymin;
			integrals[6] = (t110*tix*tix+(t71*tix+t64*Ri02)*Ri02+(t70*tix+(t74*xmax+t82*xmin)*Ri02+t62*Ri00)*Ri00+((t93*tix+t83*Ri02+t84*Ri00)*ymax+(t92*tix+t80*Ri02+t73*Ri00)*ymin+t63*Ri01)*Ri01)*t318;
			integrals[7] = ((t106*tiy+t69+t71*Ri12+t70*Ri10)*tix+(t86*Ri12+(t90*tiy+t76*Ri11+t81*Ri10)*zmin+(t91*tiy+Ri12*t108+t75*Ri11+t78*Ri10)*zmax)*Ri02+(t85*Ri10+t54+(Ri10*t109+t59)*xmin)*Ri00+(t87*Ri11+(t93*tiy+t83*Ri12+t84*Ri10)*ymax+(t92*tiy+Ri11*t107+t80*Ri12+t73*Ri10)*ymin)*Ri01)*t317;
			integrals[8] = (t49*tix+t45*Ri02+t44*Ri00+t37*Ri01)*t317;
			integrals[9] = (t63*Ri11*Ri11+(t69+t110*tiy)*tiy+(t71*tiy+t60*Ri11+t64*Ri12)*Ri12+(t54+t59*xmin+t62*Ri10)*Ri10)*t318;
			integrals[10] = (t49*tiy+t37*Ri11+t45*Ri12+t44*Ri10)*t317;
			integrals[11] = (t63*Ri21*Ri21+(t68+t110*tiz)*tiz+(t70*tiz+t61*Ri21+t62*Ri20)*Ri20+(t52+t58*zmax+t64*Ri22)*Ri22)*t318;
			integrals[12] = (t48*tix+t43*Ri02+t42*Ri00+t36*Ri01)*t317;
			integrals[13] = (t47*tix+t41*Ri02+t40*Ri00+t35*Ri01)*t317;
			integrals[14] = (t46*tix+t39*Ri02+t38*Ri00+t34*Ri01)*t317;
			integrals[15] = (t48*tiy+t36*Ri11+t43*Ri12+t42*Ri10)*t317;
			integrals[16] = (t47*tiy+t35*Ri11+t41*Ri12+t40*Ri10)*t317;
			integrals[17] = (t46*tiy+t34*Ri11+t39*Ri12+t38*Ri10)*t317;
			integrals[18] = (t48*tiz+t36*Ri21+t42*Ri20+t43*Ri22)*t317;
			integrals[19] = (t47*tiz+t35*Ri21+t40*Ri20+t41*Ri22)*t317;
			integrals[20] = (t46*tiz+t34*Ri21+t38*Ri20+t39*Ri22)*t317;
			integrals[21] = (t63*Rj01*Rj01+(t67+t110*tjx)*tjx+(t70*tjx+t61*Rj01+t62*Rj00)*Rj00+(t51+t57*zmax+t64*Rj02)*Rj02)*t318;
			integrals[22] = (t47*tjx+t35*Rj01+t40*Rj00+t41*Rj02)*t317;
			integrals[23] = (t46*tjx+t34*Rj01+t38*Rj00+t39*Rj02)*t317;
			integrals[24] = (t64*Rj12*Rj12+(t65+t110*tjy)*tjy+(t72*tjy+t60*Rj12+t63*Rj11)*Rj11+(t53+t56*xmin+t62*Rj10)*Rj10)*t318;
			integrals[25] = (t46*tjy+t39*Rj12+t34*Rj11+t38*Rj10)*t317;
			integrals[26] = (t63*Rj21*Rj21+(t66+t110*tjz)*tjz+(t70*tjz+t61*Rj21+t62*Rj20)*Rj20+(t50+t55*zmax+t64*Rj22)*Rj22)*t318;
		}

		// Normalize each integral
		for (int j = 0; j < integrals.size(); j++) {
			integrals[j] = integrals[j] * tau_ij;
		}

		// Finally, assign the result
		result.push_back(make_pair(bonePairs[i], integrals));
	}

	return bonePairs.size();
}

/**
 * Scan through all cells and bookkeep the interior unmarked cells.
 * Perform this after the marking process.
 */
void MarkedUniformGrid3::bookkeepInteriorCells()
{
	// 1. Scan through all original grid points and expand all the unmarked points
	int3_type gridIndex = UniformGrid3::scanStart();
	do {
		// Test if the point is not marked
		if (!validPoint(gridIndex)) {
			addExpandedPoint(gridIndex);
		}
	} while(UniformGrid3::scanNext(gridIndex));


	// 2. Scan through all original grid cells and expand all the unmarked cells
	int3_type cellIndex = UniformGrid3::scanCellStart();
	do {
		// Test if the cell is not marked
		if (!validCell(cellIndex)) {
			addExpandedCell(cellIndex);
		}
	} while(UniformGrid3::scanNextCell(cellIndex));
}

bool MarkedUniformGrid3::validPoint( const int3_type &gridIndex ) const
{
	return (inBounds(gridIndex) && markedPoints[UniformGrid3::getLinearIndex(gridIndex)]);
}

bool MarkedUniformGrid3::validCell( const int3_type &cellIndex ) const
{
	return (inCellBounds(cellIndex) && isMarked(cellIndex));
}
