///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __MARKED_UNIFORM_GRID_H__
#define __MARKED_UNIFORM_GRID_H__

#include "UniformGrid.h"
#include <hash_map>

class Mesh3_real;

class MarkedUniformGrid3 : public UniformGrid3 {
public:
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Creation
	MarkedUniformGrid3( int dataDim = 1 );


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Display
	void drawSortedSubset( int comp, const std::vector<int> &sortedLinearIndex );
	void drawEdges( bool deformed = true );
	void drawFaces( int comp = -1, bool deformed = true, bool showWireframe = false );


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Size & Format
	virtual index_type totalGridPoints() const;
	virtual index_type totalGridCells() const;
	virtual bool validPoint( const int3_type &gridIndex ) const;
	virtual bool validCell( const int3_type &cellIndex ) const;


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Marking
	void resetMark();
	void countMarked();
	void markCells( const Mesh3_real &mesh );
	void markCells( const PointList &points );
	void markCells( const std::vector<PointList*> &points, const std::vector< std::vector<bool> > &valid );
	bool isMarked( real3_type query ) const;
	bool isMarked( int3_type cellIndex ) const;
	void markCell( real3_type query );
	void markCell( int3_type cellIndex );


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Traversal
	virtual int3_type scanStart() const;
	virtual int3_type scanCellStart() const;
	virtual bool scanNext(int3_type &gridIndex) const;							/**< Replaces the argument with the next grid point, and returns false if we are at the end. */
	virtual bool scanNextCell(int3_type &cellIndex) const;						/**< Replaces the argument with the next grid cell, and returns false if we are at the last cell. */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Indexing
	virtual index_type getLinearIndex( const int3_type &gridIndex ) const;
	virtual index_type getCellLinearIndex( const int3_type &cellIndex ) const;
	virtual int3_type translateLinearIndex( const index_type &pointIndex ) const; 
	virtual int3_type translateCellLinearIndex( const index_type &cellIndex ) const; 


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid edges
	// Create a unique list of edges in the marked grid cells
	// Represent this as a hash table to efficiently look up existing edges
	void findEdges();
	stdext::hash_map<int,std::vector<int> > allEdges;							/**< List of unique edges */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Numerical Computation
	void computeIntegrals( const int3_type cellIndex, std::vector< std::vector<double> > &result ) const;
	int computeOverlapIntegrals( const int3_type cellIndex, const RigidDeform &def, std::vector< std::pair<index2_type, std::vector<double> > > &result, bool justCount = false ) const;

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//// Grid Expansion
	void bookkeepInteriorCells();												/**< Bookkeep the interior unmarked cells */


protected:

	size_type totalMarkedCells, totalMarkedPoints;
	std::vector<bool> markedCells, markedPoints;
	std::vector<int> markedCellLinearIndex, markedPointLinearIndex;

	// Translating linear indices to grid indices
	stdext::hash_map<int,int3_type> linearIndexToIndex;							/**< Linear index to int3_type index for grid points*/
	stdext::hash_map<int,int3_type> cellLinearIndexToIndex;						/**< Linear index to int3_type index for grid cells */

};

#endif
