///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "ParamSelect.h"
#include "util.h"

ParamSelect::ParamSelect( QWidget *ui, QString controlName, int id ) : name(controlName), paramId(id),
label(NULL), minRange(NULL), maxRange(NULL), scroll(NULL), paramValue(NULL),
minValue(0), maxValue(100), currentValue(0), editingValue(false)
{
	// Find the components in the ui
	QString componentName;
	int height = 40 + 30*(id-1);

	// Remove all spaces to obtain object name
	name.remove(QString(" "));

	//////////////////////////////////////////////////////////////////////////
	// Label
	componentName = name + QString("Label");
	//label = &(ui->findChild<QLabel*>(componentName)[0]);
	label = new QLabel(ui);
	label->setObjectName(componentName);
	label->setGeometry(QRect(20, height-1, 111, 20));
	label->setText(controlName);

	//////////////////////////////////////////////////////////////////////////
	// Minimum Scroll Range
	componentName = name + QString("Min");
	//minRange = &(ui->findChild<QPushButton*>(componentName)[0]);
	minRange = new QPushButton(ui);
	minRange->setObjectName(componentName);
	minRange->setGeometry(QRect(145, height, 60, 20));

	//////////////////////////////////////////////////////////////////////////
	// Maximum Scroll Range
	componentName = name + QString("Max");
	//maxRange = &(ui->findChild<QPushButton*>(componentName)[0]);
	maxRange = new QPushButton(ui);
	maxRange->setObjectName(componentName);
	maxRange->setGeometry(QRect(485, height, 60, 20));

	//////////////////////////////////////////////////////////////////////////
	// Scroll Bar
	componentName = name + QString("Scroll");
	//scroll = &(ui->findChild<QScrollBar*>(componentName)[0]);
	scroll = new QScrollBar(ui);
	scroll->setObjectName(componentName);
	scroll->setGeometry(QRect(210, height+2, 269, 17));
	scroll->setMaximum(100);
	scroll->setOrientation(Qt::Horizontal);

	//////////////////////////////////////////////////////////////////////////
	// Value
	componentName = name + QString("Value");
	//paramValue = &(ui->findChild<QLineEdit*>(componentName)[0]);
	paramValue = new QLineEdit(ui);
	paramValue->setObjectName(componentName);
	paramValue->setGeometry(QRect(557, height, 71, 20));

	//////////////////////////////////////////////////////////////////////////
	updateDisplay();

	//////////////////////////////////////////////////////////////////////////
	// Connect signals
	connect(minRange, SIGNAL(clicked()), this, SLOT(minRangeClicked()));
	connect(maxRange, SIGNAL(clicked()), this, SLOT(maxRangeClicked()));
	connect(scroll, SIGNAL(valueChanged(int)), this, SLOT(scrollChanged(int)));
	connect(paramValue, SIGNAL(editingFinished()), this, SLOT(editChanged()));
}

ParamSelect::~ParamSelect()
{
	delete label;
	delete minRange;
	delete maxRange;
	delete scroll;
	delete paramValue;
}

void ParamSelect::minRangeClicked()
{
	// Pop-up dialog box for the minimum value
	bool ok;
	double result = QInputDialog::getDouble(this, tr("Set minimum..."),
		tr("Minimum Value:"), minValue, -1e5, 1e5, 3, &ok);
	if (!ok) return;

	// Set the minimum value
	minValue = std::min(result, maxValue);

	// Modify the current value if necessary
	if (currentValue < minValue) {
		currentValue = minValue;
	}

	// Update the scroll bar
	editingValue = true;
	updateDisplay();
	editingValue = false;
}

void ParamSelect::maxRangeClicked()
{
	// Pop-up dialog box for the maximum value
	bool ok;
	double result = QInputDialog::getDouble(this, tr("Set maximum..."),
		tr("Maximum Value:"), maxValue, -1e5, 1e5, 3, &ok);
	if (!ok) return;

	// Set the minimum value
	maxValue = result;

	// If the current value is beyond this new range, then clamp the value.
	if (currentValue > maxValue) {
		currentValue = maxValue;
	}

	// Update the scroll bar
	editingValue = true;
	updateDisplay();
	editingValue = false;
}

void ParamSelect::scrollChanged( int v )
{
	if (!editingValue) {
		currentValue = minValue + ((maxValue - minValue) * ((double)interpSlider(scroll)));
		updateDisplay(false);
		emit valueChanged(paramId, currentValue);
	}
}

void ParamSelect::editChanged()
{
	bool ok;
	double result = paramValue->text().toDouble(&ok);

	if (!ok) {
		// Reset the value
		paramValue->setText(doubleText(currentValue));
		updateDisplay(false);		/**< Don't update the scroll bar */
		return;
	}

	currentValue = result;

	// If the current value exceeds the current range, then modify the range.
	if (currentValue < minValue) {
		minValue = currentValue;
	}
	else if (currentValue > maxValue) {
		maxValue = currentValue;
	}

	// Update UI
	// Flag that we are editing the value via the text box.
	// This is to prevent scrollChanged() from emitting a signal when the scroll bar is set.
	editingValue = true;
	updateDisplay();
	editingValue = false;

	emit valueChanged(paramId, result);
}

QString ParamSelect::doubleText( double x )
{
	return tr("%1").arg(x);
}

void ParamSelect::updateDisplay( bool updateScrollBar/* = true*/ )
{
	minRange->setText(doubleText(minValue));
	maxRange->setText(doubleText(maxValue));
	paramValue->setText(doubleText(currentValue));

	if (updateScrollBar) {
		setInterpSlider(scroll, (currentValue - minValue) / (maxValue - minValue));
	}
}

int ParamSelect::getId() const
{
	return paramId;
}

void ParamSelect::initRange( double minRange, double maxRange, double initValue )
{
	minValue = minRange;
	maxValue = maxRange;
	currentValue = initValue;

	// Fix potentially bad range values
	if (minValue >= maxValue) {
		minValue = maxValue;
	}

	// If the current value exceeds the current range, then modify the range.
	if (currentValue < minValue) {
		minValue = currentValue;
	}
	else if (currentValue > maxValue) {
		maxValue = currentValue;
	}

	editingValue = true;
	updateDisplay();
	editingValue = false;
}
