///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __SPIN_IMAGE_H__
#define __SPIN_IMAGE_H__

#include "types.h"

/**
 * Apply the spin map to a point.
 */
real2_type SpinMap( const real3_type &p, const real3_type &n, const real3_type &x );
real_type GeometricConsistency( const real3_type &p_s1, const real3_type &n_s1,
							    const real3_type &p_s2, const real3_type &n_s2,
								const real3_type &p_t1, const real3_type &n_t1,
								const real3_type &p_t2, const real3_type &n_t2 );

class ShapeModel;
class CorrespondenceList;
real_type GeometricConsistency_Diameter( ShapeModel &source, ShapeModel &target, CorrespondenceList &cl );


class SpinImage
{
public:

	void accumulate( real2_type coord, real_type binSize );			/**< Accumulate the given value in the spin image. */
	int countBins() const;											/**< Count the number of non-empty bins. */

	void resize( int width );				/**< Allocate the space for the spin image. */
	void reset();							/**< Reset all the values of the spin image. */
	bool invalid() const;					/**< Check if there's no data in this spin image. */

	int width() const {
		return data.size();
	}

	real_type& operator()( int i, int j ) {
		return data[i][j];
	}
	const real_type& operator()( int i, int j ) const {
		return data[i][j];
	}

private:
	std::vector< std::vector<real_type> > data;
};


real_type ComputeSimilarity( const SpinImage &P, const SpinImage &Q, real_type lambda );


class SpinImageStack
{
public:

	size_type size() const {
		return data.size();
	}

	void resize( size_type n ) {
		data.resize(n);
	}

	void clear() {
		data.clear();
	}

	void add( const SpinImage &si ) {
		data.push_back(si);
	}

	SpinImage& operator[]( int i ) {
		return data[i];
	}

	const SpinImage& operator[]( int i ) const {
		return data[i];
	}

	void querySimilar( const SpinImage& query, idxList &result, const char *scoreOutput = NULL ) const;
	void querySubsetSimilar( const SpinImage& query, const idxList &subset, idxList &result, distList &similarityScores ) const;
	void copyParams( const SpinImageStack& sis );

	real_type binSize;
	int imageWidth;
	real_type supportAngle;
	real_type lambda;				/**< Half of median bin size */

private:
	std::vector<SpinImage> data;
};

#endif
