///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "UniformGrid.h"
#include <iostream>
#include <algorithm>
#include "Correspondence.h"
#include "util.h"
#include "geom_util.h"

using namespace std;

UniformGrid3::UniformGrid3( int dataDim /*= 1*/ ) : dataDim(dataDim), numExpandedPts(0), maxExpandLevel(0)
{
	numDivisions[0] = numDivisions[1] = numDivisions[2] = 0;
}

void UniformGrid3::queryCell( const real3_type query, int3_type &index, real3_type &remainder ) const
{
	// Compute the grid cell that encloses this point
	real3_type realIndex = (query - minPos) / sideLength;

	// Compute the index and remainder
	for (int i = 0; i < dim; i++) {
		index[i] = floor(realIndex[i]);
		remainder[i] = realIndex[i] - index[i];
	}
}

/**
 * Compute the value at an arbitrary position using trilinear interpolation of the enclosing grid cell.
 */
data_type UniformGrid3::getValue( real3_type query ) const
{
	int3_type index;
	real3_type remainder;
	queryCell(query, index, remainder);

	return getValue(index, remainder);
}

data_type UniformGrid3::getValue( const int3_type& cellIndex, const real3_type& weights ) const
{
	// Find the function values and interpolate them
	data_type value(dataDim);
	for (int i = 0; i < (1 << dim); i++) {
		int3_type offset = offset2sub(i);
		real_type weight = 1;

		// Compute the weight for this grid point (trilinear interpolation)
		for (int j = 0; j < dim; j++) {
			if (offset[j]) {
				weight *= weights[j];
			}
			else {
				weight *= (1-weights[j]);
			}
		}

		// Look up the grid point value and add to the total
		value += weight * data[getLinearIndex(cellIndex + offset)];
	}

	//// Test that the L1 norm is normalized
	//if (value.normL1() < 1) {
	//	cout << "Un-normalized value: " << value << " (sum=" << value.normL1() << ") interpolated from: " << endl;
	//	for (int i = 0; i < (1 << dim); i++) {
	//		int3_type offset = offset2sub(i);
	//		
	//		cout << "   " << data[sub2ind(index + offset)] << endl;
	//	}
	//	cout << endl;
	//}

	return value;
}


void UniformGrid3::getWeights( real3_type query, idxList &gridIndices, distList &gridWeights )
{
	int3_type index;
	real3_type remainder;
	queryCell(query, index, remainder);

	// Find the function values and interpolate them
	gridIndices.clear();
	gridWeights.clear();
	gridIndices.resize(1 << dim);
	gridWeights.resize(1 << dim);

	for (int i = 0; i < (1 << dim); i++) {
		int3_type offset = offset2sub(i);
		real_type weight = 1;

		// Compute the weight for this grid point (trilinear interpolation)
		for (int j = 0; j < dim; j++) {
			if (offset[j]) {
				weight *= remainder[j];
			}
			else {
				weight *= (1-remainder[j]);
			}
		}

		gridIndices[i] = getLinearIndex(index + offset);
		gridWeights[i] = weight;
	}
}


void UniformGrid3::setValues( real3_type query, int comp, real_type value )
{
	int3_type index;
	real3_type remainder;
	queryCell(query, index, remainder);

	for (int i = 0; i < (1 << dim); i++) {
		int3_type offset = offset2sub(i);
		setValue(index + offset, comp, value);
	}
}


/**
 * Create a uniform grid from a bounding box.
 */
void UniformGrid3::createFromBBox( real3_type bboxMin, real3_type bboxMax, int div )
{
	real3_type length = bboxMax - bboxMin;

	// Define the side length to be such that there are "div" divisions on the longest axis.
	sideLength = max(length[0], max(length[1], length[2])) / ((real_type) div);
	minPos = bboxMin;

	// Calculate the number of divisions per axis
	for (int i = 0; i < dim; i++) {
		numDivisions[i] = ceil(length[i] / sideLength) + 1;
	}

	cout << "****** Grid divisions: " << numDivisions << endl;
}


/**
* Create a uniform grid from a bounding box, with side length parameter specified.
*/
void UniformGrid3::createFromBBox( real3_type bboxMin, real3_type bboxMax, real_type sideLength )
{
	real3_type length = bboxMax - bboxMin;
	minPos = bboxMin;
	this->sideLength = sideLength;

	// Calculate the number of divisions per axis
	for (int i = 0; i < dim; i++) {
		numDivisions[i] = ceil(length[i] / sideLength) + 1;
	}

	cout << "****** Grid divisions: " << numDivisions << endl;
}


void UniformGrid3::drawGridPoint( int3_type gridIndex, int comp /*= 0*/, bool deformed /*= true */ ) const
{
	real3_type gridPos;
	
	if (deformed) {
		gridPos	= deformedGridLocation(gridIndex);
	}
	else {
		gridPos	= gridLocation(gridIndex);
	}

	// If the component is negative, color the vertex with the weighted combination
	if (comp < 0) {

		RealColor color(0.0, 0.0, 0.0);

		for (int i = 0; i < dataDim; i++) {
			real_type weight = getValue(gridIndex, i);
			color += weight * compColors[i];
		}

		// Clamp the color within the limits of [0.0, 1.0]
		for (int i = 0; i < 3; i++) {
			color[i] = min(1.0f, max(0.0f, color[i]));
		}

		// Use the color
		glColor3f(color[0], color[1], color[2]);
	}
	else {
		real_type value = getValue(gridIndex, comp);

		if ((value >= 0.0) && (value <= 1.0)) {
			// We have nice weights between 0 and 1 -- color it gray
			// Darker shade: weights closer to 1
			// Brighter shade: weights closer to 0
			value = 1.0 - value;
			glColor3f(value, value, value);
		}
		else if (value < 0.0) {
			// If we have negative weights -- color it red
			// Darker red: less negative
			// Brighter red: more negative
			glColor3f(-value, 0.f, 0.f);
		}
		else {
			// If our weights are greater than 1 -- color it blue
			// Darker blue: greater than 1 a little
			// Brighter blue: greater than 1 a lot
			glColor3f(0.f, 0.f, value - 1.0);
		}
	}

	// Draw the point
	glVertex3f(gridPos[0], gridPos[1], gridPos[2]);
}

void UniformGrid3::draw( int comp /*= 0*/, bool deformed /*= true */ ) const
{
	glDisable(GL_LIGHTING);
	int3_type gridIndex;

	//////////////////////////////////////////////////////////////////////////
	// Draw each grid point
	glBegin(GL_POINTS);

	gridIndex = scanStart();
	do {
		drawGridPoint(gridIndex, comp, deformed);
	} while(scanNext(gridIndex));

	glEnd();


	//////////////////////////////////////////////////////////////////////////
	// Next, draw the border if the component is positive
	if (comp < 0) {
		return;
	}

	GLint pointSize;
	glGetIntegerv(GL_POINT_SIZE, &pointSize);

	glPointSize(pointSize + 4);			/**< Border width = 2 pixels */
	glBegin(GL_POINTS);
	gridIndex = scanStart();
	do {
		real_type value = getValue(gridIndex, comp);
		real3_type gridPos;
		if (deformed) {
			gridPos	= deformedGridLocation(gridIndex);
		}
		else {
			gridPos	 = gridLocation(gridIndex);
		}

		if (value > 1.0) {
			glColor3f(0.0, 0.0, 1.0);			/**< Blue color border */
			glVertex3f(gridPos[0], gridPos[1], gridPos[2]);
		}
		else if (value >= REAL_TYPE_EPSILON) {
			glColor3f(1.0, 0.9, 0.2);			/**< Yellowish color border */
			glVertex3f(gridPos[0], gridPos[1], gridPos[2]);
		}
		else if (value < 0) {
			glColor3f(1.0, 0.0, 0.0);			/**< Red color border */
			glVertex3f(gridPos[0], gridPos[1], gridPos[2]);
		}
		else {
			// Don't draw the border
		}
	} while(scanNext(gridIndex));
	glEnd();

	// Restore the original point size
	glPointSize(pointSize);
}

void UniformGrid3::resetData()
{
	// Calculate the number of divisions per axis
	index_type numPoints = totalGridPoints();

	// Allocate data
	if (numPoints > 0) {
		data_type temp(dataDim);
		data.clear();
		data.resize(numPoints, temp);

		// Generate colors for each bone
		generateLightColors(dataDim, compColors);
	}
}

real3_type UniformGrid3::gridLocation( const int3_type &ind ) const
{
	real3_type pos;
	for (int i = 0; i < dim; i++) {
		pos[i] = minPos[i] + sideLength*ind[i];
	}
	return pos;
}

real_type UniformGrid3::getValue( int3_type ind, index_type comp ) const
{
	return data[getLinearIndex(ind)].val[comp];
}


void UniformGrid3::setAllValues( index_type linearIndex, const data_type &value )
{
	// Check bounds
	if (linearIndex >= totalGridPoints()) {
		cerr << "Error!" << endl;
		return;
	}
	data[linearIndex] = value;
}

void UniformGrid3::setValue( int3_type ind, index_type comp, real_type value )
{
	index_type linInd = getLinearIndex(ind);
	if (linInd < 0) {
		cerr << "Error!" << endl;
	}
	data[getLinearIndex(ind)].val[comp] = value;
}

int3_type UniformGrid3::size() const
{
	return numDivisions;
}

size_type UniformGrid3::getDataDim() const
{
	return dataDim;
}

/**
 * Normalize the values at each grid point;
 * i.e. scale the data vector so that the components add up to a total of 1.
 */
void UniformGrid3::normalizeValues()
{
	int3_type gridIndex = scanStart();
	do {

		data[getLinearIndex(gridIndex)].normalizeL1();

	} while(scanNext(gridIndex));
}


void UniformGrid3::clampMaxValues()
{
	for (int i = 0; i < totalGridPoints(); i++) {
		data_type &value = data[i];

		// Find maximum component
		int maxIndex;
		value.findMax(maxIndex);

		// Set the maximum component to 1, rest to 0.
		value.setZero();
		value.val[maxIndex] = 1.0;
	}
}


index_type UniformGrid3::totalGridPoints() const
{
	index_type numPoints = 1;
	for (int i = 0; i < dim; i++) {
		numPoints *= numDivisions[i];
	}
	return numPoints;
}

index_type UniformGrid3::totalGridCells() const
{
	index_type numPoints = 1;
	for (int i = 0; i < dim; i++) {
		numPoints *= numDivisions[i]-1;
	}
	return numPoints;
}


bool UniformGrid3::inBounds( const int3_type &gridIndex ) const
{
	bool minOK = ((gridIndex[0] >= 0) && (gridIndex[1] >= 0) && (gridIndex[2] >= 0));
	bool maxOK = ((gridIndex[0] < numDivisions[0]) && (gridIndex[1] < numDivisions[1]) && (gridIndex[2] < numDivisions[2]));
	return (minOK && maxOK);
}

bool UniformGrid3::inCellBounds( const int3_type &cellIndex ) const
{
	bool minOK = ((cellIndex[0] >= 0) && (cellIndex[1] >= 0) && (cellIndex[2] >= 0));
	bool maxOK = ((cellIndex[0] < numDivisions[0]-1) && (cellIndex[1] < numDivisions[1]-1) && (cellIndex[2] < numDivisions[2]-1));
	return (minOK && maxOK);
}

//////////////////////////////////////////////////////////////////////////
// Traversal code implementation

int3_type UniformGrid3::scanStart() const
{
	return int3_type(0,0,0);
}

int3_type UniformGrid3::scanCellStart() const
{
	return int3_type(0,0,0);
}

bool UniformGrid3::scanNext( int3_type &gridIndex ) const
{
	// Increment x, then y, then z
	for (int d = 0; d < dim; d++) {
		gridIndex[d]++;

		if (gridIndex[d] < numDivisions[d]) {
			return true;
		}

		// We are moving to the next dimension: Reset this coordinate.
		gridIndex[d] = 0;
	}

	return false;
}

/**
 * Returns the minimum positioned grid point of the next cell in which to iterate.
 */
bool UniformGrid3::scanNextCell( int3_type &cellIndex ) const
{
	// Increment x, then y, then z
	for (int d = 0; d < dim; d++) {
		cellIndex[d]++;

		// We want to iterate over the cells, so iterate the minimum position over all cells.
		if (cellIndex[d] < (numDivisions[d]-1)) {
			return true;
		}

		// We are moving to the next dimension: Reset this coordinate.
		cellIndex[d] = 0;
	}

	return false;
}

int3List UniformGrid3::findNeighbors( int3_type gridIndex ) const
{
	int3List result;

	// Neighbor in each dimension
	for (int d = 0; d < dim; d++) {
		// Top neighbor & Bottom neighbor
		for (int ni = -1; ni <= 1; ni +=2) {
			int3_type neighborIndex = gridIndex;
			neighborIndex[d] += ni;

			if (validPoint(neighborIndex)) {
				result.push_back(neighborIndex);
			}
		}
	}

	return result;
}


int3List UniformGrid3::findAdjacentCellsToPoint( int3_type gridPointIndex ) const
{
	int3List result;

	// All neighboring cells
	for (int i = 0; i < (1 << dim); i++) {
		int3_type cellIndex = gridPointIndex - offset2sub(i);

		if (validCell(cellIndex)) {
			result.push_back(cellIndex);
		}
	}

	return result;
}

int3List UniformGrid3::findAdjacentCellsToCell( int3_type gridCellIndex ) const
{
	int3List result;

	// Neighbor in each dimension (4-connected neighbors)
	for (int d = 0; d < dim; d++) {
		// Top neighbor & Bottom neighbor
		for (int ni = -1; ni <= 1; ni +=2) {
			int3_type neighborIndex = gridCellIndex;
			neighborIndex[d] += ni;

			if (validCell(neighborIndex)) {
				result.push_back(neighborIndex);
			}
		}
	}

	return result;
}

void UniformGrid3::getCellBox( const int3_type &cellIndex, real3_type &minPt, real3_type &maxPt ) const
{
	static const int3_type offset(1,1,1);

	minPt = gridLocation(cellIndex);
	maxPt = gridLocation(cellIndex + offset);
}

index_type UniformGrid3::getLinearIndex( const int3_type &gridIndex ) const
{
	return sub2ind(gridIndex, numDivisions);
}

index_type UniformGrid3::getCellLinearIndex( const int3_type &cellIndex ) const
{
	static const int3_type offset(1,1,1);
	return sub2ind(cellIndex, numDivisions - offset);
}

bool UniformGrid3::withinGrid( real3_type query ) const
{
	static const int3_type offset(1,1,1);
	real3_type minOffset = query - minPos;
	real3_type maxOffset = gridLocation(numDivisions - offset) - query;

	for (int d = 0; d < 3; d++) {
		if ((minOffset[d] < 0) || (maxOffset[d] < 0)) {
			return false;
		}
	}

	return true;
}

data_type UniformGrid3::cellNonzeroValue( int3_type cellIndex ) const
{
	data_type result(dataDim);

	for (int i = 0; i < (1 << dim); i++) {
		int3_type offset = offset2sub(i);
		data_type value = data[getLinearIndex(cellIndex + offset)];

		for (int j = 0; j < dataDim; j++) {
			if (value.val[j] != 0.0) {
				result.val[j] += 1.0;
			}
		}
	}

	return result;
}

void UniformGrid3::getCellGridPoints( int3_type cellIndex, idxList &ptIndices ) const
{
	ptIndices.clear();

	if (validCell(cellIndex)) {
		ptIndices.reserve(1 << dim);

		for (int i = 0; i < (1 << dim); i++) {
			int3_type offset = offset2sub(i);
			index_type ptIndex = getLinearIndex(cellIndex + offset);

			if (validPoint(cellIndex + offset)) {
				ptIndices.push_back(ptIndex);
			}
			else {
				// If it's a valid cell, then all points in it should be valid!
				cerr << "Error in UniformGrid3::getCellGridPoints(): invalid point in cell!" << endl;
			}
		}
	}
}


/**
* Get the vertex weights for each vertex constraint.  The last 2 arguments should be empty lists, which will be filled with
*   nodeList:      Contains for each vertex, a list of which deformation nodes influence it.
*   vertexWeight:  Contains for each vertex, the weight of the deformation nodes.
*/
void UniformGrid3::getVertexWeights( const CorrespondenceList *con, std::vector<idxList> &nodeList, std::vector<distList> &vertexWeight ) const
{
	nodeList.clear();
	vertexWeight.clear();
	nodeList.resize(con->size());
	vertexWeight.resize(con->size());

	// Gather up the list of bones and weights affecting each constraint
	for (index_type ci = 0; ci < con->size(); ci++) {
		if (!con->valid(ci)) {
			continue;
		}

		real3_type pos = con->sourcePos(ci);
		idxList affectingNodes;
		distList affectingWeights;

		getVertexWeights(pos, affectingNodes, affectingWeights);

		nodeList[ci] = affectingNodes;
		vertexWeight[ci] = affectingWeights;
	}
}


void UniformGrid3::getVertexWeights( const real3_type& pt, idxList &bones, distList &weights ) const
{
	bones.clear();
	weights.clear();

	// Get the weight of this vertex for each node
	data_type allWeights = getValue(pt);

	// Verify that the weights sum up to 1
	//cout << "Weight of constraint: " << allWeights << " sums up to: " << allWeights.normL1() << endl;
	//allWeights.normalizeL1();

	for (index_type bi = 0; bi < dataDim; bi++) {
		if (allWeights.val[bi] > 0) {
			weights.push_back(allWeights.val[bi]);
			bones.push_back(bi);
		}
	}

	if (bones.size() == 0) {
		cerr << "Error in LinearSystemE::getVertexWeights(): No bone affects this point!" << endl;
		return;
	}
}

void UniformGrid3::updateDeformation( const RigidDeform &tfList )
{
	if (tfList.size() != dataDim) {
		// Print something
		cerr << "Warning in UniformGrid3::updateDeformation(): tfList size doesn't match the number of weights/gridPoint!" << endl;
		return;
	}

	// Update the transformation applied to the grid
	currentTransformation = tfList;

	// Update the list of grid points
	// (don't clear the array, because another thread might be trying to access the deformed grid position)
	if (deformedGrid.size() != totalGridPoints()) {
		deformedGrid.resize(totalGridPoints());
	}

	//// Clear ANN data structure
	//deformedGrid.clearANNData();

	int linearIndex = 0;
	int3_type gridIndex = scanStart();
	do {
		// Find transformation
		SE3 tf;
		tf.setZero();
		for (int i = 0; i < dataDim; i++) {
			real_type weight = getValue(gridIndex,i);
			tf.rot += weight * tfList[i].rot;
			tf.tr += weight * tfList[i].tr;
		}

		// Apply transformation
		real3_type tpos = tf*gridLocation(gridIndex);

		// Store transformed point
		deformedGrid[linearIndex] = tpos;
		linearIndex++;
		
	} while(scanNext(gridIndex));

	// Update expanded points
	updateExpandedDeformation();
}

real3_type UniformGrid3::deformedGridLocation( const int3_type &ind ) const
{
	// Determine if it's part of the original grid
	if (validPoint(ind)) {
		// Return the deformed grid location from the pre-computed list
		index_type linInd = getLinearIndex(ind);

		if (deformedGrid.size() > linInd) {
			return deformedGrid[linInd];
		}
		else {
			return gridLocation(ind);
		}
	}
	else {
		int3map::const_iterator it = expandedPts.find(ind);

		// Look up the position
		if ((it != expandedPts.end()) && (deformedExpandedPts.size() > (*it).second)) {
			return deformedExpandedPts[(*it).second];
		}
		else {
			return gridLocation(ind);
		}
	}
}

void UniformGrid3::inverseDeform( ShapeModel &shape, int maxSearchLevel, int gridExpandStep )
{
	// Algorithm: For each point on the shape
	// a. Search for the grid cell containing the point
	// b. Invert the deformation by computing trilinear coordinates
	// c. (Expand grid if point was outside the grid).

	bool warningPrinted = false;
	cerr << "*** Performing inverse deformation..." << endl;

	//// Clear the ANN data of the precomputed point locations
	//deformedGrid.clearANNData();
	//deformedExpandedPts.clearANNData();
	//originalExistingPts.clearANNData();

	// Temporary debugging: store the linear index of the cell containing each vertex
	shape.cellIndex.resize(shape.numVertices());

	// Reset bounding box information
	shape.box_max[0] = shape.box_max[1] = shape.box_max[2] = -REAL_TYPE_MAX;
	shape.box_min[0] = shape.box_min[1] = shape.box_min[2] = REAL_TYPE_MAX;

	for (index_type i = 0; i < shape.numVertices(); i++) {
		real3_type vertex = shape.vertex[i];
		real3_type weights;
		int3_type cellFoundIndex;

		if (inverseDeformPoint(vertex, cellFoundIndex, weights, maxSearchLevel, gridExpandStep)) {
			// 3. Find the original coordinate location
			shape.vertex[i] = trilinearInterpPosition(weights, cellFoundIndex);

			// Store the 3D index
			shape.cellIndex[i] = cellFoundIndex;

			// Update the bounding box
			for (int j = 0; j < 3; j++) {
				if (shape.vertex[i][j] < shape.box_min[j]) {
					shape.box_min[j] = shape.vertex[i][j];
				}
				if (shape.vertex[i][j] > shape.box_max[j]) {
					shape.box_max[j] = shape.vertex[i][j];
				}
			}
		}
		else {
			// Print an error
			//cout << "Warning in UniformGrid3::inverseDeform(): Containing cell not found for vertex " << i << " in search through " << maxExpandLevel << " levels." << endl;
			if (!warningPrinted) {
				cout << "Warning in UniformGrid3::inverseDeform(): in search through " << maxExpandLevel << " levels, containing cell not found for vertex";
				warningPrinted = true;
			}
			cout << " " << i;

			// Mark the vertex to be invalid
			shape.markVertex[i] = VERTEX_INVALID;
		}
	}

	cerr << " done!" << endl;
}


bool UniformGrid3::inverseDeformPoint( const real3_type &p, int3_type &cellFoundIndex, real3_type &weights, int maxSearchLevel, int gridExpandStep )
{
	//////////////////////////////////////////////////////////////////////////
	// a. Search for the grid cell containing the point
	bool cellFound = false;
	int currentSearchLevel = 0;

	while (!cellFound && (currentSearchLevel <= maxSearchLevel)) {
		// 1) Grab a list of closest deformed grid cells
		// TODO: how to eliminate the constant 12 here?
		int3List cellList = grabClosestDeformedCells(p, currentSearchLevel, 12);		/**< WARNING!  Last parameter may vary depending on cell density */

		// 2) Test cells to see if they contain the point
		// --> Also invert the deformation using computed trilinear coordinates
		for (int j = 0; j < cellList.size(); j++) {
			if (isContainedInDeformedCell(p, cellList[j], weights)) {
				// Found a cell containing the point!
				cellFound = true;
				cellFoundIndex = cellList[j];
				break;
			}
		}

		currentSearchLevel++;
		// If've we've hit the limits of the grid, expand the grid to widen the search
		if (currentSearchLevel > maxExpandLevel) {
			expandGrid(gridExpandStep);
			updateExpandedDeformation();
		}
	}

	// Check to see that the weights aren't infinity
	bool weightsOK = true;
	for (int j = 0; j < 3; j++) {
		if (!_finite(weights[j])) {
			weightsOK = false;
		}
	}

	return (cellFound && weightsOK);
}


bool UniformGrid3::isContainedInDeformedCell( const real3_type &p, const int3_type &cellIndex, real3_type &weights ) const
{
	// 1. Find the eight cell vertices
	PointList cellPts;

	if (validCell(cellIndex) || (expandedCellLevel.find(cellIndex) != expandedCellLevel.end())) {
		real3_type* hexVerts[8];
		cellPts.resize(1 << dim);

		// Gather the vertices
		for (int i = 0; i < (1 << dim); i++) {
			int3_type offset = offset2sub(i);
			cellPts[i] = deformedGridLocation(cellIndex + offset);		/**< Handles both original and expanded points */
			hexVerts[i] = &(cellPts[i]);
		}

		// Test containment in the cell
		if (hexahedronTestInside(p, hexVerts)) {
			// Also return trilinear coordinates
			weights = invertTrilinear(p, hexVerts);
			return true;
		}
	}

	return false;
}


int3_type UniformGrid3::translateLinearIndex( const index_type &pointIndex ) const
{
	cerr << "Error in  UniformGrid3::translateLinearIndex(): Method not implemented yet!" << endl;

	return int3_type();
}

int3_type UniformGrid3::translateCellLinearIndex( const index_type &cellIndex ) const
{
	cerr << "Error in  UniformGrid3::translateCellLinearIndex(): Method not implemented yet!" << endl;

	return int3_type();
}

real3_type UniformGrid3::trilinearInterpPosition( const real3_type &weights, const int3_type &cellIndex ) const
{
	// Find the function values and interpolate them
	real3_type result;
	result[0] = result[1] = result[2] = 0.0;

	for (int i = 0; i < (1 << dim); i++) {
		int3_type offset = offset2sub(i);
		real_type weight = 1.0;

		// Compute the weight for this grid point (trilinear interpolation)
		for (int j = 0; j < dim; j++) {
			if (offset[j]) {
				weight *= weights[j];
			}
			else {
				weight *= (1-weights[j]);
			}
		}

		// Look up the grid point value and add to the total
		result += weight * gridLocation(cellIndex + offset);
	}

	return result;
}


void UniformGrid3::drawExpandedPts( int level/* = -1*/ ) const
{
#if 0
	// Rotate through colors
	std::vector<RealColor> colorSet;
	colorSet.resize(11);
	colorSet[0] = RealColor(0.0, 0.0, 0.563);
	colorSet[1] = RealColor(0.0, 0.0, 0.938);
	colorSet[2] = RealColor(0.0, 0.313, 1.0);
	colorSet[3] = RealColor(0.0, 0.75, 1.0);
	colorSet[4] = RealColor(0.125, 1.0, 0.875);
	colorSet[5] = RealColor(0.563, 1, 0.438);
	colorSet[6] = RealColor(0.938, 1, 0.0625);
	colorSet[7] = RealColor(1.0, 0.688, 0.0);
	colorSet[8] = RealColor(1.0, 0.25, 0.0);
	colorSet[9] = RealColor(0.875, 0.0, 0.0);
	colorSet[10] = RealColor(0.5, 0.0, 0.0);

	// Highlight interior points
	glDisable(GL_LIGHTING);
	glPointSize(5.f);
	glBegin(GL_POINTS);
	index_type linearIndex;
	int3_type gridIndex;
	real3_type gridPt;
	for ( int3map::const_iterator it = expandedPts.begin(); it != expandedPts.end(); it++ ) {
		gridIndex = (*it).first;
		linearIndex = (*it).second;

		// Check that the point is under the specified maximum level for display
		if ((level < 0) || (expandedPtsLevel[linearIndex] > level)) {
			continue;
		}

		// Color according to the level
		RealColor &picked = colorSet[expandedPtsLevel[linearIndex] % colorSet.size()];
		glColor3f(picked[0], picked[1], picked[2]);

		// Draw the point
		gridPt = deformedGridLocation(gridIndex);
		glVertex3f(gridPt[0], gridPt[1], gridPt[2]);
	}
	glEnd();
#endif

	// Draw expanded cells
	int3_type cellIndex;
	int cellLevel;
	glColor3f(0.8,0.8,0.8);
	glBegin(GL_LINES);
	for ( int3map::const_iterator it = expandedCellLevel.begin(); it != expandedCellLevel.end(); it++ ) {
		cellIndex = (*it).first;
		cellLevel = (*it).second;

		// Check that the cell is under the specified maximum level for display
		if ((level < 0) || (cellLevel > level)) {
			continue;
		}

		// Gather the constituent points
		PointList cellPoints;
		for (int i = 0; i < (1 << dim); i++) {
			cellPoints.add(deformedGridLocation(cellIndex + offset2sub(i)));
		}

		// Draw the edges of the cell
		drawCellEdges(cellPoints);
	}
	glEnd();
}

void UniformGrid3::updateExpandedDeformation()
{
	if (numExpandedPts == 0) {
		return;
	}

	if (currentTransformation.size() != dataDim) {
		// Print something
		cerr << "Warning in UniformGrid3::updateExpandedDeformation(): currentTransformation size doesn't match the number of weights/gridPoint!" << endl;
		return;
	}

	// Initialize list of weights and grid points
	if (expandedData.size() != numExpandedPts) {
		// Create dummy variable
		data_type temp(dataDim);
		expandedData.resize(numExpandedPts, temp);
	}
	if (deformedExpandedPts.size() != numExpandedPts) {
		deformedExpandedPts.resize(numExpandedPts);
	}

	//// Clear ANN data structure
	//deformedExpandedPts.clearANNData();

	// Iterate through each expanded grid point
	data_type weights(dataDim);
	int3_type gridIndex;
	index_type linearIndex;
	for ( int3map::const_iterator it = expandedPts.begin(); it != expandedPts.end(); it++ ) {
		gridIndex = (*it).first;
		linearIndex = (*it).second;

		// Update the weights: take a weighted average of influencing grid points
		weights.setZero();
		int numInfluencingPoints = extrapolatedFrom[linearIndex].size();
		for (int i = 0; i < numInfluencingPoints; i++) {
			weights += data[extrapolatedFrom[linearIndex][i]];
		}
		weights *= (1.0 / ((real_type)numInfluencingPoints));
		expandedData[linearIndex] = weights;

		// Find transformation
		SE3 tf;
		tf.setZero();
		for (int i = 0; i < dataDim; i++) {
			tf.rot += weights.val[i] * currentTransformation[i].rot;
			tf.tr += weights.val[i] * currentTransformation[i].tr;
		}

		// Apply transformation
		real3_type tpos = tf*gridLocation(gridIndex);

		// Store transformed point
		deformedExpandedPts[linearIndex] = tpos;
	}
}

bool UniformGrid3::validPoint( const int3_type &gridIndex ) const
{
	return inBounds(gridIndex);
}

bool UniformGrid3::validCell( const int3_type &cellIndex ) const
{
	return inCellBounds(cellIndex);
}

void UniformGrid3::drawCellEdges( const PointList &pts ) const
{
	// Draw the lines in succession.
	const static int hexEdges[12][2] = {{0,1},{0,2},{0,4},{1,3},{1,5},{2,3},{2,6},{3,7},{4,5},{4,6},{5,7},{6,7}};

	for (int i = 0; i < 12; i++) {
		int i1 = hexEdges[i][0];
		int i2 = hexEdges[i][1];

		glVertex3f(pts[i1][0],pts[i1][1],pts[i1][2]);
		glVertex3f(pts[i2][0],pts[i2][1],pts[i2][2]);
	}
}


void UniformGrid3::expandGrid_OneLayer()
{
	// Expand the grid by one layer
	maxExpandLevel++;

	// 1. Add new grid points
	int3List addedPoints;
	getExpansionList(maxExpandLevel, numDivisions, addedPoints);
	for (int i = 0; i < addedPoints.size(); i++) {
		addExpandedPoint(addedPoints[i]);

		// Debugging code
		//cout << "Added point " << addedPoints[i] << " with influence";
		//
		//const idxList &temp = extrapolatedFrom[expandedPts[addedPoints[i]]];
		//for (int j = 0; j < temp.size(); j++ ){
		//	cout << " " << temp[j];
		//}
		//cout << endl;
	}

	// 2. Add new grid cells
	int3List addedCells;
	getExpansionList(maxExpandLevel, numDivisions - int3_type(1,1,1), addedCells);
	for (int i = 0; i < addedCells.size(); i++) {
		addExpandedCell(addedCells[i]);
	}

	cout << "Expanded layer " << maxExpandLevel << ": " << addedPoints.size() << " new points, " << addedCells.size() << " new cells" << endl;
}


void UniformGrid3::getExpansionList( int expandLevel, int3_type divisions, int3List &result )
{
	int3_type gridIndex;
	result.clear();

	// a) Add bottom floor & top ceiling
	for (int i = (-expandLevel); i < (divisions[0]+expandLevel); i++) {
		for (int j = (-expandLevel); j < (divisions[1]+expandLevel); j++) {
			gridIndex[0] = i; gridIndex[1] = j;
			gridIndex[2] = (-expandLevel);
			result.push_back(gridIndex);
			gridIndex[2] = (divisions[2]+expandLevel-1);
			result.push_back(gridIndex);
		}
	}

	// Add north and south walls
	for (int i = (-expandLevel); i < (divisions[0]+expandLevel); i++) {
		for (int k = (-expandLevel+1); k < (divisions[2]+expandLevel-1); k++) {
			gridIndex[0] = i; gridIndex[2] = k;
			gridIndex[1] = (-expandLevel);
			result.push_back(gridIndex);
			gridIndex[1] = (divisions[1]+expandLevel-1);
			result.push_back(gridIndex);
		}
	}

	// Add east and west walls
	for (int j = (-expandLevel+1); j < (divisions[1]+expandLevel-1); j++) {
		for (int k = (-expandLevel+1); k < (divisions[2]+expandLevel-1); k++) {
			gridIndex[1] = j; gridIndex[2] = k;
			gridIndex[0] = (-expandLevel);
			result.push_back(gridIndex);
			gridIndex[0] = (divisions[0]+expandLevel-1);
			result.push_back(gridIndex);
		}
	}
}

void UniformGrid3::addExpandedPoint( const int3_type &gridIndex )
{
	//////////////////////////////////////////////////////////////////////////
	// 1. Need to find which of the marked grid points influence each interior cell.
	// So first, let's keep a list so that we can query the closest points
	if (!originalExistingPts.size()) {
		int3_type gridIndex = scanStart();
		do {
			// Add the original location
			originalExistingPts.add(gridLocation(gridIndex));
		} while(scanNext(gridIndex));
	}

	//////////////////////////////////////////////////////////////////////////
	// a) Add to the list of expanded points
	expandedPts[gridIndex] = numExpandedPts;

	// b) Query which grid points influence this interior point.
	real3_type location = gridLocation(gridIndex);
	idxList nearest;
	distList dist;
	real_type nearestDist, nearestDistSq;

	// Debugging code
	//cout << "****** Adding expanded point ******" << endl;
	//cout << "Current location: " << gridIndex << endl;

	// Method 1: use the weight at the nearest neighbor
	originalExistingPts.queryKNN(location, 1, nearest, dist, true);					/**< 1. KNN with k=1 to find distance d to closest point */

	// Debugging code
	//cout << "Nearest existing point: " << nearest[0] << " with distance " << dist[0] << endl;

	// This code is in case there are multiple equivalent nearest neighbors (possible since it's a grid).
	// --> In that case, we average the values of all equivalent nearest neighbors.
	nearestDist = dist[0];
	nearestDistSq = dist[0]*dist[0] + 1e-5;
	originalExistingPts.queryRange(location, nearestDistSq, nearest, dist, true);		/**< 2. Range query with d+epsilon to get equidistant closest points */

	// Method 2: average weights at k-nearest neighbors: seems smoother, but may not necessarily be better.
	//originalExistingPts.queryKNN(location, 6, nearest, dist, true);

	// c) Add these grid points to the list
	extrapolatedFrom.push_back(nearest);

	// d) Compute level of these points
	expandedPtsLevel.push_back(ceil(nearestDist / (sqrt(3.0) * sideLength) - 1.0e-5));

	// Increment total number of expanded points
	numExpandedPts++;
}

void UniformGrid3::addExpandedCell( const int3_type &cellIndex )
{
	// a) Compute the level of this cell by taking the maximum level of its constituent points
	int cellLevel = 0;
	for (int i = 0; i < (1 << dim); i++) {
		int3_type ptIndex = cellIndex + offset2sub(i);

		// Find the level of this point
		int level = 1;
		int3map::const_iterator it = expandedPts.find(ptIndex);
		if (it != expandedPts.end()) {
			level = expandedPtsLevel[(*it).second];
		}

		if (level > cellLevel) {
			cellLevel = level;
		}
	}

	// b) Add to the list of expanded cells
	expandedCellLevel[cellIndex] = cellLevel;
}


void UniformGrid3::addExpandedPoint( const int3_type &gridIndex, int level )
{
	//////////////////////////////////////////////////////////////////////////
	// 1. Need to find which of the marked grid points influence each interior cell.
	// So first, let's keep a list so that we can query the closest points
	if (!originalExistingPts.size()) {
		int3_type gridIndex = scanStart();
		do {
			// Add the original location
			originalExistingPts.add(gridLocation(gridIndex));
		} while(scanNext(gridIndex));
	}

	//////////////////////////////////////////////////////////////////////////
	// a) Add to the list of expanded points
	expandedPts[gridIndex] = numExpandedPts;

	// b) Query which grid points influence this interior point.
	real3_type location = gridLocation(gridIndex);
	idxList nearest;
	distList dist;
	real_type nearestDistSq;

	// Debugging code
	//cout << "****** Adding expanded point ******" << endl;
	//cout << "Current location: " << gridIndex << endl;

	// Method 1: use the weight at the nearest neighbor
	originalExistingPts.queryKNN(location, 1, nearest, dist, true);					/**< 1. KNN with k=1 to find distance d to closest point */

	// Debugging code
	//cout << "Nearest existing point: " << nearest[0] << " with distance " << dist[0] << endl;

	// This code is in case there are multiple equivalent nearest neighbors (possible since it's a grid).
	// --> In that case, we average the values of all equivalent nearest neighbors.
	nearestDistSq = dist[0]*dist[0] + 1e-5;
	originalExistingPts.queryRange(location, nearestDistSq, nearest, dist, true);		/**< 2. Range query with d+epsilon to get equidistant closest points */

	// Method 2: average weights at k-nearest neighbors: seems smoother, but may not necessarily be better.
	//originalExistingPts.queryKNN(location, 6, nearest, dist, true);

	// c) Add these grid points to the list
	extrapolatedFrom.push_back(nearest);

	// d) Assign level of this points
	expandedPtsLevel.push_back(level);

	// e) Also store inverse map back to 3d indices
	int3_indices.push_back(gridIndex);

	// Increment total number of expanded points
	numExpandedPts++;
}

void UniformGrid3::expandCell( const int3_type &cellIndex, const int cellLevel )
{
	// Neighbor in each dimension (4-connected neighbors)
	for (int d = 0; d < dim; d++) {
		// Top neighbor & Bottom neighbor
		for (int ni = -1; ni <= 1; ni +=2) {
			// Find index of the neighbor
			int3_type neighborIndex = cellIndex;
			neighborIndex[d] += ni;

			//////////////////////////////////////////////////////////////////////////
			// If it doesn't already exist, expand the cell and the points belonging to the cell
			if (!validCell(neighborIndex)) {
				// If cell doesn't exist among expanded cells
				if (expandedCellLevel.find(neighborIndex) == expandedCellLevel.end()) {
					// Add the cell with the corresponding level
					expandedCellLevel.insert(std::make_pair(neighborIndex, cellLevel+1));

					// Add the points belonging to the cell
					for (int i = 0; i < (1 << dim); i++) {
						int3_type gridIndex = neighborIndex + offset2sub(i);

						// Check if this point already exists
						if (!validPoint(gridIndex)) {
							// If point doesn't exist among expanded point
							if (expandedPts.find(gridIndex) == expandedPts.end()) {
								addExpandedPoint(gridIndex, cellLevel+1);
							}
						}
					}
				}
			}
			//////////////////////////////////////////////////////////////////////////
		}
	}
}


void UniformGrid3::expandGrid( int numLevels )
{
	for (int i = 0; i < numLevels; i++) {
		if (maxExpandLevel == 0) {
			// Iterate through all the current cells once and expand towards its neighbor
			int3_type cellIndex = scanCellStart();
			do {
				expandCell(cellIndex, maxExpandLevel);
			} while(scanNextCell(cellIndex));
		}
		else {
			// Scan through all the expanded cells once
			int3map::const_iterator it;
			for (it = expandedCellLevel.begin(); it != expandedCellLevel.end(); it++) {
				// If the cell has the same level as the current max level, expand from this cell
				if ((*it).second == maxExpandLevel) {
					expandCell((*it).first, maxExpandLevel);
				}
			}
		}

		// Increment current expansion level
		maxExpandLevel++;
		cout << "Expanded grid to level " << maxExpandLevel << ", total expanded cells: " << expandedCellLevel.size() << ", points: " << expandedPts.size() << endl;
	}
}


int3List UniformGrid3::grabClosestDeformedCells( const real3_type &p, int maxLevel, int pointSearchKNN ) const
{
	//////////////////////////////////////////////////////////////////////////
	// First, grab the list of closest original & expanded grid points
	idxList originalNearest, expandedNearest;
	distList originalDists, expandedDists;
	originalNearest.clear(); expandedNearest.clear();
	originalDists.clear(); expandedDists.clear();
	deformedGrid.queryKNN(p, pointSearchKNN, originalNearest, originalDists, true);
	if (deformedExpandedPts.size() > pointSearchKNN) {
		deformedExpandedPts.queryKNN(p, pointSearchKNN, expandedNearest, expandedDists, true);
	}

	//////////////////////////////////////////////////////////////////////////
	// Merge the two lists and map the linear indices back to the 3d indices
	std::vector< std::pair<int3_type,real_type> > allPoints;
	allPoints.reserve(originalNearest.size()+expandedNearest.size());
	for (int i = 0; i < originalNearest.size(); i++) {
		allPoints.push_back(std::make_pair(translateLinearIndex(originalNearest[i]),originalDists[i]));
	}
	for (int i = 0; i < expandedNearest.size(); i++) {
		allPoints.push_back(std::make_pair(int3_indices[expandedNearest[i]],expandedDists[i]));
	}

	//////////////////////////////////////////////////////////////////////////
	// Sort the list according to distance
	struct SortedPointsCompare : public std::binary_function<std::pair<int3_type,real_type>,std::pair<int3_type,real_type>,bool> {
		bool operator()( const std::pair<int3_type,real_type> &x, const std::pair<int3_type,real_type> &y ) {
			// Sort based on distance
			return (x.second < y.second);
		}
	};
	std::sort(allPoints.begin(), allPoints.end(), SortedPointsCompare());

	//////////////////////////////////////////////////////////////////////////
	// Find the cells adjacent to each point that is within the desired level
	int3List cellList;

	for (int i = 0; i < allPoints.size(); i++) {
		for (int j = 0; j < (1 << dim); j++) {
			int3_type cellIndex = allPoints[i].first - offset2sub(j);

			// Find the level
			int level = 0;
			if (!validCell(cellIndex)) {
				int3map::const_iterator it = expandedCellLevel.find(cellIndex);
				if (it != expandedCellLevel.end()) {
					level = (*it).second;
				}
				else {
					level = -1;
				}
			}

			// Add to the list if the level is valid & within the maximum
			if ((level >= 0) && (level <= maxLevel)) {
				cellList.push_back(cellIndex);
			}
		}
	}

	return cellList;
}

const std::vector<data_type>& UniformGrid3::getData() const
{
	return data;
}

void UniformGrid3::setData( const std::vector<data_type>& data )
{
	this->data = data;
}

real_type UniformGrid3::getCellSize() const
{
	return sideLength;
}

data_type UniformGrid3::getValues_beyondGrid( const int3_type &gridIndex )
{
	// 1. Try to see if it's in the original grid
	// 2. If not, find the closest point in the original grid and use its weights
	data_type result(dataDim);

	if (validPoint(gridIndex)) {
		// Grab the vertex weight value directly
		for (int i = 0; i < dataDim; ++i) {
			result.val[i] = getValue(gridIndex, i);
		}
	}
	else {
		//////////////////////////////////////////////////////////////////////////
		// Otherwise, the grid index is beyond the existing set of points.
		// IMPORTANT NOTE: This code is borrowed from code used to extend the grid
		// i.e. void addExpandedPoint( const int3_type &gridIndex, int level ).
		//////////////////////////////////////////////////////////////////////////

		// 1. Need to find which of the marked grid points influence each interior cell.
		// So first, let's keep a list so that we can query the closest points
		if (!originalExistingPts.size()) {
			int3_type gridIndex = scanStart();
			do {
				// Add the original location
				originalExistingPts.add(gridLocation(gridIndex));
			} while(scanNext(gridIndex));
		}

		// Query which grid points influence this interior point.
		real3_type location = gridLocation(gridIndex);
		idxList nearest;
		distList dist;

		// Method 1: use the weight at the nearest neighbor(s).
		originalExistingPts.queryKNN(location, 1, nearest, dist, true);						/**< 1. KNN with k=1 to find distance d to closest point */

		// This code is in case there are multiple equivalent nearest neighbors (possible since it's a grid).
		real_type nearestDistSq = dist[0]*dist[0] + REAL_TYPE_EPSILON;
		originalExistingPts.queryRange(location, nearestDistSq, nearest, dist, true);		/**< 2. Range query with d+epsilon to get equidistant closest points */

		// --> In that case, we average the values of all equivalent nearest neighbors.
		result = averageExtrapolatedPoints(nearest);
	}

	return result;
}

data_type UniformGrid3::averageExtrapolatedPoints( const idxList& extrapLinearIndices ) const
{
	data_type result(dataDim);

	int numInfluencingPoints = extrapLinearIndices.size();
	for (int i = 0; i < numInfluencingPoints; i++) {
		result += data[extrapLinearIndices[i]];
	}
	result *= (1.0 / ((real_type)numInfluencingPoints));

	return result;
}


void UniformGrid3::copyColors( const UniformGrid3 &grid )
{
	compColors = grid.compColors;
}

void UniformGrid3::generateColors()
{
	// Generate colors for each bone
	generateLightColors(dataDim, compColors);
}

const std::vector<RealColor>& UniformGrid3::getCompColors() const
{
	return compColors;
}

data_type UniformGrid3::getExtendedCellValue( const int3_type& cellIndex, const real3_type& weights ) const
{
	if (validCell(cellIndex)) {
		return getValue(cellIndex, weights);
	}
	else if (expandedCellLevel.find(cellIndex) != expandedCellLevel.end()) {
		// Find the function values and interpolate them
		data_type value(dataDim);
		for (int i = 0; i < (1 << dim); i++) {
			int3_type offset = offset2sub(i);
			real_type weight = 1;

			// Compute the weight for this grid point (trilinear interpolation)
			for (int j = 0; j < dim; j++) {
				if (offset[j]) {
					weight *= weights[j];
				}
				else {
					weight *= (1-weights[j]);
				}
			}

			// Look up the grid point value and add to the total
			int3_type gridIndex = cellIndex + offset;
			if (validPoint(gridIndex)) {
				value += weight * data[getLinearIndex(gridIndex)];
			}
			else {
				int3map::const_iterator it = expandedPts.find(gridIndex);
				if (it == expandedPts.end()) {
					cerr << "Error in getExtendedCellValue(): can't find point " << offset << " for extended cell " << cellIndex << endl;
				}

				value += weight * averageExtrapolatedPoints(extrapolatedFrom[it->second]);
			}
		}

		return value;
	}

	cerr << "Error in getExtendedCellValue(): cell index doesn't exist in original or extended grid." << endl;
	return data_type(dataDim);
}

void sortedListUnion( const std::vector<int> &first, const std::vector<int> &second, std::vector<int> &result )
{
	result.resize(first.size() + second.size());

	std::vector<int>::iterator it = set_union(first.begin(), first.end(), second.begin(), second.end(), result.begin());

	result.resize(int(it - result.begin()));
}

void sortedListUnionMerge( const std::vector<int> &from, std::vector<int> &to ) {
	std::vector<int> temp;
	sortedListUnion(from, to, temp);
	to = temp;
}
