///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __UNIFORM_GRID_H__
#define __UNIFORM_GRID_H__

#include "types.h"
#include "FindCorrespondence.h"
#include <QtOpenGL>

class CorrespondenceList;

/**
 * A uniform grid structure to interpolate data.
 * The template parameter is for the number of dimensions of each data point.
 *
 * Traversal code template:
 * 
 * 	 int3_type gridIndex = vWeightGrid->scanStart();
 * 	 do {
 *
 *       // Insert per-grid point operation here!
 * 	
 * 	 } while(vWeightGrid->scanNext(gridIndex));
 * 
 *
 * Neighbor traversal template:
 * 
 * 	int3List neighbors = vWeightGrid->findNeighbors(gridIndex);
 * 	for (unsigned int i = 0; i < neighbors.size(); i++) {
 * 	
 * 		// Insert per-neighbor operation here!
 * 	
 * 	}
 * 	
 */
class UniformGrid3 {
public:
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Creation
	UniformGrid3(int dataDim = 1);															/**< Empty grid default constructor. */
	void createFromBBox( real3_type bboxMin, real3_type bboxMax, int div );					/**< Create a grid with specified bounding box and number of divisions. */
	void createFromBBox( real3_type bboxMin, real3_type bboxMax, real_type sideLength );	/**< Create a grid with specified bounding box and cell size. */
	void copyColors( const UniformGrid3 &grid );											/**< Copy the colors from another grid. */
	void generateColors();


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Display
	virtual void draw( int comp = 0, bool deformed = true ) const;							/**< Draw the grid in OpenGL. */
	void drawGridPoint( int3_type gridIndex, int comp = 0, bool deformed = true ) const;	/**< Draw an individual grid point. */
	void drawCellEdges( const PointList &pts ) const;										/**< Draw an outline of a given cell. */
	void drawExpandedPts( int level = -1 ) const;											/**< Draw expanded points (deformed version, if available). */
	const std::vector<RealColor>& getCompColors() const;


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Size & Format
	int3_type size() const;
	virtual void resetData();
	size_type getDataDim() const;
	virtual index_type totalGridPoints() const;
	virtual index_type totalGridCells() const;
	bool withinGrid( real3_type query ) const;									/**< Is the position within the grid? */
	virtual bool validPoint( const int3_type &gridIndex ) const;
	virtual bool validCell( const int3_type &cellIndex ) const;
	real_type getCellSize() const;


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Geometry
	real3_type gridLocation( const int3_type &ind ) const;													/**< The 3D location of a grid point */
	virtual void getCellGridPoints( int3_type cellIndex, idxList &ptIndices ) const;						/**< Get grid points belonging to this cell  (Note: Not safe for extended grid). */
	virtual void getCellBox( const int3_type &cellIndex, real3_type &minPt, real3_type &maxPt ) const;		/**< Returns the minimum and maximum position of the grid cell in question. */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Values
	void queryCell( const real3_type query, int3_type &index, real3_type &remainder ) const;
	data_type getValue(real3_type query) const;
	data_type getValue( const int3_type& cellIndex, const real3_type& weights ) const;
	void getWeights(real3_type query, idxList &gridIndices, distList &gridWeights);
	void setAllValues( index_type linearIndex, const data_type &value );
	void setValues( real3_type query, int comp, real_type value );
	void setValue(int3_type ind, index_type comp, real_type value);				/**< Set a component of the data value at a grid point (Note: Not safe for extended grid). */
	real_type getValue(int3_type ind, index_type comp) const;					/**< Get a component of the data value (Note: Not safe for extended grid). */
	data_type cellNonzeroValue( int3_type cellIndex ) const;					/**< Find non-zero data value components for this cell (Note: Not safe for extended grid). */
	void normalizeValues();														/**< Normalize the data values at each grid point. */
	void clampMaxValues();														/**< Make the data values binary with maximum component = 1, rest = 0. */
	void getVertexWeights( const CorrespondenceList *con, std::vector<idxList> &nodeList, std::vector<distList> &vertexWeight ) const;
	void getVertexWeights( const real3_type& pt, idxList &bones, distList &weights ) const;
	const std::vector<data_type>& getData() const;									/**< Access the raw data */
	void setData( const std::vector<data_type>& data );								/**< Set the raw data */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Traversal
	virtual int3_type scanStart() const;
	virtual int3_type scanCellStart() const;
	virtual bool scanNext(int3_type &gridIndex) const;										/**< Replaces the argument with the next grid point, and returns false if we are at the end. */
	virtual bool scanNextCell(int3_type &cellIndex) const;									/**< Replaces the argument with the next grid cell, and returns false if we are at the last cell. */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Indexing
	virtual index_type getLinearIndex( const int3_type &gridIndex ) const;					/**< Get the linear index corresponding to the specified grid point index. */
	virtual index_type getCellLinearIndex( const int3_type &cellIndex ) const;				/**< Get the linear index corresponding to the specified grid cell index. */
	virtual int3_type translateLinearIndex( const index_type &pointIndex ) const; 
	virtual int3_type translateCellLinearIndex( const index_type &cellIndex ) const; 


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Neighbors
	virtual int3List findNeighbors(int3_type gridIndex) const;								/**< Find points neighboring the specified grid point (within the grid) */
	virtual int3List findAdjacentCellsToPoint(int3_type gridPointIndex) const;				/**< Find cells adjacent to the specified grid point (within the grid) */
	virtual int3List findAdjacentCellsToCell(int3_type gridCellIndex) const;				/**< Find cells adjacent to the specified grid cell (within the grid) */


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Deformation
	void updateDeformation( const RigidDeform &tfList );
	real3_type deformedGridLocation( const int3_type &ind ) const;
	void inverseDeform( ShapeModel &shape, int maxSearchLevel, int gridExpandStep );
	bool inverseDeformPoint( const real3_type &p, int3_type &cellFoundIndex, real3_type &weights, int maxSearchLevel, int gridExpandStep );
	bool isContainedInDeformedCell( const real3_type &p, const int3_type &cellIndex, real3_type &weights ) const;
	real3_type trilinearInterpPosition( const real3_type &weights, const int3_type &cellIndex ) const;
	int3List grabClosestDeformedCells( const real3_type &p, int maxLevel, int pointSearchKNN ) const;


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Expansion Notes
	// 1. The idea is that we keep a list of points that are outside the grid.
	// 2. We map the 3d coordinates of these points into linear indices, which are used to access the following data:
	//    a. List of existing grid points that contribute their value to the expanded point
	//    b. Precomputed values at expanded grid points
	//    c. Precomputed deformed location of expanded grid points
	//
	// (a) is used for updating the data values at each grid point, which can then be used to precompute the values and the deformed locations.
	// Once we have these expanded cells, we can use them to inversely map points outside the original grid.
	// This will require updating the grid neighbors code to handle the expanded set of points.

	// MarkedUniformGrid just relies on the code in UniformGrid to handle new cells outside the original bounds.
	// By default, MarkedUniformGrid keeps track of in all interior cells that are not marked as expanded cells,
	// And keeps it in the same data structure.

	// Implementation schedule
	// 1. Code for MarkedUniformGrid to automatically handle unmarked interior cells
	// 2. Code to update values / deformed locations, debug and display (1)
	// 3. Code to expand beyond original bounds (in UniformGrid)
	// 4. Code to update values / deformed locations, debug and display (3)
	// 5. Modified neighbor query code
	// 6. Update inverse deformation code

	void updateExpandedDeformation();														/**< Use the list of current transformations to update the deformation at expanded points */
	void expandGrid( int numLevels );														/**< Expand the grid a specified number of levels. */
	data_type getValues_beyondGrid( const int3_type &gridIndex );							/**< Get data values, possibly beyond the existing grid. */
	data_type getExtendedCellValue( const int3_type& cellIndex, const real3_type& weights ) const;		/**< Get data values from one of the extended cells */

	// Deprecated
	void expandGrid_OneLayer();																/**< Expand the grid by one more layer of cells */
	// Deprecated
	void getExpansionList( int expandLevel, int3_type divisions, int3List &result );		/**< Get a list of points to expand */


protected:
	bool inBounds( const int3_type &gridIndex ) const;					/**< Is the index within bounds? */
	bool inCellBounds( const int3_type &cellIndex ) const;				/**< Is the cell index within bounds? */
	inline index_type sub2ind(int3_type sub, int3_type size) const;		/**< Convert to a linear index. */
	inline int3_type offset2sub(index_type ind) const;					/**< Convert integer to a 3D offset. */

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Dimensions
	static const int dim = 3;					/**< Number of Spatial Dimensions */
	const int dataDim;							/**< Number of Dimensions for grid values */
	real3_type minPos;							/**< The minimum position of the grid (the starting point) */ 
	real_type sideLength;						/**< The extent of each grid cell */
	int3_type numDivisions;						/**< The extent of the entire grid as # points per axis */

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Grid Data
	std::vector<data_type> data;				/**< Stores the vector of values at each grid point */
	std::vector<RealColor> compColors;			/**< Colors for each value component */

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Deformation Data
	RigidDeform currentTransformation;			/**< Current transformation applied to the grid */
	PointList deformedGrid;						/**< Precomputed deformed grid points */

	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Expansion Data: for inverse map outside the original grid
	typedef std::map<int3_type,index_type,int3Comp> int3map;	/**< Mapping from 3d indices to linear indices */
	int maxExpandLevel;											/**< Keep track of how many layers we've expanded the grid so far. */
	int numExpandedPts;											/**< Number of expanded grid points */
	int3map expandedPts;										/**< Expanded grid points (mapping to linear index) */
	std::vector<idxList> extrapolatedFrom;						/**< List of existing grid points that contribute their value to the expanded point */
	std::vector<data_type> expandedData;						/**< Precomputed values at expanded grid points */
	PointList deformedExpandedPts;								/**< Precomputed deformed location of expanded grid points */
	std::vector<int> expandedPtsLevel;							/**< Level of each expanded point */
	std::vector<int3_type> int3_indices;						/**< Mapping from linear indices to 3d indices (inverse of expandedPts) */
	int3map expandedCellLevel;									/**< Level of expanded cells */

	// Deprecated
	void addExpandedPoint( const int3_type &gridIndex );		/**< Add a new point to the list of expanded points */
	// Deprecated
	void addExpandedCell( const int3_type &cellIndex );			/**< Add a new cell to the list of expanded cells */

	void addExpandedPoint( const int3_type &gridIndex, int level );
	void expandCell( const int3_type &cellIndex, const int cellLevel );
	data_type averageExtrapolatedPoints( const idxList& extrapLinearIndices ) const;

	PointList originalExistingPts;								/**< Helper variable for addExpandedPoint(): list of original points in the grid, so that we can determine the level of each point/cell */
};


/**
* Intended to create an offset of the index for values of ind = 0 ... 7.
*/
inline int3_type UniformGrid3::offset2sub( index_type ind ) const
{
	int3_type sub;

	sub[2] = ((int) ind) & 1;
	sub[1] = (((int) ind) >> 1) & 1;
	sub[0] = (((int) ind) >> 2) & 1;

	return sub;
}

inline index_type UniformGrid3::sub2ind( int3_type sub, int3_type size ) const
{
	return sub[2] + size[2]*(sub[1] + size[1]*sub[0]);
}


// Utility function for union-ing two integer lists
void sortedListUnion( const std::vector<int> &first, const std::vector<int> &second, std::vector<int> &result );
void sortedListUnionMerge( const std::vector<int> &from, std::vector<int> &to );

#endif

