///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include <QString>
#include "deform.h"
#include "util.h"
#include <iostream>
#include <iomanip>

#undef min
#undef max

using namespace std;
using namespace Deform;

template class SparseMatrix<float>;

template<typename T>
SparseMatrix<T>::SparseMatrix() : maxRow(0), maxCol(0), currentSize(0)
{
	data.clear();
	index.clear();
}

template<typename T>
size_type SparseMatrix<T>::dataSize()
{
	if (data.size() != index.size()) {
		throw "SparseMatrix<T>::size(): Data and index sizes do not match!";
	}

	return data.size();
}

template<typename T>
size_type SparseMatrix<T>::size()
{
	return currentSize;
}

template<typename T>
void SparseMatrix<T>::fill( index_type row, index_type col, T val )
{
	fill(index2_type(row, col), val);
}

template<typename T>
void SparseMatrix<T>::fill( index2_type idx, T val )
{
	// Push back the value and the index
	if (size() < dataSize()) {
		// Just fill in the value
		data[currentSize] = val;
		index[currentSize] = idx;
	}
	else {
		data.push_back(val);
		index.push_back(idx);
	}
	if (idx[0] > maxRow) {
		maxRow = idx[0];
	}
	if (idx[1] > maxCol) {
		maxCol = idx[1];
	}

	currentSize++;
}

template<typename T>
void SparseMatrix<T>::get( size_type seq, index2_type &idx, T &val ) const
{
	if (seq >= currentSize) {
		throw "SparseMatrix<T>::get(): non-zero element out of range.";
	}
	idx = index[seq];
	val = data[seq];
}

template<typename T>
void SparseMatrix<T>::clear()
{
	data.clear();
	index.clear();
	maxCol = maxRow = currentSize = 0;
}

template<typename T>
void SparseMatrix<T>::resize( size_type n )
{
	clear();
	data.resize(n);
	index.resize(n);
}

template<typename T>
void SparseMatrix<T>::print()
{
	// Print in Matlab format; change to 1-based indices
	for (unsigned int i = 0; i < size(); i++) {
		cout << "M(" << index[i][0]+1 << "," << index[i][1]+1 << ") = " << data[i] << ";" << endl;
	}
}

template<typename T>
void SparseMatrix<T>::write( const char *filename ) const
{
	std::ofstream out;
	out.open(filename);

	if (out.fail()) {
		std::cerr << "Error in SparseMatrix<T>::write(): Could not open output file " << filename << " for writing!" << std::endl;
		return;
	}

	// Print in Matlab format; change to 1-based indices
	for (unsigned int i = 0; i < currentSize; i++) {
		out << index[i][0]+1 << " " << index[i][1]+1 << " " << (taucsType) data[i] << endl;
	}

	out.close();
}

template<typename T>
void SparseMatrix<T>::getTaucsColumns( std::vector< std::map<int,taucsType> > &result, size_type nCols ) const
{
	result.clear();
	result.resize(nCols);

	// gather the data and put them into columns
	for (size_type i = 0; i < currentSize; i++) {
		index2_type idx = index[i];
		result[idx[1]][idx[0]] = (taucsType) data[i];
	}
}

// Sort ascending: first column, then row
struct index2_colAsc : public std::binary_function<index2_type,index2_type,bool> {
	bool operator()(const index2_type &x, const index2_type &y) const {
		if (x[1] == y[1])
			return (x[0] < y[0]);
		else
			return (x[1] < y[1]);
	}
};

// Sort ascending: first row, then column
struct index2_rowAsc : public std::binary_function<index2_type,index2_type,bool> {
	bool operator()(const index2_type &x, const index2_type &y) const {
		if (x[0] == y[0])
			return (x[1] < y[1]);
		else
			return (x[0] < y[0]);
	}
};

// Help from http://objectmix.com/c/358470-how-parallel-sort.html to perform sort on indices of the array
template < typename Sequence, class Predicate >
class ForwardCompare {
	Sequence const & the_sequence;
	Predicate const & the_predicate;

public:
	ForwardCompare ( Sequence const & ref, Predicate & pred ): the_sequence ( ref ), the_predicate ( pred ) { }
	bool operator() ( typename Sequence::size_type lhs,
		typename Sequence::size_type rhs ) const {
			return the_predicate(the_sequence[lhs], the_sequence[rhs]);
	}
};

template < typename Sequence, class Predicate >
ForwardCompare< Sequence, Predicate > make_ForwardCompare ( Sequence const & seq, Predicate & pred ) {
	return ( ForwardCompare< Sequence, Predicate >( seq, pred ) );
}


template<typename T>
taucs_ccs_matrix * SparseMatrix<T>::writeTaucsMatrix(size_type nRows, size_type nCols, bool transposed/* = false*/ ) const
{
	//// Find the columns and use the CreateTaucsMatrixFromColumns() utility function.
	//std::vector< std::map<int,taucsType> > cols;
	//getTaucsColumns(cols, nCols);
	//return CreateTaucsMatrixFromColumns(cols, nRows, TAUCS_DOUBLE);		/**< Currently the taucsType is double; should follow the type defined in taucsaddon.h */

	//cout << "Total number of non-zeros: " << currentSize << endl;

	// Alternative algorithm: avoid creating a vector of maps
	std::vector<int> linearIndex(currentSize);
	for (int i = 0; i < linearIndex.size(); ++i) {
		linearIndex[i] = i;
	}

	// Sort the entries
	if (transposed) {
		std::sort(linearIndex.begin(), linearIndex.end(), make_ForwardCompare(index, index2_rowAsc()));
	}
	else {
		std::sort(linearIndex.begin(), linearIndex.end(), make_ForwardCompare(index, index2_colAsc()));
	}

	// count nnz:
	int nnz = currentSize;

	taucs_ccs_matrix *matC;
	if (transposed) {
		matC = taucs_ccs_create(nCols,nRows,nnz,TAUCS_DOUBLE);
	}
	else {
		matC = taucs_ccs_create(nRows,nCols,nnz,TAUCS_DOUBLE);
	}
	if (! matC)
		return NULL;

	// copy cols into matC
	std::map<int,taucsType>::const_iterator rit;
	int lastColumn = -1;
	int rowptrC = 0;
	for (int i = 0; i < linearIndex.size(); ++i) {
		index_type row, col;
		if (transposed) {
			row = index[linearIndex[i]][1];
			col = index[linearIndex[i]][0];
		}
		else {
			row = index[linearIndex[i]][0];
			col = index[linearIndex[i]][1];
		}

		// Check if we're beginning a new column
		if (lastColumn != col) {
			for (int c = lastColumn+1; c <= col; ++c) {
				matC->colptr[c] = rowptrC;
			}
			lastColumn = col;
		}

		// Populate the entry
		matC->rowind[rowptrC] = row;
		matC->taucs_values[rowptrC] = data[linearIndex[i]];
		++rowptrC;
	}
	if (transposed) {
		for (int c = lastColumn+1; c <= nRows; ++c) {
			matC->colptr[c] = rowptrC;
		}
	}
	else {
		for (int c = lastColumn+1; c <= nCols; ++c) {
			matC->colptr[c] = rowptrC;
		}
	}

	return matC;
}


ErrorTerm::ErrorTerm() : rows(0), rhs(NULL), weight(0)
{

}

ErrorTerm::ErrorTerm( real_type weight, size_type rows, size_type totalNonZero ) :
weight(weight), rows(rows)
{
	rhs = new real_type[rows];
	lhs.resize(totalNonZero);
	rowWeight.resize(rows, 1.0);			/**< By default, the rows are weight = 1.0 */
}

ErrorTerm::~ErrorTerm()
{
	clear();
}

void ErrorTerm::clear()
{
	rows = 0;
	weight = 0;
	lhs.clear();
	if (rhs) {
		delete[] rhs;
		rhs = NULL;
	}
	rowWeight.clear();
}

ErrorTerm* Deform::mergeTerms( TermList terms )
{
	// Merge all the terms together
	ErrorTerm *result = new ErrorTerm;

	// Determine the total number of terms
	size_type totalRows, totalNNZ;
	totalNNZ = totalRows = 0;

	for (size_type i = 0; i < terms.size(); i++) {
		totalRows += terms[i]->rows;
		totalNNZ += terms[i]->lhs.size();
	}

	// Allocate the space
	result->weight = 1;
	result->rows = totalRows;
	result->rhs = new real_type[result->rows];
	result->lhs.resize(totalNNZ);

	index2_type idx;
	real_type val;

	// Copy the terms with the weights multiplied
	size_type trackRow = 0;
	size_type startRow = 0;
	for (size_type i = 0; i < terms.size(); i++) {
		real_type termWeight = terms[i]->weight;
		real_type rowWeight = 1.0;
		bool rowWeightExists = (terms[i]->rowWeight.size() == terms[i]->rows);
		std::vector<int> trackRowMapping(terms[i]->rows, -1);

		//// Debugging
		//if (rowWeightExists) {
		//	for (int row = 0; row < terms[i]->rows; ++row) {
		//		cout << "Weight of row " << row << ": " << terms[i]->rowWeight[row] << endl;
		//	}
		//}

		for (size_type row = 0; row < terms[i]->rows; row++) {
			if (rowWeightExists) {
				rowWeight = terms[i]->rowWeight[row];
			}
			if (rowWeight > 0.0) {
				result->rhs[trackRow] = termWeight * rowWeight * terms[i]->rhs[row];
				trackRowMapping[row] = trackRow;
				trackRow++;
			}
		}

		//// Debugging
		//for (int row = 0; row < terms[i]->rows; ++row) {
		//	cout << "Mapping " << row << " -> " << trackRowMapping[row] << endl;
		//}

		SparseMatrix<real_type> &thisLHS = terms[i]->lhs;
		for (size_type elt = 0; elt < thisLHS.size(); elt++) {
			thisLHS.get(elt, idx, val);

			if (rowWeightExists) {
				rowWeight = terms[i]->rowWeight[idx[0]];
			}

			if (rowWeight > 0.0) {
				// To get the index in the merged system: modify just the row
				//idx[0] += startRow;
				idx[0] = trackRowMapping[idx[0]];

				// Populate the matrix
				result->lhs.fill(idx, termWeight * rowWeight * val);
			}
		}

		startRow = trackRow;
	}

	////////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < result->rows; ci++) {
	//	cout << "V[" << ci << "] = " << result->rhs[ci] << ";" << endl;
	//}
	//result->lhs.print();
	//cout << "Estimated size for the LHS (all terms): " << totalNNZ
	//	 << " vs. Actual size: " << result->lhs.size() << endl;

	return result;
}
