///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __DEFORM_H__
#define __DEFORM_H__

#include <QFile>
#include <QTextStream>
#include "AbstractDeform.h"
#include "util.h"
#include "CorrespViewer.h"
#include "PointList.h"
#include "ShapeModel.h"
#include "FindCorrespondence.h"
#include <SylvainTools/simple_mesh.h>
#include <taucsaddon.h>

namespace Deform {

	/**
	* Sparse matrix class for generating matrices in CCS format.
	* This is a dynamically growing matrix, with 0-based indexing.
	*/
	template<typename T>
	class SparseMatrix {
	public:
		SparseMatrix();

		size_type size();				/**< Number of non-zero elements */
		void resize(size_type n);		/**< Reserve space for n elements */
		void clear();
		void fill(index_type row, index_type col, T val);
		void fill(index2_type idx, T val);
		void print();
		void write( const char *filename ) const;
		void get(size_type seq, index2_type &idx, T &val) const;
		void getTaucsColumns( std::vector< std::map<int,taucsType> > &result, size_type nCols ) const;
		taucs_ccs_matrix * writeTaucsMatrix(size_type nRows, size_type nCols, bool transposed = false ) const;

		inline size_type getMaxRow() const {
			return maxRow;
		}

		inline size_type getMaxCol() const {
			return maxCol;
		}

	private:
		size_type dataSize();			/**< Total reserved space for storage */

		size_type currentSize;			/**< Keep track of the number of non-zero elements */
		size_type maxRow;
		size_type maxCol;
		std::vector<T> data;
		std::vector<index2_type> index;
	};

	/**
	* Class for holding each error term.
	*/
	class ErrorTerm {
	public:
		ErrorTerm();
		ErrorTerm( real_type weight, size_type rows, size_type totalNonZero );
		~ErrorTerm();
		void clear();

		SparseMatrix<real_type> lhs;
		real_type *rhs;
		size_type rows;
		real_type weight;
		std::vector<real_type> rowWeight;
	};

	typedef std::vector<ErrorTerm*> TermList;

	// Merge multiple error terms
	ErrorTerm* mergeTerms(TermList terms);


	/**
	* A class useful for passing drawing options for the various DeformAlgorithm classes.
	*/
	class DrawOptions
	{
	public:
		bool showSourceModel;
		bool showTargetModel;
		bool showDeformedModel;
		bool showSourceCorresp;
		bool showTargetCorresp;
		bool showDeformedCorresp;
		bool showShapeWeights;
		bool showAllFrames;
		bool showGraph;
		bool showEffRad;
		bool renderTransparent;
		int showComponent;
		bool showGraphDeformed;
		bool showWireframe;
		bool showGraphFaces;
		int showExtendedGrid;
		float openGLPointSize;
		RealColor sourceModelColor;
		RealColor targetModelColor;
		RealColor deformedModelColor;
		RealColor pointColor;
		bool showDeformedModelPoints;
		int showErrorComponent;
		int displayFrame;
		bool ignoreDisplayInterface;
	};
}


#endif
