///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <limits>
#include "geom_util.h"
#include "ShapeModel.h"

using namespace std;

real_type pointBoxSqDistance ( const real3_type &p, const real3_type &boxMin, const real3_type &boxMax, real3_type &closestPoint )
{
	// work in the box's coordinate system
	real3_type boxCenter = (boxMin + boxMax) / (real_type)2.0;
	real3_type halfLength = (boxMax - boxMin) / (real_type)2.0;
	real3_type kDiff = p - boxCenter;

	// compute squared distance and closest point on box
	real_type fSqrDistance = (real_type)0.0, fDelta;
	real3_type kClosest;

	for (int i = 0; i < 3; i++)
	{
		kClosest[i] = kDiff[i];
		if (kClosest[i] < -halfLength[i])
		{
			fDelta = kClosest[i] + halfLength[i];
			fSqrDistance += fDelta*fDelta;
			kClosest[i] = -halfLength[i];
		}
		else if (kClosest[i] > halfLength[i])
		{
			fDelta = kClosest[i] - halfLength[i];
			fSqrDistance += fDelta*fDelta;
			kClosest[i] = halfLength[i];
		}
	}

	closestPoint = boxCenter + kClosest;

	return fSqrDistance;
}


real_type lowerBoundTriangleSqDistance( const real3_type &p, const real3_type &t0, const real3_type &t1, const real3_type &t2 )
{
	// Find bounding box of triangle
	static real3_type boxMin, boxMax;
	boxMin[0] = boxMin[1] = boxMin[2] = REAL_TYPE_MAX;
	boxMax[0] = boxMax[1] = boxMax[2] = -REAL_TYPE_MAX;

	for (int d = 0; d < 3; d++) {
		if (t0[d] < boxMin[d])      boxMin[d] = t0[d];
		else if (t0[d] > boxMax[d]) boxMax[d] = t0[d];
		if (t1[d] < boxMin[d])      boxMin[d] = t1[d];
		else if (t1[d] > boxMax[d]) boxMax[d] = t1[d];
		if (t2[d] < boxMin[d])      boxMin[d] = t2[d];
		else if (t2[d] > boxMax[d]) boxMax[d] = t2[d];
	}

	//////////////////////////////////////////////////////////////////////////
	// Method 1: Lower-bound via distance to bounding box of triangle
	static real3_type temp;
	return pointBoxSqDistance(p, boxMin, boxMax, temp);

	//////////////////////////////////////////////////////////////////////////
	// Method 2: Approximate bounding sphere as center of box with length half of diagonal
	//real3_type center = (boxMax + boxMin) / (real_type)2.0;
	//real_type r = (boxMax - boxMin).norm() / (real_type)2.0;
	//real3_type diff = p - center;
	//return max(0.0, diff*diff - r*r);
}


// inverts a 3x3 matrix
void invert(Geometry::Matrix33f &mat) {
	float a00, a01, a02;
	float a10, a11, a12;
	float a20, a21, a22;

	a00 = mat(0,0); a10 = mat(1,0); a20 = mat(2,0);
	a01 = mat(0,1); a11 = mat(1,1); a21 = mat(2,1);
	a02 = mat(0,2); a12 = mat(1,2); a22 = mat(2,2);

	float div = 1/(a00*a11*a22-a00*a21*a12-a10*a01*a22+a10*a21*a02+a20*a01*a12-a20*a11*a02);

	mat(0,0) = (a11*a22-a21*a12)*div;
	mat(1,0) = -(a10*a22-a20*a12)*div;
	mat(2,0) = (a10*a21-a20*a11)*div;

	mat(0,1) = -(a01*a22-a21*a02)*div;
	mat(1,1) = (a00*a22-a20*a02)*div;
	mat(2,1) = -(a00*a21-a20*a01)*div;

	mat(0,2) = (a01*a12-a11*a02)*div;
	mat(1,2) = -(a00*a12-a10*a02)*div;
	mat(2,2) = (a00*a11-a10*a01)*div;
}



void bilinearInvert3D(const real3_type &p,
					  const real3_type &v0,
					  const real3_type &v1,
					  const real3_type &v2,
					  const real3_type &v3,
					  real_type &alpha,
					  real_type &beta)
{
	real3_type n = cross(v1-v2,v0-v3);
	real_type A = n * (cross(v0-v2,v1-v3));
	real_type B = n * (cross(v1-v3,p-v2) - cross(v0-v2,p-v3));
	real_type C = n * (cross(p-v2,p-v3));

	if (B > 0) {
		beta = 0.5 * (-B - sqrt(B*B-4.0*A*C)) / A;
	}
	else {
		beta = 2.0 * C / (-B + sqrt(B*B-4.0*A*C));
	}

	real3_type s1 = (1.0 - beta)*v2 + beta*v0;
	real3_type s2 = (1.0 - beta)*v3 + beta*v1;

	alpha = ((p-s1)*(s2-s1)) / ((s2-s1)*(s2-s1));
}

real3_type trilinearNewtonUpdate( const real3_type &p, real3_type **verts, const real3_type &weights, real3_type &f )
{
	Geometry::Matrix33f J;
	real_type t1, t2, t3;
	const real_type &x = p[0];
	const real_type &y = p[1];
	const real_type &z = p[2];
	const real_type &a = weights[0];
	const real_type &b = weights[1];
	const real_type &c = weights[2];
	const real_type &x_0 = (*verts[0])[0]; const real_type &y_0 = (*verts[0])[1]; const real_type &z_0 = (*verts[0])[2];
	const real_type &x_1 = (*verts[1])[0]; const real_type &y_1 = (*verts[1])[1]; const real_type &z_1 = (*verts[1])[2];
	const real_type &x_2 = (*verts[2])[0]; const real_type &y_2 = (*verts[2])[1]; const real_type &z_2 = (*verts[2])[2];
	const real_type &x_3 = (*verts[3])[0]; const real_type &y_3 = (*verts[3])[1]; const real_type &z_3 = (*verts[3])[2];
	const real_type &x_4 = (*verts[4])[0]; const real_type &y_4 = (*verts[4])[1]; const real_type &z_4 = (*verts[4])[2];
	const real_type &x_5 = (*verts[5])[0]; const real_type &y_5 = (*verts[5])[1]; const real_type &z_5 = (*verts[5])[2];
	const real_type &x_6 = (*verts[6])[0]; const real_type &y_6 = (*verts[6])[1]; const real_type &z_6 = (*verts[6])[2];
	const real_type &x_7 = (*verts[7])[0]; const real_type &y_7 = (*verts[7])[1]; const real_type &z_7 = (*verts[7])[2];

	t1 = 1.0-a;
	t2 = 1.0-b;
	t3 = 1.0-c;
	f[0] = -x+((t1*x_3+a*x_7)*c+(t1*x_2+a*x_6)*t3)*b+((t1*x_1+a*x_5)*c+(t1*x_0+a*x_4)*t3)*t2;
	f[1] = -y+((t1*y_3+a*y_7)*c+(t1*y_2+a*y_6)*t3)*b+((t1*y_1+a*y_5)*c+(t1*y_0+a*y_4)*t3)*t2;
	f[2] = -z+((t1*z_3+a*z_7)*c+(t1*z_2+a*z_6)*t3)*b+((t1*z_1+a*z_5)*c+(t1*z_0+a*z_4)*t3)*t2;

	t1 = -1.0+b;
	t2 = 1.0-c;
	t3 = 1.0-a;
	J(0,0) = ((-x_3+x_7)*b+(x_1-x_5)*t1)*c+((-x_2+x_6)*b+(x_0-x_4)*t1)*t2;
	J(0,1) = ((-x_5+x_7)*c+(-x_4+x_6)*t2)*a+((-x_1+x_3)*c+(-x_0+x_2)*t2)*t3;
	J(0,2) = ((-x_6+x_7)*b+(x_4-x_5)*t1)*a+((-x_2+x_3)*b+(x_0-x_1)*t1)*t3;
	J(1,0) = ((-y_3+y_7)*b+(y_1-y_5)*t1)*c+((-y_2+y_6)*b+(y_0-y_4)*t1)*t2;
	J(1,1) = ((-y_5+y_7)*c+(-y_4+y_6)*t2)*a+((-y_1+y_3)*c+(-y_0+y_2)*t2)*t3;
	J(1,2) = ((-y_6+y_7)*b+(y_4-y_5)*t1)*a+((-y_2+y_3)*b+(y_0-y_1)*t1)*t3;
	J(2,0) = ((-z_3+z_7)*b+(z_1-z_5)*t1)*c+((-z_2+z_6)*b+(z_0-z_4)*t1)*t2;
	J(2,1) = ((-z_5+z_7)*c+(-z_4+z_6)*t2)*a+((-z_1+z_3)*c+(-z_0+z_2)*t2)*t3;
	J(2,2) = ((-z_6+z_7)*b+(z_4-z_5)*t1)*a+((-z_2+z_3)*b+(z_0-z_1)*t1)*t3; 

	// Invert and solve the system
	invert(J);
	return -J*f;
}

real3_type invertTrilinear(const real3_type &p, real3_type **verts, int maxItr/* = 10*/)
{
	int i;
	real3_type weights, update, f;
	weights[0] = weights[1] = weights[2] = 0.0;

	for (i = 0; i < maxItr; i++) {
		update = trilinearNewtonUpdate(p, verts, weights, f);
		weights = weights + update;

		if (f.norm() < 1e-5) {
			break;
		}
	}

	//cerr << "Terminated in iteration " << i << " with error " << f.norm() << endl;

	if (f.norm() > 1e-5) {
		// Set weights to infinity
		cerr << "Warning: invertTrilinear() failed; larger error after iteration " << i << " with error " << f.norm() << endl;
		weights[0] = weights[1] = weights[2] = REAL_TYPE_INF;
	}
	return weights;
}

bool hexahedronTestInside( const real3_type &p, real3_type **verts )
{
	// Find the projected position under each face
	static const int hexFaces[6][4] = {{0,1,2,3},{4,6,5,7},{0,4,1,5},{2,3,6,7},{0,2,4,6},{1,5,3,7}};

	for (int i = 0; i < 6; i++) {
		real_type alpha = 0.0, beta = 0.0;
		real3_type s1, s2, projected, normal;

		bilinearInvert3D(p,
			*verts[hexFaces[i][0]],
			*verts[hexFaces[i][1]],
			*verts[hexFaces[i][2]],
			*verts[hexFaces[i][3]],
			alpha, beta);

		// Compute the clamped projected location
		alpha = std::min(1.0f, std::max(0.0f, alpha));
		beta = std::min(1.0f, std::max(0.0f, beta));
		s1 = (1.0 - beta)*(*verts[hexFaces[i][2]]) + beta*(*verts[hexFaces[i][0]]);
		s2 = (1.0 - beta)*(*verts[hexFaces[i][3]]) + beta*(*verts[hexFaces[i][1]]);
		projected = (1.0 - alpha)*s1 + alpha*s2;

		// Compute normal using alpha, beta values
		s1 = (1.0 - beta) * (*verts[hexFaces[i][3]] - *verts[hexFaces[i][2]]) +
				beta      * (*verts[hexFaces[i][1]] - *verts[hexFaces[i][0]]);
		s2 = (1.0 - alpha) * (*verts[hexFaces[i][0]] - *verts[hexFaces[i][2]]) +
				alpha      * (*verts[hexFaces[i][1]] - *verts[hexFaces[i][3]]);
		normal = cross(s1,s2);
		normal.normalize();

		// Test if it's on the correct side of the surface
		if ((normal * (p - projected)) < 0) {
			return false;
		}
	}

	return true;
}


/**
 * Create a list of edges (along with their adjacent triangles) for the input mesh.
 *
 * We actually maintain two lists: the adjacency list for vertices, and the list
 * of edges.  The adjacency list stores, for each vertex v0, a list of adjacent
 * vertices v_adj along with the edge index # for the edge (v0 -> v_adj).
 * We use this adjacency list to check if the edge is already present in the
 * edge list, avoiding duplicates.
 *
 * To generate the list of edges, for each triangle, we add the three edges into
 * the edge list while checking for duplicates in the adjacency list.
 *
 */
void createEdgeList( const Mesh3_real &input, EdgeList &edgeList, AdjList &adjList )
{
	edgeList.clear();
	adjList.clear();

	index_type numVertices = input.numVertices();
	index_type numTriangles = input.triangle_vertex_index.size();

	adjList.resize(numVertices);

	// Construct a list of edges
	for (index_type ti = 0, tNum = numTriangles;
		 ti < tNum; ti++)
	{
		// Grab the vertices
		index3_type tri = input.triangle_vertex_index[ti];

		// For each edge
		for (int i = 1; i <= 3; i++) {
			int i0 = i - min(i,2); int i1 = i - i0;		// iterates in (0, 1), (0, 2), (1, 2)
			int v0, v1;									// grabs the vertices of the edges
			bool found;
			index_type j;

			// Check to see if the edge is already in the adjList
			// Do both orders of the edge (v0->v1 and v1->v0)
			for (int temp = 0; temp < 2; temp++) {
				// The endpoints of the edge
				if (temp == 0) {
					v0 = tri[i0]; v1 = tri[i1];
				}
				else {
					v0 = tri[i1]; v1 = tri[i0];
				}

				found = false;
				for (j = 0; j < adjList[v0].size(); j++) {
					if (adjList[v0][j].x() == v1) {
						found = true;
						break;
					}
				}

				if (found) {
					// Update the edge with the current triangle index
					edgeList[adjList[v0][j].y()].ti[1] = ti;
				}
				else {
					index_type ei;

					if (temp == 0) {
						// Create a new edge
						Edge ne;
						ne.vi[0] = v0; ne.vi[1] = v1;
						ne.ti[0] = ti;

						// push back onto the edgeList
						edgeList.push_back(ne);

						ei = edgeList.size()-1;
					}
					else {
						// The edge must be stored already in the adjList
						found = false;
						for (j = 0; j < adjList[v1].size(); j++) {
							if (adjList[v1][j].x() == v0) {
								found = true;
								break;
							}
						}
						if (found) {
							ei = adjList[v1][j].y();
						}
						else {
							// This case should not happen
							throw "createEdgeList(): Reverse edge not found in adjList!";
						}
					}

					// Add to the adjacency list
					adjList[v0].push_back(index2_type(v1, ei));
				}
			}
		}
	}
}
