///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "scene.h"
#include <ctime>
#include <utility>
#include <iostream>
#include "DeformEM.h"
#include "FileSequence.h"
#include "DeformEdit.h"

using namespace std;
using namespace Deform;

extern "C" {
	#include <SFMT/SFMT.h>
}

const char *globalForegroundColor = "black";
const char *globalBackgroundColor = "white";

Scene::Scene() : animation(NULL), deform(NULL), target(NULL), source(NULL), t(NULL), cl(NULL),
graph_n(10), graph_k(8), overlapf(0.001), solveInitialRigid(false), solveFinalNNLS(false),
markCellsWithPoints(false), useAccumulatedGeometry(false), maxCorrespondenceDistance(0.0), boundaryCorrespondenceDistance(0.0),
solverMaxIterations(50), estepMaxIterations(20), projMaxIterations(50), IRLSMaxIterations(50),
gridDivisions(10),
numCorrespondences(20), bonesPerGridPoint(4),
maxAutoBoneNum(50), minBoneVertexSupport(20), numGridCellsExpand(1),
defaultLoadFrames(0), defaultSkipFrames(1),
totalScene_boundingBoxDiagonal(1.0), depthScan_gridSpacing(0.0),
sourceFrame(0), targetFrame(0), autoLoadFrameData(1),
autoProcessMode(AUTO_PROCESS_NONE),
currentBone(-1), colormapMin(0), colormapMax(0),
correspondenceType(CORRESPONDENCE_NONE),
robust_PC(false), ANNEpsilon_MC(0),
featureSamplesSource(1000), featureSamplesTarget(1000), featureType(FEATURE_PRINCIPAL_CURVATURES),
ANNEpsilon_FM(0), radius_FM(0), maxCandidates_FM(100), threshold_FM(0),
maxDist_FM(100), pruneMaxRatio_FM(1.0),
ANNEpsilon_MS(0), radius_MS(0), maxIter_MS(100), rate_MS(1),
ANNEpsilon_VR(0), minVerifyClusterSize(0), minMatchSize_VR(0), maxNumVerifyClusters_VR(-1), maxNumLabels_GC(-1),
growRegionDistThreshold_VR(1), skipVerification_VR(false), graphCutSolutionType_GC(2), useDummyLabel_GC(false),
mismatchError_GC(10000), dummyLabelSmoothError_GC(100), dummyLabelDataError_GC(100), bdySmoothnessKNN_GC(0), closestPtDistThreshold_GC(1),
smoothWeight_GC(1.0), dataWeight_GC(1.0), smoothTruncateValue_GC(100.0), dataTruncateValue_GC(100.0),
dataTermType_GC(GRAPH_CUT_DATA_POINT_TO_POINT), graphCutInstance_GC(GRAPH_CUT_INSTANCE_SYMMETRIC),
RANSAC_max_itr(0), RANSAC_sourceModel_ANNEpsilon(0), RANSAC_targetModel_ANNEpsilon(0),
RANSAC_inlier_queryRadius(0), RANSAC_inlier_epsilonBand(0), RANSAC_minimumInliersPercentage(0),
showClusterMode(0)
{
	//////////////////////////////////////////////////////////////////////////
	// Pick random numbers
	//////////////////////////////////////////////////////////////////////////
	// Initialize random number generator
	init_gen_rand((unsigned)time( NULL ));
	for (index_type i = 0; i < 10000; i++) {
		// generate a random number
		cpi.push_back(genrand_real1());
	}

	//////////////////////////////////////////////////////////////////////////
	// Initialize all the weights
	edWeights.resize(3);
	wdWeights.resize(2);

	//////////////////////////////////////////////////////////////////////////
	// Initialize color defaults
	sourceColor.resize(3);
	targetColor.resize(3);
	pointColor.resize(3);
	selectedColor.resize(3);
	sourceColor[0] = 0.7; sourceColor[1] = 0.7; sourceColor[2] = 1.0;
	targetColor[0] = 1.0; targetColor[1] = 0.7; targetColor[2] = 0.7;
	pointColor[0] = 1.0; pointColor[1] = 1.0; pointColor[2] = 0.0;
	selectedColor[0] = 0.0; selectedColor[1] = 1.0; selectedColor[2] = 1.0;
}

Scene::~Scene()
{
	clear();
}

void Scene::setupDeform( int deformType )
{
	if (animation && (animation->numFrames() > 0) && source && target) {
		if (source == target) {
			cout << "Warning: source and target meshes are the same!" << endl;
		}

		//////////////////////////////////////////////////////////////////////////
		// Code useful for older experiments:
		// This stuff is commented out for now
		//// Build on the previous deformation
		//if (deformResult() && deform->finished()) {
		//	source = deformResult();
		//	source->computeFeatures(featureType, source->featureScales, ANNEpsilon_MC);
		//}
		//////////////////////////////////////////////////////////////////////////

		switch(deformType)
		{
		case DEFORM_EM:
			{
				// Delete the existing deformation object
				if (deform) {
					deform->Detach(this);
					delete deform;
					deform = NULL;
				}

				deform = new DeformEM();
				deform->Attach(this);

				perfTimers["DeformEM Setup"].Start();

				// If we have an initialization of the bone locations
				if (sourceSample.size() == graph_n) {
					((DeformEM*)deform)->initialize(source, this, target, &sourceSample, getDeformEMOptions());
				}
				else {
					((DeformEM*)deform)->initialize(source, this, target, NULL, getDeformEMOptions());
				}

				perfTimers["DeformEM Setup"].Stop();

				// Make sure we're not displaying out of bounds
				currentBone = min(graph_n-1, currentBone);
			}
			break;

		case DEFORM_ALL:
			break;

		default:
			deformType = 0;
			break;
		}
	}
}

DeformEMOptions Scene::getDeformEMOptions()
{
	// Populate options
	DeformEMOptions options;
	options.maxEstepItr = estepMaxIterations;
	options.maxIRLSItr = IRLSMaxIterations;
	options.maxProjItr = projMaxIterations;
	options.updateCorrespType = correspondenceType;
	options.solveInitialRigid = solveInitialRigid;
	options.EStepWeights = em_EWeights;
	options.MStepWeights = em_MWeights;
	options.DiscreteMStepWeights = em_DiscreteMWeights;
	options.useAccumulatedGeometry = useAccumulatedGeometry;
	options.markCellsWithPoints = markCellsWithPoints;
	options.inverse_deformation_maxSearchLevel;
	options.inverse_deformation_gridExpandStep;
	options.maxCorrespondenceDistance = maxCorrespondenceDistance;
	options.boundaryCorrespondenceDistance = boundaryCorrespondenceDistance;
	options.solveFinalNNLS = solveFinalNNLS;
	options.numBones = graph_n;
	options.numInflNeighbors = graph_k;
	options.gridDivisions = gridDivisions;
	options.numCorrespondences = numCorrespondences;

	return options;
}

void Scene::setDeformOptions()
{
	if (deform && (deform->getType() == DEFORM_EM)) {
		((DeformEM*)deform)->setOptions(getDeformEMOptions());
	}
}

void Scene::findCorrespondences()
{
	switch(correspondenceType) {
		case CORRESPONDENCE_LOAD_FROM_FILE:
			// 1. Try loading correspondences from file
			clearCorrespondenceList();
			cl = new CorrespondenceList();
			cl->load(correspondenceFile);
			break;
		case CORRESPONDENCE_GROUND_TRUTH:
			setGroundTruthCorrespondences();
			break;
		case CORRESPONDENCE_CLOSEST_POINTS:
		case CORRESPONDENCE_CLOSEST_POINTS_SPIN_IMAGE_WEIGHTS:
			setClosestPointCorrespondences();
			break;
		case CORRESPONDENCE_GRAPH_CUT:
			break;
		case CORRESPONDENCE_SPIN_IMAGES:
			// Just set closest points for now, DeformEM will take care of actually assigning spin image correspondences
			setClosestPointCorrespondences();
			break;
		default:
			break;
	}
}

void Scene::setGroundTruthCorrespondences()
{
	// Assumes point-to-point correspondences given by vertex indices
	clearCorrespondenceList();
	cl = new CorrespondenceList();

	cout << "*** Generating ground truth point correspondences...";

	index_type numVertices = source->numVertices();
	index_type numPick = min(numCorrespondences, cpi.size());

	if (numPick > numVertices) {
		// Set all vertices
		for (index_type i = 0; i < numVertices; i++) {
			cl->add(source->vertex[i], target->vertex[i], i, i);
		}
	}
	else {
		// Otherwise, select a random set of vertices using best candidate sampling of all source vertices
		idxList pickList = bestCandidateSample(source->vertex, numPick);
		for (index_type i = 0; i < numPick; i++) {
			cl->add(source->vertex[pickList[i]], target->vertex[pickList[i]], pickList[i], pickList[i]);
		}
	}

	cout << "done!" << endl;
}

void Scene::setClosestPointCorrespondences()
{
	// Assumes point-to-point correspondences given by closest points
	clearCorrespondenceList();
	cl = new CorrespondenceList();

	cout << "*** Generating closest point correspondences...";

	const PointList &sourceVertex = source->vertex;
	const PointList &targetVertex = target->vertex;
	index_type numVertices = source->numVertices();
	index_type numPick = numCorrespondences;
	idxList nearest;
	distList dists;

	if (numPick >= numVertices) {
		// Set all vertices
		for (index_type i = 0; i < numVertices; i++) {
			real3_type p = sourceVertex[i];

			// Find closest point in the target surface
			targetVertex.queryKNN(p, 1, nearest, dists, true);

			// Add to the list of correspondences
			cl->add(p, targetVertex[nearest[0]], i, nearest[0]);
		}
	}
	else {
		// Otherwise, select a random set of vertices using best candidate sampling of all source vertices
		idxList pickList = bestCandidateSample(sourceVertex, numPick);

		for (index_type i = 0; i < numPick; i++) {
			real3_type p = sourceVertex[pickList[i]];

			// Find closest point in the target surface
			targetVertex.queryKNN(p, 1, nearest, dists, true);

			// Add to the list of correspondences
			cl->add(p, targetVertex[nearest[0]], pickList[i], nearest[0]);
		}
	}

	cout << "done!" << endl;
}

void Scene::clear()
{
	clearAnimation();
	clearDeform();
}

void Scene::clearAnimation()
{
	if (animation) {
		delete animation;
		animation = NULL;
		target = source = NULL;
	}
}

void Scene::createNewAnimation() {
	if (animation) {
		clearAnimation();
	}
	animation = new ShapeAnimation();
}

void Scene::loadNewAnimation( const FileSequence &fs )
{
	createNewAnimation();
	animation->load(QString(rootDir.c_str()), fs, *this);
}

void Scene::addAnimationFrame( QString filename, int processFrames )
{
	animation->add(QString(rootDir.c_str()), filename, *this, processFrames);
}

bool Scene::is_deformation_set_up() const
{
	if (deform) {
		return true;
	}
	return false;
}

void Scene::clearDeform()
{
	// if there's a dangling solver thread
	if (t) {
		cout << "Warning in Scene::clearDeform(): dangling solver thread!  Please finish its execution.";
		t->wait();
		delete t;
		t = NULL;
	}

	if (deform) {
		// Detach from being notified by this deformation
		deform->Detach(this);

		// Delete the object
		delete deform;
		deform = NULL;
	}

	clearCorrespondenceList();
}

void Scene::solveDeformation()
{
	if (animation && animation->numFrames() > 0) {
		if (!deform) {
			cerr << "Error! Deformation not initialized properly." << endl;
		}
		else {
			// Check if there's already a thread
			if (t && !deform->finished()) {
				return;
			}

			// Create temporary variable to return solver options
			SolveOptions temp;
			temp.maxItr = solverMaxIterations;

			// Create new solver thread object
			t = new Deform::SolveThread(deform, temp);
	
			cerr << "*** Solver thread " << t << " created!" << endl;

			t->start();
		}
	}
}

void Scene::controlSolveThread( int signal, int type )
{
	if (type == DEFORM_EM) {
		setSolveStepType(DEFORM_EM_AUTO_STEP);
		DeformEM_ControlSolveThread(signal);
	}
	else if (type == DEFORM_ALL) {
		if (signal == SETUP_DEFORM) {
			cout << "-------------------------------------------------------------------------------" << endl;
			cout << "Setting Up Simultaneous Alignment..." << endl;
			setupDeform(type);
		}
		else if (signal == SOLVE_DEFORM_AUTO) {
			solveDeformation();
		}
	}
}

void Scene::DeformEM_ControlSolveThread( int signal )
{
	// Set up the deformation if necessary.
	if (!is_deformation_set_up()) {
		if ((signal == SOLVE_DEFORM_AUTO) || (signal == SETUP_DEFORM)) {
			setupDeform(DEFORM_EM);
			if (signal == SETUP_DEFORM) {
				return;
			}
		}
	}
	else if (deform->getType() != DEFORM_EM) {
		cerr << "Error in Scene::controlSolveThread(): Only DeformEM is supported!" << endl;
		return;
	}


	// Execute solver step.
	if (t && !deform->finished()) {
		switch(signal) {
			case SOLVE_DEFORM_AUTO:
				cerr << endl << "Resuming" << endl;
				((DeformEM*)deform)->controlSolver(SOLVER_RESUME);
				break;
			case PAUSE_SOLVER:
				cerr << endl << "Pausing" << endl;
				((DeformEM*)deform)->controlSolver(SOLVER_PAUSE);
				break;
			case STEP_SOLVER:
				cerr << endl << "Stepping" << endl;
				((DeformEM*)deform)->controlSolver(SOLVER_STEP);
				break;
			case SETUP_DEFORM:			/**< If the solver is already setup and running, just stop it */
			case STOP_SOLVER:
				cerr << endl << "Stopping" << endl;
				((DeformEM*)deform)->controlSolver(SOLVER_STOP);
				break;
			default:
				break;
		}
	}
	else {
		if (signal == SOLVE_DEFORM_AUTO) {
			cerr << endl << "Initial Solve" << endl;
			solveDeformation();
		}
		if (signal == SETUP_DEFORM) {
			cerr << endl << "Setup Solve" << endl;
			setupDeform(DEFORM_EM);
		}
	}
}

void Scene::checkExit()
{
	if (deform->finished()) {
		// Delete the solver thread if it's done
		if (t) {
			// Wait till the solver thread exits
			t->wait();

			cerr << "-------------------------------------------------------------------------------" << endl;
			cerr << "*** Solver thread " << t << " finished!" << endl;

			// Delete the thread
			delete t;
			t = NULL;
		}

		//// Check the final error between the deformed and target meshes
		//real_type finalError = meshNorm(*target, *(deform->deformationResult()));
		//cout << "Final Error between target and deformed meshes: " << finalError << endl;

		//// If we have embedded deformation, save the previous solution
		//if (deformType == DEFORM_EMBEDDED) {
		//	lastSolution = ((Deform::EmbeddedDeform*)deform)->getDeformation();
		//	cout << "Saved last solution of " << lastSolution.size() << "elements." << endl;
		//	cout << "First rotation: " << endl << lastSolution[0].rot << endl;
		//}

		printTime("DeformEM Setup");
		printTime("DeformEM RANSAC Correspondences");
		printTime("DeformEM E-step");
		printTime("DeformEM M-step");
		//printTime("DeformEM Registration");
		if (solveFinalNNLS) {
			printTime("DeformEM Final NNLS Update");
		}
		if (solveInitialRigid) {
			cout << "****** RANSAC Total Candidates: " << perfTimers["RANSAC Total Candidates"].iterations << endl;
			cout << "****** RANSAC Total Outliers: " << perfTimers["RANSAC Total Outliers"].iterations << endl;
			double percentage = perfTimers["RANSAC Total Outliers"].iterations / perfTimers["RANSAC Total Candidates"].iterations * 100.0;
			cout << "****** RANSAC Outlier percentage: " << percentage << "%" << endl;
		}

		if ((autoProcessMode == AUTO_PROCESS_DEFORMATION) && !autoSaveDeformEMResult.empty()) {
			// Save the result
			saveDeformation(QString(autoSaveDeformEMResult.c_str()));

			//////////////////////////////////////////////////////////////////////////
			// Gather timings and statistics
			if (!autoSaveDeformEMTimings.empty()) {
				
				// Save out statistics in text format
				DeformEM &dem = *((DeformEM*)deform);
				ofstream out;
				out.open(autoSaveDeformEMTimings.c_str(), ios_base::out);

				if (out.fail()) {
					cerr << "Error in registration timings: could not open output file " << autoSaveDeformEMTimings << " for writing!" << endl;
				}
				else {
					out << "# Vertices (source)" << "\t" << source->numVertices() << "\n";
					out << "# Vertices (target)" << "\t" << target->numVertices() << "\n";
					out << "# Correspondences" << "\t" << cl->size() << "\n";
					if (solveInitialRigid) {
						out << "RANSAC Total Candidates" << "\t" << perfTimers["RANSAC Total Candidates"].iterations << "\n";
						out << "RANSAC Total Outliers" << "\t" << perfTimers["RANSAC Total Outliers"].iterations << "\n";
						double percentage = perfTimers["RANSAC Total Outliers"].iterations / perfTimers["RANSAC Total Candidates"].iterations * 100.0;
						out << "RANSAC Outlier percentage" << "\t" << percentage << "%\n";
					}
					dem.writeStats(out);

					out << "Setup Time (secs)" << "\t" << perfTimers["DeformEM Setup"].GetDurationInSecs() << "\n";
					out << "RANSAC Correspondence Time (secs)" << "\t" << perfTimers["DeformEM RANSAC Correspondences"].GetDurationInSecs() << "\n";
					out << "E-Step Time (secs)" << "\t" << perfTimers["DeformEM E-step"].GetDurationInSecs() << "\n";
					out << "E-Step Iterations" << "\t" << perfTimers["DeformEM E-step"].iterations << "\n";
					out << "M-Step Time (secs)" << "\t" << perfTimers["DeformEM M-step"].GetDurationInSecs() << "\n";
					out << "M-Step Iterations" << "\t" << perfTimers["DeformEM M-step"].iterations << "\n";
					out << "Total Registration Time (secs)" << "\t" << perfTimers["DeformEM Registration"].GetDurationInSecs() << "\n";
					if (solveFinalNNLS) {
						out << "Final NNLS Weight Time (secs)" << "\t" << perfTimers["DeformEM Final NNLS Update"].GetDurationInSecs() << "\n";
					}

					out.close();
				}
			}

			// Exit
			exit(0);
		}
	}
}


ShapeAnimation* Scene::getAnimation()
{
	return animation;
}

void Scene::drawDeformation( const Deform::DrawOptions& options ) const
{
	// Draw the deformed result and related visualizations
	if (deform) {
		deform->draw(options);
	}
}

void Scene::drawAccumulatedGeometry( Deform::DrawOptions *options ) const
{
	if (deform && (deform->getType() == DEFORM_EM)) {
		((DeformEM*)deform)->drawAccumulatedGeometry(*options);
	}
	delete options;
}

void Scene::Update( Subject* theChangedSubject )
{
	// Notify the observers of the scene
	Notify();
}


void Scene::resampleSource()
{
	if (source) {
		source->sampleVertices(graph_n, sourceSample);
	}
}

void Scene::setXMLParamList( xmlParam &p )
{
	// Old deformation parameters
	p.addParam("Graph_Nodes", "int", (void*)&graph_n);
	p.addParam("Graph_Neighbors", "int", (void*)&graph_k);
	p.addParam("WeightedGraph_OverlapFactor", "float", (void*)&overlapf);
	p.addParam("EmbeddedDeform_Weights", "floatArray", (void*)&edWeights);
	p.addParam("WeightedDeform_Weights", "floatArray", (void*)&wdWeights);

	// DeformEM parameters
	p.addParam("solverMaxIterations", "int", (void*)&solverMaxIterations);
	p.addParam("estepMaxIterations", "int", (void*)&estepMaxIterations);
	p.addParam("projMaxIterations", "int", (void*)&projMaxIterations);
	p.addParam("IRLSMaxIterations", "int", (void*)&IRLSMaxIterations);
	p.addParam("DeformEM_EWeights", "floatArray", (void*)&em_EWeights);
	p.addParam("DeformEM_MWeights", "floatArray", (void*)&em_MWeights);
	p.addParam("DeformEM_DiscreteMWeights", "floatArray", (void*)&em_DiscreteMWeights);
	p.addParam("GridDivisions", "int", (void*)&gridDivisions);
	p.addParam("numCorrespondences", "int", (void*)&numCorrespondences);
	p.addParam("bonesPerGridPoint", "int", (void*)&bonesPerGridPoint);
	p.addParam("maxAutoBoneNum", "int", (void*)&maxAutoBoneNum);
	p.addParam("minBoneVertexSupport", "int", (void*)&minBoneVertexSupport);
	p.addParam("numGridCellsExpand", "int", (void*)&numGridCellsExpand);
	p.addParam("solveInitialRigid", "bool", (void*)&solveInitialRigid);
	p.addParam("solveFinalNNLS", "bool", (void*)&solveFinalNNLS);
	p.addParam("markCellsWithPoints", "bool", (void*)&markCellsWithPoints);
	p.addParam("useAccumulatedGeometry", "bool", (void*)&useAccumulatedGeometry);
	p.addParam("maxCorrespondenceDistance", "float", (void*)&maxCorrespondenceDistance);
	p.addParam("boundaryCorrespondenceDistance", "float", (void*)&boundaryCorrespondenceDistance);
	p.addParam("autoSaveDeformEMResult", "string", (void*)&autoSaveDeformEMResult);
	p.addParam("autoSaveDeformEMTimings", "string", (void*)&autoSaveDeformEMTimings);

	// Display parameters
	p.addParam("sourceColor", "floatArray", (void*)&sourceColor);
	p.addParam("targetColor", "floatArray", (void*)&targetColor);
	p.addParam("pointColor", "floatArray", (void*)&pointColor);
	p.addParam("selectedColor", "floatArray", (void*)&selectedColor);

	// Data files
	p.addParam("rootDir", "string", (void*)&rootDir);
	p.addParam("defaultFile", "string", (void*)&defaultFile);
	p.addParam("defaultLoadFrames", "int", (void*)&defaultLoadFrames);
	p.addParam("defaultSkipFrames", "int", (void*)&defaultSkipFrames);
	p.addParam("sourceFrame", "int", (void*)&sourceFrame);
	p.addParam("targetFrame", "int", (void*)&targetFrame);
	p.addParam("autoLoadFrameData", "int", (void*)&autoLoadFrameData);

	p.addParam("correspondenceFile", "string", (void*)&correspondenceFile);
	p.addParam("defaultColormapFile", "string", (void*)&defaultColormapFile);
	p.addParam("defaultSourceFile", "string", (void*)&defaultSourceFile);
	p.addParam("defaultTargetFile", "string", (void*)&defaultTargetFile);
	p.addParam("autoProcessMode", "int", (void*)&autoProcessMode);
	p.addParam("autoSaveCorrespondenceResult", "string", (void*)&autoSaveCorrespondenceResult);

	// Correspondence parameters
	p.addParam("correspondenceType", "int", (void*)&correspondenceType);
	p.addParam("scaleFactors_MC", "floatArray", (void*)&scaleFactors_MC);
	p.addParam("ANNEpsilon_MC", "float", (void*)&ANNEpsilon_MC);
	p.addParam("colormapMin", "float", (void*)&colormapMin);
	p.addParam("colormapMax", "float", (void*)&colormapMax);
	p.addParam("featureSamplesSource", "int", (void*)&featureSamplesSource);
	p.addParam("featureSamplesTarget", "int", (void*)&featureSamplesTarget);
	p.addParam("featureType", "int", (void*)&featureType);
	p.addParam("ANNEpsilon_FM", "float", (void*)&ANNEpsilon_FM);
	p.addParam("radius_FM", "float", (void*)&radius_FM);
	p.addParam("maxCandidates_FM", "int", (void*)&maxCandidates_FM);
	p.addParam("threshold_FM", "float", (void*)&threshold_FM);
	p.addParam("ANNEpsilon_MS", "float", (void*)&ANNEpsilon_MS);
	p.addParam("radius_MS", "float", (void*)&radius_MS);
	p.addParam("maxIter_MS", "float", (void*)&maxIter_MS);
	p.addParam("rate_MS", "float", (void*)&rate_MS);
	p.addParam("scaleFactors_PC", "floatArray", (void*)&scaleFactors_PC);
	p.addParam("robust_PC", "bool", (void*)&robust_PC);
	p.addParam("minVerifyClusterSize", "int", (void*)&minVerifyClusterSize);
	p.addParam("maxDist_FM", "float", (void*)&maxDist_FM);
	p.addParam("ANNEpsilon_VR", "float", (void*)&ANNEpsilon_VR);
	
	p.addParam("totalScene_boundingBoxDiagonal", "float", (void*)&totalScene_boundingBoxDiagonal);
	p.addParam("depthScan_gridSpacing", "float", (void*)&depthScan_gridSpacing);

	p.addParam("RANSAC_max_itr", "int", (void*)&RANSAC_max_itr);
	p.addParam("RANSAC_sourceModel_ANNEpsilon", "float", (void*)&RANSAC_sourceModel_ANNEpsilon);
	p.addParam("RANSAC_targetModel_ANNEpsilon", "float", (void*)&RANSAC_targetModel_ANNEpsilon);
	p.addParam("RANSAC_inlier_queryRadius", "float", (void*)&RANSAC_inlier_queryRadius);
	p.addParam("RANSAC_inlier_epsilonBand", "float", (void*)&RANSAC_inlier_epsilonBand);
	p.addParam("RANSAC_minimumInliersPercentage", "float", (void*)&RANSAC_minimumInliersPercentage);

	p.addParam("pruneMaxRatio_FM", "float", (void*)&pruneMaxRatio_FM);
	p.addParam("minMatchSize_VR", "int", (void*)&minMatchSize_VR);
	p.addParam("maxNumVerifyClusters_VR", "int", (void*)&maxNumVerifyClusters_VR);
	p.addParam("skipVerification_VR", "bool", (void*)&skipVerification_VR);
	p.addParam("maxNumLabels_GC", "int", (void*)&maxNumLabels_GC);
	p.addParam("growRegionDistThreshold_VR", "float", (void*)&growRegionDistThreshold_VR);
	p.addParam("graphCutSolutionType_GC", "int", (void*)&graphCutSolutionType_GC);
	p.addParam("useDummyLabel_GC", "bool", (void*)&useDummyLabel_GC);
	p.addParam("bdySmoothnessKNN_GC", "int", (void*)&bdySmoothnessKNN_GC);
	p.addParam("closestPtDistThreshold_GC", "float", (void*)&closestPtDistThreshold_GC);
	p.addParam("mismatchError_GC", "float", (void*)&mismatchError_GC);
	p.addParam("dummyLabelSmoothError_GC", "float", (void*)&dummyLabelSmoothError_GC);
	p.addParam("dummyLabelDataError_GC", "float", (void*)&dummyLabelDataError_GC);
	p.addParam("smoothWeight_GC", "float", (void*)&smoothWeight_GC);
	p.addParam("dataWeight_GC", "float", (void*)&dataWeight_GC);
	p.addParam("smoothTruncateValue_GC", "float", (void*)&smoothTruncateValue_GC);
	p.addParam("dataTruncateValue_GC", "float", (void*)&dataTruncateValue_GC);
	p.addParam("dataTermType_GC", "int", (void*)&dataTermType_GC);
	p.addParam("graphCutInstance_GC", "int", (void*)&graphCutInstance_GC);

	// Spin Image Parameters
	p.addParam("binSizeMultiple_SI", "float", (void*)&binSizeMultiple_SI);
	p.addParam("imageWidth_SI", "int", (void*)&imageWidth_SI);
	p.addParam("maxSupportAngle_SI", "float", (void*)&maxSupportAngle_SI);

	// Compute alignment distance
	p.addParam("alignDistanceFile", "string", (void*)&alignDistanceFile);
}

RealColor Scene::getColormapColor( real_type val ) const
{
	return colormap.getColor((val - colormapMin) / (colormapMax - colormapMin));
}

void Scene::regenerateColors()
{
	for (index_type l = 0; l < clusterColors.size(); l++) {
		// Get a random color from the colormap
		clusterColors[l] = colormap.getColor(genrand_real2());
	}
}

void Scene::generateClusterColors( size_type n )
{
	// Pick colors from the colormap for each label
	clusterColors.resize(n);

	for (index_type l = 0; l < n; l++) {
		// Get a random color from the colormap
		clusterColors[l] = colormap.getColor(genrand_real2());
	}
}

RealColor Scene::getClusterColor( index_type i )
{
	return clusterColors[i];
}

real_type Scene::makeSceneRadius( real_type percent ) const
{
	return totalScene_boundingBoxDiagonal * percent / 100.0;
}

real_type Scene::makeSceneRadiusSq( real_type percent ) const
{
	real_type radius = makeSceneRadius(percent);
	return (radius * radius);
}

distList Scene::makeSceneSqRadii( const FloatArray &percentList ) const
{
	// Generate scales as a percentage of the bounding box diagonal
	distList result;
	result.resize(percentList.size());

	for (int i = 0; i < percentList.size(); i++) {
		result[i] = makeSceneRadiusSq(percentList[i]);
	}

	return result;
}

void Scene::clearCorrespondenceList()
{
	if (cl) {
		delete cl;
		cl = NULL;
	}
}

real_type Scene::pointSize( real_type frac /*= 0.01 */ ) const
{
	if (depthScan_gridSpacing > 0.0) {
		return depthScan_gridSpacing;
	}
	return frac*totalScene_boundingBoxDiagonal;
}

void Scene::init()
{
	lastDir = QString(rootDir.c_str());

	if (defaultColormapFile.length() > 0) {
		cout << "Opening default colormap: " << defaultColormapFile << endl;
		colormap.load(defaultColormapFile.c_str());
	}
}

void Scene::setClosestPointCorrespondences_RejectBoundary()
{
	// Assumes point-to-point correspondences given by closest points
	clearCorrespondenceList();
	cl = new CorrespondenceList();

	cout << "*** Generating closest point correspondences (Rejecting Boundary)...";

	index_type numVertices = source->numVertices();
	index_type numPick = min(numCorrespondences, cpi.size());
	idxList nearest;
	distList dists;

	if (numPick > numVertices) {
		// Set all vertices
		for (index_type i = 0; i < numVertices; i++) {
			real3_type sourceVertex = source->vertex[i];

			// Find closest point in the target surface
			target->vertex.queryKNN(sourceVertex, 1, nearest, dists, true);

			// Add to the list of correspondences
			cl->add(sourceVertex, target->vertex[nearest[0]], i, nearest[0]);
		}
	}
	else {
		// Otherwise, select a random set of vertices using best candidate sampling of all source vertices
		idxList pickList = bestCandidateSample(source->vertex, numPick);
		for (index_type i = 0; i < numPick; i++) {
			real3_type sourceVertex = source->vertex[pickList[i]];

			// Find closest point in the target surface
			target->vertex.queryKNN(sourceVertex, 1, nearest, dists, true);

			// Check that target vertex is not on boundary
			if (target->markVertex[nearest[0]] == VERTEX_VALID) {
				// Add to the list of correspondences
				cl->add(sourceVertex, target->vertex[nearest[0]], pickList[i], nearest[0]);
			}
		}
	}

	cout << "done!" << endl;
}


void Scene::saveDeformation( QString saveFilename )
{
	if (!deform || (deform->getType() != DEFORM_EM)) {
		cerr << "Error in Scene::saveDeformation(): DeformEM not initialized!" << endl;
		return;
	}

	//QString saveFilename;
	//if (prompt) {
	//	// Prompt for the filename
	//	saveFilename = QFileDialog::getSaveFileName(this, QString("Save Registration Result"), this->lastDir, QString("Data (*.data)"));
	//}
	//else {
	//	saveFilename = QString(autoSaveDeformEMResult.c_str());
	//}

	// Check if cancelled
	if (saveFilename.isEmpty()) {
		return;
	}

	// Update last dir
	lastDir = grabPath(saveFilename);

	// Save the result
	((DeformEM*)deform)->saveResult(qPrintable(saveFilename));
}

void Scene::loadDeformation( QString loadFilename )
{
	if (!deform || (deform->getType() != DEFORM_EM)) {
		cerr << "Error in Scene::loadDeformation(): DeformEM not initialized!" << endl;
		return;
	}


	//QString loadFilename;
	//if (prompt) {
	//	// Prompt for the filename
	//	loadFilename = QFileDialog::getOpenFileName(this, QString("Load Registration Result"), this->lastDir, QString("Data (*.data)"));
	//}
	//else {
	//	loadFilename = QString(autoSaveDeformEMResult.c_str());
	//}

	// Check if cancelled
	if (loadFilename.isEmpty()) {
		return;
	}

	// Update last dir
	lastDir = grabPath(loadFilename);

	// Load the result
	((DeformEM*)deform)->loadResult(qPrintable(loadFilename));
}

void Scene::generateGridColors()
{
	if (!deform) {
		cerr << "Error in Scene::generateGridColors(): Deform not initialized!" << endl;
		return;
	}

	if (deform->getType() == DEFORM_EM) {
		((DeformEM*)deform)->generateGridColors();
	}
}

std::pair<QAbstractItemModel*,DeformEM_GUI_Communicator*> Scene::getSolutionViewModel()
{
	if (!deform || (deform->getType() != DEFORM_EM)) {
		cerr << "Error in Scene::getSolutionViewModel(): DeformEM not initialized!" << endl;
		return std::make_pair((QAbstractItemModel*)NULL, (DeformEM_GUI_Communicator*)NULL);
	}

	return ((DeformEM*)deform)->solutionViewModel();
}


void Scene::setSolveStepType( int solveStepType )
{
	if (deform && (deform->getType() == DEFORM_EM)) {
		((DeformEM*)deform)->setStepType(solveStepType);
	}
}

void Scene::browseDeformation( int step )
{
	////if (!deform || (deformType != DEFORM_EM) || (!deform->finished())) {
	////	return;
	////}

	////// Get the maximum number of deformations to browse
	////const std::vector<RigidDeform> &storedDeformations = ((DeformEM*)deform)->getSolvedDeformations();
	////int browseMax = storedDeformations.size();

	////if (browseMax == 0) {
	////	return;
	////}

	////// Step and apply the deformation
	////int previousIndex = browseCurrentDeformation;
	////browseCurrentDeformation += step;

	////// Check limits
	////if (browseCurrentDeformation < 0) {
	////	browseCurrentDeformation = 0;
	////}
	////if (browseCurrentDeformation >= browseMax) {
	////	browseCurrentDeformation = browseMax - 1;
	////}

	////// Apply the deformation
	////if (browseCurrentDeformation != previousIndex) {
	////	((DeformEM*)deform)->setVertexWeightsFromRawData(((DeformEM*)deform)->getSolvedWeights()[browseCurrentDeformation]);
	////	((DeformEM*)deform)->applyDeformation(storedDeformations[browseCurrentDeformation]);
	////}
}

void Scene::printCurrentDeformation()
{
	if (!deform || (deform->getType() != DEFORM_EM) || (!deform->finished())) {
		return;
	}
	((DeformEM*)deform)->printCurrentDeformation();
}

void Scene::revertDiscreteWeights()
{
	if (!deform || (deform->getType() != DEFORM_EM) || (!deform->finished())) {
		return;
	}
	((DeformEM*)deform)->revertDiscreteWeights();
}

void Scene::solveFinalNNLSWeights()
{
	if (!deform || (deform->getType() != DEFORM_EM) || (!deform->finished())) {
		return;
	}
	((DeformEM*)deform)->solveWeights_finalNNLSUpdate();
}

std::string Scene::getSolverStatus() const
{
	if (!deform || (deform->getType() != DEFORM_EM)) {
		return std::string("");
	}
	
	return ((DeformEM*)deform)->solverStatus;
}

bool Scene::saveDeformedResult( const char *file )
{
	if (!deform || (deform->getType() != DEFORM_EM)) {
		return false;
	}

	return ((DeformEM*)deform)->writeDeformedResult(file);
}
