///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __SCENE_H__
#define __SCENE_H__

#include "subject.h"
#include "ShapeAnimation.h"
#include "deform.h"
#include <vector>
#include <QString>
#include <libXMLParam/xmlparam.h>
#include "UniformGrid.h"
#include "Colormap.h"
#include "FindCorrespondence.h"


// Global foreground and background colors for OpenGL viewer windows
extern const char *globalForegroundColor;
extern const char *globalBackgroundColor;

enum AutoProcessModes {
	AUTO_PROCESS_NONE = 0,
	AUTO_PROCESS_CORRESPONDENCE,
	AUTO_READ_CORRESPONDENCE_RESULT,
	AUTO_COMPUTE_ALIGNMENT_DISTANCE,
	AUTO_PROCESS_DEFORMATION,
	AUTO_READ_DEFORMATION_RESULT,
	AUTO_SIMULTANEOUS_ALIGN
};

class QAbstractItemModel;
class DeformEM_GUI_Communicator;
class DeformEMOptions;


class Scene : public Subject, public Observer
{
public:
	Scene();
	virtual ~Scene();

	// Initializing the scene
	void init();

	// Clearing the data
	void clear();
	void clearAnimation();
	void clearDeform();

	// Loading and Accessing the Mesh Animation
	void createNewAnimation();
	void loadNewAnimation(const FileSequence &fs);
	void addAnimationFrame( QString filename, int processFrames );
	ShapeAnimation* getAnimation();


	// Accessing, initializing and solving for the deformation
	void setupDeform(int deformType);
	void solveDeformation();
	void drawDeformation( const Deform::DrawOptions& options ) const;
	void saveDeformation( QString saveFilename );
	void loadDeformation( QString loadFilename );
	void setSolveStepType( int solveStepType );
	void controlSolveThread( int signal, int type );
	void DeformEM_ControlSolveThread( int signal );
	void checkExit();														/**< Checking if the deformation thread exits */
	void browseDeformation( int step );
	void printCurrentDeformation();

	// Debugging implementation of continuous NNLS weights
	void revertDiscreteWeights();
	void solveFinalNNLSWeights();
	std::string getSolverStatus() const;


	// Deformation result
	bool is_deformation_set_up() const;
	//ShapeModel* inverseDeform( ShapeModel &shape );
	void drawAccumulatedGeometry( Deform::DrawOptions *options ) const;		/**< Draw accumulated geometry */
	void setDeformOptions();												/**< Set options for Deformation solver */
	DeformEMOptions getDeformEMOptions();
	void generateGridColors();
	std::pair<QAbstractItemModel*,DeformEM_GUI_Communicator*> getSolutionViewModel();
	void solutionViewModelChanged();
	bool saveDeformedResult( const char *file );


	// Updating the display
	virtual void Update(Subject* theChangedSubject);

	// Sampling the source mesh
	void resampleSource();

	// Deformation parameters
	ShapeModel *source, *target;		/**< Source and target meshes for the deformation */

	// Get correspondences
	CorrespondenceList* getCorrespondences();

	// Useful global variables
	QString lastDir;

	// Get a colormap color
	RealColor getColormapColor(real_type val) const;

	// Get the scene point size; automatically determined based on length of the bounding box diagonal
	real_type pointSize( real_type frac = 0.01 ) const;

	// Reading deformation parameters from the file
	void setXMLParamList(xmlParam &p);

	// Weights for Error Terms
	FloatArray edWeights, wdWeights, em_EWeights, em_MWeights, em_DiscreteMWeights;

	// Defining Maximum # of Iterations for Solver
	int solverMaxIterations;						/**< Maximum total iterations for the solver */
	int estepMaxIterations;							/**< Maximum number of iterations for solving the transformations */
	int projMaxIterations;							/**< Maximum number of iterations for projecting affine transform to rigid transform */
	int IRLSMaxIterations;							/**< Maximum number of iterations for IRLS solver */

	// Setting up the deformation solver
	int graph_n;									/**< Total number of bones */
	int gridDivisions;								/**< Spacing between grid points */
	size_type numCorrespondences;					/**< Number of correspondences to generate */
	int bonesPerGridPoint;							/**< Maximum number of bones affecting each grid point (bone-influence set size) */
	int maxAutoBoneNum;								/**< Maximum number of bones to use */
	int minBoneVertexSupport;						/**< Minimum number of vertices each bone needs to influence */
	int numGridCellsExpand;							/**< Number of grid cells to expand the influence of each bone */
	int graph_k;									/**< Initializing bone influences: set bone radius to be distance to k^th nearest neighbor */
	float overlapf;									/**< Initializing bone influences for weightedDeform: set overlap factor of bone radii */
	bool solveInitialRigid;							/**< Initially solve for a single rigid transformation */
	bool solveFinalNNLS;							/**< At the end, solve for final continuous weights */
	bool markCellsWithPoints;						/**< Initially when setting up the grid, mark cells with the mesh vertices, not triangles */
	bool useAccumulatedGeometry;					/**< Gradually accumulate geometry in the reference pose after solving each frame */
	float maxCorrespondenceDistance;				/**< Maximum distance between correspondences, given by multiple of grid resolution */
	float boundaryCorrespondenceDistance;			/**< Maximum distance between correspondences, given by multiple of grid resolution, for points on boundary */

	// Deformation solver batch processing
	std::string autoSaveDeformEMResult;				/**< Automatically save the result from DeformEM */
	std::string autoSaveDeformEMTimings;			/**< Automatically save the stats and timings from DeformEM */

	FloatArray sourceColor, targetColor, pointColor, selectedColor;		/**< Color settings */

	std::string rootDir;							/**< Root data directory */
	std::string defaultFile;						/**< File to open at startup */
	int defaultLoadFrames;							/**< How many frames (at most) to load at startup */
	int defaultSkipFrames;							/**< Skip over frames when loading */
	int sourceFrame, targetFrame;					/**< Source and target mesh frame (animation frame #) */
	int autoLoadFrameData;							/**< Automatically load extra frame data */

	std::string defaultSourceFile;					/**< Default source mesh to open at startup */
	std::string defaultTargetFile;					/**< Default target mesh to open at startup */
	int autoProcessMode;							/**< Automatic processing mode */
	std::string autoSaveCorrespondenceResult;		/**< Automatic save location for correspondence result */

	std::string correspondenceFile;					/**< File to open for correspondences */
	std::string defaultColormapFile;				/**< Colormap to use */

	// Display settings
	int currentBone;						/**< Current bone we are displaying */
	Colormap colormap;						/**< Colormap */
	float colormapMin, colormapMax;			/**< Min and max values for assigning colormap values */

	float totalScene_boundingBoxDiagonal;	/**< A measure of scene size */
	float depthScan_gridSpacing;			/**< Grid spacing size (in terms of world space coordinates) for depth scans */

	// Correspondence parameters
	int correspondenceType;					/**< What type of correspondences to generate */
	bool robust_PC;							/**< Whether to generate robust principal curvatures? */
	FloatArray scaleFactors_PC;				/**< Scales for generating the principal curvature frames (percentage of bbox diagonal) */
	FloatArray scaleFactors_MC;				/**< Sizes for the curvature descriptor (percentage of bbox diagonal) */
	float ANNEpsilon_MC;					/**< Range query epsilon for mean curvature estimation */
	int featureSamplesSource;				/**< Number of features to compute in the source mesh */
	int featureSamplesTarget;				/**< Number of features to compute in the target mesh */
	int featureType;						/**< Type of features to compute in each mesh */
	float ANNEpsilon_FM;					/**< Range query epsilon for feature matching */
	float radius_FM;						/**< Range query radius for feature matching */
	int maxCandidates_FM;					/**< Maximum number of candidates for feature matching */
	float threshold_FM;						/**< Minimum norm of a feature used for matching */
	float maxDist_FM;						/**< Maximum Euclidean distance for correspondence pairing (% of bbox diagonal) */
	float pruneMaxRatio_FM;					/**< Maximum ratio of principal curvatures for point pruning */

	float ANNEpsilon_MS;					/**< Range query epsilon for range query in feature space within mean-shift clustering */
	float radius_MS;						/**< Radius of kernel for mean-shift clustering */
	float maxIter_MS;						/**< Maximum # of iterations for mean-shift clustering */
	float rate_MS;							/**< Rate of convergence for mean-shift clustering */

	float ANNEpsilon_VR;					/**< Range query epsilon for region verification */
	int minVerifyClusterSize;					/**< Minimum size of a cluster for it to be matched & verified */
	int minMatchSize_VR;						/**< Minimum size of verified region */
	int maxNumVerifyClusters_VR;				/**< Maximum number of clusters for verification */
	int maxNumLabels_GC;						/**< Maximum number of labels for the graph cut */
	real_type growRegionDistThreshold_VR;		/**< Distance threshold for growing the verified region */
	int graphCutSolutionType_GC;				/**< What solution method to use in graph cut optimization */
	bool useDummyLabel_GC;						/**< Whether to assign a dummy label in the graph cut optimization? */
	int bdySmoothnessKNN_GC;					/**< Connecting k-nearest boundary vertices for smoothness constraints */
	real_type mismatchError_GC;					/**< Penalty for inconsistent labeling */
	real_type dummyLabelSmoothError_GC;			/**< Smoothness penalty for assigning the dummy label */
	real_type dummyLabelDataError_GC;			/**< Data penalty for assigning the dummy label */
	real_type closestPtDistThreshold_GC;		/**< Closest point distance threshold for graph cut */
	real_type smoothWeight_GC;					/**< Weight of smoothness term for graph cut error function */
	real_type dataWeight_GC;					/**< Weight of data term for graph cut error function */
	real_type smoothTruncateValue_GC;			/**< Maximum value of the smooth term for graph cut error function */
	real_type dataTruncateValue_GC;				/**< Maximum value of the data term for graph cut error function */
	int dataTermType_GC;						/**< Type of data term in error function */
	int graphCutInstance_GC;					/**< Graph cut instance setup: symmetric, forward, or backward */

	int RANSAC_max_itr;							/**< RANSAC correspondence: Number of iterations */
	float RANSAC_sourceModel_ANNEpsilon;		/**< RANSAC correspondence: ANN Epsilon for closest distance query */
	float RANSAC_targetModel_ANNEpsilon;		/**< RANSAC correspondence: ANN Epsilon for closest distance query */
	float RANSAC_inlier_queryRadius;			/**< RANSAC correspondence: Radius for gathering inlier candidates */
	float RANSAC_inlier_epsilonBand;			/**< RANSAC correspondence: Epsilon-band distance threshold for inliers */
	float RANSAC_minimumInliersPercentage;		/**< RANSAC correspondence: Minimum percentage of inliers for good fit */

	// Spin image parameters
	real_type binSizeMultiple_SI;				/**< Spin Images: Bin Size parameter as a multiple of the mesh resolution */
	int imageWidth_SI;							/**< Spin Images: Image Width */
	real_type maxSupportAngle_SI;				/**< Spin Images: Maximum Support Angle (in degrees) */

	// Skip verification step?
	bool skipVerification_VR;					/**< Skip verification step? */

	// Compute and save alignment distance
	std::string alignDistanceFile;				/**< Text filename to store the resulting distance */

	// Create radius values from the total scene size
	real_type makeSceneRadius( real_type percent ) const;
	real_type makeSceneRadiusSq( real_type percent ) const;
	distList makeSceneSqRadii( const FloatArray &percentList ) const;

	// Store colors for each cluster
	void regenerateColors();
	void generateClusterColors( size_type n );
	RealColor getClusterColor( index_type i );
	size_type getClusterSize( index_type i );
	std::vector<RealColor> clusterColors;

	// Store clustering data
	void initMatchCandidates( ShapeModel &sm, ShapeModel &tm, bool forceReset = false );
	bool clusterResultExists() const;

	// Debugging: Visualizing the cluster modes.
	int showClusterMode;

	// Store initial bone locations and influence radii.
	distList initialRadii;
	RigidDeform initialTransforms;

private:
	ShapeAnimation *animation;				/**< A mesh animation sequence */

	Deform::DeformAlgorithm *deform;		/**< Deformation strategy */
	Deform::SolveThread *t;					/**< A separate thread used to solve the deformation */

	// Sampling correspondences
	std::vector<double> cpi;				/**< Random numbers to pick correspondence indices */
	CorrespondenceList *cl;					/**< New correspondence class (with only positions, not any mesh vertex indices) */

	// Sampling the mesh
	PointList sourceSample;
	void setGroundTruthCorrespondences();
	void setClosestPointCorrespondences();
	void setClosestPointCorrespondences_RejectBoundary();
	void findCorrespondences();
	void clearCorrespondenceList();

	// Debugging: save the previous solution
	RigidDeform lastSolution;
};

#endif
