///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include <iostream>
#include "taucsUtil.h"
#include <cmath>

using namespace std;

void taucsPrintResultCodeMessage( int result )
{
	char *message;

	switch(result) {
		case TAUCS_SUCCESS:
			message = "*** TAUCS Solve Success!";
			break;
		case TAUCS_ERROR:
			message = "*** TAUCS Error: Generic error.";
			break;
		case TAUCS_ERROR_NOMEM:
			message = "*** TAUCS Error: NOMEM error.";
			break;
		case TAUCS_ERROR_BADARGS:
			message = "*** TAUCS Error: BADARGS error.";
			break;
		case TAUCS_ERROR_MAXDEPTH:
			message = "*** TAUCS Error: MAXDEPTH error.";
			break;
		case TAUCS_ERROR_INDEFINITE:
			message = "*** TAUCS Error: INDEFINITE error.";
			break;
		default:
			break;
	}

	cout << message << endl;
}

void taucsSubmatrixSymmetric( const taucs_ccs_matrix *M, taucs_ccs_matrix *result, std::vector<bool> keep )
{
	const int numCols = M->n;
	
	// Check flags and size
	if (!(M->flags & TAUCS_SYMMETRIC) || (keep.size() != numCols)) {
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Keep track of new indices
	vector<int> newIndices(keep.size());
	int count = 0;
	for (int i = 0; i < keep.size(); i++) {
		if (keep[i]) {
			newIndices[i] = count;
			count++;
		}
		else {
			newIndices[i] = -1;
		}
	}

	// Create result matrix (assume enough memory is allocated in M)
	int newDim = count;
	result->n = result->m = newDim;

	// Keep selected elements
	count = 0;
	for (int j = 0; j < numCols; j++) {
		// For each column to keep
		if (keep[j]) {

			result->colptr[newIndices[j]] = count;

			for (int li = M->colptr[j]; li < M->colptr[j+1]; li++) {

				int i = M->rowind[li];

				// For each row to keep
				if (keep[i]) {
					// Add the value and row index
					result->taucs_values[count] = M->taucs_values[li];
					result->rowind[count] = newIndices[i];
					count++;
				}
			}
		}
	}
	result->colptr[newDim] = count;

	//cerr << "*** Creating submatrix with dim=" << newDim << " and nnz=" << count << endl;
}

void taucsSubvector( const taucsType *x, taucsType *result, std::vector<bool> keep, int numActive )
{
	// Assume the dimensions match
	// Allocate matrix
	//taucsType *result = new taucsType[numActive];

	// Copy data
	int count = 0;
	for (int i = 0; i < keep.size(); i++) {
		if (keep[i]) {
			result[count] = x[i];
			count++;
		}
	}
}


taucsType computeL2Error( const taucs_ccs_matrix *A, const taucsType *x, const taucsType *b, taucsType *residual )
{
	taucsType L2 = 0.0;
	MulNonSymmMatrixVector(A, x, residual);
	for (int i = 0; i < A->m; i++) {
		residual[i] = residual[i] - b[i];
		L2 += residual[i]*residual[i];
	}

	return sqrt(L2);
}


// For re-weighting A^T * A using a squared diagonal weighting matrix W^2 to get result: A^T * W^2 * A
// Assumes A^T * A was computed already before and just replaces the values
// Assumes W^2 is given by the values in the diagonal
void reweightATA( const taucs_ccs_matrix *matA, const taucs_ccs_matrix *matB, const taucsType *weightSq, taucs_ccs_matrix *result )
{
	// Compatibility of dimensions        
	if ((matA->m != matB->n) || (matA->n != matB->m) || (result->n != result->m))
		return;
	if ((matA->flags & TAUCS_SYMMETRIC) || (matB->flags & TAUCS_LOWER) )
		return;
	if (!(result->flags & TAUCS_SYMMETRIC))
		return;

	// Reset memory to zeros
	memset(result->taucs_values, 0, sizeof(taucsType) * (result->colptr[result->n]));

	// (m x n)*(n x m) = (m x m)
	int m=matA->m;
	int n=matA->n;

	for (int i=0; i<m; ++i) {
		// Creating column i of the result
		int colBegin = result->colptr[i];
		int colEnd = result->colptr[i+1];

		// travel on bi
		for (int rowptrBi = matB->colptr[i];rowptrBi < matB->colptr[i+1];++rowptrBi) {
			int rowInd = matB->rowind[rowptrBi];
			taucsType biv = weightSq[rowInd] * matB->taucs_values[rowptrBi];			/**< Multiply with the weight as well */

			// make biv*a_{rowInd} and insert into column
			// Ignore anything above the diagonal!!
			for (int rowptrA=matA->colptr[rowInd];rowptrA<matA->colptr[rowInd+1];++rowptrA) {
				int rowA = matA->rowind[rowptrA];
				if (rowA >= i) {
					taucsType valA = matA->taucs_values[rowptrA];

					// insert valA*biv into the appropriate location (search for it)
					int rowptrC;
					for ( rowptrC = colBegin; rowptrC < colEnd; ++rowptrC ) {
						// Check if the row location is the same
						if (result->rowind[rowptrC] == rowA) {
							result->taucs_values[rowptrC] += valA*biv;
							break;
						}
					}
					if (rowptrC == colEnd) {
						cerr << "Warning in reweightATA(): couldn't find appropriate location for the value!" << endl;
					}
				}
			}
		}
		// now column i is created
	}
}

// For re-weighting A^T * b using a squared diagonal weighting matrix W^2 to get result: A^T * W^2 * b
// Assumes W^2 is given by the values in the diagonal
void reweightATb(const taucs_ccs_matrix *matA, const taucsType *b, const taucsType *weightSq, taucsType *result)
{
	// make result all zero
	memset(result, 0, matA->m * sizeof(taucsType));

	for (int col = 0; col < matA->n; ++col) {
		// going over column col of matA, multiplying it by b[col] and setting the appropriate values of vector result
		// Also weight vector b according to the weight matrix
		for (int p = matA->colptr[col]; p < matA->colptr[col+1]; ++p) {
			result[matA->rowind[p]] += weightSq[col] * b[col] * matA->taucs_values[p];
		}
	}
}


// Compute the squared re-weighting based on the residual, and also compute and return the L1 norm on the way
taucsType computeIRLSWeightsSq( taucsType *residual, int n, taucsType invEps )
{
	taucsType L1 = 0.0;
	for (int i = 0; i < n; i++) {
		L1 += abs(residual[i]);

		// Compute weight = 1 ./ (1 + (res / eps).^2);
		residual[i] = residual[i] * invEps;
		residual[i] = 1.0 / (1.0 + residual[i]*residual[i]);
	}

	return L1;
}

// Compute the squared re-weighting based on the residual, and also compute and return the L1 norm on the way
taucsType computeIRLSWeightsSq( taucsType *residual, int n, taucsType invEps, const std::vector<bool> &selectiveReweight )
{
	taucsType L1 = 0.0;
	for (int i = 0; i < n; i++) {
		L1 += abs(residual[i]);

		if (selectiveReweight[i]) {
			// Compute weight = 1 ./ (1 + (res / eps).^2);
			residual[i] = residual[i] * invEps;
			residual[i] = 1.0 / (1.0 + residual[i]*residual[i]);
		}
		else {
			residual[i] = 1.0;
		}
	}

	return L1;
}
