///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __TYPES_H__
#define __TYPES_H__

#include <SylvainTools/geom.h>
#include <QTextStream>
#include <cfloat>
#include <functional>

#undef min
#undef max

// Type definitions
typedef float real_type;
typedef Geometry::Vec2<real_type> real2_type;
typedef Geometry::Vec3<real_type> real3_type;
typedef Geometry::Vec<4,real_type> real4_type;
typedef Geometry::Square_matrix<3,real_type> matrix3_type;

// Unsigned index type
typedef unsigned int index_type;
typedef Geometry::Vec2<index_type> index2_type;
typedef Geometry::Vec3<index_type> index3_type;

// Signed index type
typedef Geometry::Vec2<int> int2_type;
typedef Geometry::Vec3<int> int3_type;

// Size type
typedef unsigned int size_type;

typedef std::vector<index_type> idxList;
typedef std::vector<real_type> distList;
typedef std::vector<index3_type> idx3List;
typedef std::vector<int3_type> int3List;
typedef std::vector<real3_type> ptList;

#define REAL_TYPE_EPSILON FLT_EPSILON
#define REAL_TYPE_MAX FLT_MAX
#define REAL_TYPE_QNAN std::numeric_limits<float>::quiet_NaN()
#define REAL_TYPE_INF std::numeric_limits<float>::infinity()

typedef real3_type RealColor;
typedef Geometry::Vec3i IntColor;


// int3_type comparison operator
struct int3Comp : public std::binary_function<int3_type,int3_type,bool> {
	bool operator()(const int3_type &x, const int3_type &y) const {
		// Dictionary Ordering
		if (x[0] == y[0])
			if (x[1] == y[1])
				return (x[2] < y[2]);
			else
				return (x[1] < y[1]);
		else
			return (x[0] < y[0]);
	}
};


// Data type definition.
class data_type {
public:
	inline data_type(int d) {
		val.resize(d);
		setZero();
	}

	inline void setZero() {
		for (int i = 0; i < val.size(); i++) {
			val[i] = 0;
		}
	}

	inline bool checkFinite() {
		for (int i = 0; i < val.size(); i++) {
			if (!_finite(val[i])) {
				return false;
			}
		}
		return true;
	}

	inline data_type& operator+=(const data_type& v) {
		for (unsigned int i = 0; i < val.size(); i++) {
			val[i] += v.val[i];
		}
		return *this;
	}

	inline data_type& operator*=(const real_type r) {
		for (unsigned int i = 0; i < val.size(); i++) {
			val[i] *= r;
		}
		return *this;
	}

	inline int normL0() {
		int total = 0;
		for (unsigned int i = 0; i < val.size(); i++) {
			if (abs(val[i]) > 0.0) {
				++total;
			}
		}

		return total;
	}

	inline void normalizeL1() {
		real_type total = normL1();

		if (total > 0) {
			for (unsigned int i = 0; i < val.size(); i++) {
				val[i] /= total;
			}
		}
	}

	inline real_type normL1() {
		real_type total = 0;
		for (unsigned int i = 0; i < val.size(); i++) {
			total += abs(val[i]);
		}

		return total;
	}

	inline void normalizeL2() {
		real_type total = normL2();

		if (total > 0) {
			for (unsigned int i = 0; i < val.size(); i++) {
				val[i] /= total;
			}
		}
	}

	inline real_type normSqrL2() {
		real_type total = 0;
		for (unsigned int i = 0; i < val.size(); i++) {
			total += val[i]*val[i];
		}

		return total;
	}

	inline real_type normL2() {
		return sqrt(normSqrL2());
	}

	inline real_type findMin( int &minIndex ) const
	{
		real_type minVal = REAL_TYPE_MAX;

		for (int i = 0; i < val.size(); i++) {
			if (val[i] < minVal) {
				minVal = val[i];
				minIndex = i;
			}
		}

		return minVal;
	}

	inline real_type findMax( int &maxIndex ) const
	{
		real_type maxVal = -REAL_TYPE_MAX;

		for (int i = 0; i < val.size(); i++) {
			if (val[i] > maxVal) {
				maxVal = val[i];
				maxIndex = i;
			}
		}

		return maxVal;
	}

	std::vector<real_type> val;
};

/**
 * Add two data vectors.
 */
inline data_type operator+(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] + v2.val[i];
	}
	return res;
}

/**
* Subtract two data vectors.
*/
inline data_type operator-(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] - v2.val[i];
	}
	return res;
}

/**
 * Component-wise multiply.
 */
inline data_type operator*(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] * v2.val[i];
	}
	return res;
}

/**
 * Component-wise divide.
 */
inline data_type operator/(const data_type& v1, const data_type& v2) {
	// No size checking here!
	data_type res(v1.val.size());
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res.val[i] = v1.val[i] / v2.val[i];
	}
	return res;
}


/**
 * Dot product.
 */
inline real_type dot(const data_type& v1, const data_type& v2) {
	// No size checking here!
	real_type res = 0;
	for (unsigned int i = 0; i < v1.val.size(); i++) {
		res += v1.val[i] * v2.val[i];
	}
	return res;
}

inline data_type operator*(const real_type r, const data_type& v) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); i++) {
		res.val[i] = r * v.val[i];
	}
	return res;
}

inline data_type operator*(const data_type& v, const real_type r) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); i++) {
		res.val[i] = v.val[i] * r;
	}
	return res;
}

inline data_type operator/(const data_type& v, const real_type r) {
	data_type res(v.val.size());
	for (unsigned int i = 0; i < v.val.size(); i++) {
		res.val[i] = v.val[i] / r;
	}
	return res;
}


inline std::ostream& operator<<(std::ostream& out, const data_type &v) {
	out << "[";
	for (unsigned int i = 0; i < v.val.size(); i++) {
		if (i > 0) {
			out << ",";
		}
		out << v.val[i];
	}
	out << "]";
	return out;
}

inline QTextStream& operator<<(QTextStream& out, const data_type &v) {
	out << "[";
	for (int i = 0; i < v.val.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v.val[i];
	}
	out << "]";
	return out;
}



/**
* Edge type.
* Contains the endpoints of the edge (two vertices) and the two adjacent triangles.
* In particular we know that the edge is on the boundary if there is only one
* adjacent triangle.
*/
class Edge {
public:
	
	Edge() : vi(0, 0), ti(0, 0) {

	}

	/**
	 * Determine if the edge is on the boundary.
	 *
	 * This method assumes that the edge list has been constructed in full.
	 * After construction of the edge list, triangle index 0 can only be on the first element,
	 * since triangle 0 is the first triangle to be processed and its three edges are added in the first iteration.
	 * Therefore, if the second triangle index is the default value 0, then it is a boundary edge.
	 */
	bool isBoundary()
	{
		return (ti.y() == 0);
	}

	index2_type vi;		// adjacent vertices
	index2_type ti;		// adjacent triangles
};

typedef std::vector<Edge> EdgeList;

// Adjacency list
typedef std::vector<std::vector<index2_type> > AdjList;

#endif
