///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "util.h"
#include <iomanip>
#include <QRegExp>
#include <QFile>
#include <QScrollBar>
#include <QLineEdit>
#include <glut.h>
#include "ShapeModel.h"
#include <fstream>

extern "C" {
#include <SFMT/SFMT.h>
}


stdext::hash_map<std::string, TimerPair> perfTimers;

/**
* Write the contents of a vector to the stream.
*/
std::ostream& operator<<(std::ostream& out, const Geometry::Vec3f &v) {
	out << "[";
	for (int i = 0; i < 3; i++) {
		//out << std::setprecision(2) << setiosflags(std::ios_base::fixed)<< v[i] << ",";
		//out << std::setprecision(5) << setiosflags(std::ios_base::scientific)<< v[i] << ",";
		out << v[i] << ",";
	}
	out << "\b]";
	return out;
}

QTextStream& operator<<(QTextStream& out, const Geometry::Vec3f &v) {
	out << "[" << v[0] << "," << v[1] << "," << v[2] << "]";
	return out;
}

std::ostream& operator<<(std::ostream& out, const FloatArray &v) {
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

QTextStream& operator<<(QTextStream& out, const FloatArray &v) {
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

std::ostream& operator<<( std::ostream& out, const std::vector<int> &v )
{
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

std::ostream& operator<<( std::ostream& out, const std::vector<bool> &v )
{
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << (v[i]?1:0);
	}
	out << "]";
	return out;
}

std::ostream& operator<<( std::ostream& out, const idxList &v )
{
	out << "[";
	for (int i = 0; i < v.size(); i++) {
		if (i > 0) {
			out << ", ";
		}
		out << v[i];
	}
	out << "]";
	return out;
}

//std::ostream& operator<<( std::ostream& out, const distList &v )
//{
//	out << "[";
//	for (int i = 0; i < v.size(); i++) {
//		if (i > 0) {
//			out << ", ";
//		}
//		out << v[i];
//	}
//	out << "]";
//	return out;
//}


void setOpenGLColor( QRgb color )
{
	unsigned char r = (unsigned char) qRed(color);
	unsigned char g = (unsigned char) qGreen(color);
	unsigned char b = (unsigned char) qBlue(color);
	unsigned char a = (unsigned char) qAlpha(color);
	glColor4ub(r,g,b,a);
}

void setOpenGLColor( const RealColor &color, real_type alpha/* = 1.0*/ )
{
	glColor4f(color[0], color[1], color[2], alpha);
}

void setOpenGLColor( const FloatArray &color, real_type alpha/* = 1.0*/ )
{
	if (color.size() < 3) {
		std::cerr << "Error in setOpenGLMaterialColor(): not enough colors!" << std::endl;
		return;
	}

	glColor4f(color[0], color[1], color[2], alpha);
}

float interpSlider(QScrollBar *slider) {
	return ((float)slider->value()) / ((float)(slider->maximum() - slider->minimum()));
}

void setInterpSlider(QScrollBar *slider, float interp) {
	slider->setValue((slider->maximum() - slider->minimum())*interp + slider->minimum());
}

float QLineEditFloatValue(QLineEdit *edit) {
	bool ok;
	float value = edit->text().toFloat(&ok);

	if (!ok) {
		// Just set the value to zero
		value = 0.f;
		edit->setText(QString("%1").arg(value));
	}

	return value;
}

void outputToTextStream(QTextStream &out, int ind, Geometry::Matrix33f rot, Geometry::Vec3f tr) {
	// Print
	out << "R{" << ind+1 << "} = ";
	out << "[" << rot(0,0) << ", " << rot(0,1) << ", " << rot(0,2) << "\n        ";
	out <<        rot(1,0) << ", " << rot(1,1) << ", " << rot(1,2) << "\n        ";
	out <<        rot(2,0) << ", " << rot(2,1) << ", " << rot(2,2) << "];\n";

	out << "T{" << ind+1 << "} = ";
	out << "[" << tr[0] << ", " << tr[1] << ", " << tr[2] << "];\n";
	out << "\n";
}

void drawBoxOpenGL( const real3_type &minPos, const real3_type &maxPos )
{
	real3_type centerPos = (minPos + maxPos) / 2;

	glBegin(GL_LINES);
	// diagonal
	//glVertex3f(minPos[0], minPos[1], minPos[2]);
	//glVertex3f(maxPos[0], maxPos[1], maxPos[2]);

	// 3 lines from min pos
	//glColor3f(1.f,0.f,0.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], minPos[1], minPos[2]);

	//glColor3f(0.f,1.f,0.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(minPos[0], maxPos[1], minPos[2]);

	//glColor3f(0.f,0.f,1.f);
	glVertex3f(minPos[0], minPos[1], minPos[2]);
	glVertex3f(minPos[0], minPos[1], maxPos[2]);

	// 2 lines from min pos with x set to max pos
	//glColor3f(1.f,1.f,1.f);
	glVertex3f(maxPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);

	glVertex3f(maxPos[0], minPos[1], minPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	// 2 lines from min pos with y set to max pos
	glVertex3f(minPos[0], maxPos[1], minPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);

	glVertex3f(minPos[0], maxPos[1], minPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	// 2 lines from min pos with z set to max pos
	glVertex3f(minPos[0], minPos[1], maxPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	glVertex3f(minPos[0], minPos[1], maxPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	// 3 lines from max pos
	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(minPos[0], maxPos[1], maxPos[2]);

	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(maxPos[0], minPos[1], maxPos[2]);

	glVertex3f(maxPos[0], maxPos[1], maxPos[2]);
	glVertex3f(maxPos[0], maxPos[1], minPos[2]);
	glEnd();	// End of: bounding box drawing

	//glBegin(GL_LINES);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0]-1, centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1]-1, centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]);
	//glVertex3f(centerPos[0], centerPos[1], centerPos[2]-1);
	//glEnd();
}

void drawSphere( const real3_type &pos, const float radius, GLint slices/* = 10*/, GLint stacks/* = 10*/ )
{
	glPushMatrix();
		glTranslatef(pos[0], pos[1], pos[2]);
		glutSolidSphere(radius, slices, stacks);
	glPopMatrix();
}

void drawBoxFilled( const real3_type& min_pos, const real3_type& max_pos )
{
	glBegin(GL_QUADS);
	// Front Face
	glNormal3f( 0.0f, 0.0f, 1.0f);// Normal Pointing Towards Viewer
	glTexCoord2f(0.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], max_pos[2]);// Point 1 (Front)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], max_pos[2]);// Point 2 (Front)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], max_pos[2]);// Point 3 (Front)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], max_pos[2]);// Point 4 (Front)
	// Back Face
	glNormal3f( 0.0f, 0.0f,-1.0f);// Normal Pointing Away From Viewer
	glTexCoord2f(1.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], min_pos[2]);// Point 1 (Back)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], min_pos[2]);// Point 2 (Back)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], min_pos[2]);// Point 3 (Back)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], min_pos[2]);// Point 4 (Back)
	// Top Face
	glNormal3f( 0.0f, 1.0f, 0.0f);// Normal Pointing Up
	glTexCoord2f(0.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], min_pos[2]);// Point 1 (Top)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(min_pos[0], max_pos[1], max_pos[2]);// Point 2 (Top)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(max_pos[0], max_pos[1], max_pos[2]);// Point 3 (Top)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], min_pos[2]);// Point 4 (Top)
	// Bottom Face
	glNormal3f( 0.0f,-1.0f, 0.0f);// Normal Pointing Down
	glTexCoord2f(1.0f, 1.0f); glVertex3f(min_pos[0], min_pos[1], min_pos[2]);// Point 1 (Bottom)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(max_pos[0], min_pos[1], min_pos[2]);// Point 2 (Bottom)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], max_pos[2]);// Point 3 (Bottom)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], max_pos[2]);// Point 4 (Bottom)
	// Right face
	glNormal3f( 1.0f, 0.0f, 0.0f);// Normal Pointing Right
	glTexCoord2f(1.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], min_pos[2]);// Point 1 (Right)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], min_pos[2]);// Point 2 (Right)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(max_pos[0], max_pos[1], max_pos[2]);// Point 3 (Right)
	glTexCoord2f(0.0f, 0.0f); glVertex3f(max_pos[0], min_pos[1], max_pos[2]);// Point 4 (Right)
	// Left Face
	glNormal3f(-1.0f, 0.0f, 0.0f);// Normal Pointing Left
	glTexCoord2f(0.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], min_pos[2]);// Point 1 (Left)
	glTexCoord2f(1.0f, 0.0f); glVertex3f(min_pos[0], min_pos[1], max_pos[2]);// Point 2 (Left)
	glTexCoord2f(1.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], max_pos[2]);// Point 3 (Left)
	glTexCoord2f(0.0f, 1.0f); glVertex3f(min_pos[0], max_pos[1], min_pos[2]);// Point 4 (Left)
	glEnd();// Done Drawing Quads
}


Geometry::Vec3f VecToVec3f( qglviewer::Vec v )
{
	return Geometry::Vec3f(v[0], v[1], v[2]);
}

qglviewer::Vec Vec3fToVec( Geometry::Vec3f v )
{
	return qglviewer::Vec(v[0], v[1], v[2]);
}

real_type meshNorm( const ShapeModel &mesh1, const ShapeModel &mesh2 )
{
	size_type numVertices = mesh1.numVertices();
	real_type totalError = 0;

	if (numVertices != mesh2.numVertices()) {
		std::cerr << "Error in meshNorm(): shapes have different number of vertices!";
		return -1;
	}

	for (index_type i = 0; i < numVertices; i++) {
		real3_type diff = mesh1.vertex[i] - mesh2.vertex[i];
		totalError += diff.square_norm();
	}

	return totalError;
}

void setOpenGLMaterialColor( const FloatArray floatColor, real_type alpha /*= 1.f*/ )
{
	if (floatColor.size() < 3) {
		std::cerr << "Error in setOpenGLMaterialColor(): not enough colors!" << std::endl;
		return;
	}

	static GLfloat color[4];
	for (int d = 0; d < 3; d++) {
		color[d] = floatColor[d];
	}
	color[3] = alpha;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
}

void setOpenGLMaterialColor( const RealColor floatColor, real_type alpha /*= 1.f */ )
{
	static GLfloat color[4];
	for (int d = 0; d < 3; d++) {
		color[d] = floatColor[d];
	}
	color[3] = alpha;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
}

void TextProgress( double x, bool init/* = false*/ )
{
	static const int bins = 100;
	static int currentBin, binIncrement;
	static double binSize;

	// Initialize progress display
	if (init) {
		binSize = x / ((double)bins);
		currentBin = 0;
		binIncrement = std::max(1, (int)(1.f/binSize));
		std::cout << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
	}
	// Update progress display
	else {
		if (x >= ((double)currentBin)*binSize) {
			currentBin += binIncrement;
			std::cout << "\b\b\b\b";
			if (currentBin >= 1000.1) {
				std::cout << "\b";
			}
			if (currentBin >= 10000.1) {
				std::cout << "\b";
			}
			std::cout << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
		}
	}
}

void TextProgressWithLabel( double x, bool init /*= false*/, QString id /*= QString() */ )
{
	static const int bins = 100;
	static int currentBin, binIncrement;
	static double binSize;
	int idLength = id.length();

	// Initialize progress display
	if (init) {
		binSize = x / ((double)bins);
		currentBin = 0;
		binIncrement = std::max(1, (int)(1.f/binSize));
		std::cout << "id " << qPrintable(id) << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
	}
	// Update progress display
	else {
		if (x >= ((double)currentBin)*binSize) {
			currentBin += binIncrement;

			// Backspace
			for (int i = 0; i < idLength+10; i++) {
				std::cout << "\b";
			}

			// Print out next progress
			std::cout << "id " << qPrintable(id) << " - " << qPrintable(QString("%1%").arg(currentBin, 3, 10, QLatin1Char('0')));
		}
	}
}


Geometry::Matrix33f rotateAxis(const Geometry::Vec3f &axis, const float cosAngle) {
	float sinAngle = sqrtf(1.f - cosAngle*cosAngle);

	// use rodrigues' formula
	Geometry::Matrix33f ssm;
	ssm(0,0) = 0.f;   ssm(0,1) = -axis[2]; ssm(0,2) = axis[1];
	ssm(1,0) = axis[2];  ssm(1,1) = 0.f;   ssm(1,2) = -axis[0];
	ssm(2,0) = -axis[1]; ssm(2,1) = axis[0];  ssm(2,2) = 0.f;

	Geometry::Matrix33f rot, id;
	id = Geometry::Matrix33f::identity();
	rot = id + ssm*sinAngle + ssm*ssm*(1.f - cosAngle);

	return rot;
}

void writeBinary( std::ofstream &out, const Geometry::Matrix33f m )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			temp = m(i,j);
			out.write((char*) &temp, sizeof(float));
		}
	}
}

void writeBinary( std::ofstream &out, const Geometry::Vec3f v )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		temp = v[i];
		out.write((char*) &temp, sizeof(float));
	}
}

void readBinary( std::ifstream &in, Geometry::Matrix33f &m )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 3; j++) {
			in.read((char*) &temp, sizeof(float));
			m(i,j) = temp;
		}
	}
}

void readBinary( std::ifstream &in, Geometry::Vec3f &v )
{
	float temp;

	// Write one row at a time
	for (int i = 0; i < 3; i++) {
		in.read((char*) &temp, sizeof(float));
		v[i] = temp;
	}
}



idxList bestCandidateSample( const PointList &pts, int n )
{
	idxList indices;

	if ((n <= 0) || (n >= pts.size()) || (pts.size() <= 10)) {
		// Just use all the vertices in the mesh
		indices.resize(pts.size());
		for (index_type i = 0; i < pts.size(); i++) {
			indices[i] = i;
		}
	}
	else {
		indices.resize(n);

		int pick, bestPick;
		idxList nearest;
		distList dist;
		real_type maxDist;
		const int pickNum = 10;
		PointList pickedPts;
		int remainingPoints = pts.size();

		// Initialize array marking which vertices are picked
		std::vector<bool> markedPicked;
		std::vector<int> pickCandidates;
		markedPicked.resize(pts.size(), false);
		pickCandidates.resize(pickNum);

		// Initialize by adding one random point
		pick = floor(((double)pts.size()) * genrand_real1());
		pickedPts.clear();
		pickedPts.reserve(n);
		pickedPts.add(pts[pick]);
		markedPicked[pick] = true;
		indices[0] = pick;

		// Remaining samples
		for (index_type i = 1; i < n; i++) {

			maxDist = -REAL_TYPE_MAX;
			for (int j = 0; j < pickNum; j++) {
				// 1. Pick candidate
				do {
					pick = floor(((double)pts.size()) * genrand_real1());
				} while (markedPicked[pick]);

				pickCandidates[j] = pick;
				markedPicked[pick] = true;

				// 2. Compute distance to current selected point set
				pickedPts.queryKNN(pts[pick], 1, nearest, dist, true);

				if (dist[0] > maxDist) {
					maxDist = dist[0];
					bestPick = pick;
				}
			}

			// Add the best candidate
			pickedPts.add(pts[bestPick]);
			indices[i] = bestPick;
			--remainingPoints;
			//pickedPts.clearANNData();

			// Un-mark the discarded candidates
			for (int j = 0; j < pickNum; j++) {
				pick = pickCandidates[j];
				if (pick != bestPick) {
					markedPicked[pick] = false;
				}
			}
		}
	}

	return indices;
}


void generateLightColors( int n, std::vector<RealColor> &colors )
{
	// Generate a set of random light colors using best candidate sampling
	const real_type maxThreshold = 0.7;
	const int numCandidates = n*20;
	real_type maxValue;
	PointList candidates;
	candidates.resize(numCandidates);

	for (int i = 0; i < numCandidates; i++) {
		// Use rejection sampling:  Just make sure that the maximum rgb value is at least ~0.7
		do {
			candidates[i][0] = genrand_real1();
			candidates[i][1] = genrand_real1();
			candidates[i][2] = genrand_real1();
			maxValue = std::max(candidates[i][0], std::max(candidates[i][1],candidates[i][2]));
		} while ( maxValue < maxThreshold );
	}

	// Apply best candidate sampling
	idxList best = bestCandidateSample(candidates, n);

	// Store the result
	colors.resize(n);
	for (int i = 0; i < n; i++) {
		colors[i] = candidates[best[i]];
	}
}


void writeVector( double* x, int size, const char *filename )
{
	std::ofstream out;
	out.open(filename);

	if (out.fail()) {
		std::cerr << "Error in writeVector(): Could not open output file " << filename << " for writing!" << std::endl;
		return;
	}

	// Write data
	for (index_type i = 0; i < size; i++) {
		out << x[i] << std::endl;
	}
	out.close();
}

