///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#ifndef __UTIL_H__
#define __UTIL_H__

#include <algorithm>
#include <SylvainTools/geom.h>
#include <iostream>
#include <QString>
#include <QtOpenGL>
#include <QTextStream>
#include <Windows.h>
#include <QColor>
#include <libXMLParam/xmlParam.h>
#include "types.h"
#include <QGLViewer/qglviewer.h>
#include <hash_map>
#include <fstream>

#undef min
#undef max

// Useful functions
std::ostream& operator<<(std::ostream& out, const Geometry::Vec3f &v);
QTextStream& operator<<(QTextStream& out, const Geometry::Vec3f &v);
std::ostream& operator<<(std::ostream& out, const FloatArray &v);
QTextStream& operator<<(QTextStream& out, const FloatArray &v);
std::ostream& operator<<(std::ostream& out, const std::vector<int> &v);
std::ostream& operator<<(std::ostream& out, const std::vector<bool> &v);
std::ostream& operator<<(std::ostream& out, const idxList &v);
std::ostream& operator<<( std::ostream& out, const distList &v );

void ComputeStep1dCDF(float *f, int nSteps, float *c, float *cdf);
float SampleStep1d(float *f, float *cdf, float c, int nSteps, float u, float *pdf);



// High-performance timer collection. adapted from
// http://blog.kowalczyk.info/kb/high-resolution-timer-for-timing-code-fragments.html
// very simple, high-precision (at least in theory) timer for timing API calls
struct TimerPair {
	TimerPair() : totalDuration(0.0), iterations(0.0) {
		QueryPerformanceFrequency(&freq);
	}

	void Start() {
		totalDuration = 0.0;
		iterations = 0.0;
		QueryPerformanceCounter(&mTimeStart);
	}

	void AddStart() {
		iterations += 1.0;
		QueryPerformanceCounter(&mTimeStart);
	}

	void Stop() {
		QueryPerformanceCounter(&mTimeStop);
		totalDuration += (double) (mTimeStop.QuadPart - mTimeStart.QuadPart);
	}

	double GetDurationInSecs() {
		return (totalDuration / ((double)freq.QuadPart));
	}

	LARGE_INTEGER freq;
	LARGE_INTEGER mTimeStart;
	LARGE_INTEGER mTimeStop;
	double totalDuration;
	double iterations;
};

// A collection of performance timers
extern stdext::hash_map<std::string, TimerPair> perfTimers;

// Print timer
inline void printTime(const std::string str) {
	std::cout << "***** " << str << " took " << perfTimers[str].GetDurationInSecs() << " seconds." << std::endl;
}


void setOpenGLColor(QRgb color);
void setOpenGLColor( const RealColor &color, real_type alpha = 1.0 );
void setOpenGLColor( const FloatArray &color, real_type alpha = 1.0 );
void setOpenGLMaterialColor( const RealColor floatColor, real_type alpha = 1.f );
void setOpenGLMaterialColor( const FloatArray floatColor, real_type alpha = 1.f );

// UI controller utility functions
class QScrollBar;
class QLineEdit;
float interpSlider(QScrollBar *slider);
void setInterpSlider(QScrollBar *slider, float interp);
float QLineEditFloatValue(QLineEdit *edit);
void outputToTextStream(QTextStream &out, int ind, Geometry::Matrix33f rot, Geometry::Vec3f tr);

// Box drawing
void drawBoxOpenGL( const real3_type &minPos, const real3_type &maxPos );
void drawBoxFilled( const real3_type& min_pos, const real3_type& max_pos );

// Sphere drawing
void drawSphere( const real3_type &pos, const float radius, GLint slices = 10, GLint stacks = 10 );
qglviewer::Vec Vec3fToVec( Geometry::Vec3f v );
Geometry::Vec3f VecToVec3f( qglviewer::Vec v );
QPoint clickedCoord( QMouseEvent *e );

// Enumerate correspondence display modes.
enum DISPLAYMODE {
	DISPLAY_SOURCE = 1,
	DISPLAY_TARGET,
	DISPLAY_SOURCE_AND_TARGET
};

/**
 * Compute total error between two meshes.
 * Meshes must be in per-vertex correspondence.
 */
class ShapeModel;
real_type meshNorm(const ShapeModel &mesh1, const ShapeModel &mesh2);

/**
 * Display a text progress bar.
 * @param x     The current index (when update), or the maximum index (when init).
 * @param init  As true, this initializes the progress counter.  As false, it updates the counter.
 */
void TextProgress( double x, bool init = false );
void TextProgressWithLabel( double x, bool init = false, QString id = QString() );


// Rotation matrix about an axis.
Geometry::Matrix33f rotateAxis(const Geometry::Vec3f &axis, const float cosAngle);

void writeBinary( std::ofstream &out, const Geometry::Matrix33f m );
void writeBinary( std::ofstream &out, const Geometry::Vec3f v );
void readBinary( std::ifstream &in, Geometry::Matrix33f &m );
void readBinary( std::ifstream &in, Geometry::Vec3f &v );

// Best candidate sampling within a given point set.
class PointList;
idxList bestCandidateSample( const PointList &pts, int n );

// Generate light colors
void generateLightColors( int n, std::vector<RealColor> &colors );

// Write out the contents of a vector
void writeVector( double* x, int size, const char *filename );
#endif
