///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "viewer.h"
#include <fstream>
#include "scene.h"
#include "DeformEdit.h"
#include "CorrespViewer.h"
#include "FindCorrespondence.h"
#include "geom_util.h"
#include "FileSequence.h"
#include "DeformEM.h"

extern "C" {
#include <SFMT/SFMT.h>
}

using namespace std;

Viewer::Viewer(Scene *s) :
scene(s), displayFrame(0), sceneChanged(false),
deformEdit(NULL), maxGraphNodes(1000), maxGraphNeighbors(50), maxOverlapFactor(0.1f),
correspViewer(NULL), tempList(NULL), interactionMode_(NONE), sourceFeatures(NULL),
targetFeatures(NULL), displayScale(0),
autoLaunchDone(false), maxFeatureSpaceRange(0.03),
curvDisplayMode(1), sourceQueryIndex(-1),
showErrorComponent(0)
{
	// Attach to subject
	scene->Attach(this);
}

Viewer::~Viewer()
{
	// Detach from subject
	scene->Detach(this);

	// Close all sub-windows
	if (correspViewer) {
		correspViewer->close();
	}
	if (deformEdit) {
		deformEdit->close();
	}
}



void Viewer::drawFrame( const ShapeModel* frame )
{
	// Display the shape
	if (showFeaturesAct->isChecked()) {
		frame->drawFeatureValues(*scene, displayScale);
	}
	else if (showCurvaturesAct->isChecked()) {
		frame->drawCurvatureValues(*scene, displayScale, curvDisplayMode);
	}
	else if (showFramesAct->isChecked()) {
		frame->drawFrames(*scene, displayScale, curvDisplayMode);
	}
	else if (showStatusAct->isChecked()) {
		frame->drawWithStatus();
	}
	else if (showCellIndexAct->isChecked()) {
		if (frame->cellIndex.size()) {
			glDisable(GL_LIGHTING);
			frame->drawCellIndex();
			glEnable(GL_LIGHTING);
		}
	}
	else if (showGeometryAct->isChecked()) {
		frame->draw();
	}

	// Draw vertices
	if (showVerticesAct->isChecked()) {
		setOpenGLMaterialColor(scene->pointColor);
		for (index_type i = 0; i < frame->numVertices(); i++) {
			drawSphere(frame->vertex[i], scene->pointSize() / 2.0);
		}
	}

	// Draw frames
	if (showVertexFramesAct->isChecked()) {
		int compMax = frame->frameScales.size();
		int comp = min(compMax-1, displayScale);

		for (index_type i = 0; i < frame->numVertices(); i++) {
			const CurvInfo &coordFrame = frame->frames[i][comp];
			qglviewer::Vec from = Vec3fToVec(frame->vertex[i]);
			qglviewer::Vec vmin = Vec3fToVec(coordFrame.vmin);
			qglviewer::Vec vmax = Vec3fToVec(coordFrame.vmax);
			qglviewer::Vec vn = Vec3fToVec(coordFrame.n);

			setOpenGLMaterialColor(RealColor(0.6, 0.6, 1.0));
			drawArrow(from, from + scene->pointSize()*vmin, scene->pointSize() / 25.0);

			setOpenGLMaterialColor(RealColor(1.0, 0.6, 0.6));
			drawArrow(from, from + scene->pointSize()*vmax, scene->pointSize() / 25.0);

			setOpenGLMaterialColor(scene->pointColor);
			drawArrow(from, from + scene->pointSize()*vn, scene->pointSize() / 25.0);
		}
	}


	// Display text status
	if (showCurvaturesAct->isChecked() || showFramesAct->isChecked()) {
		setOpenGLColor(foregroundColor().rgba());

		switch(curvDisplayMode) {
			case CURVINFO_NORMALS:
				drawText(10, height() - 24, tr("Showing frame normals"));
				break;
			case CURVINFO_MIN_CURVATURE:
				drawText(10, height() - 24, tr("Showing frame minimum principal curvatures"));
				break;
			case CURVINFO_MAX_CURVATURE:
				drawText(10, height() - 24, tr("Showing frame maximum principal curvatures"));
				break;
			default:
				break;
		}
	}
}


/**
 * Draw the scene.
 */
void Viewer::draw()
{
	if (!autoLaunchDone) {
		autoLaunch();
	}

	//////////////////////////////////////////////////////////////////////////
	// Store current state
	//////////////////////////////////////////////////////////////////////////
	GLboolean colorMaterial, lighting;
	GLint polygonMode;
	glGetBooleanv(GL_LIGHTING, &lighting);
	glGetBooleanv(GL_COLOR_MATERIAL, &colorMaterial);
	glGetIntegerv(GL_POLYGON_MODE, &polygonMode);

	//////////////////////////////////////////////////////////////////////////
	// Set state
	//////////////////////////////////////////////////////////////////////////
	// Toggle wireframe display
	if (showWireframe()) {
		glDisable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	}
	else {
		//glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		//glCullFace(GL_BACK);
	}

	// Set up the material
	glEnable(GL_LIGHTING);
	glDisable(GL_COLOR_MATERIAL);
	

	// Draw the deformation result
	scene->drawDeformation(getDeformDrawOptions());


	//////////////////////////////////////////////////////////////////////////
	// Feature proximity testing
	drawSphereTest();

	// Draw correspondence candidates
	setOpenGLMaterialColor(scene->pointColor);
	PointList_Utility::draw(sourceQuery, POINT_SPHERES, scene->pointSize() * 0.5);
	setOpenGLMaterialColor(scene->selectedColor);
	PointList_Utility::draw(targetMatch, POINT_SPHERES, scene->pointSize() * 0.5);


	//////////////////////////////////////////////////////////////////////////
	// Draw axis at scene center
	if (axisIsDrawn()) {
		qglviewer::Vec center = sceneCenter();
		glPushMatrix();
		glTranslatef(center[0], center[1], center[2]);
		QGLViewer::drawAxis(sceneRadius()); // Or any scale
		glPopMatrix();
	}

	//////////////////////////////////////////////////////////////////////////
	// Restore OpenGL state
	//////////////////////////////////////////////////////////////////////////
	glPolygonMode(GL_FRONT_AND_BACK, polygonMode);

	//if (colorMaterial)glEnable(GL_COLOR_MATERIAL);
	//else glDisable(GL_COLOR_MATERIAL);

	//if (lighting) glEnable(GL_LIGHTING);
	//else glDisable(GL_LIGHTING);
	glDisable(GL_LIGHTING);
}

void Viewer::pointListTest() {
	GLfloat color[4];

	//////////////////////////////////////////////////////////////////////////
	// Unit test for testing capabilities of PointList
	if (!tempList) {
		// Create a random point list
		tempList = new PointList();
		PointList_Utility::generateRandom(*tempList, 1000);
	}
	color[0] = 1.f;  color[1] = 1.f;  color[2] = 0.0f;  color[3] = 1.0f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	PointList_Utility::draw(*tempList, POINT_SPHERES, 0.005);

	idxList list;
	distList dist;
	float rad = 1;
	real3_type q(0, 0, 0);
	//tempList->queryRange(q, rad*rad, list, dist, true);
	tempList->queryKNN(q, 100, list, dist, true);
	color[0] = 0.f;  color[1] = 0.f;  color[2] = 1.f;  color[3] = 1.f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	PointList_Utility::draw(*tempList, POINT_SPHERES, list, 0.006);

	glEnable(GL_BLEND);
	glDepthMask (GL_FALSE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	color[0] = 0.f;  color[1] = 1.f;  color[2] = 1.f;  color[3] = 0.5f;
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	drawSphere(q, rad, 15, 15);
	glDepthMask (GL_TRUE);
	glDisable(GL_BLEND);
}


void Viewer::colormapTest() {
	// Colormap rendering test using spiral
	const float nbSteps = 200.0;
	glBegin(GL_QUAD_STRIP);
	for (int i=0; i<nbSteps; ++i)
	{
		const RealColor color = scene->colormap.getColor(i/(nbSteps-1));
		const float ratio = i/nbSteps;
		const float angle = 21.0f*ratio;
		const float c = cos(angle);
		const float s = sin(angle);
		const float r1 = 1.0f - 0.8f*ratio;
		const float r2 = 0.8f - 0.8f*ratio;
		const float alt = ratio - 0.5f;
		const float nor = 0.5f;
		const float up = sqrt(1.0f-nor*nor);

		setOpenGLColor(color);
		setOpenGLMaterialColor(color);
		//cout << "Color: " << i/(nbSteps-1) << ", " << color << endl;
		glNormal3f(nor*c, up, nor*s);
		glVertex3f(r1*c, alt, r1*s);
		setOpenGLMaterialColor(color);
		glNormal3f(nor*c, up, nor*s);
		glVertex3f(r2*c, alt+0.05f, r2*s);
	}
	glEnd();
}

void Viewer::drawSphereTest()
{
	if (scene->source) {
		glDisable(GL_LIGHTING);
		glColor3f(1,1,1);

		// Plot the selected points
		glPointSize(5.f);
		glBegin(GL_POINTS);
		for (index_type i = 0; i < inRange.size(); i++) {
			real3_type p = scene->source->vertex[inRange[i]];
			glVertex3f(p[0],p[1],p[2]);
		}
		glEnd();

		// Plot the selected normals
		glBegin(GL_LINES);
		for (index_type i = 0; i < inRange.size(); i++) {
			real3_type p1 = scene->source->vertex[inRange[i]];
			real3_type p2 = p1 + 5*scene->pointSize()*scene->source->normal[inRange[i]];
			glVertex3f(p1[0],p1[1],p1[2]);
			glVertex3f(p2[0],p2[1],p2[2]);
		}
		glEnd();

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCullFace(GL_BACK);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// Draw fitted sphere
		setOpenGLMaterialColor(RealColor(0.8,0.8,1.0), 0.5);
		drawSphere(fitCenter, fitRadius, 30, 30);

		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
	}

	// Test Sphere Fitting
	if (interactionMode_ == DEFINE_SPHERE) {
		glDisable(GL_LIGHTING);
		setOpenGLColor(foregroundColor().rgba());
		glBegin(GL_LINES);
		glVertex3f(clickStartPos[0],clickStartPos[1],clickStartPos[2]);
		glVertex3f(clickEndPos[0],clickEndPos[1],clickEndPos[2]);
		glEnd();

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glCullFace(GL_BACK);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		setOpenGLMaterialColor(RealColor(0.8,0.8,1.0), 0.5);
		drawSphere(clickStartPos, (clickEndPos-clickStartPos).norm(), 30, 30);
		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
		glDisable(GL_CULL_FACE);
	}	
}


void Viewer::animate()
{
	if (sceneChanged) {
		//cerr << "Scene contents updated!" << endl;
		updateGL();

		//// Automatically save a snapshot
		//saveSnapshot();
		//cerr << "***** Saved snapshot!" << endl;

		sceneMutex.lock();
		sceneChanged = false;
		scene->checkExit();
		sceneMutex.unlock();
	}
}


/**
 * Initialize the viewer.
 */
void Viewer::init()
{
	// Create context menu.
	createContextMenu();

	// Restore previous viewer state.
	restoreStateFromFile();

	/////////////////////////////////////////////////////
	//       Keyboard shortcut customization           //
	//      Changes standard action key bindings       //
	/////////////////////////////////////////////////////

	// Define 'Control+Q' as the new exit shortcut (default was 'Escape')
	setShortcut(EXIT_VIEWER, Qt::CTRL+Qt::Key_Q);

	// Set 'Control+F' as the FPS toggle state key.
	setShortcut(DISPLAY_FPS, Qt::CTRL+Qt::Key_F);

	// Set 'Control+Y' as the camera fly mode toggle key.
	setShortcut(CAMERA_MODE, Qt::CTRL+Qt::Key_Y);

	// Disable some shortcuts
	setShortcut(MOVE_CAMERA_LEFT, Qt::NoButton);
	setShortcut(MOVE_CAMERA_RIGHT, Qt::NoButton);
	setShortcut(MOVE_CAMERA_UP, Qt::NoButton);
	setShortcut(MOVE_CAMERA_DOWN, Qt::NoButton);

	// Disable draw grid toggle shortcut (default was 'G')
	//setShortcut(DRAW_GRID, 0);

	// Add custom key description (see keyPressEvent).
	setKeyDescription(Qt::Key_Space, "Opens Viewer Menu");
	setKeyDescription(Qt::CTRL+Qt::Key_Left, "Shows previous frame in the animation");
	setKeyDescription(Qt::CTRL+Qt::Key_Right, "Shows next frame in the animation");
	setKeyDescription(Qt::KeypadModifier+Qt::Key_Enter, "Generate deformation graph");
	setKeyDescription(Qt::CTRL+Qt::Key_Return, "Generates and solves for the deformation");
	setKeyDescription(Qt::SHIFT+Qt::Key_Return, "Step one iteration in the deformation solver");
	setKeyDescription(Qt::Key_F, "Flip between displaying source and target model");

	//////////////////////////////////////////////////////////////////////////
	// Initialize rendering
	//////////////////////////////////////////////////////////////////////////
	setBackgroundColor(globalBackgroundColor);
	setForegroundColor(globalForegroundColor);

	// Set snapshot parameters
	setSnapshotQuality(0);
	setSnapshotFormat(tr("PNG"));
	//QDir::setCurrent(tr(scene->rootDir.c_str()));

	//// Opens help window
	//help();

	// Start the animation
	startAnimation();

	// Initialize scene variables
	scene->init();
}

void Viewer::autoLaunch()
{
	// Hide current viewer
	this->hide();

	// If there is a specific source and target file, load it
	if (scene->defaultSourceFile.length() && scene->defaultTargetFile.length()) {
		scene->createNewAnimation();
		scene->addAnimationFrame(tr(scene->defaultSourceFile.c_str()), scene->autoLoadFrameData);
		scene->addAnimationFrame(tr(scene->defaultTargetFile.c_str()), scene->autoLoadFrameData);

		scene->sourceFrame = 0;
		scene->targetFrame = 1;
		scene->source = (*scene->getAnimation())[scene->sourceFrame];
		scene->target = (*scene->getAnimation())[scene->targetFrame];

		// Select target mesh for viewing
		displayFrame = 1;
		fitViewToShape(scene->target);
		setTargetModelShowAct->setChecked(true);
	}
	else if (scene->defaultFile.length()) {
		// Open the default file, if specified
		cout << "Opening file: " << scene->defaultFile << endl;
		open(scene->defaultFile.c_str(), scene->defaultLoadFrames, scene->defaultSkipFrames);

		// Bounds checking on the frame number
		int maxFrame = scene->getAnimation()->numFrames() - 1;
		if (scene->sourceFrame > maxFrame) {
			scene->sourceFrame = maxFrame;
			cout << "Source Frame # too large; clamping to " << scene->sourceFrame << "." << endl;
		}
		if (scene->targetFrame > maxFrame) {
			scene->targetFrame = maxFrame;
			cout << "Target Frame # too large; clamping to " << scene->targetFrame << "." << endl;
		}

		scene->source = (*scene->getAnimation())[scene->sourceFrame];
		scene->target = (*scene->getAnimation())[scene->targetFrame];
	}

	autoLaunchDone = true;
	autoProcess();
}

void Viewer::autoProcess()
{
	if (scene->autoProcessMode == AUTO_PROCESS_NONE) {
		return;
	}
	else if (scene->autoProcessMode == AUTO_PROCESS_DEFORMATION) {

		// Use DeformEM
		useDeformEMAct->setChecked(true);

		// Display deformation editor window
		CreateDeformEditor();

		// Display correspondences window & correspondences
		CreateCorrespViewer();
		correspViewer->SwitchSelectMode();

		// Automatically position windows
		if (deformEdit) {
			deformEdit->move(16, 21);
		}
		if (correspViewer) {
			correspViewer->move(700, 21);
			correspViewer->resize(800, 600);
		}

		// Automatically setup the deformation
		DeformEditorChanged(SETUP_DEFORM, 1.0);

		if (!scene->autoSaveDeformEMResult.empty()) {
			DeformEditorChanged(SOLVE_DEFORM_AUTO, 1.0);
		}
	}
}

QString Viewer::helpString() const
{
	QString text("<h2>S i m p l e V i e w e r</h2>");
	text += "Use the mouse to move the camera around the object. ";
	text += "You can respectively revolve around, zoom and translate with the three mouse buttons. ";
	text += "Left and middle buttons pressed together rotate around the camera view direction axis<br><br>";
	text += "Pressing <b>Alt</b> and one of the function keys (<b>F1</b>..<b>F12</b>) defines a camera keyFrame. ";
	text += "Simply press the function key again to restore it. Several keyFrames define a ";
	text += "camera path. Paths are saved when you quit the application and restored at next start.<br><br>";
	text += "Press <b>F</b> to display the frame rate, <b>A</b> for the world axis, ";
	text += "<b>Alt+Return</b> for full screen mode and <b>Control+S</b> to save a snapshot. ";
	text += "See the <b>Keyboard</b> tab in this window for a complete shortcut list.<br><br>";
	text += "Double clicks automates single click actions: A left button double click aligns the closer axis with the camera (if close enough). ";
	text += "A middle button double click fits the zoom of the camera and the right button re-centers the scene.<br><br>";
	text += "A left button double click while holding right button pressed defines the camera <i>Revolve Around Point</i>. ";
	text += "See the <b>Mouse</b> tab and the documentation web pages for details.<br><br>";
	text += "Press <b>Escape</b> to exit the viewer.";
	return text;
}

void Viewer::add()
{
	// Add a new frame to the animation.
	QStringList files = QFileDialog::getOpenFileNames(this, tr("Open Mesh Animation"), scene->lastDir, tr("Meshes (*.obj)"));

	if (files.size() > 0) {
		// Input from the user: process the local frame information?
		QMessageBox::StandardButton reply = QMessageBox::question(this, tr("Process local frames?"),
												tr("Load / compute local curvature frames when adding?"),
												QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
		bool processFrames = true;

		if (reply == QMessageBox::Cancel) {
			return;
		}

		if (reply == QMessageBox::No) {
			processFrames = false;
		}

		QStringList list = files;
		QStringList::Iterator it = list.begin();
		while(it != list.end()) {
			// Add the frame to the animation
			scene->addAnimationFrame(*it, processFrames);
			++it;
		}
	}
}


void Viewer::open(const char *file/* = NULL*/, int maxFrames/* = 0*/, int skipFrames/* = 1*/) {
	QString fileName;
	if (file) {
		fileName = QString(file);
	}
	else {
		fileName = QFileDialog::getOpenFileName(this, tr("Open Mesh Animation"), scene->lastDir, tr("Meshes (*.obj)"));
	}
	if (fileName.isEmpty()) {
		return;
	}

	FileSequence fs(fileName);

	if (!fs.valid()) {
		return;
	}

	scene->lastDir = grabPath(fileName);
	int numGoodFrames = 0;
	fs.skipFrame = skipFrames;

	if (fs.hasNumber()) {
		QString newFile = fs[fs.startFrame];
		int frame = fs.startFrame;

		while(QFile::exists(newFile)) {
			//cout << "Good frame " << numGoodFrames << " in file! " << qPrintable(newFile) << endl;
			numGoodFrames++;
			frame += fs.skipFrame;
			newFile = fs[frame];
		}

		fs.endFrame = frame - fs.skipFrame;
	}

	if ((maxFrames > 0) && (numGoodFrames > maxFrames)) {
		// Clamp the number of frames we are loading, if necessary
		fs.endFrame = fs.startFrame + (maxFrames - 1)*fs.skipFrame;
	}

	//////////////////////////////////////////////////////////////////////////
	// Load all the frames

	// Clear all the current data
	scene->clear();

	// Load the new data
	scene->loadNewAnimation(fs);
	displayFrame = 0;

	// Center the view on the new data
	ShapeModel &temp = *(*scene->getAnimation())[displayFrame];
	fitViewToShape(&temp);

	// Set the target mesh
	scene->source = scene->target = &temp;
	setTargetModelShowAct->setChecked(true);
}


void Viewer::clearResult()
{
	scene->clearDeform();
}

void Viewer::setGraphNodes()
{
	// Dialog box to set the number of graph nodes
	bool ok;
	int nodes = QInputDialog::getInteger(this, tr("Set Number of Graph Nodes"),
		tr("Nodes:"), scene->graph_n, 0, 1e5, 1, &ok);

	if (ok) {
		scene->graph_n = nodes;
	}
}

void Viewer::resampleGraph()
{
	scene->resampleSource();
}

bool Viewer::showWireframe() {
	return setWireframeModeAct->isChecked();
}


bool Viewer::showBbox() {
	return setBboxModeAct->isChecked();
}

bool Viewer::showSourceModel()
{
	return setSourceModelShowAct->isChecked();
}

bool Viewer::showTargetModel()
{
	return setTargetModelShowAct->isChecked();
}

bool Viewer::showDeformedModel()
{
	return setDeformedModelShowAct->isChecked();
}

int Viewer::whichDeform() {
	if (useEDAct->isChecked()) {
		return DEFORM_EMBEDDED;
	}
	else if (useWDAct->isChecked()) {
		return DEFORM_WEIGHTED;
	}
	else if (useDeformEMAct->isChecked()) {
		return DEFORM_EM;
	}
	else if (useDeformAllAct->isChecked()) {
		return DEFORM_ALL;
	}

	return 0;
}


bool Viewer::renderTransparent() {
	return setTransparencyAct->isChecked();
}

void Viewer::createContextMenu() {
	// Create menu bar.
	contextMenu = new QMenu(tr("&Context"), this);
	fileMenu = new QMenu(tr("&File"), this);
	displayMenu = new QMenu(tr("&Display"), this);
	toolsMenu = new QMenu(tr("&Tools"), this);
	optionsMenu = new QMenu(tr("&Options"), this);

	// Top-level menus
	contextMenu->addMenu(fileMenu);
	contextMenu->addMenu(displayMenu);
	contextMenu->addMenu(toolsMenu);
	contextMenu->addMenu(optionsMenu);

	//////////////////////////////////////////////////////////////////////////
	// File Menu
	openAct = new QAction(tr("&Open Sequence..."), this);
	fileMenu->addAction(openAct);
	connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

	addAct = new QAction(tr("&Add Additional Frame..."), this);
	fileMenu->addAction(addAct);
	connect(addAct, SIGNAL(triggered()), this, SLOT(add()));

	//////////////////////////////////////////////////////////////////////////
	// Tools Menu

	captureVideoAct = new QAction(tr("&Capture Animation Sequence"), this);
	toolsMenu->addAction(captureVideoAct);
	connect(captureVideoAct, SIGNAL(triggered()), this, SLOT(captureVideo()));

	//////////////////////////////////////////////////////////////////////////
	// Display Menu
	setWireframeModeAct = new QAction(tr("&Wireframe Mode"), this);
	setWireframeModeAct->setCheckable(true);
	displayMenu->addAction(setWireframeModeAct);

	setTransparencyAct = new QAction(tr("&Transparency Mode"), this);
	setTransparencyAct->setCheckable(true);
	displayMenu->addAction(setTransparencyAct);

	fitViewToShapeAct = new QAction(tr("&Fit View to Current Frame..."), this);
	displayMenu->addAction(fitViewToShapeAct);
	connect(fitViewToShapeAct, SIGNAL(triggered()), this, SLOT(fitViewToShape()));

	displayMenu->addSeparator();

	setSourceModelShowAct = new QAction(tr("&Show Source Model"), this);
	setSourceModelShowAct->setCheckable(true);
	displayMenu->addAction(setSourceModelShowAct);

	setTargetModelShowAct = new QAction(tr("&Show Target Model"), this);
	setTargetModelShowAct->setCheckable(true);
	displayMenu->addAction(setTargetModelShowAct);

	setDeformedModelShowAct = new QAction(tr("&Show Deformed Model"), this);
	setDeformedModelShowAct->setCheckable(true);
	displayMenu->addAction(setDeformedModelShowAct);

	sourceCorrespShowAct = new QAction(tr("&Show Source Correspondence Points"), this);
	sourceCorrespShowAct->setCheckable(true);
	displayMenu->addAction(sourceCorrespShowAct);

	targetCorrespShowAct = new QAction(tr("&Show Target Correspondence Points"), this);
	targetCorrespShowAct->setCheckable(true);
	displayMenu->addAction(targetCorrespShowAct);

	deformedCorrespShowAct = new QAction(tr("&Show Deformed Correspondence Points"), this);
	deformedCorrespShowAct->setCheckable(true);
	displayMenu->addAction(deformedCorrespShowAct);

	showShapeWeightsAct = new QAction(tr("&Show Weights on Mesh"), this);
	showShapeWeightsAct->setCheckable(true);
	displayMenu->addAction(showShapeWeightsAct);
	
	showAllFramesAct = new QAction(tr("&Show All Frames"), this);
	showAllFramesAct->setCheckable(true);
	displayMenu->addAction(showAllFramesAct);

	showDeformedSameColorAct = new QAction(tr("&Deformed Model: Use target color"), this);
	showDeformedSameColorAct->setCheckable(true);
	displayMenu->addAction(showDeformedSameColorAct);

	displayMenu->addSeparator();

	showSamplesAct = new QAction(tr("&Show Samples"), this);
	showSamplesAct->setCheckable(true);
	displayMenu->addAction(showSamplesAct);

	visualizeClusterAct = new QAction(tr("&Visualize Cluster Result"), this);
	visualizeClusterAct->setCheckable(true);
	displayMenu->addAction(visualizeClusterAct);

	showVerticesAct = new QAction(tr("&Show Vertices"), this);
	showVerticesAct->setCheckable(true);
	displayMenu->addAction(showVerticesAct);

	showVertexFramesAct = new QAction(tr("&Show Vertex Frames"), this);
	showVertexFramesAct->setCheckable(true);
	displayMenu->addAction(showVertexFramesAct);

	displayMenu->addSeparator();
	modelDisplayGroup = new QActionGroup(this);

	showGeometryAct = new QAction(tr("&Show Geometry"), this);
	showGeometryAct->setCheckable(true);
	displayMenu->addAction(showGeometryAct);
	modelDisplayGroup->addAction(showGeometryAct);
	showGeometryAct->setChecked(true);

	showStatusAct = new QAction(tr("&Show Vertex Status"), this);
	showStatusAct->setCheckable(true);
	displayMenu->addAction(showStatusAct);
	modelDisplayGroup->addAction(showStatusAct);

	showFeaturesAct = new QAction(tr("&Show Features"), this);
	showFeaturesAct->setCheckable(true);
	displayMenu->addAction(showFeaturesAct);
	modelDisplayGroup->addAction(showFeaturesAct);

	showCurvaturesAct = new QAction(tr("&Show Curvatures"), this);
	showCurvaturesAct->setCheckable(true);
	displayMenu->addAction(showCurvaturesAct);
	modelDisplayGroup->addAction(showCurvaturesAct);

	showFramesAct = new QAction(tr("&Show Frames"), this);
	showFramesAct->setCheckable(true);
	displayMenu->addAction(showFramesAct);
	modelDisplayGroup->addAction(showFramesAct);

	showCellIndexAct = new QAction(tr("&Show Cell Index"), this);
	showCellIndexAct->setCheckable(true);
	displayMenu->addAction(showCellIndexAct);
	modelDisplayGroup->addAction(showCellIndexAct);

	displayMenu->addSeparator();

	setBboxModeAct = new QAction(tr("&Show Bounding Box"), this);
	setBboxModeAct->setCheckable(true);
	displayMenu->addAction(setBboxModeAct);

	setGraphShowAct = new QAction(tr("&Show Graph"), this);
	setGraphShowAct->setCheckable(true);
	displayMenu->addAction(setGraphShowAct);
	setGraphShowAct->setChecked(true);

	showDeformedGraphAct = new QAction(tr("&Show Graph Deformed"), this);
	showDeformedGraphAct->setCheckable(true);
	displayMenu->addAction(showDeformedGraphAct);
	showDeformedGraphAct->setChecked(true);
	
	showGraphFacesAct = new QAction(tr("&Show Graph Faces"), this);
	showGraphFacesAct->setCheckable(true);
	displayMenu->addAction(showGraphFacesAct);
	showGraphFacesAct->setChecked(false);

	setEffRadShowAct = new QAction(tr("&Show Influence Spheres"), this);
	setEffRadShowAct->setCheckable(true);
	displayMenu->addAction(setEffRadShowAct);

	displayMenu->addSeparator();

	changeBackgroundColorAct = new QAction(tr("&Change Background Color..."), this);
	displayMenu->addAction(changeBackgroundColorAct);
	connect(changeBackgroundColorAct, SIGNAL(triggered()), this, SLOT(changeBackgroundColor()));

	changeForegroundColorAct = new QAction(tr("&Change Foreground Color..."), this);
	displayMenu->addAction(changeForegroundColorAct);
	connect(changeForegroundColorAct, SIGNAL(triggered()), this, SLOT(changeForegroundColor()));

	saveColormapAct = new QAction(tr("&Save Colormap..."), this);
	displayMenu->addAction(saveColormapAct);
	connect(saveColormapAct, SIGNAL(triggered()), this, SLOT(saveColormap()));

	loadColormapAct = new QAction(tr("&Load Colormap..."), this);
	displayMenu->addAction(loadColormapAct);
	connect(loadColormapAct, SIGNAL(triggered()), this, SLOT(loadColormap()));

	setColormapLimitsAct = new QAction(tr("&Set Colormap Range..."), this);
	displayMenu->addAction(setColormapLimitsAct);
	connect(setColormapLimitsAct, SIGNAL(triggered()), this, SLOT(setColormapLimits()));


	//////////////////////////////////////////////////////////////////////////
	// Options menu
	clearResultAct = new QAction(tr("&Clear Result"), this);
	optionsMenu->addAction(clearResultAct);
	connect(clearResultAct, SIGNAL(triggered()), this, SLOT(clearResult()));

	setGraphNodesAct = new QAction(tr("&Graph Nodes..."), this);
	optionsMenu->addAction(setGraphNodesAct);
	connect(setGraphNodesAct, SIGNAL(triggered()), this, SLOT(setGraphNodes()));

	resampleGraphAct = new QAction(tr("&Resample Graph"), this);
	optionsMenu->addAction(resampleGraphAct);
	connect(resampleGraphAct, SIGNAL(triggered()), this, SLOT(resampleGraph()));

	showDeformEditorAct = new QAction(tr("&Edit Deformation Parameters..."), this);
	optionsMenu->addAction(showDeformEditorAct);
	connect(showDeformEditorAct, SIGNAL(triggered()), this, SLOT(CreateDeformEditor()));

	showCorrespViewerAct = new QAction(tr("&Show Correspondences..."), this);
	optionsMenu->addAction(showCorrespViewerAct);
	connect(showCorrespViewerAct, SIGNAL(triggered()), this, SLOT(CreateCorrespViewer()));

	setFeatureThresholdAct = new QAction(tr("&Set Feature L2 norm Threshold..."), this);
	optionsMenu->addAction(setFeatureThresholdAct);
	connect(setFeatureThresholdAct, SIGNAL(triggered()), this, SLOT(setFeatureThreshold()));

	optionsMenu->addSeparator()->setText(tr("&Deformation Type"));

	deformGroup = new QActionGroup(this);

	useEDAct = new QAction(tr("&Embedded Deformation"), this);
	useEDAct->setCheckable(true);
	optionsMenu->addAction(useEDAct);
	deformGroup->addAction(useEDAct);

	useWDAct = new QAction(tr("&Weighted Deformation"), this);
	useWDAct->setCheckable(true);
	optionsMenu->addAction(useWDAct);
	deformGroup->addAction(useWDAct);

	useDeformEMAct = new QAction(tr("&EM Deformation"), this);
	useDeformEMAct->setCheckable(true);
	optionsMenu->addAction(useDeformEMAct);
	deformGroup->addAction(useDeformEMAct);

	useDeformAllAct = new QAction(tr("&Simultaneous Deform"), this);
	useDeformAllAct->setCheckable(true);
	optionsMenu->addAction(useDeformAllAct);
	deformGroup->addAction(useDeformAllAct);

	// Default value
	useDeformAllAct->setChecked(true);
}

/**
 * Rotate the camera frame about an axis with angle degrees, about the revolve-around point.
 */
void Viewer::rotateAbout(qglviewer::Vec up, float angle) {
	// rotate camera about the current up vector
	qglviewer::Vec o, p, v, cup;
	qglviewer::Quaternion delta;

	up.normalize();
	o = camera()->revolveAroundPoint();
	p = camera()->position();
	v = p + camera()->viewDirection();
	cup = p + camera()->upVector();

	delta.setAxisAngle(up, angle/180.0*M_PI);

	p = delta.rotate(p-o) + o;
	v = delta.rotate(v-o) + o - p;
	cup = delta.rotate(cup-o) + o - p;
	
	camera()->setPosition(p);
	camera()->setViewDirection(v);
	camera()->setUpVector(cup);
}


void Viewer::keyPressEvent(QKeyEvent *e)
{
	// Get event modifiers key
	const Qt::KeyboardModifiers modifiers = e->modifiers();

	bool handled = false;
	if ((e->key()==Qt::Key_Space) && (modifiers==Qt::NoButton)) {
		// Compute the center location of this window
		QRect win = this->geometry();
		QPoint center;
		center.setX(win.x() + win.width() / 7);
		center.setY(win.y() + win.height() / 7);
		//cout << "(x,y,w,h) = (" << win.x() << "," << win.y() << "," << win.width() << "," << win.height() << ")" << endl;

		contextMenu->exec(center);
		handled = true;
		updateGL();
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::CTRL)) {
		// increment the frame number
		if (++displayFrame >= scene->getAnimation()->numFrames()) {
			displayFrame = 0;
		}
		displayMessage(updateFrame(displayFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::CTRL)) {
		// decrement the frame number
		if (--displayFrame < 0) {
			displayFrame = scene->getAnimation()->numFrames() - 1;
		}
		displayMessage(updateFrame(displayFrame));
		handled = true;
	}
	else if ((e->key()==Qt::Key_Plus) && (modifiers==Qt::KeypadModifier)) {
		//scene->graph_n += 10;
		//scene->setupDeform(whichDeform());
		//// Display the deformed mesh
		//setDeformedModelShowAct->setChecked(true);
		//displayMessage(tr("Generated deformation graph with %1 nodes!").arg(scene->graph_n));

		scene->currentBone += 1;
		if (scene->currentBone >= scene->graph_n) {
			scene->currentBone = scene->graph_n-1;
		}
		displayMessage(tr("Showing bone %1...").arg(scene->currentBone));


		handled = true;
	}
	else if ((e->key()==Qt::Key_Minus) && (modifiers==Qt::KeypadModifier)) {
		//scene->graph_n -= 10;
		//if (scene->graph_n < 10) {
		//	scene->graph_n = 10;
		//}
		//scene->setupDeform(whichDeform());
		//// Display the deformed mesh
		//setDeformedModelShowAct->setChecked(true);
		//displayMessage(tr("Generated deformation graph with %1 nodes!").arg(scene->graph_n));

		scene->currentBone -= 1;
		if (scene->currentBone < -1) {
			scene->currentBone = -1;
		}

		if (scene->currentBone < 0) {
			displayMessage(tr("Showing all bones..."));
		}
		else {
			displayMessage(tr("Showing bone %1...").arg(scene->currentBone));
		}

		handled = true;
	}
	else if ((e->key()==Qt::Key_Return) && (modifiers==Qt::SHIFT)) {
		scene->solveDeformation();
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Left) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(0, 1, 0), 10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Right) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(0, 1, 0), -10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Up) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(1, 0, 0), 10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_Down) && (modifiers==Qt::NoButton)) {
		rotateAbout(qglviewer::Vec(1, 0, 0), -10.0);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_S) && (modifiers==Qt::NoButton)) {
		// Toggle showing the source model
		setSourceModelShowAct->setChecked(!setSourceModelShowAct->isChecked());

		handled = true;
	}
	else if ((e->key()==Qt::Key_D) && (modifiers==Qt::NoButton)) {
		// Toggle showing the deformed model
		setDeformedModelShowAct->setChecked(!setDeformedModelShowAct->isChecked());

		//QString message;
		//// Flip between displaying the target and the deformed meshes
		//if ((showDeformedModel() && showTargetModel()) || (!showDeformedModel() && !showTargetModel()) || (!showDeformedModel() && showTargetModel())) {
		//	setDeformedModelShowAct->setChecked(true);
		//	setTargetModelShowAct->setChecked(false);
		//	message = tr("Showing deformed model");
		//}
		//else {
		//	setDeformedModelShowAct->setChecked(false);
		//	setTargetModelShowAct->setChecked(true);

		//	// Find the frame number
		//	displayFrame = scene->getAnimation()->frameNumber(scene->target);
		//	message = tr("Showing target model (frame %1)").arg(displayFrame);
		//}
		//displayMessage(message);

		handled = true;
	}
	else if ((e->key()==Qt::Key_F) && (modifiers==Qt::NoButton)) {
		// Toggle showing the target model
		setTargetModelShowAct->setChecked(!setTargetModelShowAct->isChecked());

		//QString message;
		//// Flip between displaying the source and the target meshes
		//if ((showSourceModel() && showTargetModel()) || (!showSourceModel() && !showTargetModel()) || (!showSourceModel() && showTargetModel())) {
		//	setSourceModelShowAct->setChecked(true);
		//	setTargetModelShowAct->setChecked(false);

		//	// Find the frame number
		//	displayFrame = scene->getAnimation()->frameNumber(scene->source);
		//	message = tr("Showing source model (frame %1)").arg(displayFrame);
		//}
		//else {
		//	setSourceModelShowAct->setChecked(false);
		//	setTargetModelShowAct->setChecked(true);
		//	// Find the frame number
		//	displayFrame = scene->getAnimation()->frameNumber(scene->target);
		//	message = tr("Showing target model (frame %1)").arg(displayFrame);
		//}
		//displayMessage(message);

		handled = true;
	}
	else if ((e->key()==Qt::Key_W) && (modifiers==Qt::NoButton)) {
		// Toggle showing the source correspondence points
		sourceCorrespShowAct->setChecked(!sourceCorrespShowAct->isChecked());
		handled = true;
	}
	else if ((e->key()==Qt::Key_E) && (modifiers==Qt::NoButton)) {
		// Toggle showing the deformed correspondence points
		deformedCorrespShowAct->setChecked(!deformedCorrespShowAct->isChecked());
		handled = true;
	}
	else if ((e->key()==Qt::Key_R) && (modifiers==Qt::NoButton)) {
		// Toggle showing the target correspondence points
		targetCorrespShowAct->setChecked(!targetCorrespShowAct->isChecked());
		handled = true;
	}
	else if ((e->key()==Qt::Key_H) && (modifiers==Qt::NoButton)) {
		// Toggle showing the weights painted on the shape
		showShapeWeightsAct->setChecked(!showShapeWeightsAct->isChecked());
		handled = true;
	}
	else if ((e->key()==Qt::Key_A) && (modifiers==Qt::NoButton)) {
		// Toggle showing all frames
		showAllFramesAct->setChecked(!showAllFramesAct->isChecked());
		handled = true;
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::NoButton)) {
		displayScale++;
		displayMessage(tr("Displaying feature component %1").arg(displayScale));
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::NoButton)) {
		displayScale = max(0, displayScale-1);
		displayMessage(tr("Displaying feature component %1").arg(displayScale));
	}
	else if ((e->key()==Qt::Key_PageDown) && (modifiers==Qt::CTRL)) {
		curvDisplayMode++;
		if (curvDisplayMode > 4) {
			curvDisplayMode = 1;
		}
	}
	else if ((e->key()==Qt::Key_PageUp) && (modifiers==Qt::CTRL)) {
		curvDisplayMode--;
		if (curvDisplayMode < 1) {
			curvDisplayMode = 4;
		}
	}
	// Browse solved deformations
	else if ((e->key()==Qt::Key_BracketLeft) && (modifiers==Qt::CTRL)) {
		//scene->browseDeformation(-1);
		showErrorComponent = min(2, showErrorComponent+1);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_BracketRight) && (modifiers==Qt::CTRL)) {
		//scene->browseDeformation(+1);
		showErrorComponent = max(0, showErrorComponent-1);
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_G) && (modifiers==Qt::NoButton)) {
		// Toggle showing the graph
		setGraphShowAct->setChecked(!setGraphShowAct->isChecked());
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_B) && (modifiers==Qt::NoButton)) {
		// Toggle showing the graph faces
		showGraphFacesAct->setChecked(!showGraphFacesAct->isChecked());
		updateGL();
		handled = true;
	}
	else if ((e->key()==Qt::Key_C) && (modifiers==Qt::NoButton)) {
		// Generate random grid colors
		scene->generateGridColors();
		updateGL();
		handled = true;
	}

	if (!handled) {
		QGLViewer::keyPressEvent(e);
	}

	emit updateChildDisplay();
}

real3_type Viewer::closestPointOnLine( QPoint pos, real3_type x )
{
	// Find the line in 3D corresponding to the cursor
	qglviewer::Vec o, d;
	real3_type orig, dir;
	camera()->convertClickToLine(pos, o, d);
	orig = VecToVec3f(o);
	dir = VecToVec3f(d);

	// Find closest point on line
	real_type t = - ((orig-x) * dir) / (dir*dir);
	return orig + t*dir;
}

void Viewer::mousePressEvent( QMouseEvent *e )
{
	if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
		updateGL();

		// Move the selected vertex along the shape
		bool found;
		clickStartPos = VecToVec3f(camera()->pointUnderPixel(e->pos(), found));
		clickEndPos = clickStartPos;

		if (found) {
			//cout << "Setting start position " << clickStartPos << endl;
			interactionMode_ = DEFINE_SPHERE;
		}
		else {
			interactionMode_ = NONE;
		}
	}
	else {
		QGLViewer::mousePressEvent(e);
	}
}

void Viewer::mouseMoveEvent( QMouseEvent *e )
{
	if (interactionMode_ == DEFINE_SPHERE) {
		clickEndPos = closestPointOnLine(e->pos(), clickStartPos);

		//cout << "Move: Setting end position to " << clickEndPos << endl;
		updateGL();
	}
	else {
		QGLViewer::mouseMoveEvent(e);
	}
}

void Viewer::mouseReleaseEvent( QMouseEvent *e )
{
	if (interactionMode_ == DEFINE_SPHERE) {
		if ((e->button() == Qt::LeftButton) && (e->modifiers() == Qt::AltModifier)) {
			clickEndPos = closestPointOnLine(e->pos(), clickStartPos);

			//cout << "Release: Setting end position to " << clickEndPos << endl;

			if (scene->source) {
				// Perform sphere fit
				//sphereFit();
			}

			updateGL();
		}

		interactionMode_ = NONE;
	}
	else {
		QGLViewer::mouseReleaseEvent(e);
	}
}

void Viewer::Update( Subject* theChangedSubject )
{
	// Implement updating of opengl here by setting a variable and looping in an animation loop
	sceneMutex.lock();
	sceneChanged = true;
	sceneMutex.unlock();

	//cerr << "Updated by scene at: " << theChangedSubject << endl;
	//updateGL();
}

QString Viewer::updateFrame( int frame )
{
	QString message;

	// determine the mode
	if (showSourceModel() && !showTargetModel()) {
		scene->source = (*scene->getAnimation())[displayFrame];
		message = tr("Showing source frame %1").arg(displayFrame);
	}
	else if (!showSourceModel() && showTargetModel()) {
		scene->target = (*scene->getAnimation())[displayFrame];
		message = tr("Showing target frame %1").arg(displayFrame);
	}
	else {
		message = tr("View only the source/target frame to update.");
	}

	return message;
}

void Viewer::customSaveSnapshot()
{
	cout << "Saving snapshot to file: " << qPrintable(customSaveSnapshotFile) << endl << flush;
	saveSnapshot(customSaveSnapshotFile);
}


void Viewer::captureVideo()
{
	// Iterate through each frame and take a screenshot
	// Show only the source model
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(true);
	setDeformedModelShowAct->setChecked(false);

	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the animation", scene->lastDir, tr("*.%1").arg(snapshotFormat().toLower()));
	if (savefileName.length() == 0) {
		cout << "Video capture cancelled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	const int numFrames = scene->getAnimation()->numFrames();

	FileSequence seq(savefileName);
	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	for (unsigned int i = 0; i < numFrames; i++) {
		scene->target = (*scene->getAnimation())[i];

		// Update the display and take a screenshot
		customSaveSnapshotFile = seq[i];
		updateGL();
	}
	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal

}

void Viewer::CreateDeformEditor()
{
	if (!deformEdit) {

		bool ok = true;
		// Check that we have enough weights
		if (scene->em_EWeights.size() < 2) {
			cerr << "Not enough E-step weights (only have " << scene->em_EWeights.size() << ") for DeformEM." << endl;
			ok = false;
		}
		if (scene->em_DiscreteMWeights.size() < 2) {
			cerr << "Not enough Discrete M-step weights (only have " << scene->em_DiscreteMWeights.size() << ")for DeformEM." << endl;
			ok = false;
		}
		if (scene->em_MWeights.size() < 4) {
			cerr << "Not enough M-step weights (only have " << scene->em_MWeights.size() << ")for DeformEM." << endl;
			ok = false;
		}

		if (!ok) return;

		deformEdit = new DeformEdit();
		deformEdit->initRange(NUM_BONES, 0, scene->graph_n*2, scene->graph_n);
		deformEdit->initRange(NUM_DIVISIONS, 0, scene->gridDivisions*2, scene->gridDivisions);
		deformEdit->initRange(NUM_CORRESPONDENCES, 0, scene->numCorrespondences*2, scene->numCorrespondences);
		deformEdit->initRange(MAX_CORRESP_DIST, 0, scene->maxCorrespondenceDistance*2, scene->maxCorrespondenceDistance);
		deformEdit->initRange(MAX_BDY_CORRESP_DIST, 0, scene->boundaryCorrespondenceDistance*2, scene->boundaryCorrespondenceDistance);
		deformEdit->initRange(DEFORM_EM_CONSTRAINT_E_WEIGHT, 0, scene->em_EWeights[0]*2, scene->em_EWeights[0]);
		deformEdit->initRange(DEFORM_EM_SMOOTHNESS_E_WEIGHT, 0, scene->em_EWeights[1]*2, scene->em_EWeights[1]);
		//deformEdit->initRange(DEFORM_EM_CONSTRAINT_M_WEIGHT, 0, scene->em_MWeights[0]*2, scene->em_MWeights[0]);
		//deformEdit->initRange(DEFORM_EM_SMOOTHNESS_M_WEIGHT, 0, scene->em_MWeights[1]*2, scene->em_MWeights[1]);
		//deformEdit->initRange(DEFORM_EM_UNITY_M_WEIGHT, 0, scene->em_MWeights[2]*2, scene->em_MWeights[2]);
		//deformEdit->initRange(DEFORM_EM_DISCRETE_M_WEIGHT, 0, scene->em_MWeights[3]*2, scene->em_MWeights[3]);
		//deformEdit->initRange(SOLVER_MAX_ITERATIONS, 0, scene->solverMaxIterations*2, scene->solverMaxIterations);
		//deformEdit->initRange(ESTEP_MAX_ITERATIONS, 0, scene->estepMaxIterations*2, scene->estepMaxIterations);
		//deformEdit->initRange(PROJ_MAX_ITERATIONS, 0, scene->projMaxIterations*2, scene->projMaxIterations);
		//deformEdit->initRange(IRLS_MAX_ITERATIONS, 0, scene->IRLSMaxIterations*2, scene->IRLSMaxIterations);
		deformEdit->init_solveInitialRigidCheckBox(scene->solveInitialRigid);
		deformEdit->init_solveFinalWeightsCheckBox(scene->solveFinalNNLS);
		deformEdit->init_correspondenceTypeCombo(scene->correspondenceType);

		// Show the UI dialog
		deformEdit->show();

		connect(deformEdit, SIGNAL(valueChanged(int,double)), this, SLOT(DeformEditorChanged(int,double)));
		connect(deformEdit, SIGNAL(destroyed()), this, SLOT(DeformEditorDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		deformEdit->hide();
		deformEdit->show();
	}
}

void Viewer::DeformEditorDestroyed()
{
	//cerr << "Deformation control is: " << deformEdit << endl;
	deformEdit = NULL;

	// Print the values of the parameters for verification
	cout << "----------------------------------------------" << endl;
	cout << "Final Parameter Values:" << endl;
	cout << "----------------------------------------------" << endl;
	cout << "Graph Nodes: " << scene->graph_n << endl;
	cout << "Graph Neighbors: " << scene->graph_k << endl;
	cout << "Overlap Factor: " << scene->overlapf << endl;
	cout << "Solver Max Iterations: " << scene->solverMaxIterations << endl;
	cout << "Projection Max Iterations: " << scene->projMaxIterations << endl;
	cout << "ED Weights Rotation: " << scene->edWeights[0] << endl;
	cout << "ED Weights Smooth: " << scene->edWeights[1] << endl;
	cout << "ED Weights Constraint: " << scene->edWeights[2] << endl;
	cout << "WD Weights Smooth: " << scene->wdWeights[0] << endl;
	cout << "WD Weights Constraint: " << scene->wdWeights[1] << endl;
	cout << "----------------------------------------------" << endl;
}

void Viewer::DeformEditorChanged(int which, double value)
{
	QString message;

	switch(which) {
	case OUTPUT_STATS:
		// Nothing implemented yet!
		break;
	case OTHER_TASKS:
		switch ((int)value) {
		case 1:
			CaptureResultSnapshots_AlignPair();
			break;
		case 2:
			CaptureResultSnapshots_AlignSeq();
			break;
		case 3:
			CaptureResultSnapshots_GridWeights();
			break;
		case 4:
			scene->printCurrentDeformation();
			break;
		case 5:
			{
				QString saveFilename = QFileDialog::getSaveFileName(this, QString("Save Registration Result"), scene->lastDir, QString("Data (*.data)"));
				scene->saveDeformation(saveFilename);
			}
			break;
		case 6:
			{
				QString loadFilename = QFileDialog::getOpenFileName(this, QString("Load Registration Result"), scene->lastDir, QString("Data (*.data)"));
				scene->loadDeformation(loadFilename);
			}
			break;
		default:
			break;
		}
		break;
	case NUM_BONES:
		scene->graph_n = value;
		cout << "Setting # bones: " << scene->graph_n << endl;
		break;
	case NUM_DIVISIONS:
		scene->gridDivisions = value;
		cout << "Setting # grid divisions on longest axis of bbox: " << scene->gridDivisions << endl;
		break;
	case NUM_CORRESPONDENCES:
		scene->numCorrespondences = value;
		cout << "Setting # correspondences: " << scene->numCorrespondences << endl;
		break;
	case MAX_CORRESP_DIST:
		scene->maxCorrespondenceDistance = value;
		cout << "Setting max correspondence distance: " << scene->maxCorrespondenceDistance << endl;
		scene->setDeformOptions();
		break;
	case MAX_BDY_CORRESP_DIST:
		scene->boundaryCorrespondenceDistance = value;
		cout << "Setting max boundary correspondence distance: " << scene->boundaryCorrespondenceDistance << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_CONSTRAINT_E_WEIGHT:
		scene->em_EWeights[0] = value;
		cout << "Setting DeformEM E-step constraint weight: " << scene->em_EWeights[0] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_SMOOTHNESS_E_WEIGHT:
		scene->em_EWeights[1] = value;
		cout << "Setting DeformEM E-step smoothness weight: " << scene->em_EWeights[1] << endl;
		scene->setDeformOptions();
		if (value == 0.0) {
			displayMessage(tr("Using NO Smoothness!"));
		}
		else {
			displayMessage(tr("Setting Smoothness term weight: %1").arg(value));
		}
		break;
	case DEFORM_EM_CONSTRAINT_M_WEIGHT:
		scene->em_MWeights[0] = value;
		cout << "Setting DeformEM M-step constraint weight: " << scene->em_MWeights[0] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_SMOOTHNESS_M_WEIGHT:
		scene->em_MWeights[1] = value;
		cout << "Setting DeformEM M-step smoothness weight: " << scene->em_MWeights[1] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_UNITY_M_WEIGHT:
		scene->em_MWeights[2] = value;
		cout << "Setting DeformEM M-step unity weight: " << scene->em_MWeights[2] << endl;
		scene->setDeformOptions();
		break;
	case DEFORM_EM_DISCRETE_M_WEIGHT:
		scene->em_MWeights[3] = value;
		cout << "Setting DeformEM M-step discrete weight: " << scene->em_MWeights[3] << endl;
		scene->setDeformOptions();
		break;
	case SOLVER_MAX_ITERATIONS:
		scene->solverMaxIterations = value;
		cout << "Setting Solver Max Iterations: " << scene->solverMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case ESTEP_MAX_ITERATIONS:
		scene->estepMaxIterations = value;
		cout << "Setting E-Step Max Iterations: " << scene->estepMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case PROJ_MAX_ITERATIONS:
		scene->projMaxIterations = value;
		cout << "Setting E-Step Projection Max Iterations: " << scene->projMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case IRLS_MAX_ITERATIONS:
		scene->IRLSMaxIterations = value;
		cout << "Setting M-Step IRLS Max Iterations: " << scene->IRLSMaxIterations << endl;
		scene->setDeformOptions();
		break;
	case SOLVE_INITIAL_RIGID:
		scene->solveInitialRigid = value;
		cout << "Setting solve initial correspondences: " << (scene->solveInitialRigid?"Yes":"No") << endl;
		scene->setDeformOptions();
		break;
	case SOLVE_FINAL_WEIGHTS:
		scene->solveFinalNNLS = value;
		cout << "Setting solve final continuous weights: " << (scene->solveFinalNNLS?"Yes":"No") << endl;
		scene->setDeformOptions();
		break;
	case UPDATE_CORRESPONDENCE_TYPE:
		scene->correspondenceType = value;
		cout << "Setting correspondence update type: " << scene->correspondenceType << endl;
		scene->setDeformOptions();
		break;
	case EXPORT_DEFORM_RESULT:
		// Export the deformed mesh
		if (scene->is_deformation_set_up()) {
			// Prompt for a save file
			QString savefileName;
			savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the deformed mesh...",
														scene->lastDir, QString("*.obj"));

			if (!savefileName.isEmpty()) {
				scene->saveDeformedResult(qPrintable(savefileName));
			}

			//cerr << "Not implemented yet!" << endl;
			//ShapeModel *shape = scene->deformResult();

			//if (shape->shapeType == SHAPE_MESH) {
			//	// Prompt for a save file
			//	QString savefileName;
			//	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the deformed mesh...",
			//												scene->lastDir, QString("*.obj"));
			//	if (savefileName.length() > 0) {
			//		((Mesh3_real*)shape)->writeGeometryOBJ(qPrintable(savefileName));
			//	}
			//}
		}
		break;
	case INVERSE_DEFORM:
		// Functionality not supported anymore
		cerr << "This functionality is not supported anymore!" << endl;

			////// Perform inverse deformation
			////if (scene->is_deformation_set_up()) {
			////	ShapeModel *shape = scene->inverseDeform(*(scene->target));

			////	// Add the result to the list of meshes
			////	scene->getAnimation()->add(shape);
			////}
		break;

	// Auto-solver thread control
	case SETUP_DEFORM:
		scene->controlSolveThread(which, whichDeform());
		updateGL();
		break;
	case SOLVE_DEFORM_AUTO:
	case PAUSE_SOLVER:
	case STEP_SOLVER:
	case STOP_SOLVER:
		scene->controlSolveThread(which, whichDeform());
		updateGL();
		break;

	case REVERT_PREVIOUS_SOLUTION:
		// Revert to the discrete weights
		scene->revertDiscreteWeights();
		updateGL();
		break;
	case SOLVE_DEFORM_WEIGHT:
		// Solve final NNLS weights
		scene->solveFinalNNLSWeights();
		updateGL();
		break;

	// Fine solver control: execute any step of the solver that you want
	case ACCUMULATE_GEOMETRY:
	case REQUEST_NEXT_FRAME:
	case SOLVE_DEFORM_INIT:
	case SOLVE_DEFORM_TRANSFORM:
	case SOLVE_DEFORM_CORRESP:
		cerr << "This functionality is not implemented yet!" << endl;
		break;

	////// Old stuff
	////case GRAPH_NODES:
	////	scene->graph_n = maxGraphNodes * value;
	////	deformEdit->displayValue(which, tr("%1").arg(scene->graph_n));
	////	scene->resampleSource();
	////	scene->setupDeform(whichDeform());
	////	break;
	////case GRAPH_NEIGHBORS:
	////	scene->graph_k = maxGraphNeighbors * value;
	////	deformEdit->displayValue(which, tr("%1").arg(scene->graph_k));
	////	scene->setupDeform(whichDeform());
	////	break;
	////case OVERLAP_FACTOR:
	////	scene->overlapf = maxOverlapFactor * value;
	////	deformEdit->displayValue(which, tr("%1").arg(scene->overlapf));
	////	scene->setupDeform(whichDeform());
	////	break;
	////case MAX_ITERATIONS_SOLVER:
	////	scene->solverMaxIterations = value;
	////	break;
	////case MAX_ITERATIONS_PROJECTION:
	////	scene->projMaxIterations = value;
	////	break;
	////case ED_WEIGHTS_ROTATION:
	////	scene->edWeights[0] = value;
	////	break;
	////case ED_WEIGHTS_SMOOTH:
	////	scene->edWeights[1] = value;
	////	break;
	////case ED_WEIGHTS_CONSTRAINT:
	////	scene->edWeights[2] = value;
	////	break;
	////case WD_WEIGHTS_SMOOTH:
	////	scene->wdWeights[0] = value;
	////	break;
	////case WD_WEIGHTS_CONSTRAINT:
	////	scene->wdWeights[1] = value;
	////	break;
	default:
		break;
	}

	updateGL();
}

void Viewer::CreateCorrespViewer()
{
	if (!correspViewer) {
		correspViewer = new CorrespViewer(scene, this);
		correspViewer->show();

		connect(this, SIGNAL(updateChildDisplay()), correspViewer, SLOT(updateDisplay()));
		connect(correspViewer, SIGNAL(destroyed()), this, SLOT(CorrespViewerDestroyed()));
	}
	else {
		// Focus on the current render controller
		// Raise the current path editor
		correspViewer->hide();
		correspViewer->show();
	}
}

void Viewer::CorrespViewerDestroyed()
{
	//cerr << "Correspondence Viewer is: " << correspViewer << endl;
	correspViewer = NULL;

}

void Viewer::saveColormap()
{
	// Get the save file name
	QString savefileName;
	savefileName = QFileDialog::getSaveFileName(this, "Choose a file name to save the colormap...", scene->lastDir, QString("*.txt"));
	if (savefileName.length() == 0) {
		cout << "Viewer::loadColormap(): Save canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(savefileName);
	}

	scene->colormap.save(qPrintable(savefileName));
}

void Viewer::loadColormap()
{
	QString loadfileName;
	loadfileName = QFileDialog::getOpenFileName(this, "Choose a file name to load the colormap...", scene->lastDir, tr("*.txt"));
	if (loadfileName.length() == 0) {
		cout << "Viewer::loadColormap(): Loading canceled." << endl;
		return;
	}
	else {
		scene->lastDir = grabPath(loadfileName);
	}

	scene->colormap.load(qPrintable(loadfileName));
}

void Viewer::setColormapLimits()
{
	// Set the range of the colormap
	real_type newMin, newMax;

	// Set the resolution of the screen here.
	bool ok;

	newMin = QInputDialog::getDouble(this, tr("Colormap minimum range"),
				tr("Input Minimum:"), scene->colormapMin, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	newMax = QInputDialog::getDouble(this, tr("Colormap maximum range"),
				tr("Input Maximum:"), scene->colormapMax, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	// Update the window
	scene->colormapMin = newMin;
	scene->colormapMax = newMax;
	updateGL();
}

void Viewer::setFeatureThreshold()
{
	// Set the range of the colormap
	real_type newVal;

	// Set the resolution of the screen here.
	bool ok;

	newVal = QInputDialog::getDouble(this, tr("Feature threshold display"),
		tr("Input Threshold:"), scene->threshold_FM, -DBL_MAX, DBL_MAX, 5, &ok);
	if (!ok) return;

	// Update the window
	scene->threshold_FM = newVal;
	updateGL();
}

void Viewer::sphereFit()
{
	PointList rangePts, rangeNormals;
	real3_type sample = clickStartPos;
	real_type radSq = (clickEndPos - clickStartPos).square_norm();

	cout << "Fitting sphere at: " << sample << " with radius^2 " << radSq << endl;

	scene->source->vertex.queryRange(sample, radSq, inRange, distRange, true);

	// Gather the point locations and the normals
	rangePts.resize(inRange.size());
	rangeNormals.resize(inRange.size());
	for (index_type j = 0; j < inRange.size(); j++) {
		// Throw out the point if the normal is zero
		if (scene->source->normal[inRange[j]].norm() >= REAL_TYPE_EPSILON) {
			rangePts[j] = scene->source->vertex[inRange[j]];
			rangeNormals[j] = scene->source->normal[inRange[j]];
		}
	}

	// Fit a circle at this position
	if (FitSphere(rangePts, rangeNormals, sample, radSq)) {
		// Store the mean curvature (inverse of the radius) as the feature
		//features.val[s] = 1.0 / sqrt(radSq);
		cout << "Fit sphere: center " << sample << " radius^2 " << radSq << endl;
		fitCenter = sample;
		fitRadius = sqrt(radSq);
	}
	else {
		// If the fit is degenerate, then the curvature is 0.
		//features.val[s] = 0;
		cout << "Fit plane: dist from origin " << radSq << endl;
	}
}



ShapeModel* Viewer::selectCurrentModel()
{
	if (showAllFramesAct->isChecked()) {
		return (*scene->getAnimation())[0];
	}
	else {
		if (showTargetModel()) {
			return scene->target;
		}
		if (showSourceModel()) {
			return scene->source;
		}
		if (!showSourceModel() && !showTargetModel()) {
			return scene->source;
			//cout << "Please display a mesh for computing curvatures." << endl;
		}
	}

	return NULL;
}

void Viewer::fitViewToShape( const ShapeModel *shape /*= NULL */ )
{
	if (!shape) {
		shape = selectCurrentModel();
	}

	if (shape) {
		// Fit the view to the current shape
		setSceneBoundingBox(Vec3fToVec(shape->box_min), Vec3fToVec(shape->box_max));

		camera()->setZClippingCoefficient(10);
		camera()->setZNearCoefficient(1e-2);
		showEntireScene();

		//cout << "Camera near clipping plane: " << camera()->zNear() << endl;
		//cout << "Camera far clipping plane: " << camera()->zFar() << endl;
	}
}

void Viewer::changeBackgroundColor()
{
	QColor color = QColorDialog::getColor(this->backgroundColor(), this);
	if (color.isValid()) {
		this->setBackgroundColor(color);
	}
}

void Viewer::changeForegroundColor()
{
	QColor color = QColorDialog::getColor(this->foregroundColor(), this);
	if (color.isValid()) {
		this->setForegroundColor(color);
	}
}

void Viewer::CaptureResultSnapshots_AlignPair()
{
	// Prompt for the save directory
	QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"), scene->lastDir);

	if (dir.isEmpty()) {
		return;
	}
	else {
		scene->lastDir = dir;
	}

	// Prompt for the file header
	bool ok;
	QString headerText = QInputDialog::getText(this, tr("File Header"),
		tr("Please enter the result file header:"), QLineEdit::Normal, QString(), &ok);
	if (!ok || headerText.isEmpty()) {
		return;
	}

	// Reset the view
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(false);
	setDeformedModelShowAct->setChecked(false);
	setGraphShowAct->setChecked(false);
	showGraphFacesAct->setChecked(true);
	showDeformedGraphAct->setChecked(true);
	scene->currentBone = -1;

	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	// 1. Show the source and target shapes
	setSourceModelShowAct->setChecked(true);
	setTargetModelShowAct->setChecked(true);
	customSaveSnapshotFile = tr("%1/%2 - source (blue) and target (red).png").arg(dir).arg(headerText);
	updateGL();

	// 2. Show the deformed shape
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(false);
	setDeformedModelShowAct->setChecked(true);
	customSaveSnapshotFile = tr("%1/%2 - deformed source (green).png").arg(dir).arg(headerText);
	updateGL();

	// 3. Show the target and deformed shapes
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(true);
	setDeformedModelShowAct->setChecked(true);
	customSaveSnapshotFile = tr("%1/%2 - deformed source (green) and target (red).png").arg(dir).arg(headerText);
	updateGL();

	// 4. Show the deformed shape with the graph
	setTargetModelShowAct->setChecked(false);
	setDeformedModelShowAct->setChecked(true);
	setGraphShowAct->setChecked(true);
	customSaveSnapshotFile = tr("%1/%2 - deformed source (green) with grid.png").arg(dir).arg(headerText);
	updateGL();

	// 5. Show the grid in wireframe
	showGraphFacesAct->setChecked(false);
	customSaveSnapshotFile = tr("%1/%2 - deformed source (green) with wireframe grid.png").arg(dir).arg(headerText);
	updateGL();

	//// 4. Show the inverse deformed target with the source.  Assume inverse deformed target is the last frame.
	//scene->target = (*scene->getAnimation())[scene->getAnimation()->numFrames() - 1];
	//setSourceModelShowAct->setChecked(true);
	//setTargetModelShowAct->setChecked(true);
	//setDeformedModelShowAct->setChecked(false);
	//setGraphShowAct->setChecked(false);
	//customSaveSnapshotFile = tr("%1/%2 - source (blue) and inverse deformed target (red).png").arg(dir).arg(headerText);
	//updateGL();

	//// Show with wireframe mode
	//setWireframeModeAct->setChecked(true);
	//customSaveSnapshotFile = tr("%1/%2 - source (blue) and inverse deformed target (red) wireframe.png").arg(dir).arg(headerText);
	//updateGL();

	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
}


void Viewer::CaptureResultSnapshots_AlignSeq()
{
	// Prompt for the save directory
	QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"), scene->lastDir);

	if (dir.isEmpty()) {
		return;
	}
	else {
		scene->lastDir = dir;
	}

	// Prompt for the file header
	bool ok;
	QString headerText = QInputDialog::getText(this, tr("File Header"),
		tr("Please enter the result file header:"), QLineEdit::Normal, QString(), &ok);
	if (!ok || headerText.isEmpty()) {
		return;
	}

	int numFrames = scene->getAnimation()->numFrames();
	// First, go to the first deformation
	scene->browseDeformation(-numFrames);

	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	// Show the result of registering each frame of the sequence
	for (int i = 1; i < numFrames; i++) {
		// Reset the view
		setSourceModelShowAct->setChecked(false);
		setTargetModelShowAct->setChecked(false);
		setDeformedModelShowAct->setChecked(false);
		setGraphShowAct->setChecked(false);
		showGraphFacesAct->setChecked(false);
		scene->currentBone = -1;

		// a. Deformed source and target alignment
		scene->target = (*scene->getAnimation())[i];
		setTargetModelShowAct->setChecked(true);
		setDeformedModelShowAct->setChecked(true);
		customSaveSnapshotFile = tr("%1/%2 - %3 - Frame_%4.png").arg(dir).arg(headerText).arg("alignment result").arg(i, 4, 10, QLatin1Char('0'));
		updateGL();

		// b. Deformed result
		setTargetModelShowAct->setChecked(false);
		setDeformedModelShowAct->setChecked(true);
		customSaveSnapshotFile = tr("%1/%2 - %3 - Frame_%4.png").arg(dir).arg(headerText).arg("deformed result").arg(i, 4, 10, QLatin1Char('0'));
		updateGL();

		// c. Deformed grid (all bones)
		setTargetModelShowAct->setChecked(false);
		setDeformedModelShowAct->setChecked(false);
		setGraphShowAct->setChecked(true);
		showGraphFacesAct->setChecked(true);
		showDeformedGraphAct->setChecked(true);
		customSaveSnapshotFile = tr("%1/%2 - %3 - Frame_%4.png").arg(dir).arg(headerText).arg("vertex weights").arg(i, 4, 10, QLatin1Char('0'));
		updateGL();

		// d. Undeformed grid (all bones)
		showDeformedGraphAct->setChecked(false);
		customSaveSnapshotFile = tr("%1/%2 - %3 - Frame_%4.png").arg(dir).arg(headerText).arg("vertex weights (undeformed grid)").arg(i, 4, 10, QLatin1Char('0'));
		updateGL();

		// Go to the next frame
		scene->browseDeformation(+1);
	};

	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal
}

void Viewer::CaptureResultSnapshots_GridWeights()
{
	if (!scene->is_deformation_set_up()) {
		return;
	}

	// Prompt for the save directory
	QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"), scene->lastDir);

	if (dir.isEmpty()) {
		return;
	}
	else {
		scene->lastDir = dir;
	}

	// Prompt for the file header
	bool ok;
	QString headerText = QInputDialog::getText(this, tr("File Header"),
		tr("Please enter the result file header:"), QLineEdit::Normal, QString(), &ok);
	if (!ok || headerText.isEmpty()) {
		return;
	}

	int numFrames = scene->getAnimation()->numFrames();
	// First, go to the first deformation
	scene->browseDeformation(-numFrames);

	// Reset the view
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(false);
	setDeformedModelShowAct->setChecked(false);
	setGraphShowAct->setChecked(false);
	showGraphFacesAct->setChecked(true);

	connect(this, SIGNAL(drawFinished(bool)), SLOT(customSaveSnapshot()));	// connect the signal

	// Show the result of registering each frame of the sequence
	for (int i = -1; i < scene->graph_n; i++) {
		scene->currentBone = i;

		// c. Deformed grid (all bones)
		setGraphShowAct->setChecked(true);
		showDeformedGraphAct->setChecked(false);
		if (i == -1) {
			customSaveSnapshotFile = tr("%1/%2 - Bone_%3.png").arg(dir).arg(headerText).arg("All");
		}
		else {
			customSaveSnapshotFile = tr("%1/%2 - Bone_%3.png").arg(dir).arg(headerText).arg(i, 4, 10, QLatin1Char('0'));
		}
		updateGL();
	};

	disconnect(SIGNAL(drawFinished(bool)), this, SLOT(customSaveSnapshot()));	// disconnect the signal

	// Reset the view
	setSourceModelShowAct->setChecked(false);
	setTargetModelShowAct->setChecked(true);
	setDeformedModelShowAct->setChecked(true);
	showDeformedGraphAct->setChecked(false);
	setGraphShowAct->setChecked(true);
	showGraphFacesAct->setChecked(false);
	scene->currentBone = -1;
}

Deform::DrawOptions Viewer::getDeformDrawOptions()
{
	Deform::DrawOptions options;

	// Populate option values from the viewer and the scene
	options.showSourceModel = setSourceModelShowAct->isChecked();
	options.showTargetModel = setTargetModelShowAct->isChecked();
	options.showDeformedModel = setDeformedModelShowAct->isChecked();
	options.showSourceCorresp = sourceCorrespShowAct->isChecked();
	options.showTargetCorresp = targetCorrespShowAct->isChecked();
	options.showDeformedCorresp = deformedCorrespShowAct->isChecked();
	options.showShapeWeights = showShapeWeightsAct->isChecked();
	options.showAllFrames = showAllFramesAct->isChecked();
	options.displayFrame = displayFrame;
	options.showGraph = setGraphShowAct->isChecked();
	options.showEffRad = setEffRadShowAct->isChecked();
	options.renderTransparent = renderTransparent();
	options.showComponent = scene->currentBone;
	options.showGraphDeformed = showDeformedGraphAct->isChecked();
	options.showGraphFaces = showGraphFacesAct->isChecked();
	options.showExtendedGrid = -1;
	options.showWireframe = showWireframe();
	options.openGLPointSize = 2.0;
	options.sourceModelColor = RealColor(1.0, 0.7, 0.7);
	//options.deformedModelColor = RealColor(0.7, 1.0, 0.7);
	options.deformedModelColor = RealColor(0.7, 0.7, 1.0);
	options.targetModelColor = RealColor(0.7, 0.7, 1.0);
	options.pointColor = RealColor(scene->pointColor[0],scene->pointColor[1],scene->pointColor[2]);
	options.showErrorComponent = showErrorComponent;
	options.ignoreDisplayInterface = false;

	return options;
}

