///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include <QGLViewer/qglviewer.h>
#include <QtGui>
#include <QMutex>
#include "observer.h"
#include "subject.h"
#include "PointList.h"
#include "PointListN.h"
#include "FindCorrespondence.h"

class Scene;
class CorrespViewer;
class DeformEdit;
class MatchCandidates;
namespace Deform { class DrawOptions; }

/**
 * Viewer class.
 *
 * Responsibilities:
 *  - Displaying meshes on screen
 *  - Implementing menu & menu functionality
 */
class Viewer : public QGLViewer, public Observer
{
	Q_OBJECT

public:
	Viewer(Scene* s);										/**< Constructor: initialize members here. */
	virtual ~Viewer();										/**< Destructor: destroy members here. */

	virtual void Update(Subject* theChangedSubject);		/**< Update the display according to the subject */

private slots:
	void open(const char *file = NULL, int maxFrames = 0, int skipFrames = 1);	/**< Open at most "maxFrames" frames of a mesh animation sequence. */
	void add();

	void clearResult();										/**< Clear the deformation result. */
	void setGraphNodes();									/**< Set the number of graph nodes. */
	void resampleGraph();									/**< Resample the graph node locations. */
	void customSaveSnapshot();								/**< Take a snapshot and save it in a specific file. */
	void captureVideo();									/**< Take a video of the mesh animation sequence. */

	void CreateDeformEditor();								/**< Display the dialog for editing deformation parameters */
	void DeformEditorDestroyed();							/**< Cleanup after the editor has been closed */
	void DeformEditorChanged(int which, double value);		/**< Handle changes in parameter values */

	void CreateCorrespViewer();								/**< Display the correspondence viewer */
	void CorrespViewerDestroyed();							/**< Cleanup after the correspondence viewer has been closed */

	void saveColormap();									/**< Save the current colormap */
	void loadColormap();									/**< Load the current colormap */
	void setColormapLimits();								/**< Set colormap limits */

	void setFeatureThreshold();								/**< Set threshold on feature for display */

	void autoLaunch();										/**< Code to execute after the first draw */
	void autoProcess();										/**< Automatic processing for batch jobs */

	void fitViewToShape( const ShapeModel *shape = NULL );	/**< Fit the view to the current shape. */

	void changeBackgroundColor();							/**< Change background color */
	void changeForegroundColor();							/**< Change foreground color */

	void CaptureResultSnapshots_AlignPair();				/**< Capture a sequence of snapshots showing the result. */
	void CaptureResultSnapshots_AlignSeq();					/**< Capture a sequence of snapshots showing the result. */
	void CaptureResultSnapshots_GridWeights();				/**< Capture a sequence of snapshots showing the result. */

protected:
	virtual void draw();
	virtual void init();
	virtual void animate();
	virtual QString helpString() const;
	virtual	void keyPressEvent(QKeyEvent *e);
	virtual void mousePressEvent(QMouseEvent *e);
	virtual void mouseMoveEvent(QMouseEvent *e);
	virtual void mouseReleaseEvent(QMouseEvent *e);

private:
	//////////////////////////////////////////////////////////////////////////
	// Context Menu Related Methods and Members
	//////////////////////////////////////////////////////////////////////////
	QMenu *contextMenu;
	QMenu *fileMenu;
	QMenu *displayMenu;
	QMenu *toolsMenu;
	QMenu *optionsMenu;

	QAction *openAct;
	QAction *addAct;
	QAction *setWireframeModeAct;
	QAction *setBboxModeAct;
	QAction *setSourceModelShowAct;
	QAction *setTargetModelShowAct;
	QAction *setDeformedModelShowAct;
	QAction *sourceCorrespShowAct;
	QAction *targetCorrespShowAct;
	QAction *deformedCorrespShowAct;
	QAction *setGraphShowAct;
	QAction *setEffRadShowAct;
	QAction *setTransparencyAct;
	QAction *clearResultAct;
	QAction *setGraphNodesAct;
	QAction *resampleGraphAct;
	QAction *captureVideoAct;
	QAction *showDeformEditorAct;
	QAction *showAllFramesAct;
	QAction *showCorrespViewerAct;
	QAction *showAccumulatedGeometryViewerAct;
	QAction *saveColormapAct;
	QAction *loadColormapAct;
	QAction *setColormapLimitsAct;
	QAction *setFeatureThresholdAct;
	QAction *showCorrespEditorAct;
	QAction *saveFramesAct;
	QAction *visualizeClusterAct;
	QAction *fitViewToShapeAct;
	QAction *changeBackgroundColorAct;
	QAction *changeForegroundColorAct;
	QAction *shapeSliceAct;
	QAction *shapePoseAct;
	QAction *assignTransformsAct;
	QAction *assignTransformsReverseAct;
	QAction *showVerticesAct;
	QAction *showVertexFramesAct;
	QAction *computeAlignDistanceAct;
	QAction *showDeformedSameColorAct;
	QAction *showDeformedGraphAct;
	QAction *showGraphFacesAct;
	QAction *showDisplayOptionsAct;
	QAction *showShapeWeightsAct;

	QActionGroup *modelDisplayGroup;
	QAction *showFeaturesAct;
	QAction *showCurvaturesAct;
	QAction *showFramesAct;
	QAction *showGeometryAct;
	QAction	*showSamplesAct;
	QAction *showStatusAct;
	QAction *showCellIndexAct;

	QActionGroup *deformGroup;
	QAction *useEDAct;
	QAction *useWDAct;
	QAction *useDeformEMAct;
	QAction *useDeformAllAct;

	void createContextMenu();


	//////////////////////////////////////////////////////////////////////////
	// Scene data
	//////////////////////////////////////////////////////////////////////////
	Scene *scene;

	//////////////////////////////////////////////////////////////////////////
	// State methods and parameters
	//////////////////////////////////////////////////////////////////////////
	bool showWireframe();
	bool showBbox();
	bool showSourceModel();
	bool showTargetModel();
	bool showDeformedModel();
	int whichDeform();
	bool renderTransparent();
	QString updateFrame(int frame);
	void drawSphereTest();
	void drawCurvatureScales();
	void colormapTest();
	void pointListTest();
	ShapeModel *selectCurrentModel();
	void drawFrame( const ShapeModel* frame );
	Deform::DrawOptions getDeformDrawOptions();			/**< Gather the drawing options for drawing the deformed model */
	int showErrorComponent;

	void rotateAbout(qglviewer::Vec up, float angle);
	int displayFrame;

	// Scene change notify variable
	QMutex sceneMutex;
	bool sceneChanged;

	// Snapshot file
	QString customSaveSnapshotFile;

	//////////////////////////////////////////////////////////////////////////
	// Set deform parameters
	DeformEdit* deformEdit;
	float maxGraphNodes, maxGraphNeighbors, maxOverlapFactor;

	CorrespViewer *correspViewer;

	//////////////////////////////////////////////////////////////////////////
	// Test capabilities of PointList
	PointList *tempList;

	//////////////////////////////////////////////////////////////////////////
	// Test sphere-fitting
	enum InteractionMode { NONE = 0, DEFINE_SPHERE };
	InteractionMode interactionMode_;
	real3_type clickStartPos, clickEndPos;
	real3_type fitCenter;
	real_type fitRadius;

	real3_type closestPointOnLine( QPoint pos, real3_type x );
	void sphereFit();

	// Range query for mesh
	idxList inRange;
	distList distRange;

	// Range query in feature space
	int sourceQueryIndex;
	PointList sourceQuery, targetMatch;

	// Curvature values
	int displayScale;
	PointListN *sourceFeatures, *targetFeatures;

	// Maximum values for various correspondence parameters
	float maxFeatureSpaceRange;

	// Principal curvatures testing variables
	int curvDisplayMode;

	bool autoLaunchDone;

signals:
	void updateChildDisplay();
};
