///////////////////////////////////////////////////////////////////////////
// Copyright and License Notice
///////////////////////////////////////////////////////////////////////////
// These files contain code made by Will Chang under supervision of
// Matthias Zwicker for his Ph.D. work with the University of
// California, San Diego.
//   
// Copyright (c) 2009 - 2011, William Y. Chang
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//     * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//     
//     * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
// 
//     * Neither the name of University of California, San Diego nor
//     the names of its contributors may be used to endorse or promote
//     products derived from this software without specific prior
//     written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
///////////////////////////////////////////////////////////////////////////

#include "DeformEM.h"
#include <glut.h>
#include <iostream>
#include <QtTest/QTest>
#include <tsnnls.h>
#include "scene.h"
#include "taucsUtil.h"
#include <GraphCutLib/GCoptimization.h>
#include <QStandardItemModel>
#include "geom_util.h"

#undef min
#undef max

extern "C" {
	#include <SFMT/SFMT.h>
}

using namespace Geometry;
using namespace std;

DeformEMOptions::DeformEMOptions() :
updateCorrespType(CORRESPONDENCE_NONE),
maxEstepItr(20), maxProjItr(20), maxIRLSItr(50),
solveInitialRigid(false), useAccumulatedGeometry(true),
markCellsWithPoints(true),
inverse_deformation_maxSearchLevel(8),
inverse_deformation_gridExpandStep(3),
maxErrorPerCorrespondenceThreshold(0.1),
solveFinalNNLS(false)
{

}

DeformEM::DeformEM() :
DeformAlgorithm(DEFORM_EM), vWeightGrid(NULL),
scene(NULL),
stopEpsilon(1e-9), stopEpsilon_sqRt(1e-6), stopEpsilon_cubeRt(1e-3),
stepType(DEFORM_EM_INITIALIZE), globalIterationNumber(0),
boundaryCorrespondenceDistance(0.0), maxCorrespondenceDistance(0.0), minNormalCosAngle(1.0),
currentSolveFrame(-1), solutionView(NULL), solutionViewCommunicator(NULL),
controlSignal(SOLVER_RESUME), input(NULL), deformed(NULL), cl(NULL)
{

}

DeformEM::~DeformEM()
{
	if (vWeightGrid) {
		delete vWeightGrid;
		vWeightGrid = NULL;
	}
	if (solutionView) {
		delete solutionView;
		solutionView = NULL;
	}
	if (deformed) {
		delete deformed;
		deformed = NULL;
	}
}

void DeformEM::initialize( const ShapeModel *m, Scene *s, ShapeModel *target,
						  PointList *sample, const DeformEMOptions &options )
{
	// Assign input mesh
	input = m;

	// Assign scene variable
	scene = s;

	// Set options
	solveOptions = options;

	// Initialize accumulated geometry
	accumulatedGeometry.resetInitialShape(*input);

	////else {
	////	// Use the deformed mesh
	////	if (deformed) {
	////		delete deformed;
	////		deformed = NULL;
	////	}
	////	deformed = m->makeCopy();
	////}

	// Disable using default correspondence pointer
	this->cl = NULL;

	// Initialize solutions
	solutionPerFrame.clear();
	solutionPerFrame.push_back(FrameSolution());
	FrameSolution &lastSolution = solutionPerFrame[solutionPerFrame.size()-1];
	lastSolution.solvedDeformation = RigidDeform(options.numBones);
	lastSolution.sourceModel = input;
	lastSolution.targetModel = target;
	lastSolution.deformedAccumulatedGeometry = accumulatedGeometry;
	lastSolution.generateInitialCorrespondences(options.numCorrespondences);

	// Initialize current solve frame
	currentSolveFrame = 0;


	//////////////////////////////////////////////////////////////////////////
	// Generate the grid from the mesh's bounding box
	//////////////////////////////////////////////////////////////////////////
	cout << "*** Generating vertex weight grid...";

	// Extend the bounding box along the diagonal by 10% in each direction
	real3_type extend = (input->box_max - input->box_min) * 0.1;

	perfTimers["Allocating Grid"].Start();

	vWeightGrid = new MarkedUniformGrid3(options.numBones);
	vWeightGrid->createFromBBox(input->box_min - extend, input->box_max + extend, options.gridDivisions);
	vWeightGrid->resetMark();
	if (solveOptions.markCellsWithPoints) {
		vWeightGrid->markCells(input->vertex);
	}
	else if (input->shapeType == SHAPE_MESH) {
		vWeightGrid->markCells((const Mesh3_real&)*input);
	}
	vWeightGrid->resetData();
	vWeightGrid->findEdges();				/**< Find the edges in the marked grid */

	perfTimers["Allocating Grid"].Stop();
	printTime("Allocating Grid");


	//////////////////////////////////////////////////////////////////////////
	// Populate initial vertex weights
	//////////////////////////////////////////////////////////////////////////
	cout << "*** Populating initial vertex weights...";
	int numInflNeighbors = options.numInflNeighbors;

	// Guess the initial bone positions by sampling from the mesh
	if (sample) {
		initialBoneLocation = *sample;
	}
	else {
		input->BestCandidateSampleVertices(options.numBones, initialBoneLocation);
	}
	initialBoneRadius.resize(initialBoneLocation.size());

	cout << "Initial bone positions:" << endl;
	for (index_type i = 0; i < initialBoneLocation.size(); i++) {
		cout << "Bone " << i << ": " << initialBoneLocation[i] << endl;
	}
	// Clamp the number of influential neighbors to the number of bones
	if (numInflNeighbors > initialBoneLocation.size() - 1) {
		numInflNeighbors = initialBoneLocation.size() - 1;
	}

	// Assign the influence at each bone
	// IMPORTANT TODO: optimize the bone locations so that all vertices have at least one bone affecting it
	for (size_type i = 0; i < initialBoneLocation.size(); i++) {
		idxList list;
		distList dist;
		float d_max, sigma;

		// Set the radius to be the distance to the k^th nearest neighbor
		initialBoneLocation.queryKNN(initialBoneLocation[i], numInflNeighbors+1, list, dist, true); // the closest nearest neighbor is the query point itself

		// Compute Gaussian sigma parameter based on this distance
		d_max = dist[numInflNeighbors];
		initialBoneRadius[i] = 2*d_max;

		// Assign grid values
		int3_type gridIndex = vWeightGrid->scanStart();
		do {
			real3_type gridPos = vWeightGrid->gridLocation(gridIndex);

			// Evaluate the weight function at that position
			vWeightGrid->setValue(gridIndex, i, EvaluateGaussian(gridPos, initialBoneLocation[i], initialBoneRadius[i]));

		} while(vWeightGrid->scanNext(gridIndex));
	}

	// Normalize the weights
	//vWeightGrid->normalizeValues();
	vWeightGrid->clampMaxValues();			/**< Use discrete labeling */

	//////////////////////////////////////////////////////////////////////////
	// Set the bone-influence sets so that each bone affects the entire grid
	setBoneInfluence_EntireGrid();

	//////////////////////////////////////////////////////////////////////////
	// For the data term, set up the initial list of correspondences in each cell
	for (int i = 0; i < currentCL()->size(); ++i) {
		//// Only add correspondences that are valid
		//if (!currentCL()->valid(i)) {
		//	continue;
		//}

		// For each correspondence, compute the cell containing it
		int3_type cellIndex;
		real3_type remainder;
		vWeightGrid->queryCell(currentCL()->sourcePos(i), cellIndex, remainder);

		// Add to the list of correspondences at this cell
		currentCorrespondencesPerCell()[cellIndex].push_back(i);
	}

	cout << "done!" << endl;
}


void DeformEM::initialize(ShapeModel *source, ShapeModel *alignedTarget,
						  int gridDivisions, int numBonesPerGridPoint, int numExpandInfluence,
						  RigidDeform &tfList, std::vector<int> vertexLabels,
						  RigidDeform &invTfList, std::vector<int> invVertexLabels,
						  CorrespondenceList *con,
						  FloatArray Eweights, FloatArray Mweights )
{
	// Assign input mesh and vertex constraints
	input = source;
	this->cl = con;

	// Copy the data to the deformed mesh
	if (deformed == NULL) {
		deformed = input->makeCopy();
	}

	// Set initial transformations
	setInitialDeformation(tfList);
	//initialDeform.resize(tfList.size());

	// Copy over the weights
	solveOptions.EStepWeights = Eweights;
	solveOptions.MStepWeights = Mweights;

	//////////////////////////////////////////////////////////////////////////
	// Generate the grid from the mesh's bounding box
	// Extend the bounding box along the diagonal by 10% in each direction
	real3_type extend = (input->box_max - input->box_min) * 0.1;

	cout << "Generating vertex weight grid...";
	perfTimers["Allocating Grid"].Start();

	vWeightGrid = new MarkedUniformGrid3(tfList.size());
	vWeightGrid->createFromBBox(input->box_min - extend, input->box_max + extend, gridDivisions);
	vWeightGrid->resetMark();
	if (input->shapeType == SHAPE_MESH) {
		vWeightGrid->markCells((const Mesh3_real&)*input);
		vWeightGrid->markCells((const Mesh3_real&)*alignedTarget);				// Mark also the target
	}
	vWeightGrid->resetData();
	vWeightGrid->findEdges();				/**< Find the edges in the marked grid */

	perfTimers["Allocating Grid"].Stop();
	printTime("Allocating Grid");

	//////////////////////////////////////////////////////////////////////////
	// Assign the influence for each transformation
	for (int i = 0; i < input->numVertices(); i++) {
		if ((vertexLabels[i] < 0) ||
			(input->markVertex[i] == VERTEX_NO_ADJACENT_TRIANGLE)) {
				continue;
		}

		// For each vertex:
		// 1. Find the grid cell containing the vertex
		// 2. Set the weight for each cell vertex equal to 1 for this bone
		real3_type v = input->vertex[i];
		vWeightGrid->setValues(v, vertexLabels[i], 1.0);
	}

	for (int i = 0; i < alignedTarget->numVertices(); i++) {
		if ((invVertexLabels[i] < 0) ||
			(alignedTarget->markVertex[i] == VERTEX_NO_ADJACENT_TRIANGLE)) {
				continue;
		}

		real3_type v = alignedTarget->vertex[i];
		vWeightGrid->setValues(v, invVertexLabels[i], 1.0);
	}

	// Normalize the weights
	vWeightGrid->normalizeValues();


	//////////////////////////////////////////////////////////////////////////
	// Determine bone-influence sets

	std::vector< std::vector<int> > cellConstraints;
	cellConstraints.resize(vWeightGrid->totalGridCells());
	cellsPerBone.resize(tfList.size());
	std::vector< pair<real_type,int> > residual;					/**< Store the residual of applying each transformation */
	residual.resize(tfList.size());

	// For each cell, collect the constraints that are inside the cell
	for (int i = 0; i < cl->size(); i++) {
		// Find the linear index of the cell that contains the constraint
		int3_type idx;
		real3_type rem;

		vWeightGrid->queryCell(cl->sourcePoints[i], idx, rem);
		cellConstraints[vWeightGrid->getCellLinearIndex(idx)].push_back(i);
	}

	int3_type gridCellIndex = vWeightGrid->scanCellStart();
	do {
		const std::vector<int> &constrs = cellConstraints[vWeightGrid->getCellLinearIndex(gridCellIndex)];

		//////////////////////////////////////////////////////////////////////////
		// Get the bones with non-zero weights in the grid points of this cell
		data_type nonzeroWeights = vWeightGrid->cellNonzeroValue(gridCellIndex);
		std::vector<int> bonesWithWeights;
		for (int j = 0; j < tfList.size(); j++) {
			if (nonzeroWeights.val[j] >= 8.0) {
				bonesWithWeights.push_back(j);
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// Compute the residual of applying each transformation to this cell's constraints
		for (int i = 0; i < tfList.size(); i++) {
			residual[i].first = 0.0;
			residual[i].second = i;
			const SE3 &tf = tfList[i];

			// For each constraint, add to the residual
			for (int j = 0; j < constrs.size(); j++) {
				real3_type diff = cl->targetPoints[constrs[j]] - tf * cl->sourcePoints[constrs[j]];
				residual[i].first += diff*diff;
			}
		}

		// Sort based on residual (ascending order)
		struct resComp : public binary_function<pair<real_type,int>, pair<real_type,int>, bool> {
			bool operator()(pair<real_type,int> x, pair<real_type,int> y) {
				return (x.first < y.first);
			}
		};
		sort(residual.begin(), residual.end(), resComp());

		// Find the top transformations with the least amount of residual
		std::vector<int> topBones;

		if (constrs.size()) {
			topBones.resize(numBonesPerGridPoint);

			for (int i = 0; i < numBonesPerGridPoint; i++) {
				topBones[i] = residual[i].second;
			}

			sort(topBones.begin(), topBones.end());
		}

		//////////////////////////////////////////////////////////////////////////
		// Assign the bone-influence set
		std::vector<int> affBones;

		// Option 1: Use the cells that predict the constraints well
		//affBones = topBones;

		// Option 2: Use the cells where the vertex weights were initialized to a non-zero value in all 8 corners
		//affBones = bonesWithWeights;

		// Option 3: Union of 1 and 2
		sortedListUnion(topBones, bonesWithWeights, affBones);

		//////////////////////////////////////////////////////////////////////////
		// Add this cell to the list of cells affected by these bones
		index_type gridCellLinearIndex = vWeightGrid->getCellLinearIndex(gridCellIndex);
		for (int j = 0; j < affBones.size(); j++) {
			cellsPerBone[affBones[j]].push_back(gridCellLinearIndex);
		}

	} while(vWeightGrid->scanNextCell(gridCellIndex));


	//////////////////////////////////////////////////////////////////////////
	// Expand the region of influence for each bone by a number of steps
	for (int i = 0; i < numExpandInfluence; i++) {
		expandBoneInfluence();
	}

	//////////////////////////////////////////////////////////////////////////
	// Translate into other formats
	bonesPerCell.resize(vWeightGrid->totalGridCells());
	bonesPerPoint.resize(vWeightGrid->totalGridPoints());
	pointsPerBone.resize(tfList.size());

	for (int i = 0; i < tfList.size(); i++) {
		std::vector<int> &affCells = cellsPerBone[i];

		for (int j = 0; j < affCells.size(); j++) {
			bonesPerCell[affCells[j]].push_back(i);
		}
	}

	gridCellIndex = vWeightGrid->scanCellStart();
	do {
		const std::vector<int> &affBones = bonesPerCell[vWeightGrid->getCellLinearIndex(gridCellIndex)];

		//////////////////////////////////////////////////////////////////////////
		// Based on bones affecting cells, figure out the bones affecting each grid point
		idxList gridPtIndices;
		vWeightGrid->getCellGridPoints(gridCellIndex, gridPtIndices);
		for (int i = 0; i < gridPtIndices.size(); i++) {						/**< For each grid point on this cell */
			sortedListUnionMerge(affBones, bonesPerPoint[gridPtIndices[i]]);
		}

		//////////////////////////////////////////////////////////////////////////
		// Add the grid points to the list of grid points affected by these bones
		std::vector<int> gridPtIndicesInt;					/**< Convert to list of integers */
		gridPtIndicesInt.resize(gridPtIndices.size());
		for (int i = 0; i < gridPtIndices.size(); i++) {
			gridPtIndicesInt[i] = gridPtIndices[i];
		}
		sort(gridPtIndicesInt.begin(), gridPtIndicesInt.end());
		for (int j = 0; j < affBones.size(); j++) {			/**< Add to list of cells */
			sortedListUnionMerge(gridPtIndicesInt, pointsPerBone[affBones[j]]);
		}

	} while(vWeightGrid->scanNextCell(gridCellIndex));

	// Find mapping of each bone's selected cells/points into the arrays
	boneCellLinearIndex.resize(tfList.size());
	bonePointLinearIndex.resize(tfList.size());
	for (int i = 0; i < tfList.size(); i++) {

		// Insert the linear indices into the hash map
		for (int j = 0; j < cellsPerBone[i].size(); j++) {
			boneCellLinearIndex[i][cellsPerBone[i][j]] = j;
		}

		for (int j = 0; j < pointsPerBone[i].size(); j++) {
			bonePointLinearIndex[i][pointsPerBone[i][j]] = j;
		}
	}

	cout << "done!" << endl;
}

real_type DeformEM::EvaluateGaussian( real3_type x, real3_type center, real_type d_max )
{
	real3_type diff = x - center;
	real_type distsq = diff[0]*diff[0] + diff[1]*diff[1] + diff[2]*diff[2];

	// If the point is within the maximum distance
	if (sqrt(distsq) < d_max) {
		return exp(-4.5 * distsq / (d_max*d_max));		/**< Calculate the weight with sigma = d_max/3. */
	}

	// Otherwise return 0
	return 0.0;
}


void DeformEM::draw( const DrawOptions &options ) const
{
	changingFrames.lock();

	if (solutionView) {
		real_type drawOffset = 0.0;

		// Use solutionView to choose what to display
		for (int i = 0; i < solutionView->rowCount(); ++i) {
			const FrameSolution &frameSoln = solutionPerFrame[i];
			bool showSource = solutionView->data(solutionView->index(i, VIEW_SOURCE)).toBool();
			bool showTarget = solutionView->data(solutionView->index(i, VIEW_TARGET)).toBool();
			bool showDeformed = solutionView->data(solutionView->index(i, VIEW_DEFORMED)).toBool();
			bool showInverse = solutionView->data(solutionView->index(i, VIEW_INVERSE_DEFORMED)).toBool();
			bool showAdjCorresp = solutionView->data(solutionView->index(i, VIEW_ADJ_CORRESP)).toBool();
			bool showCorrespSrc = solutionView->data(solutionView->index(i, VIEW_CORRESP_SOURCE)).toBool();
			bool showCorrespTgt = solutionView->data(solutionView->index(i, VIEW_CORRESP_TARGET)).toBool();

			if (options.ignoreDisplayInterface) {
				showSource = options.showSourceModel;
				showTarget = options.showTargetModel;
				showDeformed = options.showDeformedModel;
				showInverse = false;
				showAdjCorresp = false;
				showCorrespSrc = false;
				showCorrespTgt = false;
			}

			glEnable(GL_LIGHTING);
			glDisable(GL_COLOR_MATERIAL);
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

			glPushMatrix();
			glTranslatef(drawOffset, 0.0, 0.0);
			if (showSource) {
				setOpenGLMaterialColor(options.sourceModelColor);
				frameSoln.sourceModel->draw();
			}

			if (showTarget) {
				setOpenGLMaterialColor(options.targetModelColor);
				frameSoln.targetModel->draw();
			}

			if (showDeformed) {
				setOpenGLMaterialColor(options.deformedModelColor);
				frameSoln.deformedAccumulatedGeometry.draw(options);
			}

			if (frameSoln.inverseMappedShape && showInverse) {
				setOpenGLMaterialColor(options.deformedModelColor);
				frameSoln.inverseMappedShape->draw();
			}

			// Draw correspondences (activated = yellow, unactivated = black)
			RealColor minColor(0.0, 0.0, 0.0);
			RealColor maxColor(scene->pointColor[0],scene->pointColor[1],scene->pointColor[2]);
			RealColor maxColor2(scene->targetColor[0],scene->targetColor[1],scene->targetColor[2]);

			if (showAdjCorresp && frameSoln.aList) {
				// Display the source positions
				PointList_Utility::drawWithWeight(frameSoln.aList->sourcePoints, POINT_SPHERES, frameSoln.aList->weight, minColor, maxColor, scene->pointSize() * 0.5);
				PointList_Utility::drawWithWeight(frameSoln.aList->targetPoints, POINT_SPHERES, frameSoln.aList->weight, minColor, maxColor, scene->pointSize() * 0.5);
			}

			if (showCorrespSrc && frameSoln.cList) {
				// Display the source positions
				//PointList_Utility::drawWithWeight(frameSoln.cList->sourcePoints, POINT_SPHERES, frameSoln.cList->weight, minColor, maxColor, scene->pointSize() * 0.5);

				// Display the deformed source positions
				PointList_Utility::drawWithSubsetWeight(frameSoln.deformedAccumulatedGeometry.vertex, POINT_SPHERES, frameSoln.cList->sourceIndex, frameSoln.cList->weight, minColor, maxColor2, scene->pointSize()* 0.5);
			}

			if (showCorrespTgt && frameSoln.cList) {
				// Display the target positions
				PointList_Utility::drawWithWeight(frameSoln.cList->targetPoints, POINT_SPHERES, frameSoln.cList->weight, minColor, maxColor, scene->pointSize() * 0.5);
			}

			// Draw the graph
			if (options.showGraph && (currentSolveFrame == i)) {
				glDisable(GL_LIGHTING);

				if (options.showGraphFaces) {
					if (!options.showWireframe) {
						glColor3f(0.0, 0.0, 0.0);
						vWeightGrid->drawEdges(options.showGraphDeformed);
					}
					vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed);
				}
				else {
					vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed, true);
				}

				// Show expanded grid
				glColor3f(0.0, 0.0, 0.5);
				vWeightGrid->drawExpandedPts(options.showExtendedGrid);
			}
			glPopMatrix();

			drawOffset += frameSoln.deformedAccumulatedGeometry.box_max[0] - frameSoln.deformedAccumulatedGeometry.box_min[0];
		}
	}
	else {
		if (options.showDeformedModel) {
			currentDeformedAccumulatedGeometry().draw(options);
		}
		else {
			// Draw the deformed model
			if (deformed && options.showDeformedModel) {
				setOpenGLMaterialColor(options.deformedModelColor);
				deformed->draw();
			}
		}

		if (options.showTargetModel && currentTarget()) {
			glEnable(GL_LIGHTING);
			setOpenGLMaterialColor(RealColor(0.7, 0.7, 1.0));
			currentTarget()->draw();
		}

		if (options.showGraph) {
			glDisable(GL_LIGHTING);

			if (options.showGraphFaces) {
				if (!options.showWireframe) {
					glColor3f(0.0, 0.0, 0.0);
					vWeightGrid->drawEdges(options.showGraphDeformed);
				}
				vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed);
			}
			else {
				vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed, true);
			}

			// Show expanded grid
			glColor3f(0.0, 0.0, 0.5);
			vWeightGrid->drawExpandedPts(options.showExtendedGrid);
		}
	}


	//// Show the data error at each grid cell
	//if (currentCorrespondenceError().size() == currentCL()->size()) {
	//	glDisable(GL_LIGHTING);
	//	glEnable (GL_POLYGON_OFFSET_FILL);
	//	glPolygonOffset (1., 1.);
	//	glPointSize(10.0);
	//	glBegin(GL_POINTS);
	//	int comp = std::min(2, std::max(0, options.showErrorComponent));
	//	for (int i = 0; i < currentCorrespondenceError().size(); i++) {
	//		if (currentCL()->valid(i)) {
	//			real3_type pos = currentCL()->targetPos(i);
	//			RealColor color = scene->getColormapColor(currentCorrespondenceError()[i][comp]);
	//			setOpenGLColor(color);
	//			//glVertex3f(pos[0], pos[1], pos[2]);
	//			drawSphere(pos, scene->pointSize() * 0.5, 7, 7);
	//		}
	//	}
	//	glEnd();
	//}

	//// Draw bounding boxes for new labels to insert
	//glLineWidth(1.0);
	//glDisable(GL_LIGHTING);
	//glColor3f(0.0,0.0,0.0);
	//for (int i = 0; i < newBoneBoxes.size(); i++) {
	//	drawBoxOpenGL(newBoneBoxes[i].first, newBoneBoxes[i].second);
	//}
	//// Draw correspondences whose error is really big
	//glEnable(GL_LIGHTING);
	//setOpenGLMaterialColor(scene->pointColor);
	//for (int i = 0; i < showOutliers.size(); i++) {
	//	drawSphere(newBoneSpheres[i].first, newBoneSpheres[i].second);
	//	for (int j = 0; j < showOutliers[i].size(); j++) {
	//		drawSphere(showOutliers[i][j], scene->pointSize() * 1.5);
	//	}
	//}


	//// Draw the grid
	//if (options.showGraph) {
	//	glDisable(GL_LIGHTING);
	//	//glLineWidth(2.0);

	//	if (options.showGraphFaces) {
	//		if (!options.showWireframe) {
	//			glColor3f(0.0, 0.0, 0.0);
	//			vWeightGrid->drawEdges(options.showGraphDeformed);
	//		}
	//		vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed);
	//	}
	//	else {
	//		vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed, true);
	//		//vWeightGrid->draw(options.showComponent, options.showGraphDeformed);
	//		//glColor3f(0.0, 0.0, 0.0);
	//		//vWeightGrid->drawEdges(options.showGraphDeformed);
	//	}

	//	// Show expanded grid
	//	glColor3f(0.0, 0.0, 0.5);
	//	vWeightGrid->drawExpandedPts(options.showExtendedGrid);

	//	//vWeightGrid->draw(-1);
	//	//vWeightGrid->drawSortedSubset(comp, cellsPerBone[comp]);
	//	//vWeightGrid->drawEdges(comp);
	//	//vWeightGrid->drawDeformedBoundaryFaces();
	//	//vWeightGrid->drawDeformedExpandedPts();

	//	//glEnable(GL_LIGHTING);
	//	//// Debugging: draw the constraints in the selected cell
	//	//if (cellConstraints.size()) {
	//	//	int selected = max(0, min((int) cellConstraints.size()-1, comp));

	//	//	for (int i = 0; i < cellConstraints[selected].size(); i++) {
	//	//		const real3_type &p = cl->sourcePoints[cellConstraints[selected][i]];
	//	//		drawSphere(p, 0.01);
	//	//	}
	//	//}

	//	//// Debugging: draw the constraints in the selected point
	//	//if (pointConstraints.size()) {
	//	//	int selected = max(0, min((int) pointConstraints.size()-1, comp));

	//	//	for (int i = 0; i < pointConstraints[selected].size(); i++) {
	//	//		const real3_type &p = cl->sourcePoints[pointConstraints[selected][i]];
	//	//		drawSphere(p, 0.03);
	//	//	}
	//	//}

	//	//// Debugging: draw the grid points affected by each bone
	//	//const std::vector<int> &sortedLinearIndex = pointsPerBone[comp];
	//	//int currentPos = 0;
	//	//int3_type ptIndex = vWeightGrid->scanStart();
	//	//do {
	//	//	int ptLinearIndex = vWeightGrid->getLinearIndex(ptIndex);

	//	//	// Check if the current cell is selected
	//	//	if (ptLinearIndex == sortedLinearIndex[currentPos]) {
	//	//		drawSphere(vWeightGrid->gridLocation(ptIndex), 0.01);
	//	//		currentPos++;

	//	//		if (currentPos >= sortedLinearIndex.size()) {
	//	//			break;
	//	//		}
	//	//	}
	//	//	else if (ptLinearIndex > sortedLinearIndex[currentPos]) {
	//	//		cerr << "(this case should never happen) somehow skipped drawing a selected point!" << endl;
	//	//	}
	//	//} while(vWeightGrid->scanNext(ptIndex));
	//}

	changingFrames.unlock();


	// Draw the initial bone locations used to populate the vertex weights
	if (options.showEffRad) {
		glEnable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glDepthMask (GL_FALSE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		// Set the material
		GLfloat color[4];
		color[0] = 0.7f;  color[1] = 0.7f;  color[2] = 1.0f;  color[3] = 0.7f;
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

		// Draw spheres representing the influence of the current node
		if (options.showComponent == -1) {
			// Draw all components
			for (int i = 0; i < initialBoneLocation.size(); i++) {
				real3_type node = initialBoneLocation[i];

				glPushMatrix();
				glTranslatef(node[0], node[1], node[2]);
				glutSolidSphere(initialBoneRadius[i], 20, 10);
				glPopMatrix();
			}
		}
		else {
			real3_type node = initialBoneLocation[options.showComponent];

			glPushMatrix();
			glTranslatef(node[0], node[1], node[2]);
			glutSolidSphere(initialBoneRadius[options.showComponent], 20, 10);
			glPopMatrix();
		}

		glDepthMask (GL_TRUE);
		glDisable(GL_BLEND);
	}
}

void DeformEM::solveStep()
{
	taucsType *solution = NULL;

	if (stepType == DEFORM_EM_INITIALIZE) {
		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "*** DeformEM: Initializing solver..." << endl;

		// Update correspondences
		setCorrespondenceThresholds(globalIterationNumber);
		updateCorrespondences(solveOptions.updateCorrespType, currentSolution());
	}

	// Before moving on, check that rigid deform is correct
	if (currentDeform().size() != vWeightGrid->getDataDim()) {
		currentDeform().clear();
		currentDeform().resize(vWeightGrid->getDataDim());
	}

	if (stepType == DEFORM_EM_E_STEP) {
		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "*** DeformEM: Optimizing transformations..." << endl;

		taucsType residual, gradResidual;
		LinearSystemE *lse = new LinearSystemE(vWeightGrid->getDataDim());
		lse->clear();

		perfTimers["E-Step Populating System"].Start();

		lse->makeConstraintTerm(solveOptions.EStepWeights[0], vWeightGrid, currentCL(), currentDeform());

		perfTimers["E-Step Populating System"].Stop();
		printTime("E-Step Populating System");

		solution = lse->solve(residual, gradResidual);

		printTime("E-Step Creating System");
		printTime("E-Step Solving System");

		perfTimers["E-Step Projecting and Applying"].Start();

		RigidDeform solvedDef = lse->projectToRigid(solution, solveOptions.maxProjItr, vWeightGrid, currentCL(), currentDeform());
		concatenateDeformation(solvedDef, currentDeform());
		applyDeformation_AccumulatedGeometry(currentDeform(), currentDeformedAccumulatedGeometry());

		perfTimers["E-Step Projecting and Applying"].Stop();
		printTime("E-Step Projecting and Applying");

		delete lse;
	}
	else if (stepType == DEFORM_EM_M_STEP) {
		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "*** DeformEM: Optimizing weights..." << endl;
		cout << "*** Grid size: " << vWeightGrid->size() << endl;

		taucsType residual, gradResidual;
		LinearSystemM_Reduced *lsmr = new LinearSystemM_Reduced(vWeightGrid->getDataDim(), vWeightGrid->totalGridPoints(), pointsPerBone, cellsPerBone, bonesPerPoint, bonesPerCell, boneCellLinearIndex, bonePointLinearIndex);
		lsmr->clear();

		cout << "*** Setting up matrices..." << endl;

		perfTimers["M-Step Populating System"].Start();

		cout << "****** Constraint term (weight " << solveOptions.MStepWeights[0] << ")" << endl;
		lsmr->makeConstraintTerm(solveOptions.MStepWeights[0], vWeightGrid, currentCL(), currentDeform(), bonesPerCell);
		cout << "****** Smooth term (weight " << solveOptions.MStepWeights[1] << ")" << endl;
		lsmr->makeSmoothTerm(solveOptions.MStepWeights[1], vWeightGrid);
		cout << "****** Unity term (weight " << solveOptions.MStepWeights[2] << ")" << endl;
		lsmr->makeUnityTerm(solveOptions.MStepWeights[2], vWeightGrid);

		perfTimers["M-Step Populating System"].Stop();
		printTime("M-Step Populating System");

		solution = lsmr->solve(solveOptions.maxIRLSItr, residual, gradResidual);

		printTime("M-Step Creating System");
		printTime("M-Step Solving System");

		perfTimers["M-Step Applying"].Start();

		lsmr->updateWeights(solution, vWeightGrid);
		vWeightGrid->normalizeValues();
		applyDeformation_AccumulatedGeometry(currentDeform(), currentDeformedAccumulatedGeometry());

		perfTimers["M-Step Applying"].Stop();
		printTime("M-Step Applying");


		delete lsmr;
	}
	else if (stepType == DEFORM_EM_UPDATE_CORRESP) {
		updateCorrespondences(solveOptions.updateCorrespType, currentSolution());
	}

	//////////////////////////////////////////////////////////////////////////	
	// Clean-up
	iterationFinished = true;
	if (solution) {
		delete[] solution;			/**< Delete temporary solution */
		solution = NULL;
	}

	Notify();						/**< Notify the viewer to update the display. */
}


void DeformEM::requestInteractiveCorrespUpdate()
{
	if (!solutionPerFrame.size()) {
		return;
	}

	FrameSolution &frameSoln = solutionPerFrame[0];
	CorrespondenceList *frameCL = frameSoln.cList;


}

void DeformEM::solveDeformation( const SolveOptions &options )
{
	// Set the initial solver states
	controlSignal = SOLVER_RESUME;
	iterationFinished = false;

	if (!checkConstraints(currentCL()) || !checkWeights()) {
		cerr << "Aborting solver." << endl;
		iterationFinished = true;
		return;
	}

	if (stepType != DEFORM_EM_AUTO_STEP) {
		solveStep();
		return;
	}

	// Declare variables
	LinearSystemE *lse = NULL;
	//LinearSystemM *lsm = NULL;
	LinearSystemM_Reduced *lsmr = NULL;
	RigidDeform solvedDef;
	taucsType *solution = NULL;
	taucsType Eres, Mres, lastEres, lastMres, temp;

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Automatic solve using DeformEM..." << endl;


	bool thisFrameFinished = false;
	while (!thisFrameFinished) {
		//printCurrentDeformation();

		if (globalIterationNumber == 0) {
			cout << "-------------------------------------------------------------------------------" << endl;
			cout << "Initializing solver..." << endl;

			// Resetting termination variables
			cout << "*** Resetting termination variables..." << endl;
			lastEres = lastMres = REAL_TYPE_MAX;

			Notify();						/**< Notify the viewer to update the display. */

			// Start the timer
			perfTimers["DeformEM Registration"].Start();
		}


		//////////////////////////////////////////////////////////////////////////
		// Step 1. Solve for the deformation given the current vertex weights.
		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "*** Entering E-step optimization..." << endl;

		for (int frame = 0; frame < solutionPerFrame.size(); ++frame)
		{
			FrameSolution &frameSoln = solutionPerFrame[frame];

			currentSolveFrame = frame;

			if (frame > 0) {
				vWeightGrid->updateDeformation(solutionPerFrame[frame-1].solvedDeformation);
			}

			if (!controlCheck()) {
				break;
			}

			//////////////////////////////////////////////////////////////////////////
			// Update correspondences
			// 1. Populate correspondences from adjacency list
			updateCorrespondences_fromAdjacent(frameSoln);

			cout << "*** Updating correspondences..." << endl;
			if ((globalIterationNumber == 0) && (solveOptions.solveInitialRigid) && (frame == solutionPerFrame.size()-1)) {
				solverStatus = "Initial Correspondences";

				perfTimers["DeformEM RANSAC Correspondences"].Start();

				solveOptions.updateCorrespType = CORRESPONDENCE_SPIN_IMAGES;
				setCorrespondenceThresholds(globalIterationNumber);
				updateCorrespondences(solveOptions.updateCorrespType, frameSoln);

				perfTimers["DeformEM RANSAC Correspondences"].Stop();
			}
			else {
				setCorrespondenceThresholds(globalIterationNumber);
				updateCorrespondences(solveOptions.updateCorrespType, frameSoln);
			}

			//////////////////////////////////////////////////////////////////////////
			CorrespondenceList *frameCL = frameSoln.cList;
			const ShapeModel *frameTarget = frameSoln.targetModel;
			RigidDeform &frameDeform = frameSoln.solvedDeformation;
			AccumulatedGeometry &deformedAccumulatedGeometry = frameSoln.deformedAccumulatedGeometry;
			const ShapeModel *frameSource = frameSoln.sourceModel;

			if (!checkConstraints(frameCL)) {
				cerr << "*** Checking correspondence constraints failed for FrameSolution " << frame << "!" << endl;
				continue;
			}


			solverStatus = "T-Step";
			bool EStepIterationFinished = false;
			lastEres = REAL_TYPE_MAX;
			int EStepNumItr = 0;

			cout << "E-step iteration:";

			while (!EStepIterationFinished) {
				cout << " " << EStepNumItr+1;
				perfTimers["DeformEM E-step"].AddStart();

				if (lse) delete lse;
				lse = new LinearSystemE(vWeightGrid->getDataDim(), excludedLabels);
				lse->clear();

//perfTimers["DeformEM E-step Populating System"].Start();
				//lse->makeConstraintTerm(solveOptions.EStepWeights[0], vWeightGrid, currentCL(), currentDeform());
				lse->makeConstraintTerm_PointToPlane(solveOptions.EStepWeights[0], vWeightGrid, frameCL, frameDeform, *frameTarget);
				lse->makeSmoothnessTerm(solveOptions.EStepWeights[1], vWeightGrid, frameDeform);
//perfTimers["DeformEM E-step Populating System"].Stop();

				solution = lse->solve(Eres, temp);

//perfTimers["DeformEM E-step Projecting Rigid"].Start();
				solvedDef = lse->projectToRigid(solution, solveOptions.maxProjItr, vWeightGrid, frameCL, frameDeform);
//perfTimers["DeformEM E-step Projecting Rigid"].Stop();

//perfTimers["DeformEM E-step Applying Deformation"].Start();
				concatenateDeformation(solvedDef, frameDeform);
				applyDeformation_FrameGeometry(frameDeform, deformedAccumulatedGeometry, frameSource);
//perfTimers["DeformEM E-step Applying Deformation"].Stop();

				if (solution) {
					delete[] solution;			/**< Delete temporary solution */
					solution = NULL;
				}

				Notify();						/**< Notify the viewer to update the display. */

				// Increment the number of E-step iterations
				EStepNumItr++;

				if ((EStepNumItr > 1) && (abs(Eres - lastEres) < stopEpsilon_sqRt*(1+Eres))) {
					cout << " done! (objective)" << endl;
					EStepIterationFinished = true;
				}
				if ((solveOptions.maxEstepItr > 0) && (EStepNumItr >= solveOptions.maxEstepItr)) {
					cout << " done! (max iteration)" << endl;
					EStepIterationFinished = true;
				}

				// Save the residual
				lastEres = Eres;

				// Update correspondences
				if ((solveOptions.updateCorrespType == CORRESPONDENCE_CLOSEST_POINTS) ||
					(solveOptions.updateCorrespType == CORRESPONDENCE_CLOSEST_POINTS_SPIN_IMAGE_WEIGHTS))
				{
					updateCorrespondences(solveOptions.updateCorrespType, frameSoln);
				}

				perfTimers["DeformEM E-step"].Stop();


				//// Print timer information
				//printTime("DeformEM E-step Populating System");
				//printTime("DeformEM E-step Projecting Rigid");
				//printTime("DeformEM E-step Applying Deformation");
				//printTime("E-Step Creating System 1");
				//printTime("E-Step Creating System 2");
				//printTime("E-Step Creating System 2.1");
				//printTime("E-Step Creating System 2.2");
				//printTime("E-Step Creating System 3");
				//printTime("E-Step Creating System 4");
				//printTime("E-Step Creating System 5");
				//printTime("E-Step Solving System");
			}
		}

		// Change correspondence type
		if ((globalIterationNumber == 0) && (solveOptions.solveInitialRigid)) {
			solveOptions.updateCorrespType = CORRESPONDENCE_CLOSEST_POINTS;
			updateCorrespondences(solveOptions.updateCorrespType, currentSolution());
		}

		if (!controlCheck()) {
			break;
		}

		//////////////////////////////////////////////////////////////////////////
		// Step 2. Solve for the vertex weights given the current deformation.
		cout << "-------------------------------------------------------------------------------" << endl;
		cout << "*** Entering M-step optimization..." << endl;
		cout << "****** Discrete constraint (weight " << solveOptions.DiscreteMStepWeights[0] << ")" << endl;
		cout << "****** Discrete smoothness (weight " << solveOptions.DiscreteMStepWeights[1] << ")" << endl;

		solverStatus = "W-Step";
		perfTimers["DeformEM M-step"].AddStart();

		// Solve discrete labelling problem
		Mres = solveDiscreteLabelling();
		analyzeLabellingResult();

		// Apply the deformation induced by the binary weights
		applyDeformation_FrameGeometry(currentDeform(), currentDeformedAccumulatedGeometry(), currentSource());

		perfTimers["DeformEM M-step"].Stop();

		Notify();						/**< Notify the viewer to update the display. */

		// Increment the number of iterations
		globalIterationNumber++;

		//////////////////////////////////////////////////////////////////////////
		// Determine stop conditions
		// Stop if the change in the objective function is small
		if ((globalIterationNumber > 1) && (abs(Mres - lastMres) < stopEpsilon_cubeRt*(1+Mres))) {
			cout << " done! (objective)" << endl;
			thisFrameFinished = true;
		}

		// Maximum number of iterations
		if ((options.maxItr > 0) && (globalIterationNumber >= options.maxItr)) {
			cout << " done! (max iteration)" << endl;
			thisFrameFinished = true;
		}

		// Store the residual
		lastMres = Mres;

		// Finally, update the correspondences (optional step)
		cout << "*** Updating correspondences..." << endl;
		setCorrespondenceThresholds(globalIterationNumber);
		updateCorrespondences(solveOptions.updateCorrespType, currentSolution());


		if (!controlCheck()) {
			break;
		}


		if (thisFrameFinished) {
			ShapeModel *nextFrame = requestNextFrame();

			// Check if there's another frame to process
			if (nextFrame) {
				// Option 2. Simultaneous alignment
				updateGrid(nextFrame);

				cout << "-------------------------------------------------------------------------------" << endl;
				cout << "*** Processing next frame: " << qPrintable	(nextFrame->getFilename()) << endl;

				// Move to next frame
				++currentSolveFrame;

				// Reset iteration variables
				globalIterationNumber = 0;

				// Also set the next frame in the scene
				scene->target = nextFrame;

				// Reset some iteration variables
				thisFrameFinished = false;
				lastMres = REAL_TYPE_MAX;

				// Update solution view
				updateSolutionViewModel();
			}
			else {
				perfTimers["DeformEM Registration"].Stop();

				// Check if we do the final NNLS update
				if (solveOptions.solveFinalNNLS) {
					perfTimers["DeformEM Final NNLS Update"].Start();

					solveWeights_finalNNLSUpdate();

					perfTimers["DeformEM Final NNLS Update"].Stop();
				}

				cout << "-------------------------------------------------------------------------------" << endl;
				//cout << "*** DeformEM Reached Last Frame!" << endl;
				thisFrameFinished = true;
				iterationFinished = true;
			}

			Notify();						/**< Notify the viewer to update the display. */
		}
		else {
			// Analyze the error of the assigned labels and potentially reuse some labels
			reuseExtraLabels();
		}


		if (!controlCheck()) {
			break;
		}
	}
	cout << "*** DeformEM exited solve loop." << endl;
	solverStatus = "";

	//////////////////////////////////////////////////////////////////////////
	// Delete allocated variables
	if (lse) delete lse;
	//if (lsm) delete lsm;
	if (lsmr) delete lsmr;
	lse = NULL;
	//lsm = NULL;
	lsmr = NULL;
}


void DeformEM::solveWeights_finalNNLSUpdate()
{
	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "*** Final M-step NNLS optimization..." << endl;

	// Save discrete weights before continuing
	saveDiscreteWeights();

	// First, remove any excluded labels from the optimization
	setBoneInfluence_EntireGrid_RemoveExcludedBones();

	// Optimization setup
	LinearSystemM_Reduced *lsmr = new LinearSystemM_Reduced(vWeightGrid->getDataDim(), vWeightGrid->totalGridPoints(), pointsPerBone, cellsPerBone, bonesPerPoint, bonesPerCell, boneCellLinearIndex, bonePointLinearIndex);
	lsmr->clear();

	cout << "****** constraint (weight " << solveOptions.MStepWeights[0] << ")" << endl;
	cout << "****** smooth (weight " << solveOptions.MStepWeights[1] << ")" << endl;
	cout << "****** unity (weight " << solveOptions.MStepWeights[2] << ")" << endl;
	cout << "****** discrete (weight " << solveOptions.MStepWeights[3] << ")" << endl;

	lsmr->makeConstraintTerm(solveOptions.MStepWeights[0], vWeightGrid, currentCL(), currentDeform(), bonesPerCell);
	lsmr->makeSmoothTerm(solveOptions.MStepWeights[1], vWeightGrid);
	lsmr->makeUnityTerm(solveOptions.MStepWeights[2], vWeightGrid);
	//lsmr->makeZeroTerm(solveOptions.MStepWeights[3], vWeightGrid);
	lsmr->makeDiscreteTerm(solveOptions.MStepWeights[3], vWeightGrid);

	taucsType temp1, temp2;
	taucsType *solution;
	solution = lsmr->solve(solveOptions.maxIRLSItr, temp1, temp2);
	//solution = lsmr->solve_OLD(temp1, temp2);

	lsmr->updateWeights(solution, vWeightGrid);
	vWeightGrid->normalizeValues();
	applyDeformation_FrameGeometry(currentDeform(), currentDeformedAccumulatedGeometry(), currentSource());


	//////////////////////////////////////////////////////////////////////////
	// Clean up
	if (solution) {
		delete[] solution;			/**< Delete temporary solution */
		solution = NULL;
	}
	delete lsmr;
}


////// TODO: phase out use of applyDeformation().
////void DeformEM::applyDeformation( const RigidDeform &deform )
////{
////	size_type n_vertices = deformed->numVertices();
////
////	// Deform the output mesh
////	for (size_type i = 0; i < n_vertices; i++) {
////		idxList bones;
////		distList weights;
////		SE3 tf;
////		tf.setZero();
////
////		// Get the vertex weights
////		vWeightGrid->getVertexWeights(input->vertex[i], bones, weights);
////
////		// Transform and assign the output vertex
////		for (size_type bi = 0; bi < bones.size(); bi++) {
////			index_type j = bones[bi];
////			//out += weights[bi] * (deform[j].rot * sourcePos + deform[j].tr);
////			tf.rot += weights[bi] * deform[j].rot;
////			tf.tr += weights[bi] * deform[j].tr;
////		}
////
////		//cout << "    -- Initial: " << i << ": " << sourcePos << endl;
////		//cout << "    -- Moved:   " << i << ": " << out << endl;
////
////		//deformed->vertex[i] = out;
////		deformed->transformVertex(*input, tf, i);
////	}
////
////	// Deform the grid
////	vWeightGrid->updateDeformation(deform);
////
////	// Recompute normals
////	deformed->computeNormals(true);
////	//deformed->computeFrames(deformed->frameScales, true);
////}

void DeformEM::applyDeformation_AccumulatedGeometry( const RigidDeform &deform, AccumulatedGeometry &geometry )
{
	size_type n_vertices = geometry.vertex.size();

	//// Clear ANN data
	//geometry.vertex.clearANNData();
	//geometry.normal.clearANNData();

	// Deform the output mesh
	for (size_type i = 0; i < n_vertices; i++) {
		idxList bones;
		distList weights;
		SE3 tf;
		tf.setZero();

		// Get the vertex weights
		vWeightGrid->getVertexWeights(accumulatedGeometry.vertex[i], bones, weights);

		// Transform and assign the output vertex
		for (size_type bi = 0; bi < bones.size(); bi++) {
			index_type j = bones[bi];
			tf.rot += weights[bi] * deform[j].rot;
			tf.tr += weights[bi] * deform[j].tr;
		}

		geometry.transformVertex(accumulatedGeometry, tf, i);
	}

	// Deform the grid
	vWeightGrid->updateDeformation(deform);

	// Recompute normals
	//deformed->computeNormals(true);
}


void DeformEM::applyDeformation_FrameGeometry( const RigidDeform &deform, AccumulatedGeometry &geometry, const ShapeModel *source )
{
	size_type n_vertices = source->numVertices();

	// Check that the vertices are the same number
	if (n_vertices != geometry.vertex.size()) {
		cerr << "Error in applyDeformation_FrameGeometry(): number of vertices doesn't match!" << endl;
		return;
	}

	//cerr << "Deform has " << deform.size() << " components" << endl;

	//// Clear ANN data
	//geometry.vertex.clearANNData();
	//geometry.normal.clearANNData();

	// Deform the output mesh
	for (size_type i = 0; i < n_vertices; i++) {
		// Don't map if the point was marked "invalid" as in updateGrid()
		if (source->markVertex[i] == VERTEX_INVALID) {
			continue;
		}

		idxList bones;
		distList weights;
		SE3 tf;
		tf.setZero();

		// Get the vertex weights
		vWeightGrid->getVertexWeights(source->vertex[i], bones, weights);

		// Transform and assign the output vertex
		for (size_type bi = 0; bi < bones.size(); bi++) {
			index_type j = bones[bi];
			tf.rot += weights[bi] * deform[j].rot;
			tf.tr += weights[bi] * deform[j].tr;
		}

		// Apply the rigid transformation to the vertex and frames
		geometry.vertex[i] = tf * source->vertex[i];

		// Compute inverse transpose to transform normal
		Geometry::Matrix33f tf_invT = tf.rot.transpose();
		invert(tf_invT);
		geometry.normal[i] = tf_invT * source->normal[i];
		geometry.normal[i].normalize();
	}

	// Deform the grid
	vWeightGrid->updateDeformation(deform);

	// Recompute normals
	//deformed->computeNormals(true);
}


void DeformEM::concatenateDeformation( const RigidDeform &add, RigidDeform &cat )
{
	// Check if the sizes match
	if (add.size() != cat.size()) {
		return;
	}

	for (index_type i = 0; i < add.size(); i++) {
		// Concatenate the deformations
		cat[i].rot = add[i].rot * cat[i].rot;
		cat[i].tr = (add[i].rot * cat[i].tr) + add[i].tr;
	}
}

bool DeformEM::checkConstraints( CorrespondenceList *checkCL ) const
{
	bool result = true;

	if (!checkCL) {
		cerr << "Error in DeformEM::checkConstraints(): checkCL is NULL!" << endl;
		return false;
	}

	// Check that each constraint is within the grid
	for (index_type i = 0; i < checkCL->size(); i++) {
		// Ignore points with zero weight
		if (!checkCL->valid(i)) {
			continue;
		}

		// Check the bounds
		if (!vWeightGrid->withinGrid(checkCL->sourcePos(i))) {
			cerr << "Constraint #" << i << " " << checkCL->sourcePos(i) << " is not within the grid!" << endl;
			result = false;
		}

		// Check the cell
		int3_type cellIndex;
		real3_type remainder;
		real3_type queryPos = checkCL->sourcePos(i);
		vWeightGrid->queryCell(queryPos, cellIndex, remainder);

		if (!vWeightGrid->validCell(cellIndex)) {
			cerr << "Constraint #" << i << " accesses an invalid grid cell " << cellIndex << "!" << endl;
			result = false;
		}
		else {
			// Check each point of the cell
			for (int i = 0; i < (1 << dim); i++) {
				int3_type offset;
				offset[2] = ((int) i) & 1;
				offset[1] = (((int) i) >> 1) & 1;
				offset[0] = (((int) i) >> 2) & 1;

				if (!vWeightGrid->validPoint(cellIndex + offset)) {
					cerr << "Constraint #" << i << " accesses an invalid grid point " << cellIndex + offset << "!" << endl;
					result = false;
				}
			}
		}
	}

	return result;
}

bool DeformEM::checkWeights() const
{
	bool result = true;

	// Check that we have enough weights
	if (solveOptions.EStepWeights.size() < 2) {
		cerr << "Not enough E-step weights (only have " << solveOptions.EStepWeights.size() << ") for DeformEM." << endl;
		result = false;
	}
	if (solveOptions.DiscreteMStepWeights.size() < 2) {
		cerr << "Not enough Discrete M-step weights (only have " << solveOptions.DiscreteMStepWeights.size() << ")for DeformEM." << endl;
		result = false;
	}
	if (solveOptions.MStepWeights.size() < 4) {
		cerr << "Not enough M-step weights (only have " << solveOptions.MStepWeights.size() << ")for DeformEM." << endl;
		result = false;
	}

	return result;
}

void DeformEM::printTransforms( int itr, const RigidDeform &def ) const
{
	// Print the transformation to a text file
	QString outFile = QString("textOutput/out_EM-%1.txt").arg(itr, 3, 10, QLatin1Char('0'));
	QFile file(outFile);
	if (file.open(QIODevice::WriteOnly | QIODevice::Text)) {
		QTextStream out(&file);

		for (size_type i = 0; i < def.size(); i++) {
			// Print
			outputToTextStream(out, i, def[i].rot, def[i].tr);
		}
	}
	file.close();
}

void DeformEM::printWeights( int itr ) const
{
	// Print the vertex weights to a text file
	QString outFile = QString("textOutput/out_EM_Weights-%1.txt").arg(itr, 3, 10, QLatin1Char('0'));
	QFile file(outFile);
	if (file.open(QIODevice::WriteOnly | QIODevice::Text)) {
		QTextStream out(&file);

		data_type temp(vWeightGrid->getDataDim());
		int3_type gridIndex = vWeightGrid->scanStart();
		do {
			for (int j = 0; j < vWeightGrid->getDataDim(); ++j) {
				temp.val[j] = vWeightGrid->getValue(gridIndex, j);
			}
			out << "grid(" << gridIndex[0]+1 << "," << gridIndex[1]+1 << "," << gridIndex[2]+1 << ") = ";
			out << temp << ";\n";
		} while(vWeightGrid->scanNext(gridIndex));
	}
	file.close();
}

void DeformEM::setInitialDeformation( const RigidDeform &def )
{
	//initialDeform = def;
	cout << "*** Error in DeformEM::setInitialDeformation(): using deprecated functionality!!!" << endl;
}

void DeformEM::setStepType( int type )
{
	stepType = type;
}

void DeformEM::expandBoneInfluence()
{
	// Expand the influence region of each bone by one grid cell

	for (int i = 0; i < cellsPerBone.size(); i++) {
		std::vector<int> &affCells = cellsPerBone[i];
		std::vector<int> temp = affCells;

		for (int j = 0; j < affCells.size(); j++) {
			// Look up the index in the hash table
			int3_type cellIndex = vWeightGrid->translateCellLinearIndex(affCells[j]);

			// Get the neighbors of this cell
			int3List neighbors = vWeightGrid->findAdjacentCellsToCell(cellIndex);

			// Find linear indices
			std::vector<int> linearIndexNeighbors;
			linearIndexNeighbors.resize(neighbors.size());
			for (int k = 0; k < neighbors.size(); k++) {
				linearIndexNeighbors[k] = vWeightGrid->getCellLinearIndex(neighbors[k]);
			}

			// Sort
			sort(linearIndexNeighbors.begin(), linearIndexNeighbors.end());

			// Union with the current list of cells
			sortedListUnionMerge(linearIndexNeighbors, temp);
		}

		// Assign the new list of cells
		affCells = temp;
	}
}

void DeformEM::setBoneInfluence_EntireGrid()
{
	int i;
	bonesPerCell.resize(vWeightGrid->totalGridCells());
	bonesPerPoint.resize(vWeightGrid->totalGridPoints());
	cellsPerBone.resize(vWeightGrid->getDataDim());
	pointsPerBone.resize(vWeightGrid->getDataDim());

	//////////////////////////////////////////////////////////////////////////
	// Use all bones / cells / points
	vector<int> allBones;
	vector<int> allCells;
	vector<int> allPoints;

	allBones.resize(vWeightGrid->getDataDim());
	allCells.reserve(vWeightGrid->totalGridCells());
	allPoints.reserve(vWeightGrid->totalGridPoints());

	for (i = 0; i < allBones.size(); i++) {
		allBones[i] = i;
	}

	int3_type cellIndex = vWeightGrid->scanCellStart();
	do {
		allCells.push_back(vWeightGrid->getCellLinearIndex(cellIndex));
	} while(vWeightGrid->scanNextCell(cellIndex));

	int3_type gridIndex = vWeightGrid->scanStart();
	do {
		allPoints.push_back(vWeightGrid->getLinearIndex(gridIndex));
	} while(vWeightGrid->scanNext(gridIndex));

	//////////////////////////////////////////////////////////////////////////
	// Populate the influence sets
	for (i = 0; i < bonesPerCell.size(); i++) {
		bonesPerCell[i] = allBones;
	}

	for (i = 0; i < bonesPerPoint.size(); i++) {
		bonesPerPoint[i] = allBones;
	}

	for (i = 0; i < cellsPerBone.size(); i++) {
		cellsPerBone[i] = allCells;
	}

	for (i = 0; i < pointsPerBone.size(); i++) {
		pointsPerBone[i] = allPoints;
	}

	//////////////////////////////////////////////////////////////////////////
	// Find mapping of each bone's selected cells/points into the arrays
	boneCellLinearIndex.resize(vWeightGrid->getDataDim());
	bonePointLinearIndex.resize(vWeightGrid->getDataDim());
	for (int i = 0; i < vWeightGrid->getDataDim(); i++) {
		boneCellLinearIndex[i].clear();
		bonePointLinearIndex[i].clear();

		// Insert the linear indices into the hash map
		for (int j = 0; j < cellsPerBone[i].size(); j++) {
			boneCellLinearIndex[i][cellsPerBone[i][j]] = j;
		}

		for (int j = 0; j < pointsPerBone[i].size(); j++) {
			bonePointLinearIndex[i][pointsPerBone[i][j]] = j;
		}
	}
}


void DeformEM::setBoneInfluence_EntireGrid_RemoveExcludedBones()
{
	int i;
	bonesPerCell.resize(vWeightGrid->totalGridCells());
	bonesPerPoint.resize(vWeightGrid->totalGridPoints());
	cellsPerBone.resize(vWeightGrid->getDataDim());
	pointsPerBone.resize(vWeightGrid->getDataDim());

	//////////////////////////////////////////////////////////////////////////
	// Analyzed excluded bone list
	vector<bool> boneUsed(vWeightGrid->getDataDim(), true);
	for (i = 0; i < excludedLabels.size(); ++i) {
		boneUsed[excludedLabels[i]] = false;
	}

	//////////////////////////////////////////////////////////////////////////
	// Use unexcluded bones / all cells / all points
	vector<int> usedBones;
	vector<int> allCells;
	vector<int> allPoints;

	usedBones.clear();
	allCells.reserve(vWeightGrid->totalGridCells());
	allPoints.reserve(vWeightGrid->totalGridPoints());

	for (i = 0; i < vWeightGrid->getDataDim(); i++) {
		if (boneUsed[i]) {
			usedBones.push_back(i);
		}
	}

	int3_type cellIndex = vWeightGrid->scanCellStart();
	do {
		allCells.push_back(vWeightGrid->getCellLinearIndex(cellIndex));
	} while(vWeightGrid->scanNextCell(cellIndex));

	int3_type gridIndex = vWeightGrid->scanStart();
	do {
		allPoints.push_back(vWeightGrid->getLinearIndex(gridIndex));
	} while(vWeightGrid->scanNext(gridIndex));

	//////////////////////////////////////////////////////////////////////////
	// Populate the influence sets
	for (i = 0; i < bonesPerCell.size(); i++) {
		bonesPerCell[i] = usedBones;
	}

	for (i = 0; i < bonesPerPoint.size(); i++) {
		bonesPerPoint[i] = usedBones;
	}

	for (i = 0; i < vWeightGrid->getDataDim(); i++) {
		if (boneUsed[i]) {
			cellsPerBone[i] = allCells;
			pointsPerBone[i] = allPoints;
		}
		else {
			cellsPerBone[i].clear();
			pointsPerBone[i].clear();
		}
	}

	//////////////////////////////////////////////////////////////////////////
	// Find mapping of each bone's selected cells/points into the arrays
	boneCellLinearIndex.resize(vWeightGrid->getDataDim());
	bonePointLinearIndex.resize(vWeightGrid->getDataDim());
	for (int i = 0; i < vWeightGrid->getDataDim(); i++) {
		boneCellLinearIndex[i].clear();
		bonePointLinearIndex[i].clear();

		// Insert the linear indices into the hash map
		for (int j = 0; j < cellsPerBone[i].size(); j++) {
			boneCellLinearIndex[i][cellsPerBone[i][j]] = j;
		}

		for (int j = 0; j < pointsPerBone[i].size(); j++) {
			bonePointLinearIndex[i][pointsPerBone[i][j]] = j;
		}
	}
}


ShapeModel* DeformEM::inverseDeformation( const ShapeModel &shape )
{
	// Apply inverse of the solved deformation

	// 1. Make a copy of the shape
	ShapeModel* newShape = shape.makeCopy();

	// 2. Modify the shape
	vWeightGrid->inverseDeform(*newShape, solveOptions.inverse_deformation_maxSearchLevel, solveOptions.inverse_deformation_gridExpandStep);

	// 3. Return the shape.
	return newShape;
}

void DeformEM::accumulateFrame()
{
	if (!solveOptions.useAccumulatedGeometry) {
		cerr << "Warning in DeformEM::accumulateFrame(): accumulated geometry option turned off!" << endl;
		return;
	}

	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Accumulating new frame to reference pose..." << endl;

	//////////////////////////////////////////////////////////////////////////
	// Accumulate the geometry
	// Perform inverse deformation
	ShapeModel* mappedShape = inverseDeformation(*currentTarget());

	// Add points
	accumulatedGeometry.accumulateShape(*mappedShape, scene->pointSize() * 0.1, scene->pointSize() * 0.7);

	// Delete the inversely mapped shape
	delete mappedShape;
	mappedShape = NULL;


	//////////////////////////////////////////////////////////////////////////
	// Re-pick the source vertices for correspondences using best candidate sampling
	// Create a new set of correspondences

	idxList pickList = bestCandidateSample(accumulatedGeometry.vertex, currentCL()->size());
	CorrespondenceList *newCL = new CorrespondenceList();
	newCL->resize(currentCL()->size());

	for (index_type i = 0; i < newCL->size(); i++) {
		newCL->sourcePoints[i] = accumulatedGeometry.vertex[pickList[i]];
		newCL->sourceIndex[i] = pickList[i];
		newCL->targetPoints[i] = real3_type(0.0, 0.0, 0.0);
		newCL->targetIndex[i] = 0;
		newCL->weight[i] = 0.0;
	}

	//// Clear ANN data
	//currentCL()->sourcePoints.clearANNData();
	//currentCL()->targetPoints.clearANNData();
	//newCL->sourcePoints.clearANNData();
	//newCL->targetPoints.clearANNData();


	//////////////////////////////////////////////////////////////////////////
	// Create new grid
	std::vector<PointList*> accPoints;
	accPoints.push_back(&(accumulatedGeometry.vertex));
	std::vector< std::vector<bool> > accPointsValid(1);
	accPointsValid[0].resize(accumulatedGeometry.vertex.size(), true);

	MarkedUniformGrid3 *newGrid = constructExpandedVertexWeightGrid(accumulatedGeometry.box_min, accumulatedGeometry.box_max, accPoints, accPointsValid);

	//////////////////////////////////////////////////////////////////////////
	// Finally, quickly switch between the old grid and the new grid
	// To avoid race conditions, put this under a mutex
	//////////////////////////////////////////////////////////////////////////
	changingFrames.lock();
	//////////////////////////////////////////////////////////////////////////

	MarkedUniformGrid3 *oldGrid = vWeightGrid;
	vWeightGrid = newGrid;

	// Set the bone-influence sets so that each bone affects the entire grid
	setBoneInfluence_EntireGrid();

	// Delete the old grid
	delete oldGrid;

	// Push on a new solution variable (target model set later in requestNextFrame())
	solutionPerFrame.push_back(FrameSolution());
	FrameSolution &lastSolution = solutionPerFrame[solutionPerFrame.size()-1];
	lastSolution.solvedDeformation = currentDeform();
	lastSolution.cList = newCL;
	lastSolution.deformedAccumulatedGeometry = accumulatedGeometry;

	// Update the accumulated geometry in the new frame
	applyDeformation_AccumulatedGeometry(currentDeform(), solutionPerFrame[currentSolveFrame+1].deformedAccumulatedGeometry);

	//////////////////////////////////////////////////////////////////////////
	changingFrames.unlock();
	//////////////////////////////////////////////////////////////////////////

	// Reset grid graph structure
	cellAdjacency.clear();
}




MarkedUniformGrid3* DeformEM::constructExpandedVertexWeightGrid( real3_type newMinPos, real3_type newMaxPos, const std::vector<PointList*> &accumulationPoints, const std::vector< std::vector<bool> > &accPointsValid )
{
	// Construct a new grid to include the new points
	cout << "*** Constructing expanded vertex weight grid..." << endl;
	perfTimers["Allocating Grid"].Start();

	// Make the bounding box be exactly on locations on the existing grid
	// --> this means setting the minimum position exactly like the old grid

	// First compare the minimum positions and use the smaller
	real3_type oldGridMinPos = vWeightGrid->gridLocation(int3_type(0,0,0));
	real3_type oldGridMaxPos = vWeightGrid->gridLocation(vWeightGrid->size()-int3_type(1,1,1));
	for (int i = 0; i < 3; ++i) {
		if (oldGridMinPos[i] < newMinPos[i]) {
			newMinPos[i] = oldGridMinPos[i];
		}
		if (oldGridMaxPos[i] > newMaxPos[i]) {
			newMaxPos[i] = oldGridMaxPos[i];
		}
	}

	// Now find the existing grid cell that is contains the min pos (guaranteed to be at least smaller location since it uses floor)
	int3_type newMinIndexOffset;
	real3_type temp_remainder;
	vWeightGrid->queryCell(newMinPos, newMinIndexOffset, temp_remainder);
	newMinPos = vWeightGrid->gridLocation(newMinIndexOffset);

	cout << "****** Offset in expanded grid: " << newMinIndexOffset << endl;

	MarkedUniformGrid3 *newGrid = new MarkedUniformGrid3(vWeightGrid->getDataDim());

	// Create the new grid: it will start on an existing grid point, and cover the bounding box of the accumulated geometry
	newGrid->createFromBBox(newMinPos, newMaxPos, vWeightGrid->getCellSize());

	// Initialization
	newGrid->resetMark();
	newGrid->markCells(accumulationPoints, accPointsValid);		/**< Mark cells with all points */
	newGrid->resetData();
	newGrid->findEdges();										/**< Find the edges in the marked grid */
	newGrid->copyColors(*vWeightGrid);							/**< Copy the colors from the old grid. */

	perfTimers["Allocating Grid"].Stop();
	printTime("Allocating Grid");

	//////////////////////////////////////////////////////////////////////////
	// Derive the vertex weights from the existing grid
	// --> We can just use the newMinIndexOffset to query into the right locations!!!

	int3_type gridIndex = newGrid->scanStart();
	do {
		int3_type index_in_old_grid = gridIndex + newMinIndexOffset;
		index_type linearIndex = newGrid->getLinearIndex(gridIndex);

		// Get values, possibly beyond the marked grid.
		newGrid->setAllValues(linearIndex, vWeightGrid->getValues_beyondGrid(index_in_old_grid));

	} while(newGrid->scanNext(gridIndex));


	//// Clean label values!
	//data_type temp(newGrid->getDataDim());
	//int3_type cellIndex = newGrid->scanCellStart();
	//do {
	//	int label = getPointLabel(newGrid, cellIndex);

	//	// Reset the label at all grid points belonging to this cell
	//	idxList idx;
	//	temp.setZero();
	//	temp.val[label] = 1.0;
	//	newGrid->getCellGridPoints(cellIndex, idx);
	//	for (int j = 0; j < idx.size(); j++) {
	//		newGrid->setAllValues(idx[j], temp);
	//	}
	//} while(newGrid->scanNextCell(cellIndex));

	return newGrid;
}



void DeformEM::updateGrid( ShapeModel *newFrame )
{
	cout << "-------------------------------------------------------------------------------" << endl;
	cout << "Updating grid..." << endl;

	const int numSolns = solutionPerFrame.size();
	FrameSolution &updateSoln = solutionPerFrame[numSolns-1];			/**< The very last solution */
	std::vector<PointList*> accPoints(numSolns+1);
	std::vector< std::vector<bool> > accPointsValid(numSolns+1);
	real3_type box_max, box_min;
	PointList refPoseVertices(input->vertex);

	// Add the input frame (reference pose)
	//refPoseVertices.clearANNData();
	accPoints[0] = &refPoseVertices;
	accPointsValid[0].resize(input->vertex.size(), true);
	box_max = input->box_max;
	box_min = input->box_min;

	// Next, add all the inversely mapped frames
	for (int frame = 0, frame_max = solutionPerFrame.size(); frame < frame_max; ++frame)
	{
		FrameSolution &frameSoln = solutionPerFrame[frame];
		RigidDeform &frameDeform = frameSoln.solvedDeformation;
		const ShapeModel *frameTarget = frameSoln.targetModel;

		// 1. Perform inverse deformation of all frames
		vWeightGrid->updateDeformation(frameDeform);
		ShapeModel* mappedShape = inverseDeformation(*frameTarget);
		mappedShape->computeNormals(true);

		// Update the current frame's inversely mapped shape
		ShapeModel* oldMappedShape = frameSoln.inverseMappedShape;
		frameSoln.inverseMappedShape = mappedShape;

		// Update pointers and free old memory
		if (oldMappedShape) {
			// The next frame solution's source model points to the shape that is about to be deleted!
			// --> Update next frame's source model, if exists.
			if ((frame+1) < solutionPerFrame.size()) {
				// Run time check to see if it's the same pointer
				if (solutionPerFrame[frame+1].sourceModel == oldMappedShape) {
					// Set the pointer to the new shape
					solutionPerFrame[frame+1].sourceModel = mappedShape;
				}
				else {
					// This is an error! The next frame's source model should agree with the current inverse mapped shape.
					cerr << "Error in DeformEM::updateGrid(): next frame's source model doesn't match with the current inversely mapped shape!" << endl;
					solutionPerFrame[frame+1].sourceModel = NULL;
				}
			}
			delete oldMappedShape;
		}



		// Use the points for marking the grid
		accPoints[frame+1] = &(mappedShape->vertex);
		accPointsValid[frame+1].resize(mappedShape->vertex.size());

		// Update bounding box information
		for (int i = 0; i < mappedShape->vertex.size(); ++i) {
			const real3_type pos = mappedShape->vertex[i];
			accPointsValid[frame+1][i] = (mappedShape->markVertex[i] != VERTEX_INVALID);

			// Update bounding box
			if (accPointsValid[frame+1][i]) {
				for (int j = 0; j < 3; j++) {
					if (pos[j] < box_min[j]) {
						box_min[j] = pos[j];
					}
					if (pos[j] > box_max[j]) {
						box_max[j] = pos[j];
					}
				}
			}
		}
	}


	//////////////////////////////////////////////////////////////////////////
	// 2. Update the grid
	MarkedUniformGrid3 *newGrid = constructExpandedVertexWeightGrid(box_min, box_max, accPoints, accPointsValid);


	//// 3. Clear ANN data for correspondences
	//updateSoln.cList->sourcePoints.clearANNData();
	//updateSoln.cList->targetPoints.clearANNData();


	cout << " done!" << endl;

	//////////////////////////////////////////////////////////////////////////
	// Finally, quickly switch between the old grid and the new grid
	// To avoid race conditions, put this under a mutex
	//////////////////////////////////////////////////////////////////////////
	changingFrames.lock();
	//////////////////////////////////////////////////////////////////////////

	MarkedUniformGrid3 *oldGrid = vWeightGrid;
	vWeightGrid = newGrid;

	// Delete the old grid
	delete oldGrid;

	/**< Temporarily copy some variables (for some reason the push_back causes an error) */
	RigidDeform updateDeform(updateSoln.solvedDeformation);
	int numCorrespondences = updateSoln.cList->size();
	const ShapeModel *updateTarget = updateSoln.targetModel;

	// Push on a new solution variable (target model set later in requestNextFrame())
	solutionPerFrame.push_back(FrameSolution());
	FrameSolution &lastSolution = solutionPerFrame[solutionPerFrame.size()-1];
	lastSolution.setData(updateDeform, updateSoln.inverseMappedShape, newFrame, updateTarget);
	lastSolution.generateAdjacentCorrespondences(updateTarget, numCorrespondences);
	applyDeformation_FrameGeometry(updateDeform, lastSolution.deformedAccumulatedGeometry, lastSolution.sourceModel);

	//////////////////////////////////////////////////////////////////////////
	changingFrames.unlock();
	//////////////////////////////////////////////////////////////////////////

	// Set the bone-influence sets so that each bone affects the entire grid
	setBoneInfluence_EntireGrid();

	// Reset grid graph structure
	cellAdjacency.clear();
}



/**
* Update the cList in updateSoln from aList by inverse mapping the source points
* in aList using the current grid and its current deformation.
*/
void DeformEM::updateCorrespondences_fromAdjacent( FrameSolution &updateSoln )
{
	if (!updateSoln.aList) {
		// No adjacency correspondences to update the reference correspondences
		//cerr << "Error in DeformEM::updateCorrespondences_fromAdjacent(): no adjacency correspondences to update the reference correspondences!" << endl;
		return;
	}

	CorrespondenceList *updateCL = NULL;

	if (updateSoln.cList) {
		updateCL = updateSoln.cList;
	}
	else {
		updateCL = updateSoln.cList = new CorrespondenceList();
	}

	// Update the source positions of the correspondence list
	// --> by inverse mapping from the current grid deformation
	CorrespondenceList *updateAL = updateSoln.aList;
	updateCL->resize(updateSoln.aList->size());
	updateSoln.cList_ok.resize(updateSoln.aList->size());
	updateSoln.correspondencesPerCell.clear();

	for (int i = 0; i < updateAL->size(); ++i) {
		// 1. Inverse map the source point
		real3_type weights;
		int3_type cellFoundIndex;
		// Find index without expanding the grid
		bool ok = vWeightGrid->inverseDeformPoint(updateAL->sourcePos(i), cellFoundIndex, weights, 0, 0);

		// Debugging: check that the cellFoundIndex is actually a marked cell in the vWeightGrid!
		bool cellOk = vWeightGrid->validCell(cellFoundIndex);

		// 2. Assign the source point
		if (ok && cellOk) {
			updateAL->weight[i] = 1.0;
			updateSoln.cList_ok[i] = true;
			updateCL->sourcePoints[i] = vWeightGrid->trilinearInterpPosition(weights, cellFoundIndex);
			updateCL->sourceIndex[i] = updateAL->sourceIndex[i];
			updateCL->targetPoints[i] = real3_type(0.0, 0.0, 0.0);
			updateCL->targetIndex[i] = 0;
			updateCL->weight[i] = 0.0;

			// Also update the list of correspondences in each cell
			updateSoln.correspondencesPerCell[cellFoundIndex].push_back(i);
		}
		else {
			if (ok && !cellOk) {
				// TODO:  Why does this happen?  Investigate.
				cerr << "Warning in DeformEM::updateCorrespondences_fromAdjacent(): correspondence " << i << " doesn't inverse-map to a valid cell!" << endl;
			}

			// Make the point something degenerate and don't use it in updating correspondences
			updateAL->weight[i] = 0.0;
			updateSoln.cList_ok[i] = false;
			updateCL->sourcePoints[i] = real3_type(0.0, 0.0, 0.0);
			updateCL->sourceIndex[i] = 0;
			updateCL->targetPoints[i] = real3_type(0.0, 0.0, 0.0);
			updateCL->targetIndex[i] = 0;
			updateCL->weight[i] = 0.0;
		}
	}
}




void DeformEM::updateCorrespondences( int type, FrameSolution &updateSoln )
{
	const ShapeModel * const updateTarget = updateSoln.targetModel;
	CorrespondenceList *updateCL = updateSoln.cList;
	const AccumulatedGeometry &deformedAccumulatedGeometry = updateSoln.deformedAccumulatedGeometry;
	const std::vector<bool> &updateCL_ok = updateSoln.cList_ok;

	if (!updateTarget) {
		cerr << "Error in DeformEM::updateCorrespondences(): target model not defined!" << endl;
		return;
	}
	if (!updateCL) {
		cerr << "Error in DeformEM::updateCorrespondences(): no correspondences!" << endl;
		return;
	}

	// Setup variables
	const SpinImageStack* inputSpinImages = NULL;
	inputSpinImages = &(deformedAccumulatedGeometry.spinImageStack);

	//// Clear any ANN data
	//updateCL->sourcePoints.clearANNData();
	//updateCL->targetPoints.clearANNData();

	if (type == CORRESPONDENCE_GROUND_TRUTH) {
		// Assume correspondences are given by the indices.
		for (int i = 0; i < updateCL->size(); i++) {
			int index = updateCL->sourceIndex[i];

			if (updateCL_ok.size() && !updateCL_ok[i]) {
				// Don't update and don't use
				updateCL->weight[i] = 0.0;
				continue;
			}

			// Check if the vertex on the target is not on the boundary and not isolated
			if (updateTarget->markVertex[index] == VERTEX_VALID) {
				updateCL->targetPoints[i] = updateTarget->vertex[index];
				updateCL->targetIndex[i] = index;
				updateCL->weight[i] = 1.0;
			}
		}
	}
	if ((type == CORRESPONDENCE_CLOSEST_POINTS) ||
		(type == CORRESPONDENCE_CLOSEST_POINTS_SPIN_IMAGE_WEIGHTS)) {
			// Assumes that the source model indices are assigned properly in the correspondence list
			for (int i = 0; i < updateCL->size(); i++) {
				if (updateCL_ok.size() && !updateCL_ok[i]) {
					// Don't update and don't use
					updateCL->weight[i] = 0.0;
					continue;
				}

				// Find the vertex of the deformed mesh
				real3_type pos, n;
				pos = deformedAccumulatedGeometry.vertex[updateCL->sourceIndex[i]];
				n = deformedAccumulatedGeometry.normal[updateCL->sourceIndex[i]];

				// Find the closest point / normal combination on the target
				// This is following the Andrew Johnson's thesis: point + normal
				// where normal component is weighted 2 * grid spacing
				ClosestPointInfo cpi = updateTarget->findClosestPoint(pos, n);

				// First, just assign the point
				updateCL->targetPoints[i] = updateTarget->vertex[cpi.index];
				updateCL->targetIndex[i] = cpi.index;
				updateCL->weight[i] = 1.0;


				//////////////////////////////////////////////////////////////////////////
				// Outlier rejection code
				real3_type diff = updateCL->targetPoints[i] - pos;
				real_type cdist = diff.norm();

				// Test if boundary point is within a small distance
				if (updateTarget->markVertex[cpi.index] != VERTEX_VALID) {
					if (cdist > boundaryCorrespondenceDistance) {
						updateCL->weight[i] = 0.0;
					}
				}

				// Distance threshold
				if (cdist > maxCorrespondenceDistance) {
					updateCL->weight[i] = 0.0;
				}

				// Normal threshold: assumes that normals are already normalized
				real_type cosAngle = n * (updateTarget->normal[cpi.index]);
				if (cosAngle < minNormalCosAngle) {
					updateCL->weight[i] = 0.0;
				}


				//////////////////////////////////////////////////////////////////////////
				// Additional correspondence processing: special per-correspondence weight using feature matching
				if ((type == CORRESPONDENCE_CLOSEST_POINTS_SPIN_IMAGE_WEIGHTS) && updateCL->valid(i)) {
					// Compute the feature similarity using spin images
					if (inputSpinImages->size() && updateTarget->spinImageStack.size()) {
						const SpinImage &srimage = (*inputSpinImages)[updateCL->sourceIndex[i]];
						const SpinImage &tgtImage = updateTarget->spinImageStack[updateCL->targetIndex[i]];
						real_type score = ComputeSimilarity(srimage, tgtImage, updateTarget->spinImageStack.lambda);
						updateCL->weight[i] = std::max(score, 0.f);
						//cout << "Weight for constraint " << i << ": " << updateCL->weight[i] << endl;
					}
				}

				//////////////////////////////////////////////////////////////////////////
				// OLD STUFF
				//////////////////////////////////////////////////////////////////////////
				//// Alternative: Find the closest point on the target
				//idxList nearest;
				//distList dists;
				//updateTarget->queryKNN(pos, 1, nearest, dists, true);

				//// Check if the vertex is not on the boundary and not isolated
				//if (updateTarget->markVertex[nearest[0]] == VERTEX_VALID) {
				//	updateCL->targetPoints[i] = updateTarget->vertex[nearest[0]];
				//	updateCL->targetIndex[i] = nearest[0];
				//}
			}
	}
	if (type == CORRESPONDENCE_SPIN_IMAGES) {
		if (!inputSpinImages->size() || !updateTarget->spinImageStack.size()) {
			// No spin images to compare!
			cerr << "Error in DeformEM::updateCorrespondences(): No spin images to compare!" << endl;
			return;
		}

		// Assumes that the source model indices are assigned properly in the correspondence list
		cout << "****** Starting spin image based pairing..." << endl;
		cout << "****** Expected bin overlap: " << updateTarget->spinImageStack.lambda << endl;
		cout << "****** Progress...";

		index_type keepMax = 10;
		real_type distSq = maxCorrespondenceDistance*maxCorrespondenceDistance;
		std::vector<Candidate> candidates;
		std::vector<idxList> correspondencesPerLabel;
		correspondencesPerLabel.resize(vWeightGrid->getDataDim());
		int total_Candidates = 0, total_Outliers = 0;

		//////////////////////////////////////////////////////////////////////////
		// a. Phase 1: Find all the candidates using spin images
		cout << " candidates";
		TextProgress(updateCL->size(), true);
		for (index_type i = 0; i < updateCL->size(); ++i) {
			if (updateCL_ok.size() && !updateCL_ok[i]) {
				// Don't update and don't use
				updateCL->weight[i] = 0.0;
				continue;
			}

			// Find the vertex of the deformed mesh
			index_type sourceIndex = updateCL->sourceIndex[i];
			real3_type pos = deformedAccumulatedGeometry.vertex[sourceIndex];
			real3_type n = deformedAccumulatedGeometry.normal[sourceIndex];

			// Find the cell containing the correspondence
			int3_type cellIndex;
			real3_type remainder;
			vWeightGrid->queryCell(updateCL->sourcePos(i), cellIndex, remainder);

			// Check to see if it's a valid cell (just in case the cell is out of bounds or not marked)
			if (!vWeightGrid->validCell(cellIndex)) {
				updateCL->weight[i] = 0.0;
				continue;
			}
			
			// Find the label of the cell
			int label = getPointLabel(vWeightGrid, cellIndex);

			// Range query in target mesh
			idxList inRange;
			distList dists;
			updateTarget->queryRange(pos, distSq, inRange, dists, true);

			if (inRange.size() == 0) {
				// Remove this correspondence
				updateCL->weight[i] = 0.0;
				continue;
			}

			// Find the spin image in the original mesh and perform similarity query
			idxList similarIdx;
			const SpinImage &P = (*inputSpinImages)[sourceIndex];
			updateTarget->spinImageStack.querySubsetSimilar(P, inRange, similarIdx, dists);

			// Add the top candidates
			for (index_type j = 0, j_max = std::min(keepMax,similarIdx.size()); j < j_max; ++j) {
				correspondencesPerLabel[label].push_back(candidates.size());
				candidates.push_back(Candidate(pos, updateTarget->vertex[similarIdx[j]], similarIdx[j], dists[j], i));
			}

			TextProgress(i);
		}

		// Initially set all correspondences to weight 0
		for (index_type i = 0, i_max = updateCL->size(); i < i_max; ++i) {
			updateCL->weight[i] = 0.0;
		}

		//////////////////////////////////////////////////////////////////////////
		// b. Phase 2: RANSAC for each region's correspondences
		const int numModel = 3;
		int RANSACmaxitr = 5000;
		real_type inlierThreshold = scene->pointSize();
		std::vector<real3_type> selectedSourcePts(numModel);
		std::vector<real3_type> selectedTargetPts(numModel);
		std::vector<SE3> bestTransform(vWeightGrid->getDataDim());

		cout << " RANSAC";
		for (int region = 0; region < vWeightGrid->getDataDim(); ++region) {
			// Estimate a rigid transformation for each region
			const idxList &regionCorrespondences = correspondencesPerLabel[region];
			if (regionCorrespondences.size() <= 10) {
				continue;
			}
			cout << " region " << region;
			bool done = false;
			int itrCount = 0;
			int maxInliers = 0;
			size_type maxCandidates = regionCorrespondences.size();

			TextProgress(RANSACmaxitr, true);
			while (!done) {
				//////////////////////////////////////////////////////////////////////////
				// A. Pick some random correspondences within the region
				real3_type srcCenter(0.0,0.0,0.0);
				real3_type tgtCenter(0.0,0.0,0.0);
				for (index_type i = 0; i < numModel; ++i) {
					index_type ridx = regionCorrespondences[floor(((double)maxCandidates) * genrand_real1())];
					selectedSourcePts[i] = candidates[ridx].sourcePos;
					selectedTargetPts[i] = candidates[ridx].targetPos;
					srcCenter += selectedSourcePts[i];
					tgtCenter += selectedTargetPts[i];
				}
				srcCenter = srcCenter / ((real_type)numModel);
				tgtCenter = tgtCenter / ((real_type)numModel);

				//////////////////////////////////////////////////////////////////////////
				// B. Perform SVD registration (point-to-point)
				// TODO: implement point-to-plane here
				Geometry::Matrix33d M, S, U, VT, rotation;
				Vec3d srcCenter_d, tgtCenter_d, translation;

				// B1. Construct matrix: set to zero, accumulate value
				// Also copy source and target center vectors
				for (int r = 0; r < 3; ++r) {
					for (int c = 0; c < 3; ++c) {
						M(r,c) = 0.0;
					}
					srcCenter_d[r] = srcCenter[r];
					tgtCenter_d[r] = tgtCenter[r];
				}
				for (index_type i = 0; i < numModel; ++i) {
					// Subtract center
					real3_type srcPt = selectedSourcePts[i] - srcCenter;
					real3_type tgtPt = selectedTargetPts[i] - tgtCenter;

					// Add to matrix
					for (int r = 0; r < 3; ++r) {
						for (int c = 0; c < 3; ++c) {
							M(r,c) += srcPt[r] * tgtPt[c];
						}
					}
				}

				// B2. Compute SVD
				M.computeSVD(S, U, VT);
				rotation = (U*VT).transpose();

				// B3. Compute the translation as the right centroid minus the rotated left centroid
				translation = tgtCenter_d - (rotation * srcCenter_d);

				// B4. Assign rigid transformation
				SE3 rt;
				for (int r = 0; r < 3; r++) {
					for (int c = 0; c < 3; c++) {
						rt.rot(r,c) = rotation(r,c);
					}
					rt.tr[r] = translation[r];
				}

				//////////////////////////////////////////////////////////////////////////
				// C. Find the total number of correspondences fit using this transformation (# inliers)
				// TODO: if we use point-to-plane in step B above, then change this to find inliers using pt-to-plane
				// C1. Count the # of inliers
				int countInliers = 0;
				for (index_type i = 0; i < maxCandidates; ++i) {
					index_type ridx = regionCorrespondences[i];
					index_type tidx = candidates[ridx].targetIndex;
					const real3_type &srcPt = candidates[ridx].sourcePos;
					const real3_type &tgtPt = candidates[ridx].targetPos;
					const real3_type &tgtN = updateTarget->normal[tidx];

					// Apply transformation to source
					real3_type transf_src = rt*srcPt;

					// Find point-to-plane distance
					real_type dist = abs((transf_src - tgtPt) * tgtN);

					// Check inlier
					if (dist < inlierThreshold) {
						countInliers++;
					}
				}

				// C2. Compare this with the maximum number of inliers so far
				if (countInliers > maxInliers) {
					// update
					maxInliers = countInliers;
					bestTransform[region] = rt;
				}

				//////////////////////////////////////////////////////////////////////////
				// D. Check termination criteria

				// D1. Update number of iterations
				TextProgress(itrCount);
				itrCount++;

				// D2. Check max iteration #
				if (itrCount >= RANSACmaxitr) {
					cout << " done!" << endl;
					done = true;
				}
			}

			//////////////////////////////////////////////////////////////////////////
			// Use the best transformation to pick the correspondences
			// TODO: if we use point-to-plane in step B above, then change this to find inliers using pt-to-plane
			int inlierCount = 0;
			int outlierCount = 0;
			// Pick all inliers of the best transformation
			for (index_type i = 0; i < maxCandidates; ++i) {
				index_type ridx = regionCorrespondences[i];
				index_type tidx = candidates[ridx].targetIndex;
				const real3_type &srcPt = candidates[ridx].sourcePos;
				const real3_type &tgtPt = candidates[ridx].targetPos;
				const real3_type &tgtN = updateTarget->normal[tidx];

				// Apply transformation to source
				real3_type transf_src = bestTransform[region] * srcPt;

				// Find point-to-plane distance
				real_type dist = abs((transf_src - tgtPt) * tgtN);

				// Check inlier
				if (dist < inlierThreshold) {
					// Active original entry in correspondence list
					index_type idx = candidates[ridx].clIndex;
					updateCL->targetPoints[idx] = candidates[ridx].targetPos;
					updateCL->targetIndex[idx] = candidates[ridx].targetIndex;
					updateCL->weight[idx] = 1.0;
					++inlierCount;
				}
				else {
					++outlierCount;
				}
			}
			total_Candidates += maxCandidates;
			total_Outliers += outlierCount;
			cout << " --> " << inlierCount << " inliers, " << outlierCount << " outliers." << endl;
		}

		// Save the RANSAC stats
		// HACK: use perfTimers to store RANSAC stats.
		perfTimers["RANSAC Total Candidates"].iterations = total_Candidates;
		perfTimers["RANSAC Total Outliers"].iterations = total_Outliers;
	}


	//////////////////////////////////////////////////////////////////////////
	// Update correspondence error
	clErrorType &clError = updateSoln.correspondenceError;

	// Make sure it's the right size
	if (clError.size() != updateCL->size()) {
		clError.resize(updateCL->size());
	}

	for (int i = 0; i < updateCL->size(); i++) {
		real3_type errorValue;

		// Update the error
		if (updateCL->valid(i)) {
			real3_type sourcePos, sourceNormal, targetPos, targetNormal;
			sourcePos = deformedAccumulatedGeometry.vertex[updateCL->sourceIndex[i]];
			sourceNormal = deformedAccumulatedGeometry.normal[updateCL->sourceIndex[i]];
			targetPos = updateTarget->vertex[updateCL->targetIndex[i]];
			targetNormal = updateTarget->normal[updateCL->targetIndex[i]];

			// 1. Point-to-point
			errorValue[0] = (sourcePos - targetPos).norm();

			// 2. Point-to-plane
			errorValue[1] = abs((sourcePos - targetPos) * targetNormal);

			// 3. Normal-to-normal (scale by twice the mesh resolution)
			errorValue[2] = (2.0 * scene->pointSize()) * (sourceNormal - targetNormal).norm();
		}
		else {
			// Assign infinite error, since it's an invalid correspondence.
			// Make sure you don't use it!
			errorValue[0] = errorValue[1] = errorValue[2] = REAL_TYPE_INF;
		}

		clError[i] = errorValue;
	}
}

void DeformEM::setCorrespondenceThresholds( int iterationNumber )
{
	//////////////////////////////////////////////////////////////////////////
	// Set boundary correspondence distance
	boundaryCorrespondenceDistance = solveOptions.boundaryCorrespondenceDistance * scene->pointSize();
	cout << "****** Boundary correspondence distance: " << boundaryCorrespondenceDistance
		 << " (=" << solveOptions.boundaryCorrespondenceDistance << "x grid spacing)" << endl;


	//////////////////////////////////////////////////////////////////////////
	// Set max correspondence distance
	real_type multiplier;
	//multiplier = 3.0;
	//if (iterationNumber < 5) {
	//	multiplier = 3.0 * ((real_type)5-iterationNumber);
	//}
	//multiplier = 50.0;
	multiplier = solveOptions.maxCorrespondenceDistance;
	maxCorrespondenceDistance = multiplier * scene->pointSize();
	cout << "****** Max correspondence distance: " << maxCorrespondenceDistance << " (=" << multiplier << "x grid spacing)" << endl;


	//////////////////////////////////////////////////////////////////////////
	// Set max angle between normals
	multiplier = 1.0;
	if (iterationNumber < 4) {
		multiplier = ((real_type)4-iterationNumber);
	}
	//multiplier = 4.0;
	real_type radian = multiplier * (20.0 / 180.0 * M_PI);
	minNormalCosAngle = cos(radian);
	cout << "****** Min Normal Cos Angle: " << minNormalCosAngle << " (=max angle " << multiplier*20.0 << " degrees)" << endl;


	//////////////////////////////////////////////////////////////////////////
	// Set weight for joint constraint term
	if (solveOptions.EStepWeights[1] > 0.0) {
		double minVal = 0.05;
		double reachItr = 5.0;
		double b = exp(-log(minVal)/reachItr);
		double smooth = pow(b, -((double)iterationNumber));
		smooth = std::max(minVal, std::min(1.0, smooth));
		solveOptions.EStepWeights[1] = smooth;
		cout << "****** E-Step Smoothness: " << solveOptions.EStepWeights[1] << endl;
	}
}



ShapeModel* DeformEM::requestNextFrame()
{
	// Grab the next frame from the scene, return success/failure status
	ShapeAnimation *allFrames = scene->getAnimation();

	// First find the current frame number, and find the next
	int nextFrame = allFrames->frameNumber(currentTarget()) + 1;

	if (nextFrame < allFrames->numFrames()) {
		return (*allFrames)[nextFrame];
	}

	return NULL;
}


void DeformEM::setOptions( const DeformEMOptions &options )
{
	this->solveOptions = options;
}


real_type DeformEM::GraphCutDataFn( int p, int l, void *data )
{
	DeformEM &deform = *((DeformEM*)data);
	real_type errorValue = 0.0;

	// Idea 1: Compute data cost as
	// "How well does the transformation l satisfy the correspondences within grid cell p?"
	int numTotalFrames = deform.solutionPerFrame.size();

	for (int i = 0; i < numTotalFrames; i++) {
		int3_type cellIndex = deform.vWeightGrid->translateCellLinearIndex(p);
		const int3mapList &correspListPerCell = deform.solutionPerFrame[i].correspondencesPerCell;

		// Check to see if there are constraints for this cell
		int3mapList::const_iterator it = correspListPerCell.find(cellIndex);
		if (it == correspListPerCell.end()) {
			continue;
		}

		const CorrespondenceList *frameCL = deform.solutionPerFrame[i].cList;
		const RigidDeform &frameDeform = deform.solutionPerFrame[i].solvedDeformation;
		const ShapeModel *frameTarget = deform.solutionPerFrame[i].targetModel;

		const idxList &correspList = it->second;
		if (correspList.size()) {
			// For each correspondence in grid cell p
			for (int i = 0; i < correspList.size(); i++) {
				real3_type sourcePos = frameCL->sourcePos(correspList[i]);
				real3_type targetPos = frameCL->targetPos(correspList[i]);
				index_type targetIdx = frameCL->targetIndex[correspList[i]];

				if (frameCL->valid(correspList[i])) {
					// Apply the transformation to the source
					sourcePos = frameDeform[l] * sourcePos;

					//// Measure Euclidean distance between transformed source point and target point
					//errorValue += (sourcePos - targetPos).norm();

					// Point to plane distance (L1 norm)
					real3_type targetNormal = frameTarget->normal[targetIdx];
					errorValue += abs((sourcePos - targetPos) * targetNormal);
				}
			}
		}
		else {
			// No correspondences in this cell, thus it's unconstrained.
			//cerr << "Warning in DeformEM::GraphCutDataFn(): unconstrained cell found: " << deform.vWeightGrid->translateCellLinearIndex(p) << endl;
		}
	}

	return deform.solveOptions.DiscreteMStepWeights[0]*errorValue;
}

real_type DeformEM::GraphCutSmoothFn( int p1, int p2, int l1, int l2, void *data )
{
	DeformEM &deform = *((DeformEM*)data);
	real_type errorValue = 0.0;
	int numTotalFrames = deform.solutionPerFrame.size();

	// Constant penalty if labels are different
	if (l1 != l2) {
		//errorValue = 1.0;
		errorValue = ((real_type)numTotalFrames) * deform.scene->pointSize();
	}

	// Compute distance between cells
	real3_type p1Pos = deform.vWeightGrid->gridLocation(deform.vWeightGrid->translateCellLinearIndex(p1));
	real3_type p2Pos = deform.vWeightGrid->gridLocation(deform.vWeightGrid->translateCellLinearIndex(p2));

	for (int i = 0; i < numTotalFrames; i++) {
		const RigidDeform &frameDeform = deform.solutionPerFrame[i].solvedDeformation;
		p1Pos = frameDeform[l1] * p1Pos;
		p2Pos = frameDeform[l2] * p2Pos;
		real_type cellDist = (p1Pos - p2Pos).norm();

		// Difference between reference length and actual cell distance
		real_type refDist = deform.vWeightGrid->getCellSize();
		errorValue += abs(refDist - cellDist);
	}

	return deform.solveOptions.DiscreteMStepWeights[1]*errorValue;
}

real_type DeformEM::solveDiscreteLabelling()
{
	// Set up the problem
	real_type energy;
	const int numSites = vWeightGrid->totalGridCells();
	const int numLabels = vWeightGrid->getDataDim();
	GCoptimizationGeneralGraph *gc = new GCoptimizationGeneralGraph(numSites, numLabels);

	if (!currentCorrespondencesPerCell().size()) {
		cerr << "Error in DeformEM::solveDiscreteLabelling(): correspondences per cell not set up!" << endl;
		return -1.0;

		//int kNearest = 3;
		//cerr << "*** Setting up list of correspondences in each cell with " << kNearest << " nearest neighbors" << endl;
		//// Find cells that don't have any associated correspondences
		//for (int i = 0; i < numSites; i++) {
		//	int3_type cellIndex = vWeightGrid->translateCellLinearIndex(i);
		//	int3mapList::const_iterator it = currentCorrespondencesPerCell().find(cellIndex);

		//	if (it == currentCorrespondencesPerCell().end()) {
		//		// Gather the k-nearest correspondences to this cell
		//		real3_type cellLocation = vWeightGrid->gridLocation(cellIndex);
		//		idxList nearest;
		//		distList dists;
		//		currentCL()->sourcePoints.queryKNN(cellLocation, 10*kNearest, nearest, dists, true);

		//		int count = 0;
		//		for (int j = 0; j < nearest.size(); ++j) {
		//			// Only add correspondences that are valid
		//			if (currentCL()->valid(nearest[j])) {
		//				currentCorrespondencesPerCell()[cellIndex].push_back(nearest[j]);
		//				++count;
		//			}
		//			if (count >= kNearest) {
		//				break;
		//			}
		//		}
		//	}
		//}
	}

	if (!cellAdjacency.size()) {
		cerr << "*** Setting up cell adjacency list" << endl;

		// For the smoothness term, set of the list of cells adjacent to each cell
		int3_type gridCellIndex = vWeightGrid->scanCellStart();
		do {
			index_type linearIndex = vWeightGrid->getCellLinearIndex(gridCellIndex);

			int3List neighboringCells = vWeightGrid->findAdjacentCellsToCell(gridCellIndex);
			for (int i = 0; i < neighboringCells.size(); i++) {
				index_type linearIndexN = vWeightGrid->getCellLinearIndex(neighboringCells[i]);

				// Sort the pair
				int i1, i2;
				if (linearIndex <= linearIndexN) {
					i1 = linearIndex;
					i2 = linearIndexN;
				}
				else {
					i1 = linearIndexN;
					i2 = linearIndex;
				}

				std::vector<int> temp;
				temp.push_back(i2);
				sortedListUnionMerge(temp, cellAdjacency[i1]);
			}
		} while (vWeightGrid->scanNextCell(gridCellIndex));
	}

	try {
		// Use function pointers for the data and smoothness costs
		gc->setDataCost(&GraphCutDataFn, this);
		gc->setSmoothCost(&GraphCutSmoothFn, this);

		// Set up edges
		stdext::hash_map<int,std::vector<int> >::const_iterator it;
		for (it = cellAdjacency.begin(); it != cellAdjacency.end(); it++) {
			int i1 = it->first;
			const std::vector<int> &neighbors = it->second;
			for (int j = 0; j < neighbors.size(); j++) {
				int i2 = neighbors[j];
				gc->setNeighbors(i1,i2);
			}
		}

		// Initialize the labelling with the current vertex labels
		data_type temp(vWeightGrid->getDataDim());
		for (index_type i = 0; i < numSites; i++) {
			// Find the label of cell i
			for (int j = 0; j < vWeightGrid->getDataDim(); ++j) {
				temp.val[j] = vWeightGrid->getValue(vWeightGrid->translateCellLinearIndex(i), j);
			}
			int maxIndex;
			temp.findMax(maxIndex);
			gc->setLabel(i, maxIndex);
		}

		// Perform optimization
		//energy = gc->compute_energy();
		//cout << "*** Before optimization energy is: " << energy << endl;
		//cout << "*** Solving graph cut using swap..." << endl;
		//energy = gc->swap();
		energy = gc->expansion();
		//cout << "*** After expansion, energy is: " << energy << endl;

		// Apply the result
		updateDiscreteWeights(gc);
	}
	catch (GCException e) {
		e.Report();
	}

	// Free memory associated with graph cut data structure
	delete gc;

	// Return final energy
	return energy;
}

void DeformEM::updateDiscreteWeights( GCoptimizationGeneralGraph *gc )
{
	data_type temp(vWeightGrid->getDataDim());

	// debugging
	// keep track of which points got labeled
	std::vector<int> wasLabeled;
	wasLabeled.resize(vWeightGrid->totalGridPoints(), 0);

	// Read back the labels as binary vertex weights
	for (index_type i = 0; i < vWeightGrid->totalGridCells(); i++) {
		index_type label = gc->whatLabel(i);
		temp.setZero();
		temp.val[label] = 1.0;

		// Find grid points belonging to the label
		idxList idx;
		vWeightGrid->getCellGridPoints(vWeightGrid->translateCellLinearIndex(i), idx);
		for (int j = 0; j < idx.size(); j++) {
			vWeightGrid->setAllValues(idx[j], temp);			/**< We could mix, but just do this for now */
			++wasLabeled[idx[j]];
		}
	}

	// Debugging
	for (int i = 0; i < wasLabeled.size(); i++) {
		if (wasLabeled[i] == 0) {
			cerr << "Point " << i << " was never labeled!" << endl;
		}
	}

	//// Conventional way to read back the weights
	//std::vector<index_type> graphCutLabels;
	//graphCutLabels.resize(numSites);
	//for ( index_type i = 0; i < numSites; i++ ) {
	//	graphCutLabels[i] = gc->whatLabel(i);
	//}
}

void DeformEM::analyzeLabellingResult()
{
	// Count the number of cells each label was used
	std::vector<int> numCellsAffected(vWeightGrid->getDataDim(), 0);
	int totalNumCells = vWeightGrid->totalGridCells();

	for (int i = 0; i < totalNumCells; ++i) {
		int label = getPointLabel(vWeightGrid, vWeightGrid->translateCellLinearIndex(i));

		// Increment the label
		++numCellsAffected[label];
	}

	cout << "*** Number of cells affected per label: " << numCellsAffected << endl;


	//////////////////////////////////////////////////////////////////////////
	// Remove labels that cover less than some threshold percent of the entire grid
	//int excludeThreshold = max(0.0, 0.05 * ((double)totalNumCells));
	int excludeThreshold = 5;

	// Now, exclude some labels from being optimized
	std::vector<bool> toExclude(vWeightGrid->getDataDim(), false);
	excludedLabels.clear();
	for (int i = 0; i < vWeightGrid->getDataDim(); ++i) {
		if (numCellsAffected[i] <= excludeThreshold) {
			excludedLabels.push_back(i);
			toExclude[i] = true;
		}
	}

	cout << "*** Excluding labels: " << excludedLabels << endl;


	// Reset the transformations associated with the excluded labels
	for (int i = 0; i < excludedLabels.size(); ++i) {
		int label = excludedLabels[i];
		for (int j = 0; j < solutionPerFrame.size(); ++j) {
			solutionPerFrame[j].solvedDeformation[label].rot = Geometry::Matrix33f::identity();
			solutionPerFrame[j].solvedDeformation[label].tr = real3_type(0.0,0.0,0.0);
		}
	}


	//////////////////////////////////////////////////////////////////////////
	// Make sure that no excluded label is being used: if it's being used, replace with a closest non-excluded label

	// a. Build the search structure
	PointList nonExcludedPoints;
	std::vector<int> nonExcludedPointLabels;
	int3_type gridIndex = vWeightGrid->scanStart();
	do {
		int label = getPointLabel(vWeightGrid, gridIndex);

		if (!toExclude[label]) {
			nonExcludedPoints.add(vWeightGrid->gridLocation(gridIndex));
			nonExcludedPointLabels.push_back(label);
		}
	} while (vWeightGrid->scanNext(gridIndex));


	// b. Remove excluded labels
	for (index_type i = 0; i < vWeightGrid->totalGridPoints(); ++i) {
		gridIndex = vWeightGrid->translateLinearIndex(i);
		int label = getPointLabel(vWeightGrid, gridIndex);

		if (toExclude[label]) {
			idxList nearest;
			distList dists;
			real3_type pos = vWeightGrid->gridLocation(gridIndex);

			nonExcludedPoints.queryKNN(pos, 1, nearest, dists, true);
			int labelToReplace = nonExcludedPointLabels[nearest[0]];

			// Replace the weight at this point
			setPointLabel(vWeightGrid, gridIndex, labelToReplace);
		}
	}
}

bool DeformEM::reuseExtraLabels()
{
	const int numTotalFrames = solutionPerFrame.size();
	data_type totalLabelError0(vWeightGrid->getDataDim());
	data_type totalLabelError1(vWeightGrid->getDataDim());
	data_type totalLabelError2(vWeightGrid->getDataDim());
	data_type avgLabelError0(vWeightGrid->getDataDim());
	data_type avgLabelError1(vWeightGrid->getDataDim());
	data_type avgLabelError2(vWeightGrid->getDataDim());
	data_type countCorrespPerLabel(vWeightGrid->getDataDim());
	std::vector<distList> errorValuesPerLabel;
	std::vector<idxList> errorValues_frameIndex;
	std::vector<idxList> errorValues_indices;
	distList allErrorValues;
	errorValuesPerLabel.resize(vWeightGrid->getDataDim());
	errorValues_frameIndex.resize(vWeightGrid->getDataDim());
	errorValues_indices.resize(vWeightGrid->getDataDim());
	totalLabelError0.setZero();
	totalLabelError1.setZero();
	totalLabelError2.setZero();
	countCorrespPerLabel.setZero();

	{
		const FrameSolution &frameSoln = solutionPerFrame[currentSolveFrame];
		const CorrespondenceList *frameCL = frameSoln.cList;
		const clErrorType &frame_clError = frameSoln.correspondenceError;
		//const RigidDeform &frameDeform = frameSoln.solvedDeformation;
		//const ShapeModel *frameTarget = frameSoln.targetModel;

		// For each correspondence in grid cell p
		for (int j = 0; j < frameCL->size(); ++j) {
			if (frameCL->valid(j)) {
				// Find the label of the cell containing the correspondence
				int3_type cellIndex;
				real3_type remainder;
				vWeightGrid->queryCell(frameCL->sourcePos(j), cellIndex, remainder);

				// Check to see if it's a valid cell (just in case the cell is out of bounds or not marked)
				if (!vWeightGrid->validCell(cellIndex)) {
					continue;
				}

				int label = getPointLabel(vWeightGrid, cellIndex);

				// Accumulate error
				totalLabelError0.val[label] += frame_clError[j][0];
				totalLabelError1.val[label] += frame_clError[j][1];
				totalLabelError2.val[label] += frame_clError[j][2];

				// Increment count
				countCorrespPerLabel.val[label] += 1.0;

				// Add error value to list
				allErrorValues.push_back(frame_clError[j][1]);
				errorValuesPerLabel[label].push_back(frame_clError[j][1]);
				errorValues_frameIndex[label].push_back(currentSolveFrame);
				errorValues_indices[label].push_back(j);
			}
		}
	}

	// ---> Find the outlier threshold for all correspondences
	real_type outlierThreshold;
	{
		distList tempList(allErrorValues);
		std::sort(tempList.begin(), tempList.end());

		// Don't be careful about finding the exact median
		int totalNum = allErrorValues.size();
		real_type lowerFourth = tempList[totalNum*0.25];
		real_type upperFourth = tempList[totalNum*0.75];

		// Outlier threshold (1.5 = moderate, 3.0 = aggressive)
		outlierThreshold = upperFourth + (1.5 * (upperFourth - lowerFourth));
	}

	// Compute average error per label
	for (int i = 0; i < vWeightGrid->getDataDim(); i++) {
		// Divide by total and by mesh resolution
		if (countCorrespPerLabel.val[i] > 0.0) {
			avgLabelError0.val[i] = totalLabelError0.val[i] / (countCorrespPerLabel.val[i] * scene->pointSize());
			avgLabelError1.val[i] = totalLabelError1.val[i] / (countCorrespPerLabel.val[i] * scene->pointSize());
			avgLabelError2.val[i] = totalLabelError2.val[i] / (countCorrespPerLabel.val[i] * scene->pointSize());
		}
	}

	////cout << "*** Correspondence error (point-to-point):" << endl;
	////cout << "****** Total " << totalLabelError0 << endl << "****** Average " << avgLabelError0 << endl;
	////cout << "*** Correspondence error (point-to-plane):" << endl;
	////cout << "****** Total " << totalLabelError1 << endl << "****** Average " << avgLabelError1 << endl;
	////cout << "*** Correspondence error (normal-to-normal):" << endl;
	////cout << "****** Total " << totalLabelError2 << endl << "****** Average " << avgLabelError2 << endl;


	//////////////////////////////////////////////////////////////////////////
	// Use the average point-to-plane error
	data_type labelError = avgLabelError1;
	if (excludedLabels.size()) cout << "*** Inserting extra labels with average point-to-plane error" << endl;
	newBoneBoxes.clear();
	newBoneSpheres.clear();
	showOutliers.clear();

	// Insert all the extra labels
	while (excludedLabels.size()) {
		// a. Find the label that needs improvement
		int maxErrorLabel;
		real_type maxError = labelError.findMax(maxErrorLabel);
		cout << "*** Label with maximum error: " << maxErrorLabel << " with error " << maxError << endl;
		labelError.val[maxErrorLabel] = 0.0;		// Don't address this label any further after this iteration

		if (maxError <= solveOptions.maxErrorPerCorrespondenceThreshold) {
			break;
		}

		////// b. Insert the new label in the location where the error is high
		////// ---> Find the outliers: points with error values greater than the threshold
		////const distList &ev = errorValuesPerLabel[maxErrorLabel];
		////const idxList &ev_fi = errorValues_frameIndex[maxErrorLabel];
		////const idxList &ev_i = errorValues_indices[maxErrorLabel];

		////PointList outlierLocations;
		////real3_type avgPos, box_max, box_min;
		////avgPos[0] = avgPos[1] = avgPos[2] = 0.0;
		////box_max[0] = box_max[1] = box_max[2] = -REAL_TYPE_MAX;
		////box_min[0] = box_min[1] = box_min[2] = REAL_TYPE_MAX;
		////for (int i = 0; i < ev.size(); i++) {
		////	if (ev[i] >= outlierThreshold) {
		////		// Add the location of this outlier in the reference pose
		////		const CorrespondenceList *frameCL = solutionPerFrame[ev_fi[i]].cList;
		////		real3_type pos = frameCL->sourcePos(ev_i[i]);
		////		outlierLocations.add(pos);

		////		// Update average position
		////		avgPos += pos;

		////		// Update bounding box
		////		for (int j = 0; j < 3; j++) {
		////			if (pos[j] < box_min[j]) {
		////				box_min[j] = pos[j];
		////			}
		////			if (pos[j] > box_max[j]) {
		////				box_max[j] = pos[j];
		////			}
		////		}
		////	}
		////}
		////avgPos = avgPos / ((real_type)outlierLocations.size());
		////real_type radius = (box_max-box_min).norm() * 0.7;

		////// Push the bounding box for display
		////if (outlierLocations.size()) {
		////	newBoneBoxes.push_back(make_pair(box_min, box_max));
		////	showOutliers.push_back(outlierLocations);
		////	newBoneSpheres.push_back(make_pair(avgPos, radius));
		////}

		////// c. Perform the relabeling
		////int extraLabel = excludedLabels[0];
		////data_type extraLabelVal(vWeightGrid->getDataDim());
		////extraLabelVal.val[extraLabel] = 1.0;
		////int3_type cellIndex = vWeightGrid->scanCellStart();
		////do {
		////	int label = getPointLabel(vWeightGrid, cellIndex);

		////	if (label == maxErrorLabel) {
		////		// Set the cell to have the new label
		////		if (EvaluateGaussian(vWeightGrid->gridLocation(cellIndex), avgPos, radius) > 0.1) {
		////			idxList cellPoints;
		////			vWeightGrid->getCellGridPoints(cellIndex, cellPoints);
		////			for (int i = 0; i < cellPoints.size(); ++i) {
		////				vWeightGrid->setAllValues(cellPoints[i], extraLabelVal);
		////			}
		////		}
		////	}
		////} while (vWeightGrid->scanNextCell(cellIndex));

		////// d. Update
		////// Copy the deformation in the new label, for all frames
		////for (int i = 0; i < solutionPerFrame.size(); i++) {
		////	RigidDeform &frameDeform = solutionPerFrame[i].solvedDeformation;
		////	frameDeform[extraLabel] = frameDeform[maxErrorLabel];
		////}

		////// Reset the label use stats
		////analyzeLabellingResult();


		//////////////////////////////////////////////////////////////////////////
		// Alternative 2
		// Find the extra label
		int extraLabel = excludedLabels[0];

		// Randomly subdivide the label with the largest error
		// a. Find all points belonging to the maximum label
		PointList maxLabelPoints;
		idxList maxLabelPoints_linearIndex;
		int3_type gridIndex = vWeightGrid->scanStart();
		do {
			int label = getPointLabel(vWeightGrid, gridIndex);

			if (label == maxErrorLabel) {
				maxLabelPoints.add(vWeightGrid->gridLocation(gridIndex));
				maxLabelPoints_linearIndex.push_back(vWeightGrid->getLinearIndex(gridIndex));
			}
		} while (vWeightGrid->scanNext(gridIndex));

		// b. Find two points that are somewhat apart within the label's region
		idxList sampling = bestCandidateSample(maxLabelPoints, 2);
		real3_type center1, center2;
		center1 = maxLabelPoints[sampling[0]];
		center2 = maxLabelPoints[sampling[1]];
		real_type radius = (center1 - center2).norm();

		// c. Perform the relabeling
		data_type extraLabelVal(vWeightGrid->getDataDim());
		extraLabelVal.val[extraLabel] = 1.0;
		for (int i = 0; i < maxLabelPoints.size(); i++) {
			// Evaluate gaussian
			real_type val1 = EvaluateGaussian(maxLabelPoints[i], center1, radius);
			real_type val2 = EvaluateGaussian(maxLabelPoints[i], center2, radius);

			if (val1 < val2) {
				// Assign new label
				vWeightGrid->setAllValues(maxLabelPoints_linearIndex[i], extraLabelVal);
			}
		}

		// d. Update
		// Copy the deformation in the new label, for all frames
		for (int i = 0; i < solutionPerFrame.size(); i++) {
			RigidDeform &frameDeform = solutionPerFrame[i].solvedDeformation;
			frameDeform[extraLabel] = frameDeform[maxErrorLabel];
		}

		// Reset the label use stats
		analyzeLabellingResult();
	}

	return true;
}


void DeformEM::printCurrentDeformation() const
{
	for (int frame = 0; frame < solutionPerFrame.size(); ++frame) {
		const RigidDeform &temp = solutionPerFrame[frame].solvedDeformation;

		cout << "*** Deformation for frame " << frame << "..." << endl;
		for (int i = 0; i < temp.size(); i++) {
			//cout << "****** Transformation " << i << ":" << endl;
			//cout << "****** Rotation: " << temp[i].rot;
			cout << "****** Translation " << i << ":" << temp[i].tr << endl;
		}
	}
}

void DeformEM::drawAccumulatedGeometry( const DrawOptions &options ) const
{
	// Draw the accumulated point set
	accumulatedGeometry.draw(options);

	if (options.showGraph) {
		glDisable(GL_LIGHTING);

		if (options.showGraphFaces) {
			if (!options.showWireframe) {
				glColor3f(0.0, 0.0, 0.0);
				vWeightGrid->drawEdges(options.showGraphDeformed);
			}
			vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed);
		}
		else {
			vWeightGrid->drawFaces(options.showComponent, options.showGraphDeformed, true);
			//vWeightGrid->draw(options.showComponent, options.showGraphDeformed);
			//glColor3f(0.0, 0.0, 0.0);
			//vWeightGrid->drawEdges(options.showGraphDeformed);
		}
	}
}

int DeformEM::getPointLabel( const MarkedUniformGrid3 *grid, int3_type gridIndex ) const
{
	int maxIndex;
	real_type maxValue;
	data_type temp(grid->getDataDim());

	// Find the label at this cell
	for (int j = 0; j < grid->getDataDim(); ++j) {
		temp.val[j] = grid->getValue(gridIndex, j);
	}

	// Find the maximum value
	maxValue = temp.findMax(maxIndex);

	////////////////////////////////////////////////////////////////////////////
	//// Optional debugging: check that the maximum value is 1.0 and there is only one positive component
	//if (maxValue != 1.0) {
	//	cerr << "Warning in DeformEM::getPointLabel(): Cell " << gridIndex << " maximum weight is not 1!" << endl;
	//}
	//int count = 0;
	//for (int j = 0; j < grid->getDataDim(); ++j) {
	//	if (temp.val[j] > 0) {
	//		count++;
	//	}
	//}
	//if (count != 1) {
	//	cerr << "Warning in DeformEM::getPointLabel(): Cell " << gridIndex << " has " << count << "positive components!" << endl;
	//}
	////////////////////////////////////////////////////////////////////////////

	return maxIndex;
}

bool DeformEM::setPointLabel( MarkedUniformGrid3 *grid, int3_type gridIndex, int label ) const
{
	if ((label < 0) || (label >= grid->getDataDim())) {
		cerr << "Error in DeformEM::setPointLabel(): label out of bounds!" << endl;
		return false;
	}

	data_type temp(grid->getDataDim());
	temp.setZero();
	temp.val[label] = 1.0;

	grid->setAllValues(grid->getLinearIndex(gridIndex), temp);

	return true;
}


void DeformEM::saveResult( const char *filename ) const
{
	ofstream out;
	out.open(filename, ios_base::binary);

	if (out.fail()) {
		cout << "Error in DeformEM::saveResult(): could not open file " << filename << " for writing!" << endl;
		return;
	}

	cout << "*** Saving registration result to " << filename << "...";

	//////////////////////////////////////////////////////////////////////////
	// Note: right now we're just going to save the values, not the parameters
	// associated with the grid and transformations.  Make sure you save
	// the original configuration that generated the grid structure!
	//////////////////////////////////////////////////////////////////////////
	unsigned int numBones = vWeightGrid->getDataDim();
	unsigned int numGridPoints = vWeightGrid->totalGridPoints();
	unsigned int numFrames = 1;		/**< Save just the current deformation for now */

	// Save basic parameters
	out.write((char*) &numBones, sizeof(unsigned int));
	out.write((char*) &numGridPoints, sizeof(unsigned int));
	out.write((char*) &numFrames, sizeof(unsigned int));


	//////////////////////////////////////////////////////////////////////////
	// Save the transformations
	for (index_type i = 0; i < numBones; ++i) {
		// Store transformation
		currentDeform()[i].saveBinary(out);
	}

	//////////////////////////////////////////////////////////////////////////
	// Save the grid weights
	const std::vector<data_type> &weights = vWeightGrid->getData();

	for (index_type i = 0; i < weights.size(); ++i) {
		for (int j = 0; j < numBones; ++j) {
			real_type val = weights[i].val[j];
			out.write((char*) &val, sizeof(real_type));
		}
	}

	out.close();

	cout << "done!" << endl;
}


void DeformEM::loadResult( const char *filename )
{
	ifstream in;
	in.open(filename, ios_base::binary);

	if (in.fail()) {
		cout << "Error in DeformEM::loadResult(): could not open file " << filename << " for reading!" << endl;
		return;
	}

	cout << "Loading registration result from " << filename << "...";

	//////////////////////////////////////////////////////////////////////////
	// Read the list of all the transformations
	unsigned int numBones, numGridPoints, numFrames;
	in.read((char*) &numBones, sizeof(unsigned int));
	in.read((char*) &numGridPoints, sizeof(unsigned int));
	in.read((char*) &numFrames, sizeof(unsigned int));


	// Do basic parameter checking
	if ((vWeightGrid->getDataDim() != numBones)) {
		cerr << "Error in DeformEM::loadResult(): numBones doesn't match!" << endl;
		return;
	}
	if ((vWeightGrid->totalGridPoints() != numGridPoints)) {
		cerr << "Error in DeformEM::loadResult(): numGridPoints doesn't match!" << endl;
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// Read in a single frame
	if (currentDeform().size() != numBones) {
		currentDeform().resize(numBones);
	}

	// Read the transformations
	for (index_type i = 0; i < numBones; i++) {
		// Read transformation
		currentDeform()[i].loadBinary(in);
	}

	// Read the grid weights
	data_type temp(numBones);
	for (index_type i = 0; i < numGridPoints; ++i) {
		temp.setZero();
		for (int j = 0; j < numBones; ++j) {
			in.read((char*) &temp.val[j], sizeof(real_type));
		}
		vWeightGrid->setAllValues(i, temp);
	}

	in.close();

	cout << " ...done!" << endl;

	// Apply the loaded deformation
	applyDeformation_AccumulatedGeometry(currentDeform(), currentDeformedAccumulatedGeometry());
}

void DeformEM::writeStats( std::ofstream &out ) const
{
	out << "# Bones" << "\t" << vWeightGrid->getDataDim() << "\n";
	out << "# Grid Points" << "\t" << vWeightGrid->totalGridPoints() << "\n";
	out << "# Grid Cells" << "\t" << vWeightGrid->totalGridCells() << "\n";
}

void DeformEM::generateGridColors()
{
	vWeightGrid->generateColors();
}

void DeformEM::controlSolver( int controlType )
{
	// Implement pause, step, and stop behaviors using a mutex

	switch(controlType) {
	case SOLVER_RESUME:
		// Let go of the mutex (if held)
		if (controlSignal == SOLVER_PAUSE) {
			controlMutex.unlock();
		}
		controlSignal = SOLVER_RESUME;
		break;

	case SOLVER_PAUSE:
		// Acquire and hold the mutex
		if (controlSignal != SOLVER_PAUSE) {
			controlMutex.lock();
			controlSignal = SOLVER_PAUSE;
		}
		break;

	case SOLVER_STEP:
		// Let go of the mutex (if held) and acquire it immediately
		if (controlSignal == SOLVER_PAUSE) {
			controlMutex.unlock();
			controlMutex.lock();
		}
		break;

	case SOLVER_STOP:
		// Set the stop variable and stop at the next step
		if (controlSignal == SOLVER_PAUSE) {
			controlMutex.unlock();
		}
		controlSignal = SOLVER_STOP;
		break;
	}
}

bool DeformEM::controlCheck()
{
	// Called by the solver thread after each step
	// 1. Acquire the mutex
	controlMutex.lock();

	// 2. See the control signal and proceed
	if (controlSignal == SOLVER_STOP) {
		// Exit the solver thread
		controlMutex.unlock();
		iterationFinished = true;
		Notify();
		return false;
	}
	else {
		controlMutex.unlock();
		return true;
	}
}


std::pair<QAbstractItemModel*,DeformEM_GUI_Communicator*> DeformEM::solutionViewModel()
{
	if (solutionView) {
		return std::make_pair(solutionView, solutionViewCommunicator);
	}

	solutionView = new QStandardItemModel(0, VIEW_NUM_COLUMNS);

	solutionView->setHeaderData(0, Qt::Horizontal, QObject::tr("Title"));
	solutionView->setHeaderData(VIEW_SOURCE, Qt::Horizontal, QObject::tr("Src"));
	solutionView->setHeaderData(VIEW_TARGET, Qt::Horizontal, QObject::tr("Tgt"));
	solutionView->setHeaderData(VIEW_DEFORMED, Qt::Horizontal, QObject::tr("Def"));
	solutionView->setHeaderData(VIEW_INVERSE_DEFORMED, Qt::Horizontal, QObject::tr("Inv"));
	solutionView->setHeaderData(VIEW_ADJ_CORRESP, Qt::Horizontal, QObject::tr("ACor"));
	solutionView->setHeaderData(VIEW_CORRESP_SOURCE, Qt::Horizontal, QObject::tr("CorS"));
	solutionView->setHeaderData(VIEW_CORRESP_TARGET, Qt::Horizontal, QObject::tr("CorT"));

	updateSolutionViewModel();

	// Create communicator instance
	solutionViewCommunicator = new DeformEM_GUI_Communicator();

	return std::make_pair(solutionView, solutionViewCommunicator);
}

void DeformEM::updateSolutionViewModel()
{
	int numRows = solutionPerFrame.size();

	if (solutionViewCommunicator) {
		solutionViewCommunicator->updateGUI(numRows);
	}
	else {
		DeformEM_GUI_Communicator::updateSolutionView(solutionView, numRows);
	}
}

void DeformEM::saveDiscreteWeights()
{
	// We can only save once, so use it at the end of the discrete optimization
	if (!savedDiscreteWeights.size()) {
		savedDiscreteWeights = vWeightGrid->getData();
	}
	else {
		cerr << "Warning in saveDiscreteWeights(): Discrete weights already saved!" << endl;
	}
}

void DeformEM::revertDiscreteWeights()
{
	if (savedDiscreteWeights.size()) {
		vWeightGrid->setData(savedDiscreteWeights);
		applyDeformation_FrameGeometry(currentDeform(), currentDeformedAccumulatedGeometry(), currentSource());
	}
	else {
		cerr << "Warning in revertDiscreteWeights(): No saved discrete weights exist!" << endl;
	}
}

bool DeformEM::writeDeformedResult( const char *file ) const
{
	if (solutionPerFrame.size() <= 0) {
		return false;
	}
	const AccumulatedGeometry& shape = currentDeformedAccumulatedGeometry();

	ofstream out;
	out.open(file);

	if (out.fail()) {
		cerr << "Error: Could not open output file " << file << " for writing!" << endl;
		return false;
	}

	cout << "Saving vertex and triangle data in " << file << " ...";

	// Write vertex data
	for (index_type i = 0; i < shape.vertex.size(); i++) {
		out << "v";
		for (int c = 0; c < 3; c++) {
			out << " " << shape.vertex[i][c];
		}
		out << "\n";
	}

	// Write triangle data
	for (index_type i = 0; i < shape.triangle_vertex_index.size(); i++) {
		out << "f";
		for (int c = 0; c < 3; c++) {
			out << " " << shape.triangle_vertex_index[i][c]+1;
		}
		out << "\n";
	}

	out.close();

	cout << "...done!" << endl;

	return true;
}


void DeformEM_GUI_Communicator::updateGUI( int numRows )
{
	emit solutionIncreased(numRows);
}

void DeformEM_GUI_Communicator::updateSolutionView( QAbstractItemModel* model, int numRows )
{
	int numCurrentRows = model->rowCount();

	for (int i = numCurrentRows; i < numRows; ++i) {
		model->insertRows(i, 1);
		model->setData(model->index(i, 0), QObject::tr("Solution %1").arg(i));

		// View nothing
		for (int j = 1; j < model->columnCount(); ++j) {
			model->setData(model->index(i, j), false);
		}

		// View some as default
		model->setData(model->index(i, VIEW_DEFORMED), true);
		model->setData(model->index(i, VIEW_TARGET), true);
	}
}


DeformEM::LinearSystemE::LinearSystemE( const size_type numBones ) :
numBones(numBones), numValidBones(numBones), excludedBones(0), constraints(NULL)
{
	computeBoneIndexTranslation();
}

DeformEM::LinearSystemE::LinearSystemE( const size_type numBones, const std::vector<int> &excludedBoneList ) :
numBones(numBones), numValidBones(numBones-excludedBoneList.size()), excludedBones(excludedBoneList), constraints(NULL)
{
	computeBoneIndexTranslation();
}

void DeformEM::LinearSystemE::clear()
{
	for(int i = 0; i < terms.size(); ++i) {
		if (terms[i]) {
			delete terms[i];
			terms[i] = NULL;
		}
	}
	terms.clear();
}

void DeformEM::LinearSystemE::computeBoneIndexTranslation()
{
	// Assumes that we have the list of excluded bones ready
	excludeBone.resize(numBones, false);
	for (int i = 0; i < excludedBones.size(); i++) {
		// Set the boolean variable to be true for excluded bones
		if (excludedBones[i] < numBones) {
			excludeBone[excludedBones[i]] = true;
		}
		else {
			cerr << "Error in DeformEM::LinearSystemE::computeBoneIndexTranslation(): excluded bone " << excludedBones[i] << " out of range!" << endl;
		}
	}

	// Now assumes that the boolean array of excluded bones has been initialized.
	int numExcludedBones = excludedBones.size();

	reduced_index.resize(numBones, -1);
	back_to_original_index.resize(numBones-numExcludedBones, -1);

	// Compute translation of indices, forward and backward
	int currentIndex = 0;
	for (int i = 0; i < numBones; i++) {
		if (!excludeBone[i]) {
			reduced_index[i] = currentIndex;
			back_to_original_index[currentIndex] = i;
			++currentIndex;
		}
	}

	//// Print stats
	//cout << "*** Excluded bone list: " << excludedBones << endl;
	//cout << "*** Boolean exclusion array: " << excludeBone << endl;
	//cout << "*** Index Translation (original -> reduced): " << reduced_index << endl;
	//cout << "*** Index Translation (reduced -> original): " << back_to_original_index << endl;
}

void DeformEM::LinearSystemE::translateVertexWeights( std::vector<idxList> &nodeList, std::vector<distList> &weightList ) const
{
	const int numPoints = nodeList.size();

	if (numPoints != weightList.size()) {
		cerr << "Error in DeformEM::LinearSystemE::translateVertexWeights(): mismatch in size!" << endl;
		return;
	}

	for (int i = 0; i < numPoints; ++i) {
		idxList &affectingNodes = nodeList[i];
		distList &affectingWeights = weightList[i];

		for (int j = 0; j < affectingNodes.size(); ++j) {
			// Check whether if bone is excluded
			if (excludeBone[affectingNodes[j]]) {
				// Remove the excluded bone from affecting this constraint
				cerr << "Point " << i << " affecting nodes are: " << affectingNodes << endl;
				cerr << "Point " << i << " affecting weights are: " << affectingWeights << endl;
			}
			else {
				// Translate the index
				affectingNodes[j] = reduced_index[affectingNodes[j]];
			}
		}
	}
}

void DeformEM::LinearSystemE::translateIntegrals( std::vector< std::vector<double> > &cellIntegrals ) const
{
	std::vector< std::vector<double> > newIntegrals(numValidBones);

	for (int i = 0; i < back_to_original_index.size(); i++) {
		newIntegrals[i] = cellIntegrals[back_to_original_index[i]];
	}

	cellIntegrals = newIntegrals;
}

AffineDeform DeformEM::LinearSystemE::translateDeformation( const AffineDeform &def ) const
{
	AffineDeform result(numValidBones);

	for (int i = 0; i < back_to_original_index.size(); ++i) {
		result[i] = def[back_to_original_index[i]];
	}

	return result;
}

RigidDeform DeformEM::LinearSystemE::translateDeformation( const RigidDeform &def ) const
{
	RigidDeform result(numValidBones);

	for (int i = 0; i < back_to_original_index.size(); ++i) {
		result[i] = def[back_to_original_index[i]];
	}

	return result;
}

void DeformEM::LinearSystemE::makeConstraintTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def )
{
	// For each constraint
	//   1. Find the nodes that affect the vertex
	//   2. Calculate and add the term to the linear system
	ErrorTerm *constraintTerm = new ErrorTerm();
	const int cbStart = numValidBones*dim;
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	if ((constraints != con) || (nodeList.size() != con->size())) {
		constraints = con;
		vWeightGrid->getVertexWeights(con, nodeList, vertexWeight);
		translateVertexWeights(nodeList, vertexWeight);
	}
	RigidDeform reducedDef = translateDeformation(def);

	// Predict the number of nonzero entries.
	int totalLHSnnz = 0;
	int totalRows = 0;
	for (index_type	ci = 0; ci < nodeList.size(); ci++) {
		if (con->valid(ci)) {
			totalLHSnnz += nodeList[ci].size()*(15*nodeList[ci].size() + 9);
			totalRows += dim*(nodeList[ci].size() + 1);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	constraintTerm->weight = weight;
	constraintTerm->rows = totalRows;
	constraintTerm->lhs.resize(totalLHSnnz);
	constraintTerm->rhs = new real_type[constraintTerm->rows];
	constraintTerm->rowWeight.resize(constraintTerm->rows, 1.0);

	//////////////////////////////////////////////////////////////////////////
	for (index_type ci = 0; ci < con->size(); ci++) {
		// Don't add the constraint if the weight is zero
		if (!con->valid(ci)) {
			continue;
		}

		// Retrieve the nodes that affect this vertex from the list of precomputed weights
		real3_type pos = con->sourcePos(ci);
		real3_type newPos = con->targetPos(ci);
		idxList affectingNodes = nodeList[ci];
		distList affectingWeights = vertexWeight[ci];
		real_type xj, yj, zj, xk, yk, zk;
		real_type xp, yp, zp;
		real_type coll[dim][2*dim];
		real_type rhs[dim];

		// Precompute the current deformation applied from each node.
		std::vector<real3_type> deformedPos;
		deformedPos.resize(numValidBones);
		for (index_type bi = 0; bi < numValidBones; bi++) {
			deformedPos[bi] = reducedDef[bi].rot * pos + reducedDef[bi].tr;
		}

		// The target position, in components.
		xp = newPos[0]; yp = newPos[1]; zp = newPos[2];

		// Reset LHS
		for (int i = 0; i < dim; i++) {
			for (int j = 0; j < 2*dim; j++) {
				coll[i][j] = 0;
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// Section 1. Cross product constraints (one for each affecting node).
		for (index_type ind1 = 0; ind1 < affectingNodes.size(); ind1++) {
			index_type nj = affectingNodes[ind1];
			real_type wj = affectingWeights[ind1];

			xj = deformedPos[nj][0]; yj = deformedPos[nj][1]; zj = deformedPos[nj][2];

			// Reset RHS
			rhs[0] = wj*(yj*zp-zj*yp);
			rhs[1] = wj*(zj*xp-xj*zp);
			rhs[2] = wj*(xj*yp-yj*xp);

			for (index_type ind2 = 0; ind2 < affectingNodes.size(); ind2++) {
				index_type nk = affectingNodes[ind2];
				real_type wk = affectingWeights[ind2];

				xk = deformedPos[nk][0]; yk = deformedPos[nk][1]; zk = deformedPos[nk][2];

				// Update LHS
				coll[0][0] = yj*yk+zj*zk;
				coll[0][1] = -yj*xk;
				coll[0][2] = -zj*xk;
				coll[0][4] = -zj;
				coll[0][5] = yj;
				coll[1][0] = -xj*yk;
				coll[1][1] = zj*zk+xj*xk;
				coll[1][2] = -zj*yk;
				coll[1][3] = zj;
				coll[1][5] = -xj;
				coll[2][0] = -xj*zk;
				coll[2][1] = -yj*zk;
				coll[2][2] = xj*xk+yj*yk;
				coll[2][3] = -yj;
				coll[2][4] = xj;

				// Fill in LHS (don't forget to multiply by wk!)
				index_type cjCol = nk*dim;
				index_type cbjCol = cbStart + cjCol;

				for (int row = 0; row < dim; row++) {
					for (int col = 0; col < dim; col++) {
						// Assign the linear rotation coefficients
						if (coll[row][col] != 0) {
							constraintTerm->lhs.fill(TrackRow+row, cjCol+col, wj*wk*coll[row][col]);
						}

						// Assign the translation coefficients
						if (coll[row][col+dim] != 0) {
							constraintTerm->lhs.fill(TrackRow+row, cbjCol+col, wj*wk*coll[row][col+dim]);
						}
					}
				}

				// Update RHS
				rhs[0] += -wj*wk*(yj*zk-zj*yk);
				rhs[1] += -wj*wk*(zj*xk-xj*zk);
				rhs[2] += -wj*wk*(xj*yk-yj*xk);
			}

			// Fill in RHS
			for (int row = 0; row < dim; row++) {
				constraintTerm->rhs[TrackRow+row] = rhs[row];
			}

			// Fill in weights for this constraint
			for (int row = 0; row < dim; row++) {
				constraintTerm->rowWeight[TrackRow+row] = con->getWeight(ci);
			}

			// Increment row count
			TrackRow += dim;
		}

		//////////////////////////////////////////////////////////////////////////
		// Section 2. Deformed position constraint.

		// Reset RHS
		rhs[0] = xp;
		rhs[1] = yp;
		rhs[2] = zp;

		// Reset LHS
		for (int i = 0; i < dim; i++) {
			for (int j = 0; j < 2*dim; j++) {
				coll[i][j] = 0;
			}
		}

		for (index_type ind2 = 0; ind2 < affectingNodes.size(); ind2++) {
			index_type nk = affectingNodes[ind2];
			real_type wk = affectingWeights[ind2];

			xk = deformedPos[nk][0]; yk = deformedPos[nk][1]; zk = deformedPos[nk][2];

			// Update LHS
			coll[0][1] = zk;
			coll[0][2] = -yk;
			coll[0][3] = 1.0;
			coll[1][0] = -zk;
			coll[1][2] = xk;
			coll[1][4] = 1.0;
			coll[2][0] = yk;
			coll[2][1] = -xk;
			coll[2][5] = 1.0;

			// Fill in LHS (don't forget to multiply by wk!)
			index_type cjCol = nk*dim;
			index_type cbjCol = cbStart + cjCol;

			for (int row = 0; row < dim; row++) {
				for (int col = 0; col < dim; col++) {
					// Assign the linear rotation coefficients
					if (coll[row][col] != 0) {
						constraintTerm->lhs.fill(TrackRow+row, cjCol+col, wk*coll[row][col]);
					}

					// Assign the translation coefficients
					if (coll[row][col+dim] != 0) {
						constraintTerm->lhs.fill(TrackRow+row, cbjCol+col, wk*coll[row][col+dim]);
					}
				}
			}

			// Update RHS
			rhs[0] += -wk*xk;
			rhs[1] += -wk*yk;
			rhs[2] += -wk*zk;
		}

		// Fill in RHS
		for (int row = 0; row < dim; row++) {
			constraintTerm->rhs[TrackRow+row] = rhs[row];
		}

		// Fill in weights for this constraint
		for (int row = 0; row < dim; row++) {
			constraintTerm->rowWeight[TrackRow+row] = con->getWeight(ci);
		}

		// Increment row count
		TrackRow += dim;
	}


	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < constraintTerm->rows; ci++) {
	//	cout << "V(" << ci+1 << ") = " << constraintTerm->rhs[ci] << ";" << endl;
	//}
	//constraintTerm->lhs.print();

	//cout << "*** Estimated size for the Jacobian (energy): " << totalLHSnnz
	//	 << " vs. Actual size: " << constraintTerm->lhs.size() << endl;
	//cout << "*** Estimated Number of Rows: " << constraintTerm->rows
	//	<< " vs. Actual # rows: " << TrackRow << endl;

	// Assign the term to the list of terms
	terms.push_back(constraintTerm);
}



void DeformEM::LinearSystemE::makeConstraintTerm_PointToPlane( real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def, const ShapeModel &targetModel )
{
	// For each constraint
	//   1. Find the nodes that affect the vertex
	//   2. Calculate and add the term to the linear system
	ErrorTerm *constraintTerm = new ErrorTerm();
	const int cbStart = numValidBones*dim;
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	if (constraints != con) {
		constraints = con;
		vWeightGrid->getVertexWeights(con, nodeList, vertexWeight);
		translateVertexWeights(nodeList, vertexWeight);
	}
	RigidDeform reducedDef = translateDeformation(def);

	// Predict the number of nonzero entries.
	int totalLHSnnz = 0;
	int totalRows = 0;
	for (index_type	ci = 0; ci < nodeList.size(); ci++) {
		if (con->valid(ci)) {
			totalLHSnnz += nodeList[ci].size()*(18*nodeList[ci].size() + 18);
			totalRows += dim*(nodeList[ci].size() + 1);
		}
	}

	//////////////////////////////////////////////////////////////////////////
	constraintTerm->weight = weight;
	constraintTerm->rows = totalRows;
	constraintTerm->lhs.resize(totalLHSnnz);
	constraintTerm->rhs = new real_type[constraintTerm->rows];
	constraintTerm->rowWeight.resize(constraintTerm->rows, 1.0);

	//////////////////////////////////////////////////////////////////////////
	for (index_type ci = 0; ci < con->size(); ci++) {
		// Don't add the constraint if the weight is zero
		if (!con->valid(ci)) {
			continue;
		}

		// Retrieve the nodes that affect this vertex from the list of precomputed weights
		real3_type pos = con->sourcePos(ci);
		real3_type newPos = con->targetPos(ci);
		real3_type newNormal = targetModel.normal[con->targetIndex[ci]];
		idxList affectingNodes = nodeList[ci];
		distList affectingWeights = vertexWeight[ci];
		real_type xj, yj, zj, xk, yk, zk;
		real_type nx, ny, nz;
		real_type xp, yp, zp;
		real_type coll[dim][2*dim];
		real_type rhs[dim];

		// Precompute the current deformation applied from each node.
		std::vector<real3_type> deformedPos;
		deformedPos.resize(numValidBones);
		for (index_type bi = 0; bi < numValidBones; bi++) {
			deformedPos[bi] = reducedDef[bi].rot * pos + reducedDef[bi].tr;
		}

		// The target position and normal, in components.
		xp = newPos[0]; yp = newPos[1]; zp = newPos[2];
		nx = newNormal[0]; ny = newNormal[1]; nz = newNormal[2];

		//////////////////////////////////////////////////////////////////////////
		// Section 1. Cross product constraints (one for each affecting node).
		for (index_type ind1 = 0; ind1 < affectingNodes.size(); ind1++) {
			index_type nj = affectingNodes[ind1];
			real_type wj = affectingWeights[ind1];
			real_type t7, t12, t1, t11, t5, t10, t3, t9, t2, t8, t4;

			xj = deformedPos[nj][0]; yj = deformedPos[nj][1]; zj = deformedPos[nj][2];

			// Reset RHS
			t7 = wj*(nx*xp+ny*yp+nz*zp);
			rhs[0] = -(-yj*nz+ny*zj)*t7;
			rhs[1] = (nx*zj-xj*nz)*t7;
			rhs[2] = -(nx*yj-ny*xj)*t7;

			// Reset LHS
			for (int i = 0; i < dim; i++) {
				for (int j = 0; j < 2*dim; j++) {
					coll[i][j] = 0;
				}
			}

			for (index_type ind2 = 0; ind2 < affectingNodes.size(); ind2++) {
				index_type nk = affectingNodes[ind2];
				real_type wk = affectingWeights[ind2];

				xk = deformedPos[nk][0]; yk = deformedPos[nk][1]; zk = deformedPos[nk][2];

				// Update LHS
				t12 = wj*wk;
				t1 = nx*yj-ny*xj;
				t11 = t1*t12;
				t5 = -yj*nz+ny*zj;
				t10 = t5*t12;
				t3 = -ny*xk+nx*yk;
				t9 = t3*t12;
				t2 = nx*zj-xj*nz;
				t8 = t2*t12;
				t7 = (-nz*yk+ny*zk)*t12;
				t4 = nx*zk-nz*xk;
				coll[0][0] = t5*t7;
				coll[0][1] = -t4*t10;
				coll[0][2] = t5*t9;
				coll[0][3] = -nx*t10;
				coll[0][4] = -ny*t10;
				coll[0][5] = -nz*t10;
				coll[1][0] = -t2*t7;
				coll[1][1] = t4*t8;
				coll[1][2] = -t3*t8;
				coll[1][3] = nx*t8;
				coll[1][4] = ny*t8;
				coll[1][5] = nz*t8;
				coll[2][0] = t1*t7;
				coll[2][1] = -t4*t11;
				coll[2][2] = t1*t9;
				coll[2][3] = -nx*t11;
				coll[2][4] = -ny*t11;
				coll[2][5] = -nz*t11;

				// Fill in LHS (don't forget to multiply by wk!)
				index_type cjCol = nk*dim;
				index_type cbjCol = cbStart + cjCol;

				for (int row = 0; row < dim; row++) {
					for (int col = 0; col < dim; col++) {
						// Assign the linear rotation coefficients
						if (coll[row][col] != 0) {
							constraintTerm->lhs.fill(TrackRow+row, cjCol+col, coll[row][col]);
						}

						// Assign the translation coefficients
						if (coll[row][col+dim] != 0) {
							constraintTerm->lhs.fill(TrackRow+row, cbjCol+col, coll[row][col+dim]);
						}
					}
				}

				// Update RHS
				t7 = -wj*wk*(nx*xk+ny*yk+nz*zk);
				rhs[0] += -(-yj*nz+ny*zj)*t7;
				rhs[1] += (nx*zj-xj*nz)*t7;
				rhs[2] += -(nx*yj-ny*xj)*t7;
			}

			// Fill in RHS
			for (int row = 0; row < dim; row++) {
				constraintTerm->rhs[TrackRow+row] = rhs[row];
			}

			// Fill in weights for this constraint
			for (int row = 0; row < dim; row++) {
				constraintTerm->rowWeight[TrackRow+row] = con->getWeight(ci);
			}

			// Increment row count
			TrackRow += dim;
		}

		//////////////////////////////////////////////////////////////////////////
		// Section 2. Deformed position constraint.

		// Reset RHS
		real_type t2 = nx*xp+ny*yp+zp*nz;
		rhs[0] = t2*nx;
		rhs[1] = t2*ny;
		rhs[2] = t2*nz;

		// Reset LHS
		for (int i = 0; i < dim; i++) {
			for (int j = 0; j < 2*dim; j++) {
				coll[i][j] = 0;
			}
		}

		for (index_type ind2 = 0; ind2 < affectingNodes.size(); ind2++) {
			index_type nk = affectingNodes[ind2];
			real_type wk = affectingWeights[ind2];
			real_type t10, t9, t8, t7, t6, t5, t4;

			xk = deformedPos[nk][0]; yk = deformedPos[nk][1]; zk = deformedPos[nk][2];

			// Update LHS
			t10 = wk*nz;
			t9 = wk*(nx*zk-nz*xk);
			t8 = (-ny*xk+nx*yk)*wk;
			t7 = wk*(-nz*yk+ny*zk);
			t6 = ny*nx*wk;
			t5 = nx*t10;
			t4 = ny*t10;
			coll[0][0] = -nx*t7;
			coll[0][1] = nx*t9;
			coll[0][2] = -nx*t8;
			coll[0][3] = nx*nx*wk;
			coll[0][4] = t6;
			coll[0][5] = t5;
			coll[1][0] = -ny*t7;
			coll[1][1] = ny*t9;
			coll[1][2] = -ny*t8;
			coll[1][3] = t6;
			coll[1][4] = ny*ny*wk;
			coll[1][5] = t4;
			coll[2][0] = -nz*t7;
			coll[2][1] = nz*t9;
			coll[2][2] = -nz*t8;
			coll[2][3] = t5;
			coll[2][4] = t4;
			coll[2][5] = nz*nz*wk;

			// Fill in LHS (don't forget to multiply by wk!)
			index_type cjCol = nk*dim;
			index_type cbjCol = cbStart + cjCol;

			for (int row = 0; row < dim; row++) {
				for (int col = 0; col < dim; col++) {
					// Assign the linear rotation coefficients
					if (coll[row][col] != 0) {
						constraintTerm->lhs.fill(TrackRow+row, cjCol+col, coll[row][col]);
					}

					// Assign the translation coefficients
					if (coll[row][col+dim] != 0) {
						constraintTerm->lhs.fill(TrackRow+row, cbjCol+col, coll[row][col+dim]);
					}
				}
			}

			// Update RHS
			t2 = -wk*(nx*xk+ny*yk+nz*zk);
			rhs[0] += t2*nx;
			rhs[1] += t2*ny;
			rhs[2] += t2*nz;
		}

		// Fill in RHS
		for (int row = 0; row < dim; row++) {
			constraintTerm->rhs[TrackRow+row] = rhs[row];
		}

		// Fill in weights for this constraint
		for (int row = 0; row < dim; row++) {
			constraintTerm->rowWeight[TrackRow+row] = con->getWeight(ci);
		}

		// Increment row count
		TrackRow += dim;
	}


	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < constraintTerm->rows; ci++) {
	//	cout << "V(" << ci+1 << ") = " << constraintTerm->rhs[ci] << ";" << endl;
	//}
	//constraintTerm->lhs.print();

	//cout << "*** Estimated size for the Jacobian (energy): " << totalLHSnnz
	//	<< " vs. Actual size: " << constraintTerm->lhs.size() << endl;
	//cout << "*** Estimated Number of Rows: " << constraintTerm->rows
	//	<< " vs. Actual # rows: " << TrackRow << endl;

	// Assign the term to the list of terms
	terms.push_back(constraintTerm);
}


/**
* Assume that rd is already the reduced subset of the deformation.
* But currentDef is not reduced.
*/
real_type DeformEM::LinearSystemE::findConstraintEnergy( RigidDeform &rd, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, const RigidDeform &currentDef  )
{
	real_type totalEnergy = 0.0;

	//////////////////////////////////////////////////////////////////////////
	if (constraints != con) {
		constraints = con;
		vWeightGrid->getVertexWeights(con, nodeList, vertexWeight);
		translateVertexWeights(nodeList, vertexWeight);
	}
	RigidDeform reducedCurrentDef = translateDeformation(currentDef);

	if (rd.size() != numValidBones) {
		cerr << "Error in DeformEM::LinearSystemE::findConstraintEnergy(): rd not correct size!" << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// Calculate the difference between the initial and deformed positions
	for (index_type ci = 0; ci < con->size(); ci++) {
		if (!con->valid(ci)) {
			continue;
		}

		real3_type pos = con->sourcePos(ci);
		real3_type targetPos = con->targetPos(ci);
		real3_type diff;
		real3_type outPos(0,0,0);

		for (index_type ind = 0; ind < nodeList[ci].size(); ind++) {
			index_type j = nodeList[ci][ind];
			real_type wj = vertexWeight[ci][ind];
			real3_type currentDefPos = reducedCurrentDef[j].rot * pos + reducedCurrentDef[j].tr;

			real3_type newDefPos = rd[j].rot * currentDefPos + rd[j].tr;

			// Deform by this transformation
			outPos += wj*newDefPos;
		}


		//cout << "    -- Initial: " << ci << ": " << pos << endl;
		//cout << "    -- Moved:   " << ci << ": " << outPos << endl;


		// Add to the total energy
		diff = targetPos - outPos;
		totalEnergy += diff.square_norm();
	}

	return totalEnergy;
}


/**
* Assume that ad is already the reduced subset of the deformation.
* But currentDef is not reduced.
*/
real_type DeformEM::LinearSystemE::findConstraintEnergy( AffineDeform &ad, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, const RigidDeform &currentDef  )
{
	real_type totalEnergy = 0.0;

	//////////////////////////////////////////////////////////////////////////
	if (constraints != con) {
		constraints = con;
		vWeightGrid->getVertexWeights(con, nodeList, vertexWeight);
		translateVertexWeights(nodeList, vertexWeight);
	}
	RigidDeform reducedCurrentDef = translateDeformation(currentDef);

	if (ad.size() != numValidBones) {
		cerr << "Error in DeformEM::LinearSystemE::findConstraintEnergy(): ad not correct size!" << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// Calculate the difference between the initial and deformed positions
	for (index_type ci = 0; ci < con->size(); ci++) {
		if (!con->valid(ci)) {
			continue;
		}

		real3_type pos = con->sourcePos(ci);
		real3_type targetPos = con->targetPos(ci);
		real3_type diff;
		real3_type outPos(0,0,0);

		for (index_type ind = 0; ind < nodeList[ci].size(); ind++) {
			index_type j = nodeList[ci][ind];
			real_type wj = vertexWeight[ci][ind];
			real3_type currentDefPos = reducedCurrentDef[j].rot * pos + reducedCurrentDef[j].tr;
			real3_type affineRotation(ad[j].lr[0], ad[j].lr[1], ad[j].lr[2]);
			real3_type affineTranslation(ad[j].tr[0], ad[j].tr[1], ad[j].tr[2]);

			real3_type newDefPos = currentDefPos + affineTranslation + (affineRotation ^ currentDefPos);

			// Deform by this transformation
			outPos += wj*newDefPos;
		}

		//cout << "    -- Constraint " << ci << ": " << targetPos << endl;
		//cout << "    -- Moved      " << ci << ": " << outPos << endl;

		// Add to the total energy
		diff = targetPos - outPos;
		totalEnergy += diff.square_norm();
	}

	return totalEnergy;
}


void DeformEM::LinearSystemE::makeSmoothnessTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid, RigidDeform &def )
{
	// Check weight
	if (weight <= 0.0) {
		return;
	}

	// For each cell, compute the overlap of the vertex weight function in each cell
	ErrorTerm *smoothTerm = new ErrorTerm();
	const int cbStart = numValidBones*dim;
	index_type TrackRow = 0;
	std::vector< std::pair<index2_type, std::vector<double> > > integrals_per_pair;
	real_type coll[3*dim][4*dim];
	real_type rhs[3*dim];

	//////////////////////////////////////////////////////////////////////////
	// Predict the number of nonzero entries.
	int totalLHSnnz = 0;
	int totalRows = 0;
	int3_type cellIndex = vWeightGrid->scanCellStart();
	do {
		int pairCount = vWeightGrid->computeOverlapIntegrals(cellIndex, def, integrals_per_pair, true);		/**< Just count pairs per cell */
		totalLHSnnz += pairCount * 78;
		totalRows += pairCount * 9;
	} while(vWeightGrid->scanNextCell(cellIndex));


	//////////////////////////////////////////////////////////////////////////
	smoothTerm->weight = weight;
	smoothTerm->rows = totalRows;
	smoothTerm->lhs.resize(totalLHSnnz);
	smoothTerm->rhs = new real_type[smoothTerm->rows];

	//////////////////////////////////////////////////////////////////////////
	// Iterate over all cells
	cellIndex = vWeightGrid->scanCellStart();
	do {
		vWeightGrid->computeOverlapIntegrals(cellIndex, def, integrals_per_pair);

		if (!integrals_per_pair.size()) {
			// There's no smoothness for this cell
			continue;
		}

		// Construct smoothness for each pair
		for (int i = 0; i < integrals_per_pair.size(); i++) {
			// First, check to see that the both bones in this pair are not excluded
			// This shouldn't trigger as long as excluded bones have zero weight everywhere
			if (excludeBone[integrals_per_pair[i].first[0]] || excludeBone[integrals_per_pair[i].first[1]]) {
				cerr << "Warning in DeformEM::LinearSystemE::makeSmoothnessTerm(): Skipping overlap between pair with excluded bone." << endl;
				continue;
			}

			std::vector<double> &integrals = integrals_per_pair[i].second;
			double ix    , iy    , iz    , jx    , jy    , jz    , ix_ix , ix_iy , ix_iz , iy_iy , iy_iz , iz_iz;
			double ix_jx , ix_jy , ix_jz , iy_jx , iy_jy , iy_jz , iz_jx , iz_jy , iz_jz , jx_jx , jx_jy , jx_jz , jy_jy , jy_jz , jz_jz;

			// Reset LHS & RHS
			for (int i2 = 0; i2 < 3*dim; i2++) {
				for (int j2 = 0; j2 < 4*dim; j2++) {
					coll[i2][j2] = 0.0;
				}
				rhs[i2] = 0.0;
			}

			ix     = integrals[0];
			iy     = integrals[1];
			iz     = integrals[2];
			jx     = integrals[3];
			jy     = integrals[4];
			jz     = integrals[5];
			ix_ix  = integrals[6];
			ix_iy  = integrals[7];
			ix_iz  = integrals[8];
			iy_iy  = integrals[9];
			iy_iz  = integrals[10];
			iz_iz  = integrals[11];
			ix_jx  = integrals[12];
			ix_jy  = integrals[13];
			ix_jz  = integrals[14];
			iy_jx  = integrals[15];
			iy_jy  = integrals[16];
			iy_jz  = integrals[17];
			iz_jx  = integrals[18];
			iz_jy  = integrals[19];
			iz_jz  = integrals[20];
			jx_jx  = integrals[21];
			jx_jy  = integrals[22];
			jx_jz  = integrals[23];
			jy_jy  = integrals[24];
			jy_jz  = integrals[25];
			jz_jz  = integrals[26];

			coll[0][1] = 2.0*iz;
			coll[0][2] = -2.0*iy;
			coll[0][4] = -2.0*jz;
			coll[0][5] = 2.0*jy;
			coll[0][6] = 2.0;
			coll[0][9] = -2.0;

			coll[1][0] = -2.0*iz;
			coll[1][2] = 2.0*ix;
			coll[1][3] = 2.0*jz;
			coll[1][5] = -2.0*jx;
			coll[1][7] = 2.0;
			coll[1][10] = -2.0;

			coll[2][0] = 2.0*iy;
			coll[2][1] = -2.0*ix;
			coll[2][3] = -2.0*jy;
			coll[2][4] = 2.0*jx;
			coll[2][8] = 2.0;
			coll[2][11] = -2.0;

			coll[3][0] = iy_iy+iz_iz;
			coll[3][1] = -ix_iy;
			coll[3][2] = -ix_iz;
			coll[3][3] = -iy_jy-iz_jz;
			coll[3][4] = iy_jx;
			coll[3][5] = iz_jx;
			coll[3][7] = -iz;
			coll[3][8] = iy;
			coll[3][10] = iz;
			coll[3][11] = -iy;

			coll[4][0] = -ix_iy;
			coll[4][1] = iz_iz+ix_ix;
			coll[4][2] = -iy_iz;
			coll[4][3] = ix_jy;
			coll[4][4] = -iz_jz-ix_jx;
			coll[4][5] = iz_jy;
			coll[4][6] = iz;
			coll[4][8] = -ix;
			coll[4][9] = -iz;
			coll[4][11] = ix;

			coll[5][0] = -ix_iz;
			coll[5][1] = -iy_iz;
			coll[5][2] = ix_ix+iy_iy;
			coll[5][3] = ix_jz;
			coll[5][4] = iy_jz;
			coll[5][5] = -ix_jx-iy_jy;
			coll[5][6] = -iy;
			coll[5][7] = ix;
			coll[5][9] = iy;
			coll[5][10] = -ix;

			coll[6][0] = iy_jy+iz_jz;
			coll[6][1] = -ix_jy;
			coll[6][2] = -ix_jz;
			coll[6][3] = -jy_jy-jz_jz;
			coll[6][4] = jx_jy;
			coll[6][5] = jx_jz;
			coll[6][7] = -jz;
			coll[6][8] = jy;
			coll[6][10] = jz;
			coll[6][11] = -jy;

			coll[7][0] = -iy_jx;
			coll[7][1] = iz_jz+ix_jx;
			coll[7][2] = -iy_jz;
			coll[7][3] = jx_jy;
			coll[7][4] = -jz_jz-jx_jx;
			coll[7][5] = jy_jz;
			coll[7][6] = jz;
			coll[7][8] = -jx;
			coll[7][9] = -jz;
			coll[7][11] = jx;

			coll[8][0] = -iz_jx;
			coll[8][1] = -iz_jy;
			coll[8][2] = ix_jx+iy_jy;
			coll[8][3] = jx_jz;
			coll[8][4] = jy_jz;
			coll[8][5] = -jx_jx-jy_jy;
			coll[8][6] = -jy;
			coll[8][7] = jx;
			coll[8][9] = jy;
			coll[8][10] = -jx;

			rhs[0] = -2.0*ix+2.0*jx;
			rhs[1] = -2.0*iy+2.0*jy;
			rhs[2] = -2.0*iz+2.0*jz;
			rhs[3] = -iz_jy+iy_jz;
			rhs[4] = iz_jx-ix_jz;
			rhs[5] = -iy_jx+ix_jy;
			rhs[6] = -iz_jy+iy_jz;
			rhs[7] = iz_jx-ix_jz;
			rhs[8] = -iy_jx+ix_jy;

			// Fill in LHS
			// Make sure to translate bi, bj in to the reduced set of bone indices!
			index_type bi, bj;
			bi = reduced_index[integrals_per_pair[i].first[0]];
			bj = reduced_index[integrals_per_pair[i].first[1]];
			index_type ciCol = bi*dim;
			index_type cbiCol = cbStart + ciCol;
			index_type cjCol = bj*dim;
			index_type cbjCol = cbStart + cjCol;
			for (int row = 0; row < 3*dim; row++) {
				for (int col = 0; col < dim; col++) {
					// Assign the linear rotation coefficients for bi
					if (coll[row][col] != 0) {
						smoothTerm->lhs.fill(TrackRow+row, ciCol+col, coll[row][col]);
					}

					// Assign the linear rotation coefficients for bj
					if (coll[row][col+dim] != 0) {
						smoothTerm->lhs.fill(TrackRow+row, cjCol+col, coll[row][col+dim]);
					}

					// Assign the translation coefficients for bi
					if (coll[row][col+2*dim] != 0) {
						smoothTerm->lhs.fill(TrackRow+row, cbiCol+col, coll[row][col+2*dim]);
					}

					// Assign the translation coefficients for bj
					if (coll[row][col+3*dim] != 0) {
						smoothTerm->lhs.fill(TrackRow+row, cbjCol+col, coll[row][col+3*dim]);
					}
				}
			}

			// Fill in RHS
			for (int row = 0; row < 3*dim; row++) {
				smoothTerm->rhs[TrackRow+row] = rhs[row];
			}

			// Increment row count
			TrackRow += 3*dim;
		}
	} while(vWeightGrid->scanNextCell(cellIndex));


	////////////////////////////////////////////////////////////////////////
	// Debugging: print out results
	//cout << "*** Estimated size for the Smoothness term: " << totalLHSnnz
	//	<< " vs. Actual size: " << smoothTerm->lhs.size() << endl;
	//cout << "*** Estimated Number of Rows: " << smoothTerm->rows
	//	<< " vs. Actual # rows: " << TrackRow << endl;

	// Assign the term to the list of terms, if there are any nonzero entries
	if (smoothTerm->rows) {
		terms.push_back(smoothTerm);
	}
}




taucsType* DeformEM::LinearSystemE::solve( taucsType &residual, taucsType &gradResidual )
{
	size_type numAllVars = numValidBones*2*dim;
	taucs_ccs_matrix *A, *A_T, *lhs;
	ErrorTerm* all;

	//////////////////////////////////////////////////////////////////////////
	// Construct the sparse matrix for the entire system
	//perfTimers["E-Step Creating System 1"].Start();

	all = mergeTerms(terms);

	//perfTimers["E-Step Creating System 1"].Stop();

	//perfTimers["E-Step Creating System 2"].Start();

	A = all->lhs.writeTaucsMatrix(all->rows, numAllVars);

	//perfTimers["E-Step Creating System 2"].Stop();

	//perfTimers["E-Step Creating System 3"].Start();

	A_T = all->lhs.writeTaucsMatrix(all->rows, numAllVars, true);

	//perfTimers["E-Step Creating System 3"].Stop();

	//perfTimers["E-Step Creating System 4"].Start();

	lhs = Mul2NonSymmMatSymmResult(A_T, A);

	//taucs_ccs_write_ijv(A, "textOutput/Estep-A.txt" );
	//taucs_ccs_write_ijv(A_T, "textOutput/A_T.txt" );
	//taucs_ccs_write_ijv(lhs, "textOutput/A_T A.txt" );
	//taucs_logfile ("textOutput/taucsLogE.txt");

	//perfTimers["E-Step Creating System 4"].Stop();

	//perfTimers["E-Step Creating System 5"].Start();

	// Allocate space for the rhs
	taucsType *b = new taucsType[all->rows];
	taucsType *rhs = new taucsType[numAllVars];
	taucsType *x = new taucsType[numAllVars];
	for (size_type i = 0; i < all->rows; i++) {
		b[i] = (taucsType) all->rhs[i];
	}

	// Calculate rhs
	MulNonSymmMatrixVector(A_T, b, rhs);

	//perfTimers["E-Step Creating System 5"].Stop();


	//////////////////////////////////////////////////////////////////////////
	// Solve the linear system
	int result;
	char* options[] = { "taucs.factor.LLT=true", NULL };

	//perfTimers["E-Step Solving System"].Start();

	result = taucs_linsolve (lhs,NULL,1,x,rhs,options,NULL);  

	//perfTimers["E-Step Solving System"].Stop();

	QString message("Error");
	switch(result) {
		case TAUCS_SUCCESS:
			message = QString("Success!");
			break;
		case TAUCS_ERROR:
			message = QString("%1 -- Generic error.").arg(message);
			break;
		case TAUCS_ERROR_NOMEM:
			message = QString("%1 -- NOMEM error.").arg(message);
			break;
		case TAUCS_ERROR_BADARGS:
			message = QString("%1 -- BADARGS error.").arg(message);
			break;
		case TAUCS_ERROR_MAXDEPTH:
			message = QString("%1 -- MAXDEPTH error.").arg(message);
			break;
		case TAUCS_ERROR_INDEFINITE:
			message = QString("%1 -- INDEFINITE error.").arg(message);
			break;
		default:
			break;
	}

	//cout << "--->>> E-Step Linear system solution result: " << qPrintable(message) << endl;

	//////////////////////////////////////////////////////////////////////////
	// Calculate residuals
	if (result == TAUCS_SUCCESS) {
		// Determine stop conditions
		taucsType *Ax = new taucsType[all->rows];
		taucsType *ATAx = new taucsType[numAllVars];

		residual = 0;
		MulNonSymmMatrixVector(A, x, Ax);
		for (size_type i = 0; i < all->rows; i++) {
			taucsType temp = Ax[i] - b[i];
			residual += temp*temp;
		}

		gradResidual = 0;
		MulNonSymmMatrixVector(A_T, Ax, ATAx);
		for (size_type i = 0; i < numAllVars; i++) {
			taucsType temp = 2.0*(ATAx[i] - rhs[i]);
			gradResidual += temp*temp;
		}

		//cout << "Residual: " << residual << ", Gradient of Residual: " << gradResidual << endl;

		// Free memory
		delete[] Ax;
		delete[] ATAx;
	}

	//////////////////////////////////////////////////////////////////////////
	// Deallocate memory
	taucs_ccs_free(A);
	taucs_ccs_free(A_T);
	taucs_ccs_free(lhs);
	delete[] b;
	delete[] rhs;
	delete all;

	return x;
}


void DeformEM::LinearSystemE::translateDeformationBack( RigidDeform &def ) const
{
	RigidDeform newDef(numBones);

	for (int i = 0; i < back_to_original_index.size(); i++) {
		newDef[back_to_original_index[i]] = def[i];
	}

	def = newDef;
}

RigidDeform DeformEM::LinearSystemE::projectToRigid( taucsType *soln, int maxItr, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *vc, const RigidDeform &currentDef )
{
	const size_type numCells = vWeightGrid->totalGridCells();

	//////////////////////////////////////////////////////////////////////////
	// 1. Compute and store the integrals
	//cout << "*** Updating the grid weight integrals over all " << numCells << " cells... ";

	// Store the integral values per cell (numCells), per bone (numValidBones),
	// ... and per component (xx, yy, zz, xy, yz, xz, x, y, z, 1)
	std::vector< std::vector< std::vector<double> > > integrals;
	std::vector<double> totalX, totalY, totalZ, total;
	integrals.resize(numCells);

	// Store the centroid of each bone
	totalX.resize(numValidBones);
	totalY.resize(numValidBones);
	totalZ.resize(numValidBones);
	total.resize(numValidBones);
	for (int bi = 0; bi < numValidBones; bi++) {
		totalX[bi] = 0.0;
		totalY[bi] = 0.0;
		totalZ[bi] = 0.0;
		total[bi] = 0.0;
	}

	int3_type cellIndex = vWeightGrid->scanCellStart();
	do {
		// 1. Integrals for projectToRigid()
		index_type lind = vWeightGrid->getCellLinearIndex(cellIndex);
		vWeightGrid->computeIntegrals(cellIndex, integrals[lind]);
		translateIntegrals(integrals[lind]);

		// Add to the total
		for (int bi = 0; bi < numValidBones; bi++) {
			totalX[bi] += integrals[lind][bi][6];
			totalY[bi] += integrals[lind][bi][7];
			totalZ[bi] += integrals[lind][bi][8];
			total[bi]  += integrals[lind][bi][9];
		}
	} while(vWeightGrid->scanNextCell(cellIndex));

	// Compute the average
	for (int bi = 0; bi < numValidBones; bi++) {
		totalX[bi] = totalX[bi] / total[bi];
		totalY[bi] = totalY[bi] / total[bi];
		totalZ[bi] = totalZ[bi] / total[bi];
	}

	//cout << "done!" << endl;

	//////////////////////////////////////////////////////////////////////////
	// 2. Compute and store the centroids of the weight function
	AffineDeform affDef = solnToAffine(soln, numValidBones);
	std::vector<Vec3d> centerLeft, centerRight;
	centerLeft.resize(numValidBones);
	centerRight.resize(numValidBones);

	//cout << "*** Computed centroids: " << endl;
	for (index_type bi = 0; bi < numValidBones; bi++) {
		centerLeft[bi][0] = totalX[bi];
		centerLeft[bi][1] = totalY[bi];
		centerLeft[bi][2] = totalZ[bi];
		//cout << "  Bone " << bi << ": " << centerLeft[bi] << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// 3. Compute the rotation for each bone
	RigidDeform defRigid;
	defRigid.resize(numValidBones);

	// Allocate space for eigenvectors / eigenvalues
	double *evals = new double[4];
	double **evecs = new double*[4];
	for (int i = 0; i < 4; i++) {
		evecs[i] = new double[4];
	}

	// Debugging: test the affine solution by comparing it with the constraints
	real_type affErr = findConstraintEnergy(affDef, vWeightGrid, vc, currentDef);
	//cout << "Total error of affine solution: " << affErr << endl;

	real_type initialErr = findConstraintEnergy(defRigid, vWeightGrid, vc, currentDef);
	//cout << "Initial error of identity deformation: " << initialErr << endl;

	// Iterate until we decrease the constraint energy
	index_type itr = 0;
	real_type step = 1.0;
	real_type rigidErr;					/**< Store current deformation error */
	do {
		// Update iteration number
		itr++;

		for (index_type bi = 0; bi < numValidBones; bi++) {
			Vec3d translation = step * affDef[bi].tr;
			Vec3d rotation = step * affDef[bi].lr;
			centerRight[bi] = centerLeft[bi] + translation + (rotation ^ centerLeft[bi]);

			//cout << "  Affine deformation: " << endl
			//	 << "    rot " << rotation << endl
			//	 << "    tr  " << translation << endl;
			//cout << "  Affine transformed Bone: " << centerRight[bi] << endl;
		}

		std::vector< std::vector<double> > total_S;
		total_S.resize(numValidBones);
		for (int bi = 0; bi < numValidBones; bi++) {
			total_S[bi].resize(9);
			for (int i = 0; i < 9; i++) {
				total_S[bi][i] = 0.0;
			}
		}

		std::vector< std::vector< std::vector<double> > > collected;
		collected.resize(9);
		for (int i = 0; i < 9; i++) {
			collected[i].resize(numValidBones);
			for (int bi = 0; bi < numValidBones; bi++) {
				collected[i][bi].resize(10);
				for (int j = 0; j < 10; j++) {
					collected[i][bi][j] = 0.0;
				}
			}
		}

		//////////////////////////////////////////////////////////////////////////
		// A. Compute quantities for estimating rotation
		// Order of the S components are: Sxx, Syx, Szx, Sxy, Syy, Szy, Sxz, Syz, Szz (columns are stacked)
		for (index_type j = 0; j < numValidBones; j++) {
			double cx, cy, cz, sx, sy, sz, xb, yb, zb;
			double t2, t3, t5, t6, t7, t11, t13, t17, t20, t21, t25, t32;

			cx = step * affDef[j].lr[0]; cy = step * affDef[j].lr[1]; cz = step * affDef[j].lr[2];
			sx = step * affDef[j].tr[0]; sy = step * affDef[j].tr[1]; sz = step * affDef[j].tr[2];
			xb = centerLeft[j][0]; yb = centerLeft[j][1]; zb = centerLeft[j][2];

			t2 = zb*cy;
			t3 = cz*yb;
			t5 = xb*cz;
			t6 = cy*xb;
			t7 = xb-sx+t2-t3;
			t11 = yb*cy;
			t13 = cz*zb;
			t17 = cx*zb;
			t20 = cx*xb;
			t21 = -yb+sy-t5+t17;
			t25 = cx*yb;
			t32 = zb-sz+t25-t6;

			collected[0][j][0] = 1.0;
			collected[0][j][3] = -cz;
			collected[0][j][5] = cy;
			collected[0][j][6] = -2.0*xb+sx-t2+t3;
			collected[0][j][7] = t5;
			collected[0][j][8] = -t6;
			collected[0][j][9] = xb*t7;
			collected[1][j][1] = -cz;
			collected[1][j][3] = 1.0;
			collected[1][j][4] = cy;
			collected[1][j][6] = -yb;
			collected[1][j][7] = -xb+sx+2.0*t3-t2;
			collected[1][j][8] = -t11;
			collected[1][j][9] = yb*t7;
			collected[2][j][2] = cy;
			collected[2][j][4] = -cz;
			collected[2][j][5] = 1.0;
			collected[2][j][6] = -zb;
			collected[2][j][7] = t13;
			collected[2][j][8] = -xb+sx+t3-2.0*t2;
			collected[2][j][9] = zb*t7;
			collected[3][j][0] = cz;
			collected[3][j][3] = 1.0;
			collected[3][j][5] = -cx;
			collected[3][j][6] = -yb+sy+t17-2.0*t5;
			collected[3][j][7] = -xb;
			collected[3][j][8] = t20;
			collected[3][j][9] = -xb*t21;
			collected[4][j][1] = 1.0;
			collected[4][j][3] = cz;
			collected[4][j][4] = -cx;
			collected[4][j][6] = -t3;
			collected[4][j][7] = -2.0*yb+sy-t5+t17;
			collected[4][j][8] = t25;
			collected[4][j][9] = -yb*t21;
			collected[5][j][2] = -cx;
			collected[5][j][4] = 1.0;
			collected[5][j][5] = cz;
			collected[5][j][6] = -t13;
			collected[5][j][7] = -zb;
			collected[5][j][8] = -yb+sy+2.0*t17-t5;
			collected[5][j][9] = -zb*t21;
			collected[6][j][0] = -cy;
			collected[6][j][3] = cx;
			collected[6][j][5] = 1.0;
			collected[6][j][6] = -zb+sz+2.0*t6-t25;
			collected[6][j][7] = -t20;
			collected[6][j][8] = -xb;
			collected[6][j][9] = xb*t32;
			collected[7][j][1] = cx;
			collected[7][j][3] = -cy;
			collected[7][j][4] = 1.0;
			collected[7][j][6] = t11;
			collected[7][j][7] = -zb+sz+t6-2.0*t25;
			collected[7][j][8] = -yb;
			collected[7][j][9] = yb*t32;
			collected[8][j][2] = 1.0;
			collected[8][j][4] = cx;
			collected[8][j][5] = -cy;
			collected[8][j][6] = t2;
			collected[8][j][7] = -t17;
			collected[8][j][8] = -2.0*zb+sz-t25+t6;
			collected[8][j][9] = zb*t32;
		}

		//////////////////////////////////////////////////////////////////////////
		// B. Iterate over each cell and accumulate the total
		int3_type gridIndex = vWeightGrid->scanCellStart();
		do {
			// Find position of the grid point
			index_type lind = vWeightGrid->getCellLinearIndex(gridIndex);

			for (index_type j = 0; j < numValidBones; j++) {
				// i. Gather the computed integral values for this bone component
				const std::vector<double> &cellIntegralBone = integrals[lind][j];

				// ii. Compute the product and add to the total
				for (int l = 0; l < 9; l++) {
					double prod = 0.0;
					for (int k = 0; k < 10; k++) {
						prod += collected[l][j][k] * cellIntegralBone[k];
					}

					total_S[j][l] += prod;
				}
			}
		} while (vWeightGrid->scanNextCell(gridIndex));


		//////////////////////////////////////////////////////////////////////////
		// C. For each bone, form the matrix N and solve the eigenvalue problem
		for (index_type j = 0; j < numValidBones; j++) {
			//////////////////////////////////////////////////////////////////////////
			//// SVD solution for finding the rotation
			//Matrix33d M, S, U, VT, rotation;
			//M(0,0) = total_S[j][0];
			//M(1,0) = total_S[j][1];
			//M(2,0) = total_S[j][2];
			//M(0,1) = total_S[j][3];
			//M(1,1) = total_S[j][4];
			//M(2,1) = total_S[j][5];
			//M(0,2) = total_S[j][6];
			//M(1,2) = total_S[j][7];
			//M(2,2) = total_S[j][8];

			////cout << "Matrix for bone " << j << ":" << endl << M << endl;

			//// Compute SVD
			//M.computeSVD(S, U, VT);
			//rotation = (U*VT).transpose();

			//// 2. Compute the translation as the right centroid minus the rotated left centroid
			//Vec3d translation = centerRight[j] - (rotation * centerLeft[j]);

			//// 1. Assign rigid transformation
			//SE3 rt;

			//for (int r = 0; r < 3; r++) {
			//	for (int c = 0; c < 3; c++) {
			//		rt.rot(r,c) = rotation(r,c);
			//	}
			//	rt.tr[r] = translation[r];
			//}

			//////////////////////////////////////////////////////////////////////////
			// C. For each bone, form the matrix N and solve the eigenvalue problem
			Matrix44d Nj;
			Nj(0,0) =   total_S[j][0] + total_S[j][4] + total_S[j][8];
			Nj(1,1) =   total_S[j][0] - total_S[j][4] - total_S[j][8];
			Nj(2,2) = - total_S[j][0] + total_S[j][4] - total_S[j][8];
			Nj(3,3) = - total_S[j][0] - total_S[j][4] + total_S[j][8];
			Nj(0,1) = Nj(1,0) = total_S[j][7] - total_S[j][5];
			Nj(0,2) = Nj(2,0) = total_S[j][2] - total_S[j][6];
			Nj(0,3) = Nj(3,0) = total_S[j][3] - total_S[j][1];
			Nj(1,2) = Nj(2,1) = total_S[j][3] + total_S[j][1];
			Nj(1,3) = Nj(3,1) = total_S[j][2] + total_S[j][6];
			Nj(2,3) = Nj(3,2) = total_S[j][7] + total_S[j][5];


			//// Debugging:
			//cout << " M in vector form for bone " << j << ": " << endl;
			//for (int k = 0; k < 9; k++) {
			//	cout << "   M(" << k+1 << ") = " << totalS[k][j] << ";" << endl;
			//}
			//cout << " Resulting N: " << endl;
			//cout << Nj << endl;

			// Compute the eigenvector
			Nj.eigSymmetric(evals, evecs);


			// Find the index of the maximum eigenvalue
			double maxEval = -DBL_MAX;
			int maxIndex = -1;
			for (int i=0; i<4; i++) {
				if (evals[i] > maxEval) {
					maxIndex = i;
					maxEval = evals[i];
				}
				//cout << "Eigenvalue  " << i << ": " << evals[i] << "\n";
				//cout << "Eigenvector " << i << ": ("
				//	<< evecs[0][i] << ", "
				//	<< evecs[1][i] << ", "
				//	<< evecs[2][i] << ", "
				//	<< evecs[3][i] << ")\n";
			}

			//////////////////////////////////////////////////////////////////////////
			// Compute the bone's deformation
			SE3 rt;

			// 1. Compute the rotation according to the maximum eigenvalue's eigenvector
			double q0, qx, qy, qz;
			Geometry::Vec<4,double> q;
			for (int i = 0; i < 4; i++) q[i] = evecs[i][maxIndex];
			q.normalize();
			q0 = q[0]; qx = q[1]; qy = q[2]; qz = q[3];

			rt.rot(0,0) = q0*q0 + qx*qx - qy*qy - qz*qz;
			rt.rot(1,1) = q0*q0 - qx*qx + qy*qy - qz*qz;
			rt.rot(2,2) = q0*q0 - qx*qx - qy*qy + qz*qz;
			rt.rot(0,1) = 2.0*(qx*qy - q0*qz);
			rt.rot(1,0) = 2.0*(qy*qx + q0*qz);
			rt.rot(0,2) = 2.0*(qx*qz + q0*qy);
			rt.rot(2,0) = 2.0*(qz*qx - q0*qy);
			rt.rot(1,2) = 2.0*(qy*qz - q0*qx);
			rt.rot(2,1) = 2.0*(qz*qy + q0*qx);

			// 2. Compute the translation as the right centroid minus the rotated left centroid
			real3_type leftCenter(centerLeft[j][0], centerLeft[j][1], centerLeft[j][2]);
			real3_type rightCenter(centerRight[j][0], centerRight[j][1], centerRight[j][2]);
			rt.tr = rightCenter - (rt.rot * leftCenter);

			// Store the bone's rigid transformation
			defRigid[j] = rt;
		}

		// Repeat until we decrease the error
		rigidErr = findConstraintEnergy(defRigid, vWeightGrid, vc, currentDef);
		//cout << "  -- Current error with step " << step << ": " << rigidErr << endl;
		step /= 2.0;

	} while((rigidErr > initialErr) && (itr < maxItr));

	//////////////////////////////////////////////////////////////////////////
	// Delete the temporary variables
	delete[] evals;
	for (int i = 0; i < 4; i++) {
		delete[] evecs[i];
	}
	delete[] evecs;


	// Translate back to original number of bones
	translateDeformationBack(defRigid);

	return defRigid;
}

AffineDeform DeformEM::LinearSystemE::solnToAffine( taucsType *soln, size_type numBones )
{
	AffineDeform def;
	const index_type cbStart = numBones*dim;
	def.resize(numBones);

	for (index_type j = 0; j < numBones; j++) {
		index_type ciRow = j*dim;
		index_type cbiRow = cbStart + ciRow;

		def[j].lr[0] = soln[ciRow];  def[j].lr[1] = soln[ciRow + 1];  def[j].lr[2] = soln[ciRow + 2];
		def[j].tr[0] = soln[cbiRow]; def[j].tr[1] = soln[cbiRow + 1]; def[j].tr[2] = soln[cbiRow + 2];
	}

	return def;
}

DeformEM::LinearSystemM::LinearSystemM( const size_type numBones, const size_type numGridPoints ) : numBones(numBones), numGridPoints(numGridPoints)
{
	cout << "*** Initializing M-step system: Number of variables [" << numBones << " bones, " << numGridPoints << " grid points] = total " << numBones*numGridPoints << endl;
}

void DeformEM::LinearSystemM::clear()
{
	terms.clear();
}

void DeformEM::LinearSystemM::makeConstraintTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def, const std::vector< std::vector<int> > &affectingBones )
{
	// For each constraint
	//   1. Find the nodes that affect the vertex
	//   2. Calculate and add the term to the linear system

	ErrorTerm *constraintTerm = new ErrorTerm();
	const int3_type gridSize = vWeightGrid->size();
	const int numVars = numGridPoints*numBones;
	const int numConstraints = con->size();
	const int nnzPerRow = 8*numBones;
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	constraintTerm->weight = weight;
	constraintTerm->rows = numConstraints*dim;
	constraintTerm->lhs.resize(nnzPerRow * constraintTerm->rows);
	constraintTerm->rhs = new real_type[constraintTerm->rows];

	//////////////////////////////////////////////////////////////////////////
	for (index_type ci = 0; ci < numConstraints; ci++) {
		// Don't add the constraint if the weight is zero
		if (!con->valid(ci)) {
			continue;
		}

		// Retrieve the nodes that affect this vertex from the list of precomputed weights
		real3_type sourcePos = con->sourcePos(ci);
		real3_type newPos = con->targetPos(ci);

		// 1. Use source positions to look up vertex weights
		int3_type idx;
		real3_type rem;
		idxList gridIndices;
		distList gridWeights;
		vWeightGrid->queryCell(sourcePos, idx, rem);
		// Check to see if it's a valid cell (just in case the cell is out of bounds or not marked)
		if (!vWeightGrid->validCell(idx)) {
			continue;
		}
		vWeightGrid->getWeights(sourcePos, gridIndices, gridWeights);

		// 2. Use deformed positions for the actual constraints in the system
		//for (index_type bi = 0; bi < numBones; bi++) {

		const std::vector<int> &boneList = affectingBones[vWeightGrid->getCellLinearIndex(idx)];
		for (index_type i = 0; i < boneList.size(); i++) {
			index_type bi = boneList[i];
			index_type boneOffset = numGridPoints*bi;
			real3_type boneDeformedPos = def[bi] * sourcePos;

			for (index_type gi = 0; gi < gridIndices.size(); gi++) {
				for (int d = 0; d < dim; d++) {
					real_type fillValue = gridWeights[gi] * boneDeformedPos[d];

					// Only populate non-zero terms
					if (fillValue != 0) {
						constraintTerm->lhs.fill(TrackRow+d, boneOffset + gridIndices[gi], fillValue);
					}
				}
			}
		}

		// 3. Fill the right-hand side
		for (int d = 0; d < dim; d++) {
			constraintTerm->rhs[TrackRow+d] = newPos[d];
		}

		TrackRow += dim;
	}

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < constraintTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << constraintTerm->rhs[ci] << ";" << endl;
	//}
	//constraintTerm->lhs.print();
	//cout << "Estimated size for the M-step constraint term: " << nnzPerRow * constraintTerm->rows
	//	 << " vs. Actual size: " << constraintTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(constraintTerm);
}

void DeformEM::LinearSystemM::makeSmoothTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid )
{
	// For each constraint
	//   1. Find the nodes that affect the vertex
	//   2. Calculate and add the term to the linear system

	ErrorTerm *smoothTerm = new ErrorTerm();
	const int nnzPerRow = 2;
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	smoothTerm->weight = weight;
	smoothTerm->lhs.resize(nnzPerRow * (2*dim)*numBones*numGridPoints);			/**< Estimated size of LHS */

	//////////////////////////////////////////////////////////////////////////
	// Iterate over all grid cells
	index_type linearIndex, linearIndexN;

	// The difference between neighboring grid points
	int3_type gridIndex = vWeightGrid->scanStart();
	do {
		linearIndex = vWeightGrid->getLinearIndex(gridIndex);

		if (linearIndex == 4294967295) {
			cerr << "Error: negative index detected in DeformEM::LinearSystemM::makeSmoothTerm()!" << endl;
		}

		int3List neighbors = vWeightGrid->findNeighbors(gridIndex);
		for (unsigned int i = 0; i < neighbors.size(); i++) {

			linearIndexN = vWeightGrid->getLinearIndex(neighbors[i]);
			if (linearIndexN == 4294967295) {
				cerr << "Error: negative neighbor index detected in DeformEM::LinearSystemM::makeSmoothTerm()!" << endl;
			}

			for (index_type bi = 0; bi < numBones; bi++) {
				index_type boneOffset = numGridPoints*bi;

				smoothTerm->lhs.fill(TrackRow, boneOffset + linearIndex, 1.0);
				smoothTerm->lhs.fill(TrackRow, boneOffset + linearIndexN, -1.0);

				// Increment the row
				TrackRow++;
			}
		}
	} while(vWeightGrid->scanNext(gridIndex));

	smoothTerm->rows = TrackRow;
	smoothTerm->rhs = new real_type[smoothTerm->rows];
	for (int ci = 0; ci < smoothTerm->rows; ci++) {
		smoothTerm->rhs[ci] = 0.0;
	}

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < smoothTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << smoothTerm->rhs[ci] << ";" << endl;
	//}
	//smoothTerm->lhs.print();
	//cout << "Estimated size for the vweight smooth term: " << nnzPerRow * (2*dim)*numBones*numGridPoints
	//	 << " vs. Actual size: " << smoothTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(smoothTerm);
}

void DeformEM::LinearSystemM::makeUnityTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid )
{
	ErrorTerm *unityTerm = new ErrorTerm();
	const int3_type gridSize = vWeightGrid->size();
	const int nnzPerRow = numBones;
	index_type TrackRow = 0;

	//////////////////////////////////////////////////////////////////////////
	unityTerm->weight = weight;
	unityTerm->rows = numGridPoints;
	unityTerm->lhs.resize(nnzPerRow * unityTerm->rows);
	unityTerm->rhs = new real_type[unityTerm->rows];

	//////////////////////////////////////////////////////////////////////////
	// Iterate over all grid cells
	int3_type gridIndex;
	index_type linearIndex;

	// Sum of weight vector at each grid cell should add up to 1
	gridIndex = vWeightGrid->scanStart();
	do {

		// Fill in the current grid point for all bones
		//cout << "Index: " << gridIndex << endl;
		linearIndex = vWeightGrid->getLinearIndex(gridIndex);

		for (index_type bi = 0; bi < numBones; bi++) {
			index_type boneOffset = numGridPoints*bi;

			unityTerm->lhs.fill(TrackRow, boneOffset + linearIndex, 1.0);
		}

		// Set the RHS
		unityTerm->rhs[TrackRow] = 1.0;

		// Increment the row
		TrackRow++;

	} while(vWeightGrid->scanNext(gridIndex));

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < unityTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << unityTerm->rhs[ci] << ";" << endl;
	//}
	//unityTerm->lhs.print();
	//cout << "Estimated size for the vweight unity term: " << nnzPerRow * unityTerm->rows
	//	 << " vs. Actual size: " << unityTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(unityTerm);
}


taucsType* DeformEM::LinearSystemM::solve( taucsType &residual, taucsType &gradResidual )
{
	size_type numAllVars = numBones*numGridPoints;
	taucs_ccs_matrix *A, *A_T, *lhs;
	ErrorTerm* all;

	//perfTimers["M-Step Creating System"].Start();

		//////////////////////////////////////////////////////////////////////////
		// Create the sparse matrix for the entire system
		all = mergeTerms(terms);
		A = all->lhs.writeTaucsMatrix(all->rows, numAllVars);

		//taucs_ccs_write_ijv(A, "C:\\A_M-step.txt" );
		//taucs_ccs_write_ijv(A_T, "A_T.txt" );
		//taucs_ccs_write_ijv(lhs, "A_T A.txt" );
		//taucs_logfile ("C:\\taucsLog_M-step.txt");

		// Allocate space for the rhs
		taucsType *b = new taucsType[all->rows];
		for (size_type i = 0; i < all->rows; i++) {
			b[i] = (taucsType) all->rhs[i];
		}

	//perfTimers["M-Step Creating System"].Stop();

	//perfTimers["M-Step Solving System"].Start();

		//////////////////////////////////////////////////////////////////////////
		// Solve the linear system
		taucsType *x = t_snnls(A, b, &residual, 2.0, 1 );

	//perfTimers["M-Step Solving System"].Stop();

	if( x == NULL )	{
		cout << "Solve failed! (Matrix probably not positive definite)" << endl;
	}
	else {
		cout << "Solve success! (Residual = " << residual << ")" << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// Deallocate memory
	taucs_ccs_free(A);
	delete[] b;
	delete all;

	return x;
}

void DeformEM::LinearSystemM::updateWeights( taucsType *soln, MarkedUniformGrid3 *vWeightGrid )
{
	// Read the weights from the solution vector
	const int3_type gridSize = vWeightGrid->size();
	int3_type gridIndex;
	index_type linearIndex;
	bool negativeWeight = false;

	gridIndex = vWeightGrid->scanStart();
	do {
		linearIndex = vWeightGrid->getLinearIndex(gridIndex);

		for (index_type bi = 0; bi < numBones; bi++) {
			index_type boneOffset = numGridPoints*bi;
			vWeightGrid->setValue(gridIndex, bi, soln[boneOffset + linearIndex]);
			if (!negativeWeight && (soln[boneOffset + linearIndex] < 0)) {
				negativeWeight = true;
			}
		}

	} while(vWeightGrid->scanNext(gridIndex));

	if (negativeWeight) {
		cout << "*** Negative weights detected!" << endl;
	}
}



DeformEM::LinearSystemM_Reduced::LinearSystemM_Reduced( const size_type numBones, const size_type numGridPoints, const std::vector< std::vector<int> > &pointsPerBone, const std::vector< std::vector<int> > &cellsPerBone, const std::vector< std::vector<int> > &bonesPerPoint, const std::vector< std::vector<int> > &bonesPerCell, const std::vector< stdext::hash_map<int,int> > &boneCellLinearIndex, const std::vector< stdext::hash_map<int,int> > &bonePointLinearIndex ) :
numBones(numBones), numGridPoints(numGridPoints),
cellsPerBone(cellsPerBone), boneCellLinearIndex(boneCellLinearIndex),
pointsPerBone(pointsPerBone), bonePointLinearIndex(bonePointLinearIndex),
bonesPerPoint(bonesPerPoint), bonesPerCell(bonesPerCell)
{
	cout << "*** Initializing M-step system:" << endl;
	cout << "****** Total number of variables [" << numBones << " bones, " << numGridPoints << " grid points] = total " << numBones*numGridPoints << endl;

	// Compute the bone offsets
	cout << "****** Points per bone:";
	boneOffsets.resize(numBones+1, 0);
	int currentOffset = 0;
	for (int i = 0; i < numBones; i++) {
		currentOffset += pointsPerBone[i].size();
		boneOffsets[i+1] = currentOffset;
		cout << " " << pointsPerBone[i].size();
	}
	cout << endl << "****** Total reduced number of variables: " << boneOffsets[numBones] << endl;
}

void DeformEM::LinearSystemM_Reduced::clear()
{
	terms.clear();
}

int DeformEM::LinearSystemM_Reduced::getBoneOffset( int bone )
{
	return boneOffsets[bone];
}

int DeformEM::LinearSystemM_Reduced::findVariableCellIndex( int bone, int gridCellIndex )
{
	// Look up the index in the hash table
	stdext::hash_map<int,int>::const_iterator it;
	it = boneCellLinearIndex[bone].find(gridCellIndex);

	if (it == boneCellLinearIndex[bone].end()) {
		return -1;
	}
	else {
		return it->second;
	}
}

int DeformEM::LinearSystemM_Reduced::findVariableIndex( int bone, int gridPointIndex )
{
	// Look up the index in the hash table
	stdext::hash_map<int,int>::const_iterator it;
	it = bonePointLinearIndex[bone].find(gridPointIndex);

	if (it == bonePointLinearIndex[bone].end()) {
		return -1;
	}
	else {
		return it->second;
	}
}

void DeformEM::LinearSystemM_Reduced::makeConstraintTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid, const CorrespondenceList *con, RigidDeform &def, const std::vector< std::vector<int> > &affectingBones )
{
	// For each constraint
	//   1. Find the nodes that affect the vertex
	//   2. Calculate and add the term to the linear system

	const int3_type gridSize = vWeightGrid->size();
	int numConstraints = 0;

	// Count the number of valid constraints
	for (index_type ci = 0; ci < con->size(); ci++) {
		if (con->valid(ci)) {
			++numConstraints;
		}
	}

	const int numRows = numConstraints*dim;
	const int numNonZero = numConstraints*numBones*8*dim;			/**< Estimate the number of non-zero entries */
	index_type TrackRow = 0;

	// Construct the error term
	ErrorTerm *constraintTerm = new ErrorTerm(weight, numRows, numNonZero);

	//////////////////////////////////////////////////////////////////////////
	for (index_type ci = 0; ci < con->size(); ci++) {
		// Don't add the constraint if the weight is zero
		if (!con->valid(ci)) {
			continue;
		}

		// Retrieve the nodes that affect this vertex from the list of precomputed weights
		real3_type sourcePos = con->sourcePos(ci);
		real3_type newPos = con->targetPos(ci);

		// 1. Use source positions to look up vertex weights
		int3_type idx;
		real3_type rem;
		idxList gridIndices;
		distList gridWeights;
		vWeightGrid->queryCell(sourcePos, idx, rem);
		// Check to see if it's a valid cell (just in case the cell is out of bounds or not marked)
		if (!vWeightGrid->validCell(idx)) {
			continue;
		}
		vWeightGrid->getWeights(sourcePos, gridIndices, gridWeights);

		int cellLinearIndex = vWeightGrid->getCellLinearIndex(idx);
		const std::vector<int> &boneList = affectingBones[cellLinearIndex];

		// 2. Use deformed positions for the actual constraints in the system
		for (index_type i = 0; i < boneList.size(); i++) {
			index_type bi = boneList[i];
			real3_type boneDeformedPos = def[bi] * sourcePos;

			// Check that the cell falls within the bone's influence domain
			if (findVariableCellIndex(bi, cellLinearIndex) < 0) {
				cerr << "Warning: constraint " << ci << " in grid cell " << cellLinearIndex << " is outside of bone " << bi << "'s influence." << endl;
				continue;
			}

			for (index_type gi = 0; gi < gridIndices.size(); gi++) {
				// Check that the grid point falls within the bone's influence domain
				int pointBoneDomainIndex = findVariableIndex(bi, gridIndices[gi]);

				if (pointBoneDomainIndex < 0) {
					cerr << "Warning: constraint " << ci << " with grid point " << gridIndices[gi] << " is outside of bone " << bi << "'s influence." << endl;
					continue;
				}

				// Otherwise, fill in the value
				for (int d = 0; d < dim; d++) {
					real_type fillValue = gridWeights[gi] * boneDeformedPos[d];

					// Only populate non-zero terms
					if (fillValue != 0.0) {
						constraintTerm->lhs.fill(TrackRow+d, getBoneOffset(bi) + pointBoneDomainIndex, fillValue);
					}
				}
			}
		}

		// 3. Fill the right-hand side
		for (int d = 0; d < dim; d++) {
			constraintTerm->rhs[TrackRow+d] = newPos[d];
		}

		TrackRow += dim;
	}

	////////////////////////////////////////////////////////////////////////
	// Debugging: print out results
	//for (int ci = 0; ci < constraintTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << constraintTerm->rhs[ci] << ";" << endl;
	//}
	//constraintTerm->lhs.print();

	cout << "Estimated # rows for the M-step constraint term: " << numRows
		<< " vs. Actual size: " << TrackRow << endl;

	cout << "Estimated size for the M-step constraint term: " << numNonZero
		 << " vs. Actual size: " << constraintTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(constraintTerm);
}

void DeformEM::LinearSystemM_Reduced::makeSmoothTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid )
{
	ErrorTerm *smoothTerm = new ErrorTerm();
	const int nnzPerRow = 2;
	index_type TrackRow = 0;
	const int numNonZero = nnzPerRow * (2*dim)*numBones*numGridPoints;

	//////////////////////////////////////////////////////////////////////////
	smoothTerm->weight = weight;
	smoothTerm->lhs.resize(numNonZero);			/**< Estimated size of LHS */

	//////////////////////////////////////////////////////////////////////////
	// The difference between neighboring grid points
	stdext::hash_map<int,std::vector<int> >::const_iterator it;

	for (index_type bi = 0; bi < numBones; bi++) {

		for (it = vWeightGrid->allEdges.begin(); it != vWeightGrid->allEdges.end(); it++) {
			int i1 = it->first;
			const std::vector<int> &neighbors = it->second;

			for (int j = 0; j < neighbors.size(); j++) {
				int i2 = neighbors[j];

				int idx = findVariableIndex(bi, i1);
				int idxN = findVariableIndex(bi, i2);

				// Check if both grid points are included in this bone's influence domain
				if ((idx < 0) || (idxN < 0)) {
					continue;
				}

				// Fill the matrix
				smoothTerm->lhs.fill(TrackRow, getBoneOffset(bi) + idx, 1.0);
				smoothTerm->lhs.fill(TrackRow, getBoneOffset(bi) + idxN, -1.0);

				// Increment the row
				TrackRow++;
			}
		}
	}

	////////////////////////////////////////////////////////////////////////////
	//// Alternative version: graph laplacian-like smoothness term
	//index_type linearIndex, linearIndexN;
	//int3_type gridIndex = vWeightGrid->scanStart();
	//do {
	//	linearIndex = vWeightGrid->getLinearIndex(gridIndex);

	//	if (linearIndex == 4294967295) {
	//		cerr << "Error: negative index detected in DeformEM::LinearSystemM::makeSmoothTerm()!" << endl;
	//	}

	//	int3List neighbors = vWeightGrid->findNeighbors(gridIndex);

	//	for (index_type bi = 0; bi < numBones; bi++) {
	//		// Check if this node is inside the influence
	//		int idx = findVariableIndex(bi, linearIndex);
	//		int boneOffset = getBoneOffset(bi);

	//		// Check if this grid point is included in this bone's influence domain
	//		if (idx < 0) {
	//			continue;
	//		}

	//		int numFilledNeighbors = 0;
	//		for (unsigned int i = 0; i < neighbors.size(); i++) {
	//			linearIndexN = vWeightGrid->getLinearIndex(neighbors[i]);
	//			if (linearIndexN == 4294967295) {
	//				cerr << "Error: negative neighbor index detected in DeformEM::LinearSystemM::makeSmoothTerm()!" << endl;
	//			}
	//			int idxN = findVariableIndex(bi, linearIndexN);

	//			// Check if this neighbor is included in this bone's influence domain
	//			if (idxN < 0) {
	//				continue;
	//			}

	//			smoothTerm->lhs.fill(TrackRow, boneOffset + idxN, -1.0);
	//			numFilledNeighbors++;
	//		}

	//		if (numFilledNeighbors > 0) {
	//			smoothTerm->lhs.fill(TrackRow, boneOffset + idx, (real_type) numFilledNeighbors);
	//		}

	//		// Increment the row
	//		TrackRow++;
	//	}

	//} while (vWeightGrid->scanNext(gridIndex));


	//////////////////////////////////////////////////////////////////////////
	// Fill the RHS with 0's
	smoothTerm->rows = TrackRow;
	smoothTerm->rhs = new real_type[smoothTerm->rows];
	for (int ci = 0; ci < smoothTerm->rows; ci++) {
		smoothTerm->rhs[ci] = 0.0;
	}

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < smoothTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << smoothTerm->rhs[ci] << ";" << endl;
	//}
	//smoothTerm->lhs.print();

	cout << "# rows for the M-step smooth term: " << TrackRow << endl;

	cout << "Estimated size for the M-step smooth term: " << numNonZero
		 << " vs. Actual size: " << smoothTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(smoothTerm);
}

void DeformEM::LinearSystemM_Reduced::makeUnityTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid )
{
	const int3_type gridSize = vWeightGrid->size();
	const int numRows = numGridPoints;
	const int numNonZero = boneOffsets[numBones];
	index_type TrackRow = 0;

	ErrorTerm *unityTerm = new ErrorTerm(weight, numRows, numNonZero);

	//////////////////////////////////////////////////////////////////////////
	// Iterate over all grid cells
	int3_type gridIndex;
	index_type linearIndex;

	// Sum of weight vector at each grid cell should add up to 1
	gridIndex = vWeightGrid->scanStart();
	do {
		// Fill in the current grid point for all bones
		//cout << "Index: " << gridIndex << endl;
		linearIndex = vWeightGrid->getLinearIndex(gridIndex);

		const std::vector<int> &boneList = bonesPerPoint[linearIndex];

		for (index_type i = 0; i < boneList.size(); i++) {
			int bi = boneList[i];
			int varIdx = findVariableIndex(bi, linearIndex);

			unityTerm->lhs.fill(TrackRow, getBoneOffset(bi) + varIdx, 1.0);
		}

		// Set the RHS
		unityTerm->rhs[TrackRow] = 1.0;

		// Increment the row
		TrackRow++;

	} while(vWeightGrid->scanNext(gridIndex));

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < unityTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << unityTerm->rhs[ci] << ";" << endl;
	//}
	//unityTerm->lhs.print();
	cout << "Estimated # rows for the M-step unity term: " << numRows
		 << " vs. Actual size: " << TrackRow << endl;

	cout << "Estimated size for the M-step unity term: " << numNonZero
		 << " vs. Actual size: " << unityTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(unityTerm);
}


void DeformEM::LinearSystemM_Reduced::makeZeroTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid )
{
	const int numNonZero = boneOffsets[numBones];
	const int numRows = boneOffsets[numBones];
	index_type TrackRow = 0;

	ErrorTerm *zeroTerm = new ErrorTerm(weight, numRows, numNonZero);

	//////////////////////////////////////////////////////////////////////////
	// Identity on left, 0 on right
	for (index_type i = 0; i < boneOffsets[numBones]; i++) {
		zeroTerm->lhs.fill(TrackRow, i, 1.0);

		// Set the RHS
		zeroTerm->rhs[TrackRow] = 0.0;

		// Increment the row
		TrackRow++;
	}

	// Identity on left, 1 on right
	for (index_type i = 0; i < boneOffsets[numBones]; i++) {
		zeroTerm->lhs.fill(TrackRow, i, 1.0);

		// Set the RHS
		zeroTerm->rhs[TrackRow] = 1.0;

		// Increment the row
		TrackRow++;
	}

	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < zeroTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << zeroTerm->rhs[ci] << ";" << endl;
	//}
	//zeroTerm->lhs.print();
	cout << "Estimated # rows for the M-step zero term: " << numRows
		<< " vs. Actual size: " << TrackRow << endl;

	cout << "Estimated size for the M-step zero term: " << numNonZero
		<< " vs. Actual size: " << zeroTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(zeroTerm);
}


void DeformEM::LinearSystemM_Reduced::makeDiscreteTerm( real_type weight, MarkedUniformGrid3 *vWeightGrid )
{
	const int numNonZero = boneOffsets[numBones];
	const int numRows = boneOffsets[numBones];
	index_type TrackRow = 0;

	ErrorTerm *discreteTerm = new ErrorTerm(weight, numRows, numNonZero);

	//////////////////////////////////////////////////////////////////////////
	// For each point on the grid
	// a. Look up current value on the vertex weight grid
	// b. Constrain the terms to be exactly that value
	int3_type gridIndex;
	index_type linearIndex;
	data_type gridValue(vWeightGrid->getDataDim());

	// Sum of weight vector at each grid cell should add up to 1
	gridIndex = vWeightGrid->scanStart();
	do {
		// Fill in the current grid point for all bones
		//cout << "Index: " << gridIndex << endl;
		linearIndex = vWeightGrid->getLinearIndex(gridIndex);

		// Find current grid weight values
		for (int i = 0; i < vWeightGrid->getDataDim(); ++i) {
			gridValue.val[i] = vWeightGrid->getValue(gridIndex, i);
		}

		// Constrain the grid point to have the current grid weight value
		const std::vector<int> &boneList = bonesPerPoint[linearIndex];
		for (index_type i = 0; i < boneList.size(); i++) {
			int bi = boneList[i];
			int varIdx = findVariableIndex(bi, linearIndex);

			discreteTerm->lhs.fill(TrackRow, getBoneOffset(bi) + varIdx, 1.0);

			// Set the RHS
			discreteTerm->rhs[TrackRow] = gridValue.val[bi];

			// Increment the row
			TrackRow++;
		}
	} while(vWeightGrid->scanNext(gridIndex));


	//////////////////////////////////////////////////////////////////////////
	//// Debugging: print out results
	//for (int ci = 0; ci < discreteTerm->rows; ci++) {
	//	cout << "RHS(" << ci+1 << ") = " << discreteTerm->rhs[ci] << ";" << endl;
	//}
	//discreteTerm->lhs.print();
	cout << "Estimated # rows for the M-step discrete term: " << numRows
		<< " vs. Actual size: " << TrackRow << endl;

	cout << "Estimated size for the M-step discrete term: " << numNonZero
		<< " vs. Actual size: " << discreteTerm->lhs.size() << endl;

	// Assign the term to the list of terms
	terms.push_back(discreteTerm);
}


taucsType* DeformEM::LinearSystemM_Reduced::solve( int maxIRLSItr, taucsType &resNorm, taucsType &gradResNorm )
{
	size_type numAllVars = boneOffsets[numBones];
	taucs_ccs_matrix *A, *A_T, *ATA;
	taucsType *b, *ATb, *x, *ATb2;
	ErrorTerm* all;

	//////////////////////////////////////////////////////////////////////////
	// Construct the sparse matrix for the entire system
	// Use method of normal equations to solve the least squares problem
	//perfTimers["M-Step Creating System 1"].Start();

	// Create LHS
	all = mergeTerms(terms);
	A = all->lhs.writeTaucsMatrix(all->rows, numAllVars);
	A_T = MatrixTranspose(A);
	ATA = Mul2NonSymmMatSymmResult(A_T, A);

	// Create RHS
	x = new taucsType[numAllVars];
	b = new taucsType[all->rows];
	ATb = new taucsType[numAllVars];
	taucsType b_max = -DBL_MAX;
	for (size_type i = 0; i < all->rows; i++) {
		b[i] = (taucsType) all->rhs[i];
		if (abs(b[i]) > b_max) {
			b_max = abs(b[i]);
		}
	}
	MulNonSymmMatrixVector(A_T, b, ATb);

	// Debug output
	all->lhs.write("C:\\Mstep-A-orig.txt");
	writeVector(b, all->rows, "C:\\Mstep-b.txt");
	//taucs_ccs_write_ijv(A_T, "C:\\Mstep-A_T.txt" );
	//taucs_ccs_write_ijv(A, "C:\\Mstep-A.txt" );
	//taucs_logfile ("C:\\taucsLog_Mstep.txt");

	//perfTimers["M-Step Creating System 1"].Stop();


	//////////////////////////////////////////////////////////////////////////
	// Solve the linear system
	int result;
	char* options[] = { "taucs.factor.LLT=true", NULL };

	//perfTimers["M-Step Solving System"].Start();

	// Preallocate some stuff
	taucsType *residual = new taucsType[all->rows];
	taucsType *tempX = new taucsType[numAllVars];
	taucsType *c = new taucsType[numAllVars];
	taucs_ccs_matrix *M = taucs_ccs_create(ATA->m, ATA->n, ATA->colptr[ATA->n], ATA->flags);
	char buffer[255];
	taucsType l2Error, l1Error, prevL1Error;
	const int numForceZero = floor(0.05*((double)numAllVars));		/**< NNLS parameter */
	taucsType invEps = 100.0 / b_max;								/**< IRLS parameter */

	// Start IRLS solver
	int numIRLSItr = 0;
	bool IRLSdone = false;
	taucsType IRLSConvergeThreshold = 1e-3;

	// Mark the rows involved with the smoothness term
	// Some pre-processing: figure out row positions for each error term
	std::vector<size_type> startRowList(terms.size()+1, 0);
	for (int i = 1; i <= terms.size(); i++) {
		startRowList[i] = startRowList[i-1] + terms[i-1]->rows;
		cerr << "****** Term #" << i << ": from row " << startRowList[i-1] << " to " << startRowList[i]-1 << endl;
	}
	std::vector<bool> selectiveReweight(all->rows, true);
	for (int i = startRowList[0]; i < startRowList[1]; ++i) {
		selectiveReweight[i] = false;
	}
	for (int i = startRowList[2]; i < startRowList[3]; ++i) {
		selectiveReweight[i] = false;
	}

	// Main IRLS loop
	while (!IRLSdone)
	{
		// Keep track of previous l1 residual
		prevL1Error = l1Error;

		// Start NNLS solver
		int numNNLSItr = 0;
		bool NNLSdone = false;
		int numActiveVars = numAllVars;
		std::vector<bool> keep(numAllVars, true);

		do {
			//cerr << "*** Matrix is dim=" << M->n << " and nnz=" << M->colptr[M->n] << endl;

			//////////////////////////////////////////////////////////////////////////
			// Solve the linear system
			if (numNNLSItr == 0) {
				result = taucs_linsolve (ATA,NULL,1,tempX,ATb,options,NULL);
			}
			else {
				result = taucs_linsolve (M,NULL,1,tempX,c,options,NULL);
			}
			numNNLSItr++;

			//////////////////////////////////////////////////////////////////////////
			// Map the result back and also gather stats
			int nneg = 0, nOverOne = 0, count = 0;
			for (int i = 0; i < numAllVars; i++) {
				if (keep[i]) {
					x[i] = tempX[count];
					count++;

					// Count negative values and values over one
					if (x[i] < 0) {
						nneg++;
					}
					else if (x[i] > 1) {
						nOverOne++;
					}
				}
				else {
					x[i] = 0.0;
				}
			}
			l2Error = computeL2Error(A, x, b, residual);
			cerr << "*** NNLS itr " << numNNLSItr << ": L2Error=" << l2Error << ", #negative=" << nneg
				<< ", #active=" << numActiveVars << ", #overOne=" << nOverOne << endl;


			//// Output debugging
			//if (numNNLSItr == 1) {
			//	sprintf_s(buffer, 255, "C:\\Mstep-NNLS-ATA_%03d.txt", numNNLSItr);
			//	taucs_ccs_write_ijv(ATA, buffer);
			//	sprintf_s(buffer, 255, "C:\\Mstep-NNLS-ATb_%03d.txt", numNNLSItr);
			//	writeVector(ATb, numActiveVars, buffer);
			//}
			//else {
			//	sprintf_s(buffer, 255, "C:\\Mstep-NNLS-ATA_%03d.txt", numNNLSItr);
			//	taucs_ccs_write_ijv(M, buffer);
			//	sprintf_s(buffer, 255, "C:\\Mstep-NNLS-ATb_%03d.txt", numNNLSItr);
			//	writeVector(c, numActiveVars, buffer);
			//}
			//sprintf_s(buffer, 255, "C:\\Mstep-x_%03d.txt", numNNLSItr);
			//writeVector(tempX, numActiveVars, buffer);

			//////////////////////////////////////////////////////////////////////////
			if (nneg > 0) {
				// Force a fraction of the negative variables to zero
				if (nneg > numForceZero) {
					// Sort based on residual (ascending order)
					std::vector< pair<taucsType,int> > sortWithIndex(numAllVars);
					for (int i = 0; i < numAllVars; i++) {
						sortWithIndex[i].first = x[i];
						sortWithIndex[i].second = i;
					}
					struct resComp : public binary_function< const pair<taucsType,int>&, const pair<taucsType,int>&, bool> {
						bool operator()( const pair<taucsType,int> &x, const pair<taucsType,int> &y ) {
							return (x.first < y.first);
						}
					};
					sort(sortWithIndex.begin(), sortWithIndex.end(), resComp());

					// Force only the smallest to become zero
					for (int i = 0; i < numForceZero; i++) {
						keep[sortWithIndex[i].second] = false;
						numActiveVars--;
					}
				}
				else {
					// Force all negative variables to become zero
					for(int i = 0; i < numAllVars; i++) {
						if (x[i] < 0) {
							keep[i] = false;
							numActiveVars--;
						}
					}
				}

				// Recompute the matrices
				taucsSubmatrixSymmetric(ATA, M, keep);
				taucsSubvector(ATb, c, keep, numActiveVars);
			}
			else {
				NNLSdone = true;
			}
		} while(!NNLSdone);

		// Increment number of IRLS iterations done
		++numIRLSItr;

		//// Output debugging
		//sprintf_s(buffer, 255, "C:\\Mstep-IRLS-Residuals_%03d.txt", numIRLSItr);
		//writeVector(residual, all->rows, buffer);

		//////////////////////////////////////////////////////////////////////////
		// Compute weights and l1 error
		l1Error = computeIRLSWeightsSq(residual, all->rows, invEps, selectiveReweight);

		cerr << "*** IRLS itr " << numIRLSItr << ": L1Error=" << l1Error << endl;

		// Determine stop conditions
		if ((numIRLSItr >= maxIRLSItr) || ((numIRLSItr > 1) && (abs(prevL1Error - l1Error) < IRLSConvergeThreshold))) {
			IRLSdone = true;
		}

		// Re-weight matrices
		if (!IRLSdone) {
			reweightATA(A_T, A, residual, ATA);
			reweightATb(A_T, b, residual, ATb);

			//// Output debugging
			//sprintf_s(buffer, 255, "C:\\Mstep-IRLS-Weights_%03d.txt", numIRLSItr);
			//writeVector(residual, all->rows, buffer);
			//sprintf_s(buffer, 255, "C:\\Mstep-IRLS-ATA_%03d.txt", numIRLSItr);
			//taucs_ccs_write_ijv(ATA, buffer);
			//sprintf_s(buffer, 255, "C:\\Mstep-IRLS-ATb_%03d.txt", numIRLSItr);
			//writeVector(ATb, numAllVars, buffer);
		}
	}

	// Free some memory
	delete[] tempX;
	delete[] c;
	delete[] residual;
	taucs_ccs_free(M);  /**< TODO: watch for memory leak.  M has been overwritten many times with different parameters. */

	//perfTimers["M-Step Solving System"].Stop();

	// Assign residuals
	resNorm = l2Error;
	gradResNorm = 0.0;			// Since it's convex, the global minimum must have gradient = 0.

	//////////////////////////////////////////////////////////////////////////
	// Deallocate memory
	taucs_ccs_free(A);
	taucs_ccs_free(A_T);
	taucs_ccs_free(ATA);
	delete[] b;
	delete[] ATb;
	delete all;

	// Return solution
	return x;
}


taucsType* DeformEM::LinearSystemM_Reduced::solve_OLD( taucsType &residual, taucsType &gradResidual )
{
	const int numVars = boneOffsets[numBones];
	taucs_ccs_matrix *A, *A_T, *lhs;
	ErrorTerm* all;

	perfTimers["M-Step Creating System"].Start();

	//////////////////////////////////////////////////////////////////////////
	// Create the sparse matrix for the entire system
	all = mergeTerms(terms);
	A = all->lhs.writeTaucsMatrix(all->rows, numVars);

	//taucs_ccs_write_ijv(A, "C:\\A_M-step_Reduced.txt" );
	//taucs_ccs_write_ijv(A_T, "A_T.txt" );
	//taucs_ccs_write_ijv(lhs, "A_T A.txt" );
	//taucs_logfile ("C:\\taucsLog_M-step_Reduced.txt");

	// Allocate space for the rhs
	taucsType *b = new taucsType[all->rows];
	for (size_type i = 0; i < all->rows; i++) {
		b[i] = (taucsType) all->rhs[i];
	}
	// Write out RHS
	//writeVector(b, all->rows, "C:\\A_M-step_Reduced_RHS.txt");

	perfTimers["M-Step Creating System"].Stop();

	perfTimers["M-Step Solving System"].Start();

	//////////////////////////////////////////////////////////////////////////
	// Solve the linear system
	taucsType *x = t_snnls(A, b, &residual, 2.0, 1 );

	perfTimers["M-Step Solving System"].Stop();

	if( x == NULL )	{
		cout << "Solve failed! (Matrix probably not positive definite)" << endl;
	}
	else {
		cout << "Solve success! (Residual = " << residual << ")" << endl;
	}

	//////////////////////////////////////////////////////////////////////////
	// Deallocate memory
	taucs_ccs_free(A);
	delete[] b;
	delete all;

	return x;
}

void DeformEM::LinearSystemM_Reduced::updateWeights( taucsType *soln, MarkedUniformGrid3 *vWeightGrid )
{
	// Read the weights from the solution vector
	const int3_type gridSize = vWeightGrid->size();
	bool negativeWeight = false;

	int3_type gridIndex = vWeightGrid->scanStart();
	do {
		index_type linearIndex = vWeightGrid->getLinearIndex(gridIndex);

		// Get list of bones affecting this point
		const std::vector<int> &boneList = bonesPerPoint[linearIndex];

		for (index_type i = 0; i < boneList.size(); i++) {
			int bi = boneList[i];
			int varIdx = findVariableIndex(bi, linearIndex);
			taucsType solnValue = soln[getBoneOffset(bi) + varIdx];

			vWeightGrid->setValue(gridIndex, bi, solnValue);

			// Check if the weight is negative
			if (!negativeWeight && (solnValue < 0)) {
				negativeWeight = true;
			}
		}

	} while(vWeightGrid->scanNext(gridIndex));

	if (negativeWeight) {
		cout << "*** Negative weights detected!" << endl;
	}
}

AccumulatedGeometry::AccumulatedGeometry()
{

}

AccumulatedGeometry::AccumulatedGeometry( const ShapeModel &shape )
{
	resetInitialShape(shape);
}

void AccumulatedGeometry::resetInitialShape( const ShapeModel &shape )
{
	// Clear off all data
	clearData();

	// Copy vertex, normal, spin image information
	vertex = shape.vertex;
	normal = shape.normal;
	spinImageStack = shape.spinImageStack;

	// Copy bounding box info
	box_min = shape.box_min;
	box_max = shape.box_max;

	// Copy triangle info
	if (shape.shapeType == SHAPE_MESH) {
		triangle_vertex_index = ((Mesh3_real&)shape).triangle_vertex_index;
	}
}

void AccumulatedGeometry::clearData()
{
	vertex.clear();
	normal.clear();
	spinImageStack.clear();
	box_min = box_max = real3_type(0.0,0.0,0.0);
}

void AccumulatedGeometry::accumulateShape( const ShapeModel &shape, real_type maxDistanceThreshold, real_type minDistanceThreshold )
{
	cout << "*** Accumulating points on reference pose..." << endl;

	if (shape.shapeType == SHAPE_MESH) {
		// Recompute normals of the shape to add
		((Mesh3_real&)shape).computeNormals(true);
	}

	// Mark the vertices to add to the shape. We do this first as a preprocess in order to not confuse ANN.
	std::vector<bool> markAdd;
	markAdd.resize(shape.numVertices(), false);

	for (int i = 0; i < shape.numVertices(); i++) {
		if (shape.markVertex[i] == VERTEX_INVALID) {
			// Don't add invalid points
			continue;
		}

		// Compute the distance between the vertex and the existing point set
		real3_type pos = shape.vertex[i];
		idxList nearest;
		distList dists;
		vertex.queryKNN(pos, 1, nearest, dists, true);

		// Find the point to plane distance
		real_type toPlaneDistance = abs((pos - vertex[nearest[0]]) * normal[nearest[0]]);

		// Find the distance on the plane
		real3_type projectedOnPlane = pos - normal[nearest[0]]*toPlaneDistance;
		real_type onPlaneDistance = (projectedOnPlane - vertex[nearest[0]]).norm();

		//// See if it satisfies the distance threshold
		//if ((toPlaneDistance < maxDistanceThreshold) && (onPlaneDistance > minDistanceThreshold)) {
		//	markAdd[i] = true;
		//}

		// See if it satisfies the distance threshold
		if (dists[0] > minDistanceThreshold) {
			markAdd[i] = true;
		}
	}

	// Finally, add all the marked points
	int startIndex = vertex.size();
	for (int i = 0; i < shape.numVertices(); i++) {
		if (markAdd[i]) {
			real3_type pos = shape.vertex[i];
			vertex.add(pos);
			normal.add(shape.normal[i]);

			// Update bounding box
			for (int j = 0; j < 3; j++) {
				if (pos[j] < box_min[j]) {
					box_min[j] = pos[j];
				}
				if (pos[j] > box_max[j]) {
					box_max[j] = pos[j];
				}
			}

		}
	}
	//// Reset ANN data
	//vertex.clearANNData();
	//normal.clearANNData();


	//////////////////////////////////////////////////////////////////////////
	// Optional step: add triangles where all three vertices are marked for addition
	if (shape.shapeType == SHAPE_MESH) {
		Mesh3_real &mesh = ((Mesh3_real&)shape);

		// First, translate the indices of the original shape to the accumulated shape
		stdext::hash_map<int,int> indexMap;
		int count = 0;
		for (int i = 0; i < shape.numVertices(); ++i) {
			if (markAdd[i]) {
				indexMap[i] = startIndex + count;
				++count;
			}
		}

		// Next, iterate over all triangles in the original shape
		for(size_type t = 0, t_end = mesh.triangle_vertex_index.size(); t < t_end; ++t){
			const index3_type& tvi = mesh.triangle_vertex_index[t];
			index3_type newTriangle;

			// Check to see if all three vertices are added to the accumulated shape
			bool ok = true;
			for(size_type j = 0; j < 3; ++j){
				if (!markAdd[tvi[j]]) {
					ok = false;
					break;
				}
				else {
					// Translate the triangle index
					stdext::hash_map<int,int>::const_iterator it = indexMap.find(tvi[j]);
					if (it != indexMap.end()) {
						newTriangle[j] = it->second;
					}
					else {
						cerr << "Error in AccumulatedGeometry::accumulateShape(): index translation not found in hash map!" << endl;
					}
				}
			}

			if (ok) {
				// Add the triangle
				this->triangle_vertex_index.push_back(newTriangle);
			}
		}
	}


	//////////////////////////////////////////////////////////////////////////
	// Compute the spin images for new points
	cout << "*** Computing spin image features for new points..." << endl;
	cout << "****** Bin Size: " << spinImageStack.binSize << endl;
	cout << "****** Image Width: " << spinImageStack.imageWidth << endl;
	cout << "****** Support Angle: " << spinImageStack.supportAngle << " degrees" << endl;
	cout << "****** Progress";
	TextProgress(vertex.size() - startIndex, true);
	for (int i = startIndex; i < vertex.size(); i++) {
		SpinImage new_si;

		// Store the current vertex and normal
		real3_type p = vertex[i];
		real3_type n = normal[i];
		real_type minCosAngle = cos(spinImageStack.supportAngle / 180.0 * M_PI);

		// Allocate space for the spin image
		new_si.resize(spinImageStack.imageWidth);
		new_si.reset();

		// Accumulate spin image for the rest of the vertices
		for (index_type j = 0; j < vertex.size(); j++) {
			// Don't accumulate the same point
			if (i == j) {
				continue;
			}

			// Support angle test
			if ( (n * normal[j]) > minCosAngle ) {
				// Compute spin map coordinates and accumulate
				new_si.accumulate(SpinMap(p, n, vertex[j]), spinImageStack.binSize);
			}
		}

		// Add to the stack
		spinImageStack.add(new_si);

		TextProgress(i-startIndex);
	}
	
	cout << " done!" << endl;
}

void AccumulatedGeometry::draw( const DrawOptions &options ) const
{
	//// Just draw the points without normals or lighting
	//glPointSize(options.openGLPointSize);
	//glDisable(GL_LIGHTING);
	//vertex.draw(POINT_SQUARES);

	if (triangle_vertex_index.size()) {
		// Draw triangles
		glEnable(GL_LIGHTING);
		glDisable(GL_COLOR_MATERIAL);
		setOpenGLMaterialColor(options.deformedModelColor);

		glBegin(GL_TRIANGLES);
		for( size_type t = 0,t_end = triangle_vertex_index.size(); t < t_end; ++t){
			const index3_type& tvi = triangle_vertex_index[t];

			for (size_type i = 0; i < 3; ++i) {
			 	const real3_type& v = vertex[tvi[i]];
				const real3_type& n = normal[tvi[i]];

				glNormal3f(n[0],n[1],n[2]);
				glVertex3f(v[0],v[1],v[2]);
			}
		}
		glEnd();
	}
}

void AccumulatedGeometry::transformVertex( const AccumulatedGeometry &geometry, const SE3 &tf, index_type i )
{
	// Apply the rigid transformation to the vertex and frames
	vertex[i] = tf * geometry.vertex[i];

	// Compute inverse transpose to transform normal
	Matrix33f tf_invT = tf.rot.transpose();
	invert(tf_invT);
	normal[i] = tf_invT * geometry.normal[i];
	normal[i].normalize();
}


void DeformEM::FrameSolution::generateInitialCorrespondences( int numCorrespondences )
{
	cout << "*** Picking " << numCorrespondences << " correspondences in new solution...";
	idxList pickList = bestCandidateSample(sourceModel->vertex, numCorrespondences);
	numCorrespondences = pickList.size();

	cList = new CorrespondenceList();
	cList->resize(numCorrespondences);

	for (index_type i = 0; i < numCorrespondences; i++) {
		cList->sourcePoints[i] = sourceModel->vertex[pickList[i]];
		cList->sourceIndex[i] = pickList[i];
		cList->targetPoints[i] = real3_type(0.0, 0.0, 0.0);
		cList->targetIndex[i] = 0;
		cList->weight[i] = 0.0;
	}
}


void DeformEM::FrameSolution::generateAdjacentCorrespondences( const ShapeModel *shape, int numCorrespondences )
{
	// 1. TODO: just generating new samples on the new mesh may result in a lot of samples
	// concentrated in certain regions.  Is there a better strategy than just sampling the new mesh?
	// This wasn't a problem when accumulating the geometry.

	// 2. CAUTION: when using the aList, be careful that you're not using a vertex that
	// hasn't been inversely mapped well.  So check sourceModel->markVertex before using the alist correspondence.

	cout << "*** Picking " << numCorrespondences << " correspondences in new solution...";
	idxList pickList = bestCandidateSample(shape->vertex, numCorrespondences);
	numCorrespondences = pickList.size();

	aList = new CorrespondenceList();
	aList->resize(numCorrespondences);

	for (index_type i = 0; i < numCorrespondences; i++) {
		aList->sourcePoints[i] = shape->vertex[pickList[i]];
		aList->sourceIndex[i] = pickList[i];
		aList->targetPoints[i] = real3_type(0.0, 0.0, 0.0);
		aList->targetIndex[i] = 0;
		aList->weight[i] = 0.0;
	}
}

